///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//

#include "../include/gomez.h"

#define NUMSAMPLES				8

class GThugha : public GApplication {
	private:
	GFrame		Frame;
	GSoundInput	Sound;
	GSample		Sample[NUMSAMPLES];
	static GSample *Current;

	public:
	GThugha():GApplication("GThugha") {};

	protected:
	BOOL FASTCALL OnStart(LONG );
	BOOL FASTCALL OnTerminate(void);
	BOOL FASTCALL OnChar(LONG );
	BOOL FASTCALL OnTimer(LONG ,LONG );
	BOOL FASTCALL OnResize(INT ,INT );
    BOOL FASTCALL OnActivate(void);
    BOOL FASTCALL OnDeactivate(void);

	private:
	static BOOL FASTCALL SoundProc(INT ,GSoundInput *,GSample *);
	void FASTCALL Flame0(void);
	void FASTCALL Flame1(void);
	void FASTCALL Scope0(void);
};

GSample* GThugha::Current;



BOOL FASTCALL GThugha::OnStart(LONG p) {
 	INT i;
 	GPalette pal;

    SetResizeable(TRUE);
	Open(320,200,8);

 	Frame.Alloc(GetWidth(),GetHeight(),8);

 	pal.ImportMAP("gthugha.map");
 	SetPalette(&pal);
 	RealizePalette();
 	Frame.SetPalette(&pal);
 	Frame.RealizePalette();
 	Frame.ClearImage(0);

 	for (i=0;i<NUMSAMPLES;i++)
		Sample[i].Alloc(1000,22000,8,1);
 	Current=&Sample[0];
 	Sound.SetHandler(SoundProc);
 	Sound.Open(&Sample[0]);
 	for (i=0;i<NUMSAMPLES;i++)
		Sound.AddBuffer(&Sample[i]);

 	SetTimerSpeed(50);
 	SetPriority(GMZ_PRIORITY_HIGH);

 	return TRUE;
}

BOOL FASTCALL GThugha::OnTerminate(void) {
 	INT i;

 	SetPriority(GMZ_PRIORITY_IDLE);
 	SetTimerSpeed(0);
 	Sound.Stop();
 	Sound.Close();

 	for (i=0;i<NUMSAMPLES;i++)
		Sample[i].Free();
 	Frame.Free();
 	Close();

 	return TRUE;
}

BOOL FASTCALL GThugha::SoundProc(INT msg,GSoundInput *sound,GSample *smp) {
 	if (msg==GMZ_SOUND_DONE)    {
		sound->AddBuffer(smp);
		Current=smp;
   	}
 	return TRUE;
}

BOOL FASTCALL GThugha::OnActivate(void) {
 	SetPriority(GMZ_PRIORITY_HIGH);
 	return TRUE;
}

BOOL FASTCALL GThugha::OnDeactivate(void) {
 	SetPriority(GMZ_PRIORITY_IDLE);
 	return TRUE;
}

BOOL FASTCALL GThugha::OnResize(INT width,INT height) {
    Frame.Resize(width,height);
    return TRUE;
}


void FASTCALL GThugha::Flame0(void) {
 	FAST BYTE *frm=(BYTE *)Frame.GetImage();
 	BYTE *eof;
 	INT width=Frame.GetWidth(),height=Frame.GetHeight();
 	FAST INT nfrm;

 	frm+=width;

 	for(eof=frm+width*(height-2);frm<eof;frm++)    {
		nfrm=((((UINT)*frm+1)+
			   ((UINT)*(frm-1))+
			   ((UINT)*(frm+Width))+
			   ((UINT)*(frm-Width)))>>2);
		*frm=(BYTE)(((((UINT)*frm)<<1)+nfrm)>>2);
   	}
}


void FASTCALL GThugha::Flame1(void) {
 	FAST BYTE *frm=(BYTE *)Frame.GetImage();
 	BYTE *eof;
 	INT width=Frame.GetWidth(),height=Frame.GetHeight();
 	FAST INT nfrm;

 	frm+=width;

 	for(eof=frm+width*(height-2);frm<eof;frm++)    {
		nfrm=((((UINT)*frm+1)+
			   ((UINT)*(frm-1))+
			   ((UINT)*(frm+Width))+
			   ((UINT)*(frm-Width)))>>2);
		*frm=(BYTE)(((UINT)*frm+(nfrm<<1))>>2);
   	}
}


void FASTCALL GThugha::Scope0(void) {
 	INT x=0,center=Frame.GetHeight()/2;
 	FAST INT y,s;
 	FAST BYTE *smp=(BYTE *)Current->GetData();
 	BYTE *lsmp=smp+Current->GetLength();

 	while ((*smp!=128) && (smp<lsmp))
    	smp++;
 	if (smp+Frame.GetWidth()>lsmp)
    	smp=lsmp-Frame.GetWidth();

 	y=*smp-128+center;
 	for (lsmp=smp+Frame.GetWidth();smp<lsmp;(smp++,x++)) {
/*		y=smp[i];
		if (y>=1 && y<=Frame.GetHeight()-1) {
			frm[y*width]=255;
			frm[y*width+1]=255;
			frm[y*(width+1)]=255;
			frm[y*(width+1)+1]=255;
	   	}*/

		s=*smp-128+center;
		if (y>s)
			while(y>=s)
				Frame.PutPixel8(x,y--,196);
		else
			while(y<=s)
				Frame.PutPixel8(x,y++,196);
   	}
}


BOOL FASTCALL GThugha::OnChar(LONG key) {
 	switch(key) {
		case 'x': Terminate(); break;
	}
 return TRUE;
}


BOOL FASTCALL GThugha::OnTimer(LONG id,LONG time) {
 	Scope0();
 	Flame0();
    Frame.PutPixel(0,0,196);
    Frame.PutPixel(Frame.GetWidth()-1,0,196);
    Frame.PutPixel(0,Frame.GetHeight()-1,196);
    Frame.PutPixel(Frame.GetWidth()-1,Frame.GetHeight()-1,196);
    Frame.Put(this,0,0);
 	return TRUE;
}

void main(void) {
 	GThugha Scope;
 	Scope.Start();
}
