///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#ifndef __GAPPLET_H
#define __GAPPLET_H

#include "../include/gomez.h"

#if defined(__FORWINDOWS__)
 #define GMZ_PRIORITY_IDLE			IDLE_PRIORITY_CLASS
 #define GMZ_PRIORITY_NORMAL   		NORMAL_PRIORITY_CLASS
 #define GMZ_PRIORITY_HIGH			HIGH_PRIORITY_CLASS
#else
 #define GMZ_PRIORITY_IDLE			1
 #define GMZ_PRIORITY_NORMAL   		8
 #define GMZ_PRIORITY_HIGH			16
#endif


class GTimer {
		protected:
		void		(FASTCALL *Handler)(LONG );
		LONG		Speed;
		LONG		TickCount;
		#if defined(__FORWINDOWS__)
		UINT		TimerId;
		#endif

		public:
		GTimer();
		GTimer(void (FASTCALL *)(LONG ));
		~GTimer();

		PTR			FASTCALL SetHandler(void (FASTCALL *)(LONG));
		PTR			FASTCALL GetHandler(void);
		LONG		FASTCALL SetSpeed(LONG ,LONG res=20);
		LONG		FASTCALL GetSpeed(void);
		LONG		FASTCALL SetTickCount(LONG );
		LONG		FASTCALL GetTickCount(void);

		private:
		static void	FASTCALL DefaultHandler(LONG );
		#if defined(__FORWINDOWS__)
		 static VOID CALLBACK TimerProc(UINT ,UINT ,DWORD ,DWORD ,DWORD );
		#endif
	   };

class GThread : public GObject,
				public GMessageQueue {
		public:
        enum {
        	THREAD_TYPE = MAKETYPE('T','H','R','D')
        };

		protected:
		DWORD       ThreadId;
		DWORD 		ThreadHandle;
		INT			Priority;
		LONG		SuspendCount;

		public:
		GThread(CHAR* name=NULL,DWORD typ=THREAD_TYPE);
		~GThread();
        CHAR*		FASTCALL SetDefaultName();

		BOOL		FASTCALL Start(LONG p1=0);
		BOOL		FASTCALL Suspend(void);
		BOOL        FASTCALL Resume(void);
		BOOL		FASTCALL Terminate(void);
		BOOL		FASTCALL Sleep(LONG );

		INT			FASTCALL GetPriority(void);
		INT			FASTCALL SetPriority(INT );

		DWORD		FASTCALL GetThreadId(void);
		DWORD		FASTCALL GetThreadHandle(void);
		LONG		FASTCALL GetSuspendCount(void);


		protected:
		virtual INT 		  Main(LONG p1);
		virtual BOOL FASTCALL OnStart(LONG p);
		virtual BOOL FASTCALL OnTerminate(void);
        virtual BOOL FASTCALL OnMessage(void);
        virtual BOOL FASTCALL OnSuspend(void);
        virtual BOOL FASTCALL OnResume(void);

		private:
        virtual void FASTCALL NotifyMessage(void);
		#if defined(__FORWIN32__) && defined(__FORBORLAND__) && defined(__MT__)
		 static void _USERENTRY StartProc(void *);
		 DWORD		param;
		#elif defined(__FORWIN32__)
		 static DWORD WINAPI StartProc(LPVOID );
		 DWORD		param;
		#endif
	   };

class GParameter {
		public:
		static CHAR*	FASTCALL GetParameter(INT );
        static CHAR*  	FASTCALL GetParameter(CHAR *);
        static CHAR*  	FASTCALL GetProgram(void);
	   };

class GApplication: public GDisplay {
		public:
        enum {
        	APPLICATION_TYPE = MAKETYPE('A','P','P','N')
        };

		protected:
		INT 		Priority;

        public:
        GParameter	Parameter;

		public:
		GApplication(CHAR* name=NULL,DWORD typ=APPLICATION_TYPE);
		~GApplication();

        CHAR*		FASTCALL SetDefaultName();

		INT			FASTCALL GetPriority(void);
		INT			FASTCALL SetPriority(INT );

		CHAR*		FASTCALL GetFlagsName(CHAR *);
		LONG		FASTCALL GetProcessId(void);
	   };

///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GTimer
//
inline GTimer::GTimer() {
 	Handler=DefaultHandler;
 	Speed=TickCount=0;
   #if defined(__FORWINDOWS__)
 	TimerId=0;
   #endif
}

inline GTimer::GTimer(void (FASTCALL *h)(LONG )) {
 	Handler=h;
 	Speed=TickCount=0;
   #if defined(__FORWINDOWS__)
 	TimerId=0;
   #endif
}

inline GTimer::~GTimer() {
 	Handler=DefaultHandler;
 	SetSpeed(0);
}

inline PTR FASTCALL GTimer::SetHandler(void (FASTCALL *h)(LONG )) {
 	return (PTR)Handler=h;
}

inline PTR FASTCALL GTimer::GetHandler(void) {
 	return (PTR)Handler;
}

inline LONG FASTCALL GTimer::SetTickCount(LONG h) {
 	return TickCount=h;
}

inline LONG FASTCALL GTimer::GetTickCount(void) {
 	return TickCount;
}

inline LONG FASTCALL GTimer::GetSpeed(void) {
 	return Speed;
}


///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GApplication
//
inline GApplication::GApplication(CHAR* name,DWORD typ) : GDisplay(name,typ) {
   #if defined(__FORWIN32__)
 	Priority=GetPriorityClass(GetCurrentProcess());
   #else
 	Priority=GMZ_PRIORITY_NORMAL;
 	Instance=TRUE;
   #endif
}

inline GApplication::~GApplication() {
 	Close();
}

inline CHAR* FASTCALL GApplication::SetDefaultName() {
    strcpy(Name,"application");
    return Name;
}

inline LONG FASTCALL GApplication::GetProcessId(void) {
   #if defined(__FORWINDOWS__)
 	return (LONG)GetCurrentProcess();
   #else
 	return TRUE;
   #endif
}

inline INT FASTCALL GApplication::GetPriority(void) {
 	return Priority;
}

inline INT FASTCALL GApplication::SetPriority(INT p) {
   #if defined(__FORWIN32__)
  	return (SetPriorityClass(GetCurrentProcess(),p) ? Priority=p : Priority);
   #else
  	return Priority=p;
   #endif
}


///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GThread
//
inline GThread::GThread(CHAR *name,DWORD typ) : GObject(name,typ) {
 	ThreadHandle=ThreadId=SuspendCount=Priority=NULL;
}

inline GThread::~GThread() {
 	if (ThreadHandle)
 		Terminate();
}

inline CHAR* FASTCALL GThread::SetDefaultName() {
    strcpy(Name,"thread");
    return Name;
}

#pragma argsused
inline INT GThread::Main(LONG p) {
 	return TRUE;
}

#pragma argsused
inline BOOL FASTCALL GThread::OnStart(LONG p) {
 	return TRUE;
}

inline BOOL FASTCALL GThread::OnTerminate(void) {
 	return TRUE;
}

inline BOOL FASTCALL GThread::OnMessage(void) {
 	return TRUE;
}

inline BOOL FASTCALL GThread::OnSuspend(void) {
 	return TRUE;
}

inline BOOL FASTCALL GThread::OnResume(void) {
 	return TRUE;
}

inline DWORD FASTCALL GThread::GetThreadId(void) {
 	return ThreadId;
}

inline DWORD FASTCALL GThread::GetThreadHandle(void) {
 	return ThreadHandle;
}

inline LONG FASTCALL GThread::GetSuspendCount(void) {
 	return SuspendCount;
}

inline INT FASTCALL GThread::GetPriority(void) {
 	return Priority;
}

inline INT FASTCALL GThread::SetPriority(INT p) {
   #if defined(__FORWIN32__)
  	return (SetThreadPriority((HANDLE)ThreadHandle,p) ? Priority=p : Priority);
   #else
  	return Priority=p;
   #endif
}

inline void FASTCALL GThread::NotifyMessage(void) {
   #if defined(__FORWINDOWS__)
  	PostThreadMessage(ThreadId,WM_MESSAGE_NOTIFY,0,0);
   #else
  	OnMessage();
   #endif
}


///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GParameter
//
inline CHAR* FASTCALL GParameter::GetProgram(void) {
   #if defined(__FORDOS__) || defined(__FORWINDOWS__)
  	return _argv[0];
   #else
  	return NULL;
   #endif
}

inline CHAR* FASTCALL GParameter::GetParameter(INT i) {
   #if defined(__FORDOS__) || defined(__FORWINDOWS__)
  	return _argv[i+1];
   #else
  	return NULL;
   #endif
}

inline CHAR* FASTCALL GParameter::GetParameter(CHAR* sub) {
   #if defined(__FORDOS__) || defined(__FORWINDOWS__)
  	INT i=1,len=strlen(sub);
  	while(_argv[i]!=NULL) {
    	if (memcmp(_argv[i],sub,len)==0)
        	return _argv[i];
    	i++;
   	}
  	return NULL;
   #else
  	return NULL;
   #endif
}

#endif
