///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#ifndef __GBITMAP_H
#define __GBITMAP_H

#include "../include/gomez.h"

#define DCX_Magic	       		0x3ade68b1L
#define SCR_Magic	       		0x0cd
#define PCX_Magic             	10
#define PCX_Coding            	1
#define PCX_Version           	5
#define PCX_Pal_Type          	0
#define PCX_Bitplanes         	1

DECLARE_TYPE(GPixel);
DECLARE_TYPE(GMask);

class GImage : public GObject {
		public:
        enum {
        	IMAGE_TYPE = MAKETYPE('I','M','A','G')
        };

		protected:
		INT			Width;
		INT			Height;
		INT			LineSize;
		INT			Depth;
		DOUBLE		AspectRatio;
		GRectL    	ClipWindow;

		public:
		GImage(CHAR *name=NULL,DWORD typ=IMAGE_TYPE);
		~GImage();

        CHAR*		FASTCALL SetDefaultName();

		INT 	   	FASTCALL GetWidth(void);
		INT 		FASTCALL GetHeight(void);
		INT			FASTCALL GetLineSize(void);
		INT 		FASTCALL GetDepth(void);
		LONG		FASTCALL GetSize(void);
		DOUBLE		FASTCALL GetAspectRatio(void);
		DOUBLE 		FASTCALL SetAspectRatio(DOUBLE );

		GRectL		FASTCALL SetClipWindow(INT ,INT ,INT ,INT );
		GRectL		FASTCALL SetClipWindow(GRectL& );
		GRectL		FASTCALL GetClipWindow(void);
	   };

template<class T> class GImageBuffer : public GImage {
		protected:
		T*			ImageBuffer;

		public:
		GImageBuffer(CHAR *name=NULL,DWORD typ=IMAGE_TYPE);
		~GImageBuffer();

        CHAR*		FASTCALL SetDefaultName();

		BOOL   	  	FASTCALL Free(void);
		BOOL       	FASTCALL Alloc(INT ,INT );
		BOOL       	FASTCALL Resize(INT ,INT );

		void		FASTCALL PutPixel(INT ,INT ,T );
		T			FASTCALL GetPixel(INT ,INT );
        void		FASTCALL PutFastPixel(INT ,INT ,T );
        T			FASTCALL GetFastPixel(INT ,INT );

		BOOL		FASTCALL ClearImage(T col);
		T* 	   		FASTCALL GetImage(void);
		T*	  		FASTCALL GetImage(T *,INT start=0,INT count=0);
		T*			FASTCALL SetImage(T *,INT start=0,INT count=0);
	   };

class GRasterImage : public GImage {
		friend class GBitmap;
		protected:
		GPixel*		ImageBuffer;

		public:
		GRasterImage(CHAR *name=NULL,DWORD typ=IMAGE_TYPE);
		~GRasterImage();

        CHAR*		FASTCALL SetDefaultName();

		BOOL   	  	FASTCALL Free(void);
		BOOL       	FASTCALL Alloc(INT ,INT ,INT );
		BOOL       	FASTCALL Resize(INT ,INT );
        BOOL		FASTCALL Clear(void);

		void		FASTCALL PutPixel(INT ,INT ,DWORD );
		DWORD		FASTCALL GetPixel(INT ,INT );
		void		FASTCALL PutFastPixel(INT ,INT ,DWORD );
		DWORD		FASTCALL GetFastPixel(INT ,INT );
        void		FASTCALL PutSpan(INT ,INT ,INT ,GPixel* );
        void		FASTCALL GetSpan(INT ,INT ,INT ,GPixel* );
        void		FASTCALL PutSpan(INT ,INT ,INT ,DWORD );

        BOOL		FASTCALL PutBitmap(GBitmap *,INT x,INT y);
        BOOL		FASTCALL StretchBitmap(GBitmap *,INT x0,INT y0,INT x1,INT y1);

		BOOL		FASTCALL ClearImage(DWORD col=0);
		GPixel*    	FASTCALL GetImage(void);
		GPixel*	  	FASTCALL GetImage(GPixel *,INT start=0,INT count=0);
		GPixel*		FASTCALL PutImage(GPixel *,INT start=0,INT count=0);
		CHAR*		FASTCALL GetFlagsName(CHAR *);

		void		FASTCALL PutPixel0(INT ,INT ,DWORD );
		DWORD 		FASTCALL GetPixel0(INT ,INT );
		void		FASTCALL PutPixel8(INT ,INT ,DWORD );
		DWORD 		FASTCALL GetPixel8(INT ,INT );
		void		FASTCALL PutPixel16(INT ,INT ,DWORD );
		DWORD		FASTCALL GetPixel16(INT ,INT );
		void		FASTCALL PutPixel32(INT ,INT ,DWORD );
		DWORD		FASTCALL GetPixel32(INT ,INT );
		void		FASTCALL PutFastPixel0(INT ,INT ,DWORD );
		DWORD 		FASTCALL GetFastPixel0(INT ,INT );
		void		FASTCALL PutFastPixel8(INT ,INT ,DWORD );
		DWORD 		FASTCALL GetFastPixel8(INT ,INT );
		void		FASTCALL PutFastPixel16(INT ,INT ,DWORD );
		DWORD		FASTCALL GetFastPixel16(INT ,INT );
		void		FASTCALL PutFastPixel32(INT ,INT ,DWORD );
		DWORD		FASTCALL GetFastPixel32(INT ,INT );

        protected:
        void		FASTCALL SetPixelFunctions(void);

        private:
        BOOL 		(FASTCALL GRasterImage::*PutBitmapFunc)(GBitmap *,GRectL &,GRectL &);
        DWORD		(FASTCALL GRasterImage::*GetPixelFunc)(INT ,INT );
        void 		(FASTCALL GRasterImage::*PutPixelFunc)(INT ,INT ,DWORD );
        DWORD		(FASTCALL GRasterImage::*GetFastPixelFunc)(INT ,INT );
        void 		(FASTCALL GRasterImage::*PutFastPixelFunc)(INT ,INT ,DWORD );
        void		(FASTCALL GRasterImage::*GetPixSpanFunc)(INT ,INT ,INT ,GPixel* );
        void 		(FASTCALL GRasterImage::*PutPixSpanFunc)(INT ,INT ,INT ,GPixel* );
		void 		(FASTCALL GRasterImage::*PutColSpanFunc)(INT ,INT ,INT ,DWORD );
		BOOL		FASTCALL PutBitmap8(GBitmap *,GRectL &,GRectL &);
		BOOL		FASTCALL PutBitmap16(GBitmap *,GRectL &,GRectL &);
		BOOL		FASTCALL PutBitmap32(GBitmap *,GRectL &,GRectL &);
		BOOL		FASTCALL PutBitmap0(GBitmap *,GRectL &,GRectL &);
		void		FASTCALL PutPixSpan8(INT ,INT ,INT ,GPixel* );
		void		FASTCALL PutPixSpan16(INT ,INT ,INT ,GPixel* );
		void		FASTCALL PutPixSpan32(INT ,INT ,INT ,GPixel* );
		void		FASTCALL PutPixSpan0(INT ,INT ,INT ,GPixel* );
		void		FASTCALL GetPixSpan8(INT ,INT ,INT ,GPixel* );
		void		FASTCALL GetPixSpan16(INT ,INT ,INT ,GPixel* );
		void		FASTCALL GetPixSpan32(INT ,INT ,INT ,GPixel* );
		void		FASTCALL GetPixSpan0(INT ,INT ,INT ,GPixel* );
		void		FASTCALL PutColSpan0(INT ,INT ,INT ,DWORD );
		void		FASTCALL PutColSpan8(INT ,INT ,INT ,DWORD );
		void		FASTCALL PutColSpan16(INT ,INT ,INT ,DWORD );
		void		FASTCALL PutColSpan32(INT ,INT ,INT ,DWORD );
	   };

class GBitmap : public GRasterImage {
		public:
        enum {
	        BITMAP_TYPE = MAKETYPE('B','M','P',' '),
            TEXTUREMAP_TYPE = MAKETYPE('T','M','A','P'),
            LIGHTMAP_TYPE = MAKETYPE('L','M','A','P'),
            BUMPMAP_TYPE = MAKETYPE('B','M','A','P'),
	        MASK_TYPE = MAKETYPE('M','A','S','K')
        };
        enum {
			PALETTE = 0x00000001,
			MASK = 0x00000002
        };

		public:
		#pragma optiona -a1
		typedef struct {
				BYTE 	Magic;
				BYTE 	Version;
				BYTE 	Pack_Type;
				BYTE 	Color_Depth;
				WORD   	X1,Y1,X2,Y2;
				WORD   	X_Resolution;
				WORD   	Y_Resolution;
				BYTE 	EGA_Palette[48];
				BYTE 	Reserved;
				BYTE 	Bitplanes;
				WORD   	Bytes_per_Line;
				WORD   	Palette_Type;
				BYTE 	Dummy[58];
		   } PCX_Header;
		typedef struct {
				CHAR		Name[GMZ_NAME_LENGTH];
				DWORD		Flags;
				DWORD 		Type;
				DWORD		Magic;
				LONG		Width;
				LONG		Height;
				LONG		LineSize;
				LONG		Depth;
			   } FileBitmapHead;
		#pragma option -a.

		protected:
		GPalette*	Palette;
        GMask*		MaskBuffer;

		public:
		GBitmap(CHAR *name=NULL,DWORD typ=BITMAP_TYPE);
		~GBitmap();

        CHAR*		FASTCALL SetDefaultName();

		BOOL   	  	FASTCALL Free(void);
		BOOL       	FASTCALL Alloc(INT ,INT ,INT );
        BOOL		FASTCALL Resize(INT ,INT );
		BOOL       	FASTCALL Rescale(INT ,INT );
		BOOL       	FASTCALL Remap(GPalette *dst,GPalette *src=NULL);
        BOOL		FASTCALL MakeMask(DWORD back,INT mode);
		BOOL       	FASTCALL ImportPCX(char *,GPalette* p=NULL);
		BOOL   	   	FASTCALL ExportPCX(char *,GPalette* p=NULL);
		BOOL       	FASTCALL ImportBMP(char *,GPalette* p=NULL);
		BOOL   	   	FASTCALL ExportBMP(char *,GPalette* p=NULL);

        BOOL		FASTCALL AllocMask(void);
        BOOL		FASTCALL FreeMask(void);

        void		FASTCALL SetMask(INT ,INT ,BOOL );
        BOOL		FASTCALL GetMask(INT ,INT );

		GMask*     	FASTCALL GetMask(void);
		GMask*	  	FASTCALL GetMask(GMask *,INT start=0,INT count=0);
		GMask*		FASTCALL SetMask(GMask *,INT start=0,INT count=0);
		CHAR*		FASTCALL GetFlagsName(CHAR *);

		GPalette*   FASTCALL GetPalette(void);
		GPalette*	FASTCALL GetPalette(GPalette *,INT f=0,INT count=256);
		GPalette*	FASTCALL SetPalette(GPalette *,INT f=0,INT count=256);
		GColour		FASTCALL SetColour(INT ,GColour& );
		GColour  	FASTCALL GetColour(INT );
		GPalette*	FASTCALL AllocPalette(void);
		BOOL		FASTCALL FreePalette(void);

		BOOL	  	FASTCALL Write(GFile &,DWORD );
		BOOL      	FASTCALL Read(GFile &);

		#if defined(__FORWINDOWS__)
		HBITMAP		FASTCALL CreateDIBitmap(HDC hdc=NULL);
		HBITMAP		FASTCALL CreateCompatibleBitmap(HDC hdc=NULL);
		#endif
	   };

typedef GList<GBitmap,GMZ_OBJECTLIST_SIZE> GBitmapList;


///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GImage
//
inline GImage::GImage(CHAR* name,DWORD typ) : GObject(name,typ) {
 	Flags=Width=Height=LineSize=Depth=NULL;
 	AspectRatio=1.0;
}

inline GImage::~GImage() {
}

inline CHAR* FASTCALL GImage::SetDefaultName() {
 	strcpy(Name,"image");
    return Name;
}

inline INT FASTCALL GImage::GetWidth(void) {
 	return Width;
}

inline INT FASTCALL GImage::GetLineSize(void) {
 	return LineSize;
}

inline INT FASTCALL GImage::GetHeight(void) {
 	return Height;
}

inline INT FASTCALL GImage::GetDepth(void) {
 	return Depth;
}

inline LONG FASTCALL GImage::GetSize(void) {
 	return LineSize*Height;
}

inline DOUBLE FASTCALL GImage::SetAspectRatio(DOUBLE r) {
 	return AspectRatio=r;
}

inline DOUBLE FASTCALL GImage::GetAspectRatio(void) {
 	return AspectRatio;
}

inline GRectL FASTCALL GImage::SetClipWindow(INT l,INT t,INT r,INT b) {
 	ClipWindow.Left=min(max(l,0),Width-1);
 	ClipWindow.Top=min(max(t,0),Height-1);
 	ClipWindow.Right=min(max(r,l),Width-1);
 	ClipWindow.Bottom=min(max(b,t),Height-1);
 	return ClipWindow;
}

inline GRectL FASTCALL GImage::SetClipWindow(GRectL &r) {
 	ClipWindow.Left=min(max(r.Left,0),Width-1);
 	ClipWindow.Top=min(max(r.Top,0),Height-1);
 	ClipWindow.Right=min(max(r.Right,r.Left),Width-1);
 	ClipWindow.Bottom=min(max(r.Bottom,r.Top),Height-1);
 	return ClipWindow;
}

inline GRectL FASTCALL GImage::GetClipWindow(void) {
 	return ClipWindow;
}


///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GImageBuffer
//
template<class T> inline GImageBuffer<T>::GImageBuffer(CHAR* name,DWORD typ) : GImage(name,typ) {
 	(PTR)ImageBuffer=NULL;
}

template<class T> inline GImageBuffer<T>::~GImageBuffer(void) {
 	Free();
}

template<class T> inline CHAR* FASTCALL GImageBuffer<T>::SetDefaultName() {
 	strcpy(Name,"image buffer");
    return Name;
}

template<class T> inline BOOL GImageBuffer<T>::Alloc(INT w,INT h) {
 	Width=w;
 	Height=h;
 	Depth=sizeof(T)*8;
 	LineSize=Width*sizeof(T);
 	if((ImageBuffer=(T *)malloc(w*h*sizeof(T)))==NULL) return FALSE;
 	return TRUE;
}

template<class T> inline BOOL GImageBuffer<T>::Free(void) {
 	if (ImageBuffer)
 		free(ImageBuffer);
 	ImageBuffer=NULL;
 	Width=Height=Depth=0;
 	return TRUE;
}

template<class T> inline BOOL GImageBuffer<T>::Resize(INT w,INT h) {
 	Free();
 	return Alloc(w,h);
}

template<class T> inline T* FASTCALL GImageBuffer<T>::GetImage(void) {
 	return ImageBuffer;
}

template<class T> inline T* FASTCALL GImageBuffer<T>::GetImage(T* dst,INT start,INT count) {
 	if (count==0) count=Height-start;
 	return (T *)memcpy(dst,&((BYTE *)ImageBuffer)[start*LineSize],count*Width*sizeof(T));
}

template<class T> inline T* FASTCALL GImageBuffer<T>::SetImage(T* src,INT start,INT count) {
 	if (count==0) count=Height-start;
 	return (T *)memcpy(&((BYTE *)ImageBuffer)[start*LineSize],src,count*Width*sizeof(T));
}

template<class T> inline void FASTCALL GImageBuffer<T>::PutPixel(INT x,INT y,T col) {
 	if ((x<ClipWindow.Left) || (y<ClipWindow.Top) ||
		(x>ClipWindow.Right) || (y>ClipWindow.Bottom)) return;
 	ImageBuffer[y*Width+x]=T;
}

template<class T> inline T FASTCALL GImageBuffer<T>::GetPixel(INT x,INT y) {
 	if (((UINT)x>=(UINT)Width) || ((UINT)y>=(UINT)Height)) return 0;
 	return ImageBuffer[y*Width+x];
}

template<class T> inline void FASTCALL GImageBuffer<T>::PutFastPixel(INT x,INT y,T col) {
 	ImageBuffer[y*Width+x]=T;
}

template<class T> inline T FASTCALL GImageBuffer<T>::GetFastPixel(INT x,INT y) {
 	return ImageBuffer[y*Width+x];
}


///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GRasterImage
//
inline GRasterImage::GRasterImage(CHAR* name,DWORD typ) : GImage(name,typ) {
 	(PTR)ImageBuffer=NULL;
 	GetPixelFunc=&GRasterImage::GetPixel0;
 	PutPixelFunc=&GRasterImage::PutPixel0;
 	GetFastPixelFunc=&GRasterImage::GetFastPixel0;
 	PutFastPixelFunc=&GRasterImage::PutFastPixel0;
 	PutBitmapFunc=&GRasterImage::PutBitmap0;
}

inline GRasterImage::~GRasterImage(void) {
 	Free();
}

inline CHAR* FASTCALL GRasterImage::SetDefaultName() {
 	strcpy(Name,"pixel buffer");
    return Name;
}

inline BOOL FASTCALL GRasterImage::Resize(INT w,INT h) {
 	INT d=Depth;
 	Free();
 	return Alloc(w,h,d);
}

inline BOOL FASTCALL GRasterImage::Clear(void) {
 	if (!ImageBuffer) return FALSE;
 	memset(ImageBuffer,0,LineSize*Height);
 	return TRUE;
}

inline GPixel* FASTCALL GRasterImage::GetImage(void) {
 	return ImageBuffer;
}

inline GPixel* FASTCALL GRasterImage::GetImage(GPixel* dst,INT start,INT count) {
 	if (count==0) count=Height-start;
 	return (GPixel *)memcpy(dst,&((BYTE *)ImageBuffer)[start*LineSize],count*LineSize);
}

inline GPixel* FASTCALL GRasterImage::PutImage(GPixel* src,INT start,INT count) {
 	if (count==0) count=Height-start;
 	return (GPixel *)memcpy(&((BYTE *)ImageBuffer)[start*LineSize],src,count*LineSize);
}

inline void FASTCALL GRasterImage::PutPixel(INT x,INT y,DWORD col) {
 	(this->*PutPixelFunc)(x,y,col);
}

inline DWORD FASTCALL GRasterImage::GetPixel(INT x,INT y) {
 	return (this->*GetPixelFunc)(x,y);
}

inline void FASTCALL GRasterImage::PutFastPixel(INT x,INT y,DWORD col) {
 	(this->*PutFastPixelFunc)(x,y,col);
}

inline DWORD FASTCALL GRasterImage::GetFastPixel(INT x,INT y) {
 	return (this->*GetFastPixelFunc)(x,y);
}

inline void FASTCALL GRasterImage::PutSpan(INT x,INT y,INT l,GPixel *p) {
 	if ((y<ClipWindow.Top) || (y>ClipWindow.Bottom) || (x>ClipWindow.Right)) return;
 	if (x<ClipWindow.Left) {l-=ClipWindow.Left-x; p+=ClipWindow.Left-x; x=ClipWindow.Left;}
 	if ((x+l)>ClipWindow.Right) l=ClipWindow.Right-x;
 	if (l>0) (this->*PutPixSpanFunc)(x,y,l,p);
}

inline void FASTCALL GRasterImage::GetSpan(INT x,INT y,INT l,GPixel *p) {
 	(this->*GetPixSpanFunc)(x,y,l,p);
}

inline void FASTCALL GRasterImage::PutSpan(INT x,INT y,INT l,DWORD c) {
 	if ((y<ClipWindow.Top) || (y>ClipWindow.Bottom) || (x>ClipWindow.Right)) return;
 	if (x<ClipWindow.Left) {l-=ClipWindow.Left-x; x=ClipWindow.Left;}
 	if ((x+l)>ClipWindow.Right) l=ClipWindow.Right-x;
 	if (l>0) (this->*PutColSpanFunc)(x,y,l,c);
}


///////////////////////////////////////////////////////////////////////////////
//
// Custom GetPixel / PutPixel / GetSpan / PutSpan routines
//
inline void FASTCALL GRasterImage::PutPixel0(INT x,INT y,DWORD col) {
}

inline DWORD FASTCALL GRasterImage::GetPixel0(INT x,INT y) {
 return 0;
}

inline void FASTCALL GRasterImage::PutPixel8(INT x,INT y,DWORD col) {
 	if ((x<ClipWindow.Left) || (y<ClipWindow.Top) ||
		(x>ClipWindow.Right) || (y>ClipWindow.Bottom)) return;
	((BYTE *)ImageBuffer)[y*LineSize+x]=(BYTE)col;
}

inline DWORD FASTCALL GRasterImage::GetPixel8(INT x,INT y) {
 	if (((UINT)x>=(UINT)Width) || ((UINT)y>=(UINT)Height)) return 0;
 	return ((BYTE *)ImageBuffer)[y*LineSize+x];
}

inline void FASTCALL GRasterImage::PutPixel16(INT x,INT y,DWORD col) {
 	if ((x<ClipWindow.Left) || (y<ClipWindow.Top) ||
		(x>ClipWindow.Right) || (y>ClipWindow.Bottom)) return;
 	((WORD *)ImageBuffer)[y*Width+x]=(WORD)col;
}

inline DWORD FASTCALL GRasterImage::GetPixel16(INT x,INT y) {
 	if (((UINT)x>=(UINT)Width) || ((UINT)y>=(UINT)Height)) return 0;
 	return ((WORD *)ImageBuffer)[y*Width+x];
}

inline void FASTCALL GRasterImage::PutPixel32(INT x,INT y,DWORD col) {
 	if ((x<ClipWindow.Left) || (y<ClipWindow.Top) ||
		(x>ClipWindow.Right) || (y>ClipWindow.Bottom)) return;
 	((DWORD *)ImageBuffer)[y*Width+x]=col;
}

inline DWORD FASTCALL GRasterImage::GetPixel32(INT x,INT y) {
 	if (((UINT)x>=(UINT)Width) || ((UINT)y>=(UINT)Height)) return 0;
 	return ((DWORD *)ImageBuffer)[y*Width+x];
}

inline void FASTCALL GRasterImage::PutFastPixel0(INT x,INT y,DWORD col) {
}

inline DWORD FASTCALL GRasterImage::GetFastPixel0(INT x,INT y) {
 	return 0;
}

inline void FASTCALL GRasterImage::PutFastPixel8(INT x,INT y,DWORD col) {
  	((BYTE *)ImageBuffer)[y*LineSize+x]=(BYTE)col;
}

inline DWORD FASTCALL GRasterImage::GetFastPixel8(INT x,INT y) {
 	return ((BYTE *)ImageBuffer)[y*LineSize+x];
}

inline void FASTCALL GRasterImage::PutFastPixel16(INT x,INT y,DWORD col) {
 	((WORD *)ImageBuffer)[y*Width+x]=(WORD)col;
}

inline DWORD FASTCALL GRasterImage::GetFastPixel16(INT x,INT y) {
 	return ((WORD *)ImageBuffer)[y*Width+x];
}

inline void FASTCALL GRasterImage::PutFastPixel32(INT x,INT y,DWORD col) {
 	((DWORD *)ImageBuffer)[y*Width+x]=col;
}

inline DWORD FASTCALL GRasterImage::GetFastPixel32(INT x,INT y) {
 	return ((DWORD *)ImageBuffer)[y*Width+x];
}

#pragma argsused
inline void FASTCALL GRasterImage::PutPixSpan0(INT x,INT y,INT l,GPixel *p) {
}

inline void FASTCALL GRasterImage::PutPixSpan8(INT x,INT y,INT l,GPixel *p) {
 	memcpy(&((BYTE *)ImageBuffer)[y*Width+x],p,l);
}

inline void FASTCALL GRasterImage::PutPixSpan16(INT x,INT y,INT l,GPixel *p) {
 	memcpy(&((WORD *)ImageBuffer)[y*Width+x],p,l<<1);
}

inline void FASTCALL GRasterImage::PutPixSpan32(INT x,INT y,INT l,GPixel *p) {
 	memcpy(&((DWORD *)ImageBuffer)[y*Width+x],p,l<<2);
}

#pragma argsused
inline void FASTCALL GRasterImage::PutColSpan0(INT x,INT y,INT l,DWORD c) {
}

inline void FASTCALL GRasterImage::PutColSpan8(INT x,INT y,INT l,DWORD c) {
 	memset(&((BYTE *)ImageBuffer)[y*Width+x],c,l);
}

inline void FASTCALL GRasterImage::PutColSpan16(INT x,INT y,INT l,DWORD c) {
 	for (INT i=x;i<l+x;i++)
    	((WORD *)ImageBuffer)[y*Width+l]=(WORD)c;
}

inline void FASTCALL GRasterImage::PutColSpan32(INT x,INT y,INT l,DWORD c) {
 	for (INT i=x;i<l+x;i++) ((DWORD *)ImageBuffer)[y*Width+l]=(DWORD)c;
}

#pragma argsused
inline void FASTCALL GRasterImage::GetPixSpan0(INT x,INT y,INT l,GPixel *p) {
}

inline void FASTCALL GRasterImage::GetPixSpan8(INT x,INT y,INT l,GPixel *p) {
 	memcpy(p,&((BYTE *)ImageBuffer)[y*Width+x],l);
}

inline void FASTCALL GRasterImage::GetPixSpan16(INT x,INT y,INT l,GPixel *p) {
 	memcpy(p,&((WORD *)ImageBuffer)[y*Width+x],l<<1);
}

inline void FASTCALL GRasterImage::GetPixSpan32(INT x,INT y,INT l,GPixel *p) {
 	memcpy(p,&((DWORD *)ImageBuffer)[y*Width+x],l<<2);
}


///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GBitmap
//
inline GBitmap::GBitmap(CHAR* name,DWORD typ) : GRasterImage(name,typ) {
 	(PTR)MaskBuffer=
 	(PTR)Palette=NULL;
}

inline GBitmap::~GBitmap(void) {
 	Free();
}

inline CHAR* FASTCALL GBitmap::SetDefaultName() {
 	strcpy(Name,"bitmap");
    return Name;
}

inline GPalette* FASTCALL GBitmap::SetPalette(GPalette *pal,INT first,INT count) {
 	if (Flags & PALETTE)
		Palette->SetPalette(pal,first,count);
 	else
		Palette=pal;
	return Palette;
}

inline GPalette* FASTCALL GBitmap::GetPalette(GPalette *pal,INT first,INT count) {
 	pal->SetPalette(Palette,first,count);
 	return pal;
}

inline GPalette* FASTCALL GBitmap::GetPalette(void) {
 	return Palette;
}

inline GColour FASTCALL GBitmap::SetColour(INT i,GColour &col) {
 	if(!Palette) return GColour(0,0,0);
 	return Palette->SetColour(i,col);
}

inline BOOL FASTCALL GBitmap::AllocMask(void) {
 	if (!FreeMask()) return FALSE;
 	if ((MaskBuffer=(GMask *)malloc(((Width+7)>>3)*Height))==NULL) return FALSE;
	Flags|=MASK;
 	return TRUE;
}

inline BOOL FASTCALL GBitmap::FreeMask(void) {
 	if (!(Flags & MASK) || (!MaskBuffer)) return TRUE;
 	free(MaskBuffer);
 	MaskBuffer=NULL;
 	Flags&=~MASK;
 	return TRUE;
}

inline void FASTCALL GBitmap::SetMask(INT x,INT y,BOOL v) {
 	if (!MaskBuffer) return;
 	((BYTE *)MaskBuffer)[y*((Depth+7)>>3)+(x>>3)]=
  	v ? ((BYTE *)MaskBuffer)[y*((Depth+7)>>3)+(x>>3)] | (BYTE)(1<<(x&7))
      : ((BYTE *)MaskBuffer)[y*((Depth+7)>>3)+(x>>3)] &  (BYTE)~(1<<(x&7));
}

inline BOOL FASTCALL GBitmap::GetMask(INT x,INT y) {
 	if (!MaskBuffer) return 0;
 	return (BOOL)(((BYTE *)MaskBuffer)[y*((Depth+7)>>3)+(x>>3)] & (1<<(x&7)));
}

inline GMask* FASTCALL GBitmap::GetMask(void) {
 	return MaskBuffer;
}

inline GMask* FASTCALL GBitmap::GetMask(GMask* dst,INT start,INT count) {
 	if (count==0) count=Height-start;
 	return (GMask *)memcpy(dst,&((BYTE *)MaskBuffer)[start*((Width+7)>>3)],count*((Width+7)>>3));
}

inline GMask* FASTCALL GBitmap::SetMask(GMask* src,INT start,INT count) {
 	if (count==0) count=Height-start;
 	return (GMask *)memcpy(&((BYTE *)MaskBuffer)[start*((Width+7)>>3)],src,count*((Width+7)>>3));
}

inline GPalette* FASTCALL GBitmap::AllocPalette(void) {
 	if ((Flags & PALETTE) && (Palette)) return Palette;
 	if ((Palette=new GPalette)==NULL) return NULL;
 	Flags|=PALETTE;
 	return Palette;
}

inline BOOL FASTCALL GBitmap::FreePalette(void) {
 	if ((Flags & PALETTE) && (Palette)) delete Palette;
 	Palette=NULL;
 	Flags&=~PALETTE;
 	return TRUE;
}
#endif
