///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#ifndef __GFILE_H
#define __GFILE_H

#include "../include/gomez.h"

#define GMZ_MAX_CHUNKS				40


#define GMZ_MODE_READ				1
#define GMZ_MODE_WRITE				2

#define LZW_LARGEST_CODE			4095
#define LZW_STACK_SIZE				4096
#define LZW_TABLE_SIZE				5003
#define	LZW_MIN_CODE_SIZE			8
#define LZW_MAX_CODE_SIZE			12


class LZWCodec {
		private:
		static BYTE MaskTable[8];
		static BYTE BitTable[8];
		static UINT WordMaskTable[16];

		UINT 		PrefixTable[LZW_TABLE_SIZE];
		UINT 		CurrentTable[LZW_TABLE_SIZE];
		BYTE 		SuffixTable[LZW_TABLE_SIZE];
		BYTE		CodeStack[LZW_STACK_SIZE];

		UINT 		MaxCode,FreeCode,ClearCode,EofCode;
		UINT 		PrefixCode,SuffixChar;
		UINT 		CodeSize;
		UINT 		BitOffset;
		UINT 		BitsLeft;
		UINT 		TopSlot;
		UINT 		Slot;

		void		*Destination;
		void		*Source;
		BYTE		*WriteBuffer;
		BYTE		*ReadBuffer;

		public:
		INT			FASTCALL Decompress(void *src,void *&dst,INT);
		INT			FASTCALL Compress(void *src,void *&dst,INT);

		private:
		void		FASTCALL InitReadTable(void);
		void		FASTCALL InitWriteTable(void);
		void		FASTCALL WriteCode(int);
		int			FASTCALL ReadCode(void);
	   };

class GFile {
		public:
        enum {
            END_TYPE = MAKETYPE('E','N','D',' '),
            SKIP_TYPE = MAKETYPE('S','K','I','P'),
            MAIN_TYPE = MAKETYPE('M','A','I','N'),
            SONS_TYPE = MAKETYPE('S','O','N','S'),
            ENTRY_TYPE = MAKETYPE('E','N','T','R'),
            HEAD_TYPE = MAKETYPE('H','E','A','D'),
            DATA_TYPE = MAKETYPE('D','A','T','A'),
            KNOT_TYPE = MAKETYPE('K','N','O','T')
        };
		enum {
		    COMPRESSION_STANDARD =	0,
		    COMPRESSION_NONE = MAKETYPE('N','O','N','E'),
		    COMPRESSION_LZW = MAKETYPE('L','Z','W','8'),
		    COMPRESSION_SUBS = MAKETYPE('S','U','B','S')
		};
        
		protected:
		typedef struct {
				BYTE	Type[8];
				DWORD	Version;
			   } FileHeader;
		typedef struct {
				DWORD 	Type;
				DWORD	Size;
				DWORD 	Compression;
			   } ChunkHeader;

		 protected:
		 FILE*		Handle;
		 LONG 		ChunkStart[GMZ_MAX_CHUNKS];
		 LONG 		ChunkSize[GMZ_MAX_CHUNKS];
		 LONG		ChunkPosition[GMZ_MAX_CHUNKS];
		 DWORD		ChunkCompression[GMZ_MAX_CHUNKS];
		 DWORD		ChunkType[GMZ_MAX_CHUNKS];
		 INT  		ChunkLevel;
		 INT 		Mode;
		 DWORD 		StandardCompression;


		 public:
		 BYTE 		FileType[8];
		 DWORD 		FileVersion;


		 public:
		 GFile(CHAR *typ=NULL,DWORD V=0x10000L);
		 ~GFile();

		 BOOL		FASTCALL Open(char *);
		 DWORD		FASTCALL OpenChunk(void);
		 DWORD		FASTCALL OpenMetaChunk(void);
		 INT 		FASTCALL Read(void *,INT);
		 INT 		FASTCALL ReadChunk(void **);

		 BOOL		FASTCALL Create(char *,DWORD compression=COMPRESSION_NONE);
		 INT 		FASTCALL Write(void *,INT);
		 INT 		FASTCALL WriteChunk(void *,INT);
		 BOOL		FASTCALL CreateChunk(DWORD ,DWORD compression=COMPRESSION_STANDARD);
		 BOOL		FASTCALL OpenChunk(DWORD ,DWORD compression=COMPRESSION_STANDARD);
		 BOOL		FASTCALL CreateMetaChunk(DWORD );
		 BOOL		FASTCALL OpenMetaChunk(DWORD );

		 BOOL		FASTCALL CloseChunk(void);
		 BOOL		FASTCALL CloseMetaChunk(void);
		 BOOL		FASTCALL Close();

		 INT		FASTCALL GetChunkLevel(void);
		 DWORD		FASTCALL GetChunkType(void);
		 DWORD		FASTCALL GetChunkType(INT );
		 DWORD		FASTCALL GetChunkSize(void);
		 DWORD		FASTCALL GetChunkSize(INT );
		 DWORD		FASTCALL GetChunkCompression(void);
		 DWORD		FASTCALL GetChunkCompression(INT );

		 CHAR*		FASTCALL GetFileType(void);
		 CHAR*		FASTCALL GetFileType(CHAR *);
		 DWORD		FASTCALL GetFileVersion(void);
		};


inline GFile::GFile(CHAR *typ,DWORD v) {
 	FileVersion=v;
 	Handle=NULL;
 	if (typ==NULL)
		memset(FileType,0,sizeof(FileType));
 	else
		memcpy(FileType,typ,sizeof(FileType));
}

inline GFile::~GFile() {
 	if (Handle) Close();
}

inline DWORD FASTCALL GFile::GetChunkType(void) {
 	return ChunkType[ChunkLevel];
}

inline DWORD FASTCALL GFile::GetChunkType(INT i) {
 	if ((i<=ChunkLevel) && (i>=0))
		return ChunkType[i];
 	else
		return NULL;
}

inline DWORD FASTCALL GFile::GetChunkSize(void) {
 	return ChunkSize[ChunkLevel];
}

inline DWORD FASTCALL GFile::GetChunkSize(INT i) {
 	if ((i<=ChunkLevel) && (i>=0))
		return ChunkSize[i];
 	else
		return NULL;
}

inline DWORD FASTCALL GFile::GetChunkCompression(void) {
 	return ChunkCompression[ChunkLevel];
}

inline DWORD FASTCALL GFile::GetChunkCompression(INT i) {
 	if ((i<=ChunkLevel) && (i>=0))
		return ChunkCompression[i];
 	else
		return NULL;
}

inline INT FASTCALL GFile::GetChunkLevel(void) {
 	return ChunkLevel;
}

inline CHAR* FASTCALL GFile::GetFileType(void) {
 	return (CHAR *)FileType;
}

inline CHAR* FASTCALL GFile::GetFileType(CHAR *dst) {
 	return (CHAR *)memcpy(dst,FileType,sizeof(FileType));
}

inline DWORD FASTCALL GFile::GetFileVersion(void) {
 	return FileVersion;
}

inline BOOL FASTCALL GFile::CreateMetaChunk(DWORD t) {
 	return CreateChunk(t,COMPRESSION_SUBS);
}

inline BOOL FASTCALL GFile::OpenChunk(DWORD t,DWORD c) {
 	return CreateChunk(t,c);
}

inline DWORD FASTCALL GFile::OpenMetaChunk(void) {
 	DWORD t=OpenChunk();
 	if (t==NULL) return NULL;
 	if (ChunkCompression[ChunkLevel]!=COMPRESSION_SUBS) return NULL;
 	return t;
}

inline BOOL FASTCALL GFile::OpenMetaChunk(DWORD t) {
 	return CreateChunk(t,COMPRESSION_SUBS);
}

inline BOOL FASTCALL GFile::CloseMetaChunk(void) {
 	return CloseChunk();
}

#endif