///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#ifndef __GMATH_H
#define __GMATH_H

#include "../include/gomez.h"

/* Constants rounded for 21 decimals. */
#ifndef M_E
#define M_E         ((DOUBLE)2.71828182845904523536)
#define M_LOG2E     ((DOUBLE)1.44269504088896340736)
#define M_LOG10E    ((DOUBLE)0.434294481903251827651)
#define M_LN2       ((DOUBLE)0.693147180559945309417)
#define M_LN10      ((DOUBLE)2.30258509299404568402)
#define M_PI        ((DOUBLE)3.14159265358979323846)
#define M_PI_2      ((DOUBLE)1.57079632679489661923)
#define M_PI_4      ((DOUBLE)0.785398163397448309616)
#define M_1_PI      ((DOUBLE)0.318309886183790671538)
#define M_2_PI      ((DOUBLE)0.636619772367581343076)
#define M_1_SQRTPI  ((DOUBLE)0.564189583547756286948)
#define M_2_SQRTPI  ((DOUBLE)1.12837916709551257390)
#define M_SQRT2     ((DOUBLE)1.41421356237309504880)
#define M_SQRT_2    ((DOUBLE)0.707106781186547524401)
#endif


template<int P> class GFixed {
		 public:
		 LONG value;

		 public:
		 GFixed() {value=0};
		 GFixed(LONG &x) {value=x << P; };
		 GFixed(DOUBLE &x) {value=(LONG)(x*(DOUBLE)(1<<P));};
		 GFixed(GFixed<P> &x) {value=x.value;};

		 GFixed<P> operator+(GFixed<P> &x) {GFixed<P> c; c.value=value+x.value; return c;};
		 GFixed<P> operator-(GFixed<P> &x) {GFixed<P> c; c.value=value-x.value; return c;};
		 GFixed<P> operator*(GFixed<P> &x) {GFixed<P> c; c.value=(value*x.value)>>P; return c;};
		 GFixed<P> operator/(GFixed<P> &x) {GFixed<P> c; c.value=(value<<P) / x.value; return c;};
		 GFixed<P> operator<<(LONG &x) {GFixed<P> c; c.value=value<<x; return c;};
		 GFixed<P> operator>>(LONG &x) {GFixed<P> c; c.value=value>>x; return c;};
		 GFixed<P> &operator+=(GFixed<P> &x) {value+=x.value; return *this;};
		 GFixed<P> &operator-=(GFixed<P> &x) {value-=x.value; return *this;};
		 GFixed<P> &operator*=(GFixed<P> &x) {value=(value*x.value)>>P; return *this;};
		 GFixed<P> &operator/=(GFixed<P> &x) {value=value<<P / x.value; return *this;};
		 GFixed<P> &operator<<=(LONG &x) {value<<=x; return *this;};
		 GFixed<P> &operator>>=(LONG &x) {value>>=x; return *this;};
		 GFixed<P> operator-() {GFixed<P> c; c.value=-value; return c;};
		 GFixed<P> operator++() {value+=1 << P; return *this;};
		 GFixed<P> operator--() {value-=1 << P; return *this;};

		 BOOL operator<(GFixed<P> &x) {return (value<x.value);};
		 BOOL operator>(GFixed<P> &x) {return (value>x.value);};
		 BOOL operator<=(GFixed<P> &x) {return (value<=x.value);};
		 BOOL operator>=(GFixed<P> &x) {return (value>=x.value);};
		 BOOL operator==(GFixed<P> &x) {return (value==x.value);};
		 GFixed<P> &operator=(LONG y) {value = y << P; return *this;};
		 GFixed<P> &operator=(DOUBLE y) {value = (LONG)(y*(DOUBLE)(1<<P)); return *this;};
		};

template<class T> class GPoint {
		public:
		T			x;
		T			y;

		public:
		GPoint() {};
		GPoint(T a,T b) {x=a; y=b;};
		GPoint<T>		FASTCALL operator+(GPoint<T> &m) {return GPoint<T>((T)(x+m.x),(T)(y+m.y));};
		GPoint<T>		FASTCALL operator-(GPoint<T> &m) {return GPoint<T>((T)(x-m.x),(T)(y-m.y));};
		GPoint<T>& 		FASTCALL operator+=(GPoint<T> &m) {x+=m.x; y+=m.y; return *this;};
		GPoint<T>& 		FASTCALL operator-=(GPoint<T> &m) {x-=m.x; y-=m.y; return *this;};
		BOOL 			FASTCALL operator==(GPoint<T> &m) {return ((x==m.x) && (y==m.y));};
		BOOL 			FASTCALL operator!=(GPoint<T> &m) {return ((x!=m.x) || (y!=m.y));};

		T				FASTCALL SetX(T a) {return x=a;};
		T				FASTCALL SetY(T a) {return y=a;};
		T				FASTCALL GetX(void) {return x;};
		T				FASTCALL GetY(void) {return y;};
	   };

template<class T> class GRect {
		public:
		T      		Left;
		T   		Top;
		T   		Right;
		T   		Bottom;

		public:
		GRect();
		GRect(T ,T ,T ,T );
        BOOL		FASTCALL IsInside(GPoint<T> &);
        BOOL		FASTCALL IsInside(T ,T );
        BOOL		FASTCALL IsInside(GRect<T> &);
        BOOL		FASTCALL IsEmpty(void);
        T			FASTCALL GetWidth(void);
        T			FASTCALL GetHeight(void);
        GRect<T>	FASTCALL operator+(GPoint<T> &);
        GRect<T>	FASTCALL operator-(GPoint<T> &);
        GRect<T>	FASTCALL operator&(GRect<T> &);
        GRect<T>	FASTCALL operator|(GRect<T> &);
        GRect<T>&	FASTCALL operator&=(GRect<T> &);
        GRect<T>&	FASTCALL operator|=(GRect<T> &);
        GRect<T>&	FASTCALL operator+=(GPoint<T> &);
        GRect<T>&	FASTCALL operator-=(GPoint<T> &);
        BOOL		FASTCALL operator==(GRect<T> &);
        BOOL		FASTCALL operator!=(GRect<T> &);
	   };

template<class T> class GDimension {
		public:
		T      		Width;
		T   		Height;

		public:
		GDimension();
		GDimension(T ,T );
        BOOL		FASTCALL IsEmpty(void);
        T			FASTCALL GetWidth(void);
        T			FASTCALL GetHeight(void);
        T			FASTCALL SetWidth(T );
        T			FASTCALL SetHeight(T );
        BOOL		FASTCALL operator==(GDimension<T> &);
        BOOL		FASTCALL operator!=(GDimension<T> &);
	   };

template<class T> class GTriple {
		public:
		T			 	x,y,z;

		public:
		GTriple();
		GTriple(T ,T ,T );

		T     			FASTCALL GetX(void);
		T     			FASTCALL GetY(void);
		T     			FASTCALL GetZ(void);
		T     			FASTCALL SetX(T );
		T     			FASTCALL SetY(T );
		T     			FASTCALL SetZ(T );

		GTriple<T> 		FASTCALL operator+(GTriple<T> &);
		GTriple<T> 		FASTCALL operator-(GTriple<T> &);
		GTriple<T> 		FASTCALL operator%(GTriple<T> &);
		GTriple<T>   	FASTCALL operator*(LONG );
		GTriple<T>   	FASTCALL operator*(DOUBLE );
		GTriple<T>   	FASTCALL operator/(LONG );
		GTriple<T>   	FASTCALL operator/(DOUBLE );
		GTriple<T> 		FASTCALL operator~();
		T      			FASTCALL operator*(GTriple<T> &);
		GTriple<T>& 	FASTCALL operator+=(GTriple<T> &);
		GTriple<T>& 	FASTCALL operator-=(GTriple<T> &);
		GTriple<T>& 	FASTCALL operator*=(LONG );
		GTriple<T>& 	FASTCALL operator*=(DOUBLE );
		GTriple<T>& 	FASTCALL operator/=(LONG );
		GTriple<T>& 	FASTCALL operator/=(DOUBLE );
		BOOL			FASTCALL operator==(GTriple<T> &);
		BOOL			FASTCALL operator!=(GTriple<T> &);
	   };

template<int P> class GTripleM {
		public:
		MATH		 	x,y,z;

		public:
		GTripleM();
		GTripleM(LONG ,LONG ,LONG );
		GTripleM(DOUBLE ,DOUBLE ,DOUBLE );

		DOUBLE			FASTCALL GetX(void);
		DOUBLE			FASTCALL GetY(void);
		DOUBLE			FASTCALL GetZ(void);
		DOUBLE			FASTCALL SetX(DOUBLE );
		DOUBLE			FASTCALL SetY(DOUBLE );
		DOUBLE			FASTCALL SetZ(DOUBLE );
		LONG			FASTCALL SetX(LONG );
		LONG			FASTCALL SetY(LONG );
		LONG			FASTCALL SetZ(LONG );

		GTripleM<P> 	FASTCALL operator+(GTripleM<P> &);
		GTripleM<P> 	FASTCALL operator-(GTripleM<P> &);
		GTripleM<P> 	FASTCALL operator%(GTripleM<P> &);
		GTripleM<P>   	FASTCALL operator*(LONG );
		GTripleM<P>   	FASTCALL operator*(DOUBLE );
		GTripleM<P>   	FASTCALL operator/(LONG );
		GTripleM<P>   	FASTCALL operator/(DOUBLE );
		GTripleM<P> 	FASTCALL operator~();
		MATH   			FASTCALL operator*(GTripleM<P> &);
		GTripleM<P>& 	FASTCALL operator+=(GTripleM<P> &);
		GTripleM<P>& 	FASTCALL operator-=(GTripleM<P> &);
		GTripleM<P>& 	FASTCALL operator*=(LONG );
		GTripleM<P>& 	FASTCALL operator*=(DOUBLE );
		GTripleM<P>& 	FASTCALL operator/=(LONG );
		GTripleM<P>& 	FASTCALL operator/=(DOUBLE );
		BOOL			FASTCALL operator==(GTripleM<P> &);
		BOOL			FASTCALL operator!=(GTripleM<P> &);
	   };


typedef GPoint<SHORT>		GPointS;
typedef GPoint<LONG>		GPointL;
typedef GPoint<FLOAT>		GPointF;
typedef GPoint<DOUBLE>		GPointD;
typedef GPoint<MATH>		GPointM;

typedef GRect<SHORT>		GRectS;
typedef GRect<LONG>			GRectL;
typedef GRect<FLOAT>		GRectF;
typedef GRect<DOUBLE>		GRectD;
typedef GRect<MATH>			GRectM;

typedef GDimension<SHORT>	GDimensionS;
typedef GDimension<LONG>	GDimensionL;
typedef GDimension<FLOAT>	GDimensionF;
typedef GDimension<DOUBLE>	GDimensionD;
typedef GDimension<MATH>	GDimensionM;



///////////////////////////////////////////////////////////////////////////////
//
// The GRect implementation
//
template<class T> inline GRect<T>::GRect(void) {
	Top=Left=Right=Bottom=0;
}

template<class T> inline GRect<T>::GRect(T l,T t,T r,T b) {
 	Top=t; Left=l; Right=r; Bottom=b;
}

template<class T> inline BOOL FASTCALL GRect<T>::IsInside(GPoint<T> &p) {
 	return ((p.x>=Left) && (p.x<=Right) && (p.y>=Top) && (p.y<=Bottom));
}

template<class T> inline BOOL FASTCALL GRect<T>::IsInside(T x,T y) {
 	return ((x>=Left) && (x<=Right) && (y>=Top) && (y<=Bottom));
}

template<class T> inline BOOL FASTCALL GRect<T>::IsInside(GRect<T> &r) {
 	return ((r.Left>=Left) && (r.Right<=Right) && (r.Top>=Top) && (r.Bottom<=Bottom));
}

template<class T> inline BOOL FASTCALL GRect<T>::IsEmpty(void) {
 	return ((Right<Left) || (Bottom<Top));
}

template<class T> inline T FASTCALL GRect<T>::GetWidth(void) {
 	return (T)(Right-Left+1);
}

template<class T> inline T FASTCALL GRect<T>::GetHeight(void) {
 	return (T)(Bottom-Top+1);
}

template<class T> inline GRect<T> FASTCALL GRect<T>::operator&(GRect<T> &r) {
 	return GRect<T>((T)max(Left,r.Left),(T)max(Top,r.Top),(T)min(Right,r.Right),(T)min(Bottom,r.Bottom));
}

template<class T> inline GRect<T> FASTCALL GRect<T>::operator|(GRect<T> &r) {
 	return GRect<T>((T)min(Left,r.Left),(T)min(Top,r.Top),(T)max(Right,r.Right),(T)max(Bottom,r.Bottom));
}

template<class T> inline GRect<T> FASTCALL GRect<T>::operator+(GPoint<T> &p) {
 	return GRect<T>((T)(Left+p.x),(T)(Top+p.y),(T)(Right+p.x),(T)(Bottom+p.y));
}

template<class T> inline GRect<T> FASTCALL GRect<T>::operator-(GPoint<T> &p) {
 	return GRect<T>((T)(Left-p.x),(T)(Top-p.y),(T)(Right-p.x),(T)(Bottom-p.y));
}

template<class T> inline GRect<T>& FASTCALL GRect<T>::operator&=(GRect<T> &r) {
 	Left=max(Left,r.Left);
 	Top=max(Top,r.Top);
 	Right=min(Right,r.Right);
 	Bottom=min(Bottom,r.Bottom);
 	return *this;
}

template<class T> inline GRect<T>& FASTCALL GRect<T>::operator|=(GRect<T> &r) {
 	Left=min(Left,r.Left);
 	Top=min(Top,r.Top);
 	Right=max(Right,r.Right);
 	Bottom=max(Bottom,r.Bottom);
 	return *this;
}

template<class T> inline GRect<T>& FASTCALL GRect<T>::operator+=(GPoint<T> &p) {
 	Left+=p.x;
 	Top+=p.y;
 	Right+=p.x;
 	Bottom+=p.y;
	 return *this;
}

template<class T> inline GRect<T>& FASTCALL GRect<T>::operator-=(GPoint<T> &p) {
 	Left-=p.x;
 	Top-=p.y;
	 Right-=p.x;
	 Bottom-=p.y;
	 return *this;
}

template<class T> inline BOOL FASTCALL GRect<T>::operator==(GRect<T> &r) {
 	return ((Left==r.Left) && (Right==r.Right) && (Top==r.Top) && (Bottom==r.Bottom));
}

template<class T> inline BOOL FASTCALL GRect<T>::operator!=(GRect<T> &r) {
 	return ((Left!=r.Left) || (Right!=r.Right)  || (Top!=r.Top) || (Bottom!=r.Bottom));
}

///////////////////////////////////////////////////////////////////////////////
//
// The GDimension implementation
//
template<class T> inline GDimension<T>::GDimension(void) {
	Width=Height=0;
}

template<class T> inline GDimension<T>::GDimension(T w,T h) {
	Width=w; Height=h;
}

template<class T> inline BOOL FASTCALL GDimension<T>::IsEmpty(void) {
 	return ((Width==0) || (Height==0));
}

template<class T> inline T FASTCALL GDimension<T>::GetWidth(void) {
	return Width;
}

template<class T> inline T FASTCALL GDimension<T>::GetHeight(void) {
	return Height;
}

template<class T> inline T FASTCALL GDimension<T>::SetWidth(T w) {
	return Width=w;
}

template<class T> inline T FASTCALL GDimension<T>::SetHeight(T h) {
	return Height=h;
}

template<class T> inline BOOL FASTCALL GDimension<T>::operator==(GDimension<T> &r) {
 	return ((Width==r.Width) && (Height==r.Height));
}

template<class T> inline BOOL FASTCALL GDimension<T>::operator!=(GDimension<T> &r) {
 	return ((Width!=r.Width) || (Height!=r.Height));
}

///////////////////////////////////////////////////////////////////////////////
//
// The GFixed implementation
//
template<int P> ostream& operator<< (ostream& ,GFixed<P>& );
template<int P> istream& operator>> (istream& ,GFixed<P>& );

template<int P> inline ostream& operator<< (ostream& s,GFixed<P>& f) {
 	s<<ldexp((DOUBLE)f.value,-P);
 	return s;
}

template<int P> inline istream& operator>> (istream& s,GFixed<P>& f) {
 	DOUBLE g;
 	s>>g;
	f=GFixed<P>(g);
 	return s;
}

///////////////////////////////////////////////////////////////////////////////
//
// The GTriple implementation
//
template<class T> ostream& operator<< (ostream& ,GTriple<T>& );

template<class T> inline ostream& operator<< (ostream& s,GTriple<T>& v) {
 	s<<"["<<v.x
 	 <<","<<v.y
 	 <<","<<v.z<<",1]";
 	return s;
}

template<class T> inline GTriple<T>::GTriple(void) {
	x=y=z=0;
}

template<class T> inline GTriple<T>::GTriple(T a,T b,T c) {
 	x=a;
 	y=b;
 	z=c;
}

template<class T> inline T FASTCALL GTriple<T>::GetX(void) {
 	return x;
}

template<class T> inline T FASTCALL GTriple<T>::GetY(void) {
 	return y;
}

template<class T> inline T FASTCALL GTriple<T>::GetZ(void) {
 	return z;
}

template<class T> inline T FASTCALL GTriple<T>::SetX(T d) {
 	return x=d;
}

template<class T> inline T FASTCALL GTriple<T>::SetY(T d) {
 	return y=d;
}

template<class T> inline T FASTCALL GTriple<T>::SetZ(T d) {
 	return z=d;
}

template<class T> inline GTriple<T> FASTCALL GTriple<T>::operator+(GTriple<T> &v) {
 	return GTriple<T>(x+v.x,y+v.y,z+v.z);
}

template<class T> inline GTriple<T> FASTCALL GTriple<T>::operator-(GTriple<T> &v) {
 	return GTriple<T>(x-v.x,y-v.y,z-v.z);
}

template<class T> inline GTriple<T> FASTCALL GTriple<T>::operator*(LONG s) {
 	return GTriple<T>(x*s,y*s,z*s);
}

template<class T> inline GTriple<T> FASTCALL GTriple<T>::operator*(DOUBLE s) {
 	return GTriple<T>((MATH)((DOUBLE)x*s),(MATH)((DOUBLE)y*s),(MATH)((DOUBLE)z*s));
}

template<class T> inline GTriple<T> FASTCALL GTriple<T>::operator/(LONG s) {
 	return GTriple<T>(x/s,y/s,z/s);
}

template<class T> inline GTriple<T> FASTCALL GTriple<T>::operator/(DOUBLE s) {
 	return GTriple<T>((MATH)((DOUBLE)x/s),(MATH)((DOUBLE)y/s),(MATH)((DOUBLE)z/s));
}

template<class T> inline T FASTCALL GTriple<T>::operator*(GTriple<T> &v) {
 	return (x*v.x+y*v.y+z*v.z);
}

template<class T> inline GTriple<T>& FASTCALL GTriple<T>::operator+=(GTriple<T> &v) {
 	x+=v.x;
 	y+=v.y;
 	z+=v.z;
	 return *this;
}

template<class T> inline GTriple<T>& FASTCALL GTriple<T>::operator-=(GTriple<T> &v) {
 	x-=v.x;
 	y-=v.y;
 	z-=v.z;
 	return *this;
}

template<class T> inline GTriple<T>& FASTCALL GTriple<T>::operator*=(LONG s) {
 	x*=(MATH)s;
 	y*=(MATH)s;
 	z*=(MATH)s;
 	return *this;
}

template<class T> inline GTriple<T>& FASTCALL GTriple<T>::operator*=(DOUBLE s) {
 	x=(MATH)((DOUBLE)x*s);
 	y=(MATH)((DOUBLE)y*s);
 	z=(MATH)((DOUBLE)z*s);
 	return *this;
}

template<class T> inline GTriple<T>& FASTCALL GTriple<T>::operator/=(LONG s) {
 	x/=(MATH)s;
 	y/=(MATH)s;
 	z/=(MATH)s;
 	return *this;
}

template<class T> inline GTriple<T>& FASTCALL GTriple<T>::operator/=(DOUBLE s) {
 	x=(MATH)((DOUBLE)x/s);
 	y=(MATH)((DOUBLE)y/s);
 	z=(MATH)((DOUBLE)z/s);
 	return *this;
}

template<class T> inline BOOL	FASTCALL GTriple<T>::operator==(GTriple<T> &v) {
 	return ((x==v.x) && (y==v.y) && (z==v.z));
}

template<class T> inline BOOL	FASTCALL GTriple<T>::operator!=(GTriple<T> &v) {
 	return ((x!=v.x) || (y!=v.y) || (z!=v.z));
}

template<class T> inline GTriple<T> FASTCALL GTriple<T>::operator %(GTriple<T> &v) {
 	return GTriple<T>(y*v.z-z*v.y,
					  z*v.x-x*v.z,
				      x*v.y-y*v.x);
}

template<class T> inline GTriple<T> FASTCALL GTriple<T>::operator ~() {
 	DOUBLE l=sqrt((DOUBLE)x*(DOUBLE)x+(DOUBLE)y*(DOUBLE)y+(DOUBLE)z*(DOUBLE)z);
 	return GTriple<T>((T)((DOUBLE)x/l),
	  			      (T)((DOUBLE)y/l),
				      (T)((DOUBLE)z/l));
}



///////////////////////////////////////////////////////////////////////////////
//
// The GTripleM implementation
//
template<int P> ostream& operator<< (ostream& ,GTripleM<P>& );

template<int P> inline ostream& operator<< (ostream& s,GTripleM<P>& v) {
 	s<<"["<<MATH2DOUBLE(v.x,P)
     <<","<<MATH2DOUBLE(v.y,P)
  	 <<","<<MATH2DOUBLE(v.z,P)<<",1]";
 	return s;
}

template<int P> inline GTripleM<P>::GTripleM(void) {
}

template<int P> inline GTripleM<P>::GTripleM(DOUBLE a,DOUBLE b,DOUBLE c) {
 	x=DOUBLE2MATH(a,P);
 	y=DOUBLE2MATH(b,P);
 	z=DOUBLE2MATH(c,P);
}

template<int P> inline GTripleM<P>::GTripleM(LONG a,LONG b,LONG c) {
 	x=(MATH)a;
 	y=(MATH)b;
 	z=(MATH)c;
}

template<int P> inline DOUBLE FASTCALL GTripleM<P>::GetX(void) {
 	return (DOUBLE)MATH2DOUBLE(x,P);
}

template<int P> inline DOUBLE FASTCALL GTripleM<P>::GetY(void) {
 	return (DOUBLE)MATH2DOUBLE(y,P);
}

template<int P> inline DOUBLE FASTCALL GTripleM<P>::GetZ(void) {
 	return (DOUBLE)MATH2DOUBLE(z,P);
}

template<int P> inline DOUBLE FASTCALL GTripleM<P>::SetX(DOUBLE d) {
 	x=DOUBLE2MATH(d,P);
 	return d;
}

template<int P> inline DOUBLE FASTCALL GTripleM<P>::SetY(DOUBLE d) {
 	y=DOUBLE2MATH(d,P);
 	return d;
}

template<int P> inline DOUBLE FASTCALL GTripleM<P>::SetZ(DOUBLE d) {
 	z=DOUBLE2MATH(d,P);
 	return d;
}

template<int P> inline LONG FASTCALL GTripleM<P>::SetX(LONG d) {
 	x=DOUBLE2MATH((DOUBLE)d,P);
 	return d;
}

template<int P> inline LONG FASTCALL GTripleM<P>::SetY(LONG d) {
 	y=DOUBLE2MATH((DOUBLE)d,P);
 	return d;
}

template<int P> inline LONG FASTCALL GTripleM<P>::SetZ(LONG d) {
 	z=DOUBLE2MATH((DOUBLE)d,P);
 	return d;
}

template<int P> inline GTripleM<P> FASTCALL GTripleM<P>::operator+(GTripleM<P> &v) {
 	return GTripleM<P>(x+v.x,y+v.y,z+v.z);
}

template<int P> inline GTripleM<P> FASTCALL GTripleM<P>::operator-(GTripleM<P> &v) {
 	return GTripleM<P>(x-v.x,y-v.y,z-v.z);
}

template<int P> inline GTripleM<P> FASTCALL GTripleM<P>::operator*(LONG s) {
 	return GTripleM<P>(x*s,y*s,z*s);
}

template<int P> inline GTripleM<P> FASTCALL GTripleM<P>::operator*(DOUBLE s) {
 	return GTripleM<P>((MATH)((DOUBLE)x*s),(MATH)((DOUBLE)y*s),(MATH)((DOUBLE)z*s));
}

template<int P> inline GTripleM<P> FASTCALL GTripleM<P>::operator/(LONG s) {
 	return GTripleM<P>(x/s,y/s,z/s);
}

template<int P> inline GTripleM<P> FASTCALL GTripleM<P>::operator/(DOUBLE s) {
 	return GTripleM<P>((MATH)((DOUBLE)x/s),(MATH)((DOUBLE)y/s),(MATH)((DOUBLE)z/s));
}

template<int P> inline MATH FASTCALL GTripleM<P>::operator*(GTripleM<P> &v) {
 	return (x*v.x+y*v.y+z*v.z);
}

template<int P> inline GTripleM<P>& FASTCALL GTripleM<P>::operator+=(GTripleM<P> &v) {
 	x+=v.x;
 	y+=v.y;
	 z+=v.z;
 	return *this;
}

template<int P> inline GTripleM<P>& FASTCALL GTripleM<P>::operator-=(GTripleM<P> &v) {
 	x-=v.x;
 	y-=v.y;
 	z-=v.z;
 	return *this;
}

template<int P> inline GTripleM<P>& FASTCALL GTripleM<P>::operator*=(LONG s) {
 	x*=(MATH)s;
 	y*=(MATH)s;
 	z*=(MATH)s;
	return *this;
}

template<int P> inline GTripleM<P>& FASTCALL GTripleM<P>::operator*=(DOUBLE s) {
 	x=(MATH)((DOUBLE)x*s);
 	y=(MATH)((DOUBLE)y*s);
 	z=(MATH)((DOUBLE)z*s);
 	return *this;
}

template<int P> inline GTripleM<P>& FASTCALL GTripleM<P>::operator/=(LONG s) {
 	x/=(MATH)s;
 	y/=(MATH)s;
 	z/=(MATH)s;
 	return *this;
}

template<int P> inline GTripleM<P>& FASTCALL GTripleM<P>::operator/=(DOUBLE s) {
 	x=(MATH)((DOUBLE)x/s);
 	y=(MATH)((DOUBLE)y/s);
 	z=(MATH)((DOUBLE)z/s);
 	return *this;
}

template<int P> inline BOOL	FASTCALL GTripleM<P>::operator==(GTripleM<P> &v) {
 	return ((x==v.x) && (y==v.y) && (z==v.z));
}

template<int P> inline BOOL	FASTCALL GTripleM<P>::operator!=(GTripleM<P> &v) {
 	return ((x!=v.x) || (y!=v.y) || (z!=v.z));
}

template<int P> inline GTripleM<P> FASTCALL GTripleM<P>::operator %(GTripleM<P> &v) {
   #if defined(FIXED_MATH)
  	return GTripleM<P>((MATH)(LargeShrLong(LargeMul(y,v.z),P)
							-LargeShrLong(LargeMul(z,v.y),P)),
					   (MATH)(LargeShrLong(LargeMul(z,v.x),P)
							-LargeShrLong(LargeMul(x,v.z),P)),
					   (MATH)(LargeShrLong(LargeMul(x,v.y),P)
							-LargeShrLong(LargeMul(y,v.x),P)));
   #else
  	return GTripleM<P>((MATH)MATHDIV((DOUBLE)y*(DOUBLE)v.z-(DOUBLE)z*(DOUBLE)v.y,P),
					   (MATH)MATHDIV((DOUBLE)z*(DOUBLE)v.x-(DOUBLE)x*(DOUBLE)v.z,P),
					   (MATH)MATHDIV((DOUBLE)x*(DOUBLE)v.y-(DOUBLE)y*(DOUBLE)v.x,P));
   #endif
}

template<int P> inline GTripleM<P> FASTCALL GTripleM<P>::operator ~() {
 	DOUBLE l=sqrt((DOUBLE)x*(DOUBLE)x+(DOUBLE)y*(DOUBLE)y+(DOUBLE)z*(DOUBLE)z);
 	return GTripleM<P>(DOUBLE2MATH((DOUBLE)x/l,P),
				  	   DOUBLE2MATH((DOUBLE)y/l,P),
					   DOUBLE2MATH((DOUBLE)z/l,P));
}

#endif