
///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//

#ifndef __GOBJECT_H
#define __GOBJECT_H

#include "../include/gomez.h"

class GObject {
		public:
        enum {
            OBJECT_TYPE = MAKETYPE('O','B','J',' '),
            EMPTY_TYPE = MAKETYPE('E','M','T','Y'),
            CONTAINER_TYPE = MAKETYPE('C','O','N','T'),

            VIDEO_TYPE = MAKETYPE('V','I','D','O'),

            STATE_TYPE = MAKETYPE('S','T','A','T'),
            SHADINGLUT_TYPE = MAKETYPE('S','L','U','T'),
            BUMPLUT_TYPE = MAKETYPE('B','L','U','T'),
            GLASSLUT_TYPE = MAKETYPE('G','L','U','T')
		};
        enum {
        	USERDATA = 0x80000000
        };

		public:
		#pragma option -a1
		typedef struct {
				CHAR		Name[GMZ_NAME_LENGTH];
				DWORD		Flags;
				DWORD 		Type;
				DWORD		Magic;
			   } FileObjectHead;
		#pragma option -a.

		protected:
		CHAR			Name[GMZ_NAME_LENGTH];
		DWORD			Magic;
		DWORD			Type;
		DWORD			Flags;
        PTR				UserData;


		public:
		explicit GObject(CHAR* name,DWORD typ=EMPTY_TYPE);
		explicit GObject(DWORD typ=EMPTY_TYPE);
		virtual ~GObject();

		virtual CHAR* 	FASTCALL SetDefaultName(void);
        virtual CHAR*	FASTCALL ToString(CHAR *);

		GObject& 		FASTCALL operator=(GObject &);
		BOOL  	 		FASTCALL operator==(GObject &);
		BOOL   			FASTCALL operator!=(GObject &);

		CHAR*			FASTCALL SetName(CHAR *);
		CHAR*			FASTCALL GetName(CHAR *);
		CHAR*			FASTCALL GetName(void);
		CHAR*  			FASTCALL GetTypeName(CHAR *);

		DWORD			FASTCALL SetType(DWORD );
		DWORD			FASTCALL GetType(void);
		DWORD			FASTCALL GetMagic(void);
		DWORD			FASTCALL SetMagic(DWORD );

		PTR  			FASTCALL SetUserData(PTR );
		PTR  			FASTCALL GetUserData(void);
		PTR  			FASTCALL AllocUserData(size_t );
		BOOL			FASTCALL FreeUserData(void);

		DWORD			FASTCALL SetFlags(DWORD );
		DWORD 		    FASTCALL SetFlag(DWORD );
		DWORD			FASTCALL ClearFlag(DWORD );
		DWORD			FASTCALL ChangeFlag(DWORD ,BOOL );
		DWORD			FASTCALL GetFlags(void);
		BOOL			FASTCALL GetFlag(DWORD );
		virtual CHAR* 	FASTCALL GetFlagsName(CHAR *);

		virtual BOOL 	FASTCALL Write(GFile &,DWORD );
		virtual BOOL   	FASTCALL Read(GFile &);
	   };

ostream& operator<< (ostream& ,GObject& );
istream& operator>> (istream& ,GObject& );


template<class T> class GTree : public GObject {
		public:
        enum {
            TREE_TYPE = MAKETYPE('T','R','E','E'),
            ROOT_TYPE = MAKETYPE('R','O','O','T'),
            KNOT_TYPE = MAKETYPE('K','N','O','T')
        };

		protected:
		T            	*Head;

		public:
		T            	*Current;

		public:
		GTree(CHAR* name=NULL,DWORD typ=TREE_TYPE);
		~GTree();

		CHAR* 			FASTCALL SetDefaultName(void);

		BOOL			FASTCALL Clear(void);

		T*				FASTCALL SetRoot(T *);
		T*				FASTCALL GetRoot(void);
		T* 				FASTCALL Root(void);
		T* 				FASTCALL Ascend(void);
		T* 				FASTCALL Descend(void);
		T* 				FASTCALL Brother(void);

		T* 				FASTCALL Start(void);
		T* 				FASTCALL Next(T *);
		T* 				FASTCALL Next(void);

		T* 				FASTCALL FindType(T *,DWORD );
		T* 				FASTCALL FindEntry(T *);
		T* 				FASTCALL FindEntry(DWORD );

		BOOL			FASTCALL DeleteEntry(T *);
		BOOL			FASTCALL RemoveEntry(T *);

		BOOL			FASTCALL Write(GFile &,DWORD );
		BOOL			FASTCALL Read(GFile &);

		private:
		BOOL			FASTCALL WriteObjects(GFile &,T *);
		BOOL			FASTCALL ReadObjects(GFile &,T *);
	   };

template<class T> class GKnot : public T {
		public:
        enum {
            ROOT_TYPE = MAKETYPE('R','O','O','T'),
            KNOT_TYPE = MAKETYPE('K','N','O','T')
        };

		friend class GTree<GKnot<T> >;
		protected:
		GKnot<T>    	*Father;
		GKnot<T>    	*Son;
		GKnot<T>    	*Brother;


		public:
		GKnot(GKnot<T> *f,DWORD typ=EMPTY_TYPE);
		GKnot(CHAR *name=NULL,DWORD typ=EMPTY_TYPE);
		~GKnot();
		GKnot<T>&       FASTCALL operator=(GKnot<T> &);

		GKnot<T>*		FASTCALL AddSon(void);
		GKnot<T>* 		FASTCALL AddBrother(void);
		GKnot<T>*		FASTCALL AddSon(GKnot<T> *);
		GKnot<T>* 		FASTCALL AddBrother(GKnot<T> *);
		BOOL			FASTCALL DeleteSons(void);

		GKnot<T>* 		FASTCALL GetSon(void);
		GKnot<T>* 		FASTCALL GetBrother(void);
		GKnot<T>* 		FASTCALL GetFather(void);
	   };

template<class T,int lsize> class GList : public GObject {
		public:
        enum {
        	LIST_TYPE = MAKETYPE('L','I','S','T')
        };

		protected:
		T*				List[lsize];

		public:
		GList(CHAR* name=NULL,DWORD typ=LIST_TYPE);
		~GList();

		CHAR* 			FASTCALL SetDefaultName(void);

		BOOL			FASTCALL Clear(void);
		T*				FASTCALL AddEntry(T *);
		T*       		FASTCALL AddEntry(INT ,T *);
		INT				FASTCALL GetIndex(T *);
		T*       		FASTCALL GetEntry(INT );

		T*       		FASTCALL FindEntry(T *);
		T*       		FASTCALL FindEntry(DWORD );
		T*       		FASTCALL FindType(T *,DWORD );

		BOOL			FASTCALL DeleteEntry(T *);
		BOOL			FASTCALL DeleteEntry(INT );
		BOOL			FASTCALL RemoveEntry(T *);
		BOOL			FASTCALL RemoveEntry(INT );

		BOOL	  		FASTCALL Write(GFile &,DWORD );
		BOOL      		FASTCALL Read(GFile &);
	   };

template <class T> class GStack {
		protected:
		T*       	 	Stack[GMZ_OBJECTSTACK_SIZE];
		INT				Position;

		public:
		GStack();
		~GStack();
		void 	    	FASTCALL Push(T *);
		T* 				FASTCALL Pop(void);
		void 	    	FASTCALL Clear(void);
		BOOL	    	FASTCALL IsEmpty(void);
		BOOL	    	FASTCALL IsFull(void);
	   };

class GInfoObject : public GObject {
		public:
        enum {
             COPYRIGHT_TYPE = MAKETYPE('(','C','),',' '),
             VERSION_TYPE = MAKETYPE('(','V','),',' '),
             DESCRIPTION_TYPE = MAKETYPE('(','D','S','),'),
             DATE_TYPE = MAKETYPE('(','D','),',' '),
             PROGRAM_TYPE = MAKETYPE('(','P','),',' '),
             INFO_TYPE = MAKETYPE('I','N','F','O'),
             NAME_TYPE = MAKETYPE('N','A','M','E'),
             CREATOR_TYPE = MAKETYPE('C','R','T','R')
        };

		public:
		#pragma option -a1
		typedef struct {
				CHAR		Name[GMZ_NAME_LENGTH];
				DWORD		Flags;
				DWORD		Type;
				DWORD		Magic;
			   } FileInfoHead;
		#pragma option -a.

		protected:
		CHAR			*Name;
		CHAR			*Description;
		CHAR			*Creator;
		CHAR			*Program;
		CHAR			*Copyright;
		DWORD			Version;
		struct tm		Date;

		public:
		GInfoObject(CHAR* name=NULL,DWORD typ=INFO_TYPE);
		~GInfoObject();

		GInfoObject&  	FASTCALL operator=(GInfoObject &);
		BOOL   			FASTCALL operator==(GInfoObject &);
		BOOL   			FASTCALL operator!=(GInfoObject &);

		BOOL			FASTCALL Clear(void);

		CHAR*			FASTCALL SetName(CHAR *);
		CHAR*			FASTCALL GetName(CHAR *);
		CHAR*			FASTCALL GetName(void);
		CHAR*			FASTCALL SetDescription(CHAR *);
		CHAR*			FASTCALL GetDescription(CHAR *);
		CHAR*			FASTCALL GetDescription(void);
		CHAR*			FASTCALL SetCopyright(CHAR *);
		CHAR*			FASTCALL GetCopyright(CHAR *);
		CHAR*			FASTCALL GetCopyright(void);
		CHAR*			FASTCALL SetCreator(CHAR *);
		CHAR*			FASTCALL GetCreator(CHAR *);
		CHAR*			FASTCALL GetCreator(void);
		CHAR*			FASTCALL SetProgram(CHAR *);
		CHAR*			FASTCALL GetProgram(CHAR *);
		CHAR*			FASTCALL GetProgram(void);
		DWORD			FASTCALL SetVersion(DWORD );
		DWORD			FASTCALL GetVersion(void);
		struct tm		FASTCALL SetDate(struct tm );
		struct tm		FASTCALL GetDate(void);
		CHAR*   		FASTCALL GetDate(CHAR *);

		BOOL			FASTCALL Write(GFile &,DWORD );
		BOOL			FASTCALL Read(GFile &);
	   };


typedef GStack<GObject> GObjectStack;
typedef GList<GObject,GMZ_OBJECTLIST_SIZE> GObjectList;
typedef GKnot<GObject> GObjectKnot;
typedef GTree<GObjectKnot> GObjectTree;




///////////////////////////////////////////////////////////////////////////////
//
// The GObject implementation
//
inline GObject::GObject(DWORD typ) {
 	Type=typ;
 	Magic=(DWORD)this;
 	Flags=0;
 	UserData=NULL;
 	SetDefaultName();
}

inline GObject::GObject(CHAR *name,DWORD typ) {
 	Type=typ;
 	Magic=(DWORD)this;
 	Flags=0;
 	UserData=NULL;
 	if (name)
		SetName(name);
 	else
		SetDefaultName();
}

inline GObject::~GObject() {
 	if(Flags & USERDATA)
	 	if (UserData) free(UserData);
 	UserData=NULL;
 	Flags&=~USERDATA;
}

inline DWORD FASTCALL GObject::SetType(DWORD typ) {
 	return Type=typ;
}

inline DWORD FASTCALL GObject::SetMagic(DWORD magic) {
 	return Magic=magic;
}

inline CHAR* FASTCALL GObject::ToString(CHAR *buf) {
 	return GetTypeName(buf);
}

inline CHAR* FASTCALL GObject::SetDefaultName() {
    strcpy(Name,"object");
    return Name;
}

inline CHAR* FASTCALL GObject::GetTypeName(CHAR *buf) {
	strcpy(buf,typeid(*this).name());
    return buf;
}

inline DWORD FASTCALL GObject::GetMagic(void) {
 	return Magic;
}

inline DWORD FASTCALL GObject::GetFlags(void) {
 	return Flags;
}

inline DWORD FASTCALL GObject::SetFlags(DWORD flags) {
 	return Flags=flags;
}

inline DWORD FASTCALL GObject::GetType(void) {
 	return Type;
}

inline BOOL FASTCALL GObject::GetFlag(DWORD f) {
 	return (BOOL)(((Flags & f)==f) ? TRUE : FALSE);
}

inline DWORD FASTCALL GObject::SetFlag(DWORD f) {
 	return Flags|=f;
}

inline DWORD FASTCALL GObject::ClearFlag(DWORD f) {
 	return Flags&=~f;
}

inline DWORD FASTCALL GObject::ChangeFlag(DWORD f,BOOL b) {
 	if (b)
		return SetFlag(f);
 	else
		return ClearFlag(f);
}

inline CHAR* FASTCALL GObject::SetName(CHAR *text) {
 	if (strlen(text)<sizeof(Name))
		strcpy(Name,text);
 	else
		memcpy(Name,text,sizeof(Name)-1);
 	Name[sizeof(Name)-1]=0;
	return text;
}

inline CHAR* FASTCALL GObject::GetName(CHAR *buffer) {
 	strcpy(buffer,Name);
 	return buffer;
}

inline CHAR* FASTCALL GObject::GetName(void) {
 	return Name;
}

inline CHAR* FASTCALL GObject::GetFlagsName(CHAR* b) {
 	b[0]=0;
 	return b;
}

inline GObject& FASTCALL GObject::operator=(GObject &o) {
 	strcpy(Name,o.Name);
 	Type=o.Type;
 	Flags=o.Flags;
 	Magic=(DWORD)this;
 	return *this;
}

inline BOOL FASTCALL GObject::operator==(GObject &o) {
 	return (strcmp(Name,o.Name) && (Type==o.Type) && (Flags==o.Flags));
}

inline BOOL FASTCALL GObject::operator!=(GObject &o) {
 	return (!strcmp(Name,o.Name) || (Type!=o.Type) || (Flags!=o.Flags));
}

inline PTR FASTCALL GObject::SetUserData(PTR d) {
 	return UserData=d;
}

inline PTR FASTCALL GObject::GetUserData(void) {
 	return UserData;
}

inline PTR FASTCALL GObject::AllocUserData(size_t s) {
 	Flags|=USERDATA;
 	return (UserData=(PTR)malloc(s));
}

inline BOOL FASTCALL GObject::FreeUserData(void) {
 	if (UserData) free(UserData);
 	Flags&=~USERDATA;
 	UserData=NULL;
 	return TRUE;
}

inline ostream& operator<< (ostream& strm,GObject& obj) {
	CHAR buf[1024];
	strm<<obj.ToString(buf);
 	return strm;
}

inline istream& operator>> (istream& strm,GObject& obj) {
    return strm;
}



///////////////////////////////////////////////////////////////////////////////
//
// The GKnot template implementation
//
template<class T> inline GKnot<T>::GKnot(GKnot<T> *father,DWORD typ) : T(typ) {
 	Brother=Son=NULL;
 	Father=father;
}

template<class T> inline GKnot<T>::GKnot(CHAR *name,DWORD typ) : T(name,typ) {
 	Father=Brother=Son=NULL;
}

template<class T> inline GKnot<T>::~GKnot() {
 	DeleteSons();
 	if (Father) {
		Father->Son=Brother;
   	}
}

template<class T> inline GKnot<T>* FASTCALL GKnot<T>::GetSon(void) {
 	return Son;
}

template<class T> inline GKnot<T>* FASTCALL GKnot<T>::GetFather(void) {
 	return Father;
}

template<class T> inline GKnot<T>* FASTCALL GKnot<T>::GetBrother(void) {
 	return Brother;
}

template<class T> inline GKnot<T>& FASTCALL GKnot<T>::operator=(GKnot<T> &o) {
 	*(T *)this=(T)o;
 	Brother=o.Brother;
 	Father=o.Father;
 	Son=o.Son;
 	return *this;
}

template<class T> inline GKnot<T>* FASTCALL GKnot<T>::AddSon(GKnot<T> *son) {
 	if(Son)
		return (Son->AddBrother(son));
 	else {
		Son=son;
		son->Father=this;
		return son;
   	}
}

template<class T> inline GKnot<T>* FASTCALL GKnot<T>::AddSon(void) {
 	return AddSon(new GKnot<T>);
}

template<class T> GKnot<T>* FASTCALL GKnot<T>::AddBrother(GKnot<T> *newbrother) {
 	GKnot<T> *p=this;
 	while(p->Brother)
    	p=p->Brother;
 	p->Brother=newbrother;
 	newbrother->Father=Father;
 	return newbrother;
}

template<class T> inline GKnot<T>* FASTCALL GKnot<T>::AddBrother(void) {
 	return AddBrother(new GKnot<T>);
}

template<class T> BOOL FASTCALL GKnot<T>::DeleteSons(void) {
 	GKnot<T> *p,*q;
 	p=Son;
 	while(p) {
	 	q=p->Brother;
	 	delete p;
	 	p=q;
	}
 	Son=NULL;
 	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
// The GTree template implementation
//
template<class T> inline GTree<T>::GTree(CHAR *name,DWORD typ) : GObject(name,typ) {
 	Head=Current=new T((T *)NULL,ROOT_TYPE);
}

template<class T> inline GTree<T>::~GTree(void) {
 	Clear();
}

template<class T> inline CHAR* FASTCALL GTree<T>::SetDefaultName() {
    strcpy(Name,"object tree");
    return Name;
}

template<class T> BOOL FASTCALL GTree<T>::Clear(void) {
 	T *o;
 	while (Head) {
		Head->DeleteSons();
		o=Head;
		Head=Head->Brother;
		delete o;
   	}
 	Head=Current=new T((T *)NULL,ROOT_TYPE);
 	return TRUE;
}

template<class T> inline BOOL FASTCALL GTree<T>::DeleteEntry(T *o) {
 	if (RemoveEntry(o)==FALSE) return FALSE;
 	delete o;
 	return TRUE;
}

template<class T> inline T*	FASTCALL GTree<T>::SetRoot(T *r) {
 	return Head=r;
}

template<class T> inline T*	FASTCALL GTree<T>::GetRoot(void) {
 	return Head;
}

template<class T> inline T* FASTCALL GTree<T>::Root(void) {
 	return (Current=Head);
}

template<class T> inline T* FASTCALL GTree<T>::Ascend(void) {
 	return ((Current->Father!=NULL) ? Current=Current->Father : NULL);
}

template<class T> inline T* FASTCALL GTree<T>::Descend(void) {
 	return ((Current->Son!=NULL) ? Current=Current->Son : NULL);
}

template<class T> inline T* FASTCALL GTree<T>::Brother(void){
 	return ((Current->Brother!=NULL) ? Current=Current->Brother : NULL);
}

template<class T> inline T* FASTCALL GTree<T>::Start(void) {
 	return (Current=Next(NULL));
}

template<class T> inline T* FASTCALL GTree<T>::Next(void) {
 	return (Current=Next(Current));
}

template<class T> T* FASTCALL GTree<T>::Next(T* o) {
   #if defined GMZ_WALK_POSTORDER
  	T* FAST cur;

  	if (o==NULL) {
		if ((o=Head)==NULL)
			return NULL;
		while (o->Son!=NULL)
			o=o->Son;
		return o;
    }

  	if ((cur=o->Brother)==NULL)
		return o->Father;

  	while(cur->Son!=NULL)
		cur=cur->Son;
  	return cur;
   #elif defined GMZ_WALK_PREORDER
  	if (o==NULL)
		return Head;
  	if (o->Son!=NULL)
		return o->Son;

  	while((o!=NULL) && (o->Brother==NULL))
		o=o->Father;
  	return o ? o->Brother : NULL;
   #endif
}

template<class T> T* FASTCALL GTree<T>::FindType(T* o,DWORD typ) {
 	while((o=Next(o))!=NULL)
	 	if (o->GetType()==typ) return o;
 	return NULL;
}

template<class T> T* FASTCALL GTree<T>::FindEntry(T* o) {
 	T* start=NULL;
 	while((start=Next(start))!=NULL)
	 	if (start==o) return start;
 	return NULL;
}

template<class T> T* FASTCALL GTree<T>::FindEntry(DWORD magic) {
 	T* o=NULL;
 	while((o=Next(o))!=NULL)
	 	if (o->Magic==magic) return o;
 	return NULL;
}

template<class T> BOOL FASTCALL GTree<T>::RemoveEntry(T* o) {
 	T *p;
 	if (FindEntry(o)==NULL) return FALSE;
 	if (o->Father) {
		if (o->Father->Son==o)
			o->Father->Son=o->Brother;
		else {
			p=o->Father->Son;
			while ((p!=NULL) && (p->Brother!=o))
				p=p->Brother;
			if (p)
				p->Brother=o->Brother;
	   }
   	} else {
		if (Head==o)
			Head=o->Brother;
		else {
			p=Head;
			while ((p!=NULL) && (p->Brother!=o))
				p=p->Brother;
			if (p)
				p->Brother=o->Brother;
	   	}
   	}

 	o->Father=NULL;
 	o->Brother=NULL;
 	return TRUE;
}

template<class T> BOOL FASTCALL GTree<T>::Write(GFile &file,DWORD ctyp) {
 	T *o=Head;
 	if(file.OpenMetaChunk(ctyp)!=TRUE) return FALSE;
 	if(WriteObjects(file,o)!=TRUE) return FALSE;
 	return file.CloseMetaChunk();
}

template<class T> BOOL FASTCALL GTree<T>::WriteObjects(GFile &file,T *o) {
 	while (o) {
		if(o->Write(file,KNOT_TYPE)!=TRUE) return FALSE;
		if (o->Son) {
			if(file.OpenMetaChunk(GFile::SONS_TYPE)!=TRUE) return FALSE;
			if(WriteObjects(file,o->Son)!=TRUE) return FALSE;
			if(file.CloseMetaChunk()!=TRUE) return FALSE;
	   	}
		o=o->Brother;
   	}
 	return TRUE;
}

template<class T> BOOL FASTCALL GTree<T>::Read(GFile &file) {
 	Clear();
 	return ReadObjects(file,Head);
}

template<class T> BOOL FASTCALL GTree<T>::ReadObjects(GFile &file,T *o) {
 	DWORD typ;
 	T* p=o;

 	while ((typ=file.OpenChunk())!=GFile::END_TYPE) {
		switch (typ) {
			case GFile::SONS_TYPE:
				if(ReadObjects(file,o->AddSon())!=TRUE) return FALSE;
				break;
			case GFile::KNOT_TYPE:
				o=o->AddBrother();
				if(o->Read(file)!=TRUE) return FALSE;
				break;
			default:
				if(file.CloseChunk()!=TRUE) return FALSE;
				break;
	   	}
   	}
 	return DeleteEntry(p);
}


///////////////////////////////////////////////////////////////////////////////
//
// The GList template implementation
//
template<class T,int lsize> inline GList<T,lsize>::GList(CHAR* name,DWORD typ) : GObject(name,typ) {
 	memset(List,0,sizeof(List));
}

template<class T,int lsize> GList<T,lsize>::~GList() {
 	INT i;
 	for (i=0;i<lsize;i++) {
		if (List[i]) delete List[i];
		List[i]=NULL;
   	}
}

template<class T,int lsize> inline CHAR* FASTCALL GList<T,lsize>::SetDefaultName() {
    strcpy(Name,"object list");
    return Name;
}

template<class T,int lsize> inline BOOL FASTCALL GList<T,lsize>::Clear(void) {
 	memset(List,0,sizeof(List));
 	return TRUE;
}

template<class T,int lsize> inline T* FASTCALL GList<T,lsize>::AddEntry(INT pos,T* o) {
 	if ((DWORD)pos>=lsize) return NULL;
 	return List[pos]=o;
}

template<class T,int lsize> inline T* FASTCALL GList<T,lsize>::GetEntry(INT pos) {
 	if ((DWORD)pos>=lsize) return NULL;
 	return List[pos];
}

template<class T,int lsize> inline T* FASTCALL GList<T,lsize>::AddEntry(T* o) {
 	INT i;
 	if ((i=GetIndex(NULL))==-1) return NULL;
 	return AddEntry(i,o);
}

template<class T,int lsize> inline BOOL FASTCALL GList<T,lsize>::DeleteEntry(T* o) {
 	INT i;
 	if ((i=GetIndex(o))==-1) return FALSE;
 	delete List[i];
 	List[i]=NULL;
 	return TRUE;
}

template<class T,int lsize> inline BOOL FASTCALL GList<T,lsize>::DeleteEntry(INT i) {
 	if ((UINT)i>=lsize) return FALSE;
 	delete List[i];
 	List[i]=NULL;
 	return TRUE;
}

template<class T,int lsize> inline BOOL FASTCALL GList<T,lsize>::RemoveEntry(T* o) {
 	INT i;
 	if ((i=GetIndex(o))==-1) return FALSE;
 	List[i]=NULL;
 	return TRUE;
}

template<class T,int lsize> inline BOOL FASTCALL GList<T,lsize>::RemoveEntry(INT i) {
 	if ((UINT)i>=lsize) return FALSE;
 	List[i]=NULL;
 	return TRUE;
}

template<class T,int lsize> T* FASTCALL GList<T,lsize>::FindEntry(DWORD magic) {
 	INT i;
 	for (i=0;i<lsize;i++)
	if ((List[i]) && (List[i]->GetMagic()==magic))
    	return List[i];
 	return NULL;
}

template<class T,int lsize> T* FASTCALL GList<T,lsize>::FindEntry(T* o) {
 	INT i;
 	for (i=0;i<lsize;i++)
		if (List[i]==o) return o;
 	return NULL;
}

template<class T,int lsize> T* FASTCALL GList<T,lsize>::FindType(T* start,DWORD type) {
 	INT i=-1;
 	if (start!=NULL)
	 	for (i=0;i<lsize;i++)
			if (List[i]==start) break;
 	for(++i;i<lsize;i++)
		if (List[i]->GetType()==type) return List[i];
 	return NULL;
}

template<class T,int lsize> INT FASTCALL GList<T,lsize>::GetIndex(T* o) {
 	INT i;
 	for (i=0;i<lsize;i++)
		if (List[i]==o) return i;
 	return -1;
}

template<class T,int lsize> BOOL FASTCALL GList<T,lsize>::Write(GFile &file,DWORD ctype) {
 	INT skip=0,i;
 	T *o;
 	if(file.CreateMetaChunk(ctype)!=TRUE) return FALSE;

 	for (i=0;i<lsize;i++) {
		if ((o=List[i])!=NULL) {
			if (skip) {
				DWORD v=DWORD2LSB(skip);
				if(file.CreateChunk(GFile::SKIP_TYPE,GFile::COMPRESSION_NONE)!=TRUE) return FALSE;
				if(file.WriteChunk(&v,sizeof(DWORD))!=sizeof(DWORD)) return FALSE;
				if(file.CloseChunk()!=TRUE) return FALSE;
				skip=0;
		   	}
			if(o->Write(file,GFile::ENTRY_TYPE)!=TRUE)
            	return FALSE;
	   	} else {
			skip++;
	   	}
   	}
 	return file.CloseMetaChunk();
}

template<class T,int lsize> BOOL FASTCALL GList<T,lsize>::Read(GFile &file) {
 	DWORD typ,skip;
 	PTR data;
 	INT index=0;
 	T *o;
 	Clear();

 	while ((typ=file.OpenChunk())!=GFile::END_TYPE) {
		switch (typ) {
			case GFile::SKIP_TYPE:
				if(file.ReadChunk(&data)!=sizeof(DWORD)) return FALSE;
				if(file.CloseChunk()!=TRUE) return FALSE;
				skip=*(DWORD *)data;
				index+=LSB2DWORD(skip);
				free(data);
				break;
			case GFile::ENTRY_TYPE:
				o=new T;
				if(o->Read(file)!=TRUE) return FALSE;
				AddEntry(index++,o);
				break;
			default:
				if(file.CloseChunk()!=TRUE) return FALSE;
				break;
	   	}
   	}
 	return TRUE;
}



///////////////////////////////////////////////////////////////////////////////
//
// The GStack template implementation
//
template<class T> inline GStack<T>::GStack(void) {
 	Position=0;
}

template<class T> GStack<T>::~GStack() {
}

template<class T> T* FASTCALL GStack<T>::Pop(void) {
 	if (Position>0)
		return(Stack[--Position]);
 	else
		return NULL;
}

template<class T> inline void FASTCALL GStack<T>::Push(T* o) {
 	Stack[Position++]=o;
}

template<class T> inline void FASTCALL GStack<T>::Clear(void) {
 	Position=0;
}

template<class T> inline BOOL FASTCALL GStack<T>::IsEmpty(void) {
 	return (Position==0);
}

template<class T> inline BOOL FASTCALL GStack<T>::IsFull(void) {
 	return (Position>=GMZ_OBJECTSTACK_SIZE);
}



///////////////////////////////////////////////////////////////////////////////
//
// The GInfoObject implementation
//
inline GInfoObject::GInfoObject(CHAR *name,DWORD typ) : GObject(name,typ) {
 	time_t timer=time(NULL);
 	Name=Description=Creator=Program=NULL;
 	Version=0x10000L;
 	memcpy(&Date,&timer,sizeof(struct tm));
}

inline GInfoObject::~GInfoObject() {
 	Clear();
}

inline BOOL FASTCALL GInfoObject::Clear(void) {
 	if (Name) free(Name);
 	if (Description) free(Description);
 	if (Copyright) free(Copyright);
 	if (Creator) free (Creator);
 	if (Program) free (Program);
 	Copyright=Name=Description=Creator=Program=NULL;
 	return TRUE;
}

inline CHAR* FASTCALL GInfoObject::SetName(CHAR *n) {
 	if (Name) free(Name);
 	if ((Name=(CHAR *)malloc(strlen(n)))==NULL) return NULL;
 	return strcpy(Name,n);
}

inline CHAR* FASTCALL GInfoObject::GetName(CHAR *n) {
 	if (!Name) n[0]=0;
 	else strcpy(n,Name);
 	return n;
}

inline CHAR* FASTCALL GInfoObject::GetName(void) {
 	return Name;
}

inline CHAR* FASTCALL GInfoObject::SetDescription(CHAR *n) {
 	if (Description) free(Description);
 	if ((Description=(CHAR *)malloc(strlen(n)))==NULL) return NULL;
 	return strcpy(Description,n);
}

inline CHAR* FASTCALL GInfoObject::GetDescription(CHAR *n) {
 	if (!Description) n[0]=0;
 	else strcpy(n,Description);
 	return n;
}

inline CHAR* FASTCALL GInfoObject::GetDescription(void) {
 	return Description;
}

inline CHAR* FASTCALL GInfoObject::SetCopyright(CHAR *n) {
 	if (Copyright) free(Copyright);
 	if ((Copyright=(CHAR *)malloc(strlen(n)))==NULL) return NULL;
 	return strcpy(Copyright,n);
}

inline CHAR* FASTCALL GInfoObject::GetCopyright(CHAR *n) {
 	if (!Copyright) n[0]=0;
 	else strcpy(n,Copyright);
 	return n;
}

inline CHAR* FASTCALL GInfoObject::GetCopyright(void) {
 	return Copyright;
}

inline CHAR* FASTCALL GInfoObject::SetCreator(CHAR *n) {
 	if (Creator) free(Creator);
 	if ((Creator=(CHAR *)malloc(strlen(n)))==NULL) return NULL;
 	return strcpy(Creator,n);
}

inline CHAR* FASTCALL GInfoObject::GetCreator(CHAR *n) {
 	if (!Creator) n[0]=0;
 	else strcpy(n,Creator);
 	return n;
}

inline CHAR* FASTCALL GInfoObject::GetCreator(void) {
 	return Creator;
}

inline CHAR* FASTCALL GInfoObject::SetProgram(CHAR *n) {
 	if (Program) free(Program);
 	if ((Program=(CHAR *)malloc(strlen(n)))==NULL) return NULL;
 	return strcpy(Program,n);
}

inline CHAR* FASTCALL GInfoObject::GetProgram(CHAR *n) {
 	if (!Program) n[0]=0;
 	else strcpy(n,Program);
 	return n;
}

inline CHAR* FASTCALL GInfoObject::GetProgram(void) {
 	return Program;
}

inline DWORD FASTCALL GInfoObject::SetVersion(DWORD v) {
 	return Version=v;
}

inline DWORD FASTCALL GInfoObject::GetVersion(void) {
 	return Version;
}

inline struct tm FASTCALL GInfoObject::GetDate(void) {
 	return Date;
}

inline struct tm FASTCALL GInfoObject::SetDate(struct tm d) {
 	return Date=d;
}

inline CHAR* FASTCALL GInfoObject::GetDate(CHAR *n) {
 	return strcpy(n,asctime(&Date));
}

inline GInfoObject& FASTCALL GInfoObject::operator=(GInfoObject &o) {
 	(GObject &)this=(GObject &)o;
 	SetName(o.Name);
 	SetDescription(o.Description);
 	SetProgram(o.Program);
 	SetCreator(o.Creator);
 	SetDate(o.Date);
 	return *this;
}

inline BOOL FASTCALL GInfoObject::operator==(GInfoObject &o) {
 	return (((GObject &)this==(GObject &)o) &&
			(Version==o.Version) &&
			strcmp(Name,o.Name) &&
			strcmp(Description,o.Description) &&
			strcmp(Copyright,o.Copyright) &&
			strcmp(Creator,o.Creator) &&
			strcmp(Program,o.Program) &&
			memcmp(&Date,&o.Date,sizeof(Date)));
}

inline BOOL FASTCALL GInfoObject::operator!=(GInfoObject &o) {
 	return (((GObject &)this!=(GObject &)o) ||
			(Version!=o.Version) ||
			!strcmp(Name,o.Name) ||
			!strcmp(Description,o.Description) ||
			!strcmp(Copyright,o.Copyright) ||
			!strcmp(Creator,o.Creator) ||
			!strcmp(Program,o.Program) ||
			!memcmp(&Date,&o.Date,sizeof(Date)));
}


#endif
