///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#ifndef __GPALETTE_H
#define __GPALETTE_H

#include "../include/gomez.h"

#define GMZ_PAL_RGB15				((INT)0x000f)
#define GMZ_PAL_RGB16				((INT)0x0010)
#define GMZ_PAL_RGB18				((INT)0x0012)
#define GMZ_PAL_RGB24				((INT)0x0018)
#define GMZ_PAL_RGB32				((INT)0x0020)
#define GMZ_PAL_BGR15				((INT)0x010f)
#define GMZ_PAL_BGR16				((INT)0x0110)
#define GMZ_PAL_BGR18				((INT)0x0112)
#define GMZ_PAL_BGR24				((INT)0x0118)
#define GMZ_PAL_BGR32				((INT)0x0120)

#define GMZ_PAL_DIRTY				((DWORD)0x00000001)


class GColour {
		public:
		BYTE R,G,B,A;

		public:
		GColour() {};
		GColour(BYTE ,BYTE ,BYTE );
		GColour(DWORD );

		GColour 	FASTCALL operator+ (GColour &c);
		GColour 	FASTCALL operator- (GColour &c);
		GColour&	FASTCALL operator+=(GColour &c);
		GColour& 	FASTCALL operator-=(GColour &c);
		GColour&	FASTCALL operator= (DWORD );
		BOOL		FASTCALL operator==(GColour &c);
		BOOL		FASTCALL operator!=(GColour &c);

		GColour 	FASTCALL Shade(GColour& ,GColour& ,INT ,INT);
		GColour 	FASTCALL Mix(GColour& );
		DWORD  		FASTCALL Convert(INT );
		INT  		FASTCALL Intensity(void);

        DWORD		FASTCALL ConvertRGB15(void);
        DWORD		FASTCALL ConvertRGB16(void);
        DWORD		FASTCALL ConvertRGB18(void);
        DWORD		FASTCALL ConvertRGB24(void);
        DWORD		FASTCALL ConvertRGB32(void);
        DWORD		FASTCALL ConvertBGR15(void);
        DWORD		FASTCALL ConvertBGR16(void);
        DWORD		FASTCALL ConvertBGR18(void);
        DWORD		FASTCALL ConvertBGR24(void);
        DWORD		FASTCALL ConvertBGR32(void);
	   };


class GPalette : public GObject {
		public:
        enum {
        	PALETTE_TYPE = MAKETYPE('P','A','L',' ')
        };

		public:
		#pragma option -a1
		typedef struct {
				BYTE		R,G,B,A;
			   } FileColour;
		typedef struct {
				CHAR		Name[GMZ_NAME_LENGTH];
				DWORD		Flags;
				DWORD 		Type;
				DWORD		Magic;
				LONG 		FirstEntry;
				LONG 		EntryCount;
			   } FilePaletteHead;
		#pragma option -a.

		protected:
		GColour 	Colour[256];
        DWORD		Table[256];
        INT			TableFormat;

		public:
		GPalette(CHAR *name=NULL,DWORD typ=PALETTE_TYPE);
		~GPalette();

        CHAR*		FASTCALL SetDefaultName();

		GPalette 	FASTCALL operator+ (GColour &col);
		GPalette 	FASTCALL operator- (GColour &col);
		GPalette& 	FASTCALL operator+=(GColour &col);
		GPalette& 	FASTCALL operator-=(GColour &col);
		GPalette& 	FASTCALL operator= (GPalette &);
		BOOL      	FASTCALL operator==(GPalette &);
		BOOL      	FASTCALL operator!=(GPalette &);

		INT 		FASTCALL ApproximateColour(GColour &);
		BYTE* 		FASTCALL ApproximatePalette(GPalette *);
		BOOL     	FASTCALL ImportPAL(char *);
		BOOL		FASTCALL ExportPAL(char *);
		BOOL     	FASTCALL ImportMAP(char *);
		BOOL		FASTCALL ExportMAP(char *);
		GPalette*	FASTCALL SetPalette(GPalette *,INT i=0,INT count=256);
		GPalette*	FASTCALL GetPalette(GPalette *,INT i=0,INT count=256);
		GColour		FASTCALL SetColour(INT ,GColour& );
		GColour		FASTCALL SetColour(INT ,BYTE ,BYTE ,BYTE );
		GColour		FASTCALL GetColour(INT );
		void		FASTCALL Reset(void);

        DWORD*		FASTCALL GetColourTable(INT f);

		BOOL	 	FASTCALL Write(GFile &,DWORD ,INT ,INT );
		BOOL	 	FASTCALL Write(GFile &,DWORD );
		BOOL     	FASTCALL Read(GFile &);

        private:
        DWORD*		FASTCALL UpdateTable(INT f);
	   };



inline  GColour::GColour(BYTE r,BYTE g,BYTE b) {
 	R=r;
 	G=g;
 	B=b;
 	A=0;
}

inline  GColour::GColour(DWORD c) {
 	R=(BYTE)(c);
 	G=(BYTE)((DWORD)c>>8);
 	B=(BYTE)((DWORD)c>>16);
 	A=0;
}

inline GColour FASTCALL GColour::operator +(GColour &c) {
 	return GColour((BYTE)max((INT)R+(INT)c.R,255),
				   (BYTE)max((INT)G+(INT)c.G,255),
				   (BYTE)max((INT)B+(INT)c.B,255));
}

inline GColour FASTCALL GColour::operator -(GColour &c) {
 	return GColour((BYTE)min((INT)R-(INT)c.R,0),
				   (BYTE)min((INT)G-(INT)c.G,0),
				   (BYTE)min((INT)B-(INT)c.B,0));
}

inline GColour& FASTCALL GColour::operator +=(GColour &c) {
 	R=(BYTE)max((INT)R+(INT)c.R,255);
 	G=(BYTE)max((INT)G+(INT)c.G,255);
 	B=(BYTE)max((INT)B+(INT)c.B,255);
	return *this;
}

inline GColour& FASTCALL GColour::operator -=(GColour &c) {
 	R=(BYTE)max((INT)R-(INT)c.R,255);
 	G=(BYTE)max((INT)G-(INT)c.G,255);
 	B=(BYTE)max((INT)B-(INT)c.B,255);
 	return *this;
}

inline BOOL FASTCALL GColour::operator==(GColour &c) {
 	return ((R==c.R) && (G==c.G) && (B==c.B));
}

inline BOOL FASTCALL GColour::operator!=(GColour &c) {
 	return ((R!=c.R) || (G!=c.G) || (B!=c.B));
}

inline GColour FASTCALL GColour::Shade(GColour &a,GColour &l,INT spec,INT i) {
 	if (i<32) {
		return GColour((BYTE)min(((LONG)a.R*(LONG)R+ (((LONG)i*(LONG)l.R*(LONG)R)>>5))>>8,255),
					   (BYTE)min(((LONG)a.G*(LONG)G+(((LONG)i*(LONG)l.G*(LONG)G)>>5))>>8,255),
					   (BYTE)min(((LONG)a.B*(LONG)B+(((LONG)i*(LONG)l.B*(LONG)B)>>5))>>8,255));
 	} else {
		i-=32;
		return GColour((BYTE)min((((LONG)a.R+(LONG)l.R)*(LONG)R+(((LONG)i*(LONG)l.R*(LONG)spec)>>5))>>8,255),
					   (BYTE)min((((LONG)a.G+(LONG)l.G)*(LONG)G+(((LONG)i*(LONG)l.G*(LONG)spec)>>5))>>8,255),
					   (BYTE)min((((LONG)a.B+(LONG)l.B)*(LONG)B+(((LONG)i*(LONG)l.B*(LONG)spec)>>5))>>8,255));
   	}
}

inline GColour FASTCALL GColour::Mix(GColour &col) {
 	return GColour((BYTE)(((LONG)col.R+(LONG)R)>>1),
				   (BYTE)(((LONG)col.G+(LONG)G)>>1),
				   (BYTE)(((LONG)col.B+(LONG)B)>>1));
}

inline DWORD FASTCALL GColour::ConvertRGB15(void) {
 	return ((((DWORD)R>>3)<<10) + (((DWORD)G>>3)<<5) + ((DWORD)B>>3));
}

inline DWORD FASTCALL GColour::ConvertRGB16(void) {
 	return ((((DWORD)R>>3)<<11) + (((DWORD)G>>2)<<5) + ((DWORD)B>>3));
}

inline DWORD FASTCALL GColour::ConvertRGB18(void) {
 	return ((((DWORD)R>>2)<<16) + (((DWORD)G>>2)<<6) + ((DWORD)B>>2));
}

inline DWORD FASTCALL GColour::ConvertRGB24(void) {
 	return (((DWORD)R<<16) + (((DWORD)G)<<8) + ((DWORD)B));
}

inline DWORD FASTCALL GColour::ConvertRGB32(void) {
 	return (((DWORD)R<<16) + (((DWORD)G)<<8) + ((DWORD)B));
}

inline DWORD FASTCALL GColour::ConvertBGR15(void) {
 	return ((((DWORD)B>>3)<<10) + (((DWORD)G>>3)<<5) + ((DWORD)R>>3));
}

inline DWORD FASTCALL GColour::ConvertBGR16(void) {
 	return ((((DWORD)B>>3)<<11) + (((DWORD)G>>2)<<5) + ((DWORD)R>>3));
}

inline DWORD FASTCALL GColour::ConvertBGR18(void) {
 	return ((((DWORD)B>>2)<<16) + (((DWORD)G>>2)<<6) + ((DWORD)R>>2));
}

inline DWORD FASTCALL GColour::ConvertBGR24(void) {
 	return (((DWORD)B<<16) + (((DWORD)G)<<8) + ((DWORD)R));
}

inline DWORD FASTCALL GColour::ConvertBGR32(void) {
 	return (((DWORD)B<<16) + (((DWORD)G)<<8) + ((DWORD)R));
}

inline INT FASTCALL GColour::Intensity(void) {
 	return (INT)sqrt((DOUBLE)R*(DOUBLE)R+(DOUBLE)G*(DOUBLE)G+(DOUBLE)B*(DOUBLE)B);
}



inline GPalette::GPalette(CHAR* nam,DWORD typ) : GObject(nam,typ) {
}

inline GPalette::~GPalette() {
}

inline CHAR* FASTCALL GPalette::SetDefaultName() {
 	strcpy(Name,"palette");
    return Name;
}

inline GPalette* FASTCALL GPalette::GetPalette(GPalette *pal,INT i,INT count) {
 	memcpy(&pal->Colour[i],&Colour[i],sizeof(GColour)*count);
 	return pal;
}

inline GPalette* FASTCALL GPalette::SetPalette(GPalette *pal,INT i,INT count) {
 	memcpy(&Colour[i],&pal->Colour[i],sizeof(GColour)*count);
 	Flags|=GMZ_PAL_DIRTY;
	return pal;
}

inline GColour FASTCALL GPalette::GetColour(INT i) {
 	return Colour[i];
}

inline GColour FASTCALL GPalette::SetColour(INT i,GColour &c) {
 	Colour[i]=c;
 	Flags|=GMZ_PAL_DIRTY;
 	return c;
}

inline GColour FASTCALL GPalette::SetColour(INT i,BYTE r,BYTE g,BYTE b) {
 	Colour[i].R=r;
 	Colour[i].G=g;
 	Colour[i].B=b;
 	Colour[i].A=0;
 	Flags|=GMZ_PAL_DIRTY;
 	return Colour[i];
}

inline DWORD* FASTCALL GPalette::GetColourTable(INT format) {
 	if ((TableFormat==format) && (!(Flags & GMZ_PAL_DIRTY))) return Table;
 	return UpdateTable(format);
}

inline GPalette& FASTCALL GPalette::operator=(GPalette &pal) {
 	*(GObject *)this=(GObject)pal;
 	memcpy(Colour,pal.Colour,sizeof(Colour));
 	Flags|=GMZ_PAL_DIRTY;
 	return *this;
}

inline BOOL FASTCALL GPalette::operator==(GPalette &pal) {
 	return (((GObject &)this==(GObject &)pal) &&
			 memcmp(&Colour[0],&pal.Colour[0],sizeof(GColour)*256));
}

inline BOOL FASTCALL GPalette::operator!=(GPalette &pal) {
 	return (((GObject &)this!=(GObject &)pal) ||
			 !memcmp(&Colour[0],&pal.Colour[0],sizeof(GColour)*256));
}

#endif