///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#ifndef __GSOUND_H
#define __GSOUND_H

#define GMZ_SOUNDINPUT_CLASSNAME   	((CHAR *)"GMZSOUNDINPUT")
#define GMZ_SOUNDOUTPUT_CLASSNAME  	((CHAR *)"GMZSOUNDOUTPUT")

#define GMZ_SOUND_OPEN				1
#define GMZ_SOUND_CLOSE				2
#define GMZ_SOUND_DONE				3


class GSample : public GObject {
		friend class GSoundOutput;
		friend class GSoundInput;

        public:
        enum {
        	SAMPLE_TYPE = MAKETYPE('S','M','P',' ')
        };
        enum {
            QUEUED			 	= 0x00000001,
            PREPARED			= 0x00000002,
            DONE  				= 0x00000004
        };


		public:
		#pragma option -a1

		#define	GMZ_RIFF_MAGIC			0x046464952L
		#define GMZ_WAVE_MAGIC			0x045564157L
		#define GMZ_fmt_MAGIC			0x020746d66L
		#define GMZ_data_MAGIC			0x061746164L

		typedef struct {
				DWORD		FileMagic;
				DWORD		FileSize;
				DWORD		FileType;
				DWORD		FormatMagic;
				DWORD		HeaderSize;
				WORD		Version;
				WORD		Channels;
				DWORD		SampleRate;
				DWORD		ByteRate;
				WORD		SampleSize;
				WORD		SampleDepth;
				DWORD		DataMagic;
				DWORD		DataSize;
			   } WAVE_Header;	
		typedef struct {
				CHAR		Name[GMZ_NAME_LENGTH];
				DWORD		Flags;
				DWORD 		Type;
				DWORD		Magic;
				DWORD		Length;
				LONG		Rate;
				BYTE		Channels;
				BYTE		Depth;
				WORD		Reserved;
			   } FileSampleHead;
		#pragma option -a.

		protected:
		PTR			Data;
		LONG       	Rate;
		INT			Depth;
		INT			Channels;
		LONG		Length;
		#if defined(__FORWINDOWS__)
		WAVEHDR		WaveHeader;
		#endif

		public:
		GSample(CHAR *name=NULL,DWORD typ=SAMPLE_TYPE);
		~GSample();

        CHAR*		FASTCALL SetDefaultName();

		BOOL		FASTCALL Alloc(LONG length,LONG rate,INT depth,INT chan);
		BOOL		FASTCALL Free(void);

		BOOL		FASTCALL ImportWAVE(char *);
		BOOL		FASTCALL ExportWAVE(char *);
		BOOL		FASTCALL ImportVOC(char *);
		BOOL		FASTCALL ExportVOC(char *);

		LONG		FASTCALL GetSampleRate(void);
		LONG		FASTCALL SetSampleRate(LONG );
		INT  		FASTCALL GetDepth(void);
		INT			FASTCALL GetChannels(void);
		INT			FASTCALL GetSampleSize(void);
		LONG        FASTCALL GetSize(void);
		LONG		FASTCALL GetLength(void);
		PTR			FASTCALL GetData(void);
		PTR			FASTCALL GetData(PTR );
		PTR			FASTCALL SetData(PTR );

		BOOL	  	FASTCALL Write(GFile &,DWORD );
		BOOL      	FASTCALL Read(GFile &);
	   };

class GSoundOutput : public GObject {
		#if defined(__FORWINDOWS__)
		friend INT WINAPI WinMain(HINSTANCE ,HINSTANCE ,LPSTR ,INT );
		#endif

        public:
        enum Types {
        	SOUNDSTREAM_TYPE = MAKETYPE('S','N','D','>')
        };

		protected:
		BOOL		(FASTCALL *Handler)(INT ,GSoundOutput *,GSample *);
		#if defined(__FORWINDOWS__)
		HWAVEOUT	DeviceHandle;
		HWND		WindowHandle;
		#endif

		public:
		GSoundOutput(CHAR *name=NULL,DWORD typ=SOUNDSTREAM_TYPE);
		~GSoundOutput();

        CHAR*		FASTCALL SetDefaultName();

		BOOL		FASTCALL Open(GSample *);
		BOOL		FASTCALL Open(LONG rate,INT depth,INT chan);
		BOOL		FASTCALL Close(void);
		BOOL		FASTCALL Pause(void);
		BOOL		FASTCALL Stop(void);
		BOOL		FASTCALL Continue(void);
		BOOL		FASTCALL AddBuffer(GSample *);
		PTR 		FASTCALL SetHandler(BOOL (FASTCALL *)(INT ,GSoundOutput *,GSample *));
		PTR			FASTCALL GetHandler(void);

		INT			FASTCALL GetVolume(void);
		INT			FASTCALL SetVolume(INT );
		DOUBLE		FASTCALL GetRate(void);
		DOUBLE		FASTCALL SetRate(DOUBLE );
		DOUBLE		FASTCALL GetPitch(void);
		DOUBLE		FASTCALL SetPitch(DOUBLE );

		private:
		static BOOL	FASTCALL DefaultHandler(INT msg,GSoundOutput *,GSample *);
		#if defined(__FORWINDOWS__)
		static LRESULT CALLBACK WindowProc(HWND ,UINT ,WPARAM  ,LPARAM );
		#endif
	   };

class GSoundInput : public GObject {
		#if defined(__FORWINDOWS__)
		friend INT WINAPI WinMain(HINSTANCE ,HINSTANCE ,LPSTR ,INT );
		#endif

        public:
        enum {
        	SOUNDSTREAM_TYPE = MAKETYPE('S','N','D','>')
        };

		protected:
		BOOL		(FASTCALL *Handler)(INT ,GSoundInput *,GSample *);
		#if defined(__FORWINDOWS__)
		HWAVEIN		DeviceHandle;
		HWND		WindowHandle;
		#endif

		public:
		GSoundInput(CHAR *name=NULL,DWORD typ=SOUNDSTREAM_TYPE);
		~GSoundInput();

        CHAR*		FASTCALL SetDefaultName();

		BOOL		FASTCALL Open(GSample *);
		BOOL		FASTCALL Open(LONG rate,INT depth,INT chan);
		BOOL		FASTCALL Close(void);
		BOOL		FASTCALL Pause(void);
		BOOL		FASTCALL Stop(void);
		BOOL		FASTCALL Continue(void);
		BOOL		FASTCALL AddBuffer(GSample *);
		PTR 		FASTCALL SetHandler(BOOL (FASTCALL *)(INT ,GSoundInput *,GSample *));
		PTR			FASTCALL GetHandler(void);

		private:
		static BOOL	FASTCALL DefaultHandler(INT msg,GSoundInput *,GSample *);
		#if defined(__FORWINDOWS__)
		static LRESULT CALLBACK WindowProc(HWND ,UINT ,WPARAM  ,LPARAM );
		#endif
	   };

class GSound : public GSample , protected GSoundOutput {
		public:
        enum {
        	SOUND_TYPE = MAKETYPE('S','N','D',' ')
        };

		protected:
		LONG		Position;

		public:
		GSound(CHAR *name=NULL,DWORD typ=SOUND_TYPE);
		~GSound();
        
        CHAR*		FASTCALL SetDefaultName();

		BOOL		FASTCALL Open(void);
		BOOL		FASTCALL Close(void);
		BOOL		FASTCALL Play(void);
		BOOL		FASTCALL Stop(void);
		BOOL		FASTCALL Pause(void);
		BOOL		FASTCALL Continue(void);
		LONG		FASTCALL SetPosition(LONG );
		LONG		FASTCALL GetPosition(void);

		INT			FASTCALL GetVolume(void);
		INT			FASTCALL SetVolume(INT );
		DOUBLE		FASTCALL GetRate(void);
		DOUBLE		FASTCALL SetRate(DOUBLE );
		DOUBLE		FASTCALL GetPitch(void);
		DOUBLE		FASTCALL SetPitch(DOUBLE );
	   };



inline GSample::GSample(CHAR *nam,DWORD typ) : GObject(nam,typ) {
}

inline GSample::~GSample() {
}

inline CHAR* FASTCALL GSample::SetDefaultName() {
 	strcpy(Name,"sound sample");
    return Name;
}

inline LONG FASTCALL GSample::GetSampleRate(void) {
 	return Rate;
}

inline LONG FASTCALL GSample::SetSampleRate(LONG s) {
 	return Rate=s;
}

inline LONG FASTCALL GSample::GetSize(void) {
 	return Length*(LONG)Channels*(LONG)((Depth+7)>>3);
}

inline INT FASTCALL GSample::GetSampleSize(void) {
 	return Channels*((Depth+7)>>3);
}

inline LONG FASTCALL GSample::GetLength(void) {
 	return Length;
}

inline INT FASTCALL GSample::GetDepth(void) {
 	return Depth;
}

inline INT FASTCALL GSample::GetChannels(void) {
 	return Channels;
}

inline PTR FASTCALL GSample::GetData(void) {
 	return Data;
}

inline PTR FASTCALL GSample::GetData(PTR dst) {
 	return memcpy(dst,Data,GetSize());
}

inline PTR FASTCALL GSample::SetData(PTR src) {
 	return memcpy(Data,src,GetSize());
}



inline GSound::GSound(CHAR *nam,DWORD typ) : GSoundOutput(nam,typ) {
}

inline GSound::~GSound() {
}

inline CHAR* FASTCALL GSound::SetDefaultName() {
 	strcpy(GSample::Name,"sound");
    return GSample::Name;
}

inline BOOL FASTCALL GSound::Open(void) {
 	return GSoundOutput::Open(Rate,Depth,Channels);
}

inline BOOL FASTCALL GSound::Close(void) {
 	return GSoundOutput::Close();
}

inline BOOL FASTCALL GSound::Play(void) {
 	return GSoundOutput::AddBuffer(this);
}

inline BOOL FASTCALL GSound::Stop(void) {
 	return GSoundOutput::Stop();
}

inline BOOL FASTCALL GSound::Pause(void) {
 	return GSoundOutput::Pause();
}

inline BOOL FASTCALL GSound::Continue(void) {
 	return GSoundOutput::Continue();
}

inline INT FASTCALL GSound::SetVolume(INT v) {
 	return GSoundOutput::SetVolume(v);
}

inline INT FASTCALL GSound::GetVolume(void) {
 	return GSoundOutput::GetVolume();
}

inline DOUBLE FASTCALL GSound::SetPitch(DOUBLE v) {
 	return GSoundOutput::SetPitch(v);
}

inline DOUBLE FASTCALL GSound::GetPitch(void) {
 	return GSoundOutput::GetPitch();
}

inline DOUBLE FASTCALL GSound::SetRate(DOUBLE v) {
 	return GSoundOutput::SetRate(v);
}

inline DOUBLE FASTCALL GSound::GetRate(void) {
 	return GSoundOutput::GetRate();
}




inline GSoundOutput::GSoundOutput(CHAR *nam,DWORD typ) : GObject(nam,typ) {
 	SetHandler(DefaultHandler);
}

inline GSoundOutput::~GSoundOutput() {
}

inline CHAR* FASTCALL GSoundOutput::SetDefaultName() {
 	strcpy(Name,"sound output device");
    return Name;
}

inline PTR FASTCALL GSoundOutput::GetHandler(void) {
 	return (PTR)Handler;
}

#pragma argsused
inline BOOL FASTCALL GSoundOutput::DefaultHandler(INT msg,GSoundOutput *strm,GSample *smp) {
 	return TRUE;
}

inline PTR FASTCALL GSoundOutput::SetHandler(BOOL (FASTCALL *h)(INT ,GSoundOutput *,GSample *)) {
 	return (PTR)Handler=h;
}



inline GSoundInput::GSoundInput(CHAR *nam,DWORD typ) : GObject(nam,typ) {
 	SetHandler(DefaultHandler);
}

inline GSoundInput::~GSoundInput() {
}

inline CHAR* FASTCALL GSoundInput::SetDefaultName() {
 	strcpy(Name,"sound input device");
    return Name;
}

inline PTR FASTCALL GSoundInput::GetHandler(void) {
 	return (PTR)Handler;
}

inline PTR FASTCALL GSoundInput::SetHandler(BOOL (FASTCALL *h)(INT ,GSoundInput *,GSample *)) {
 	return (PTR)Handler=h;
}

#pragma argsused
inline BOOL FASTCALL GSoundInput::DefaultHandler(INT msg,GSoundInput *strm,GSample *smp) {
 	return TRUE;
}


#endif