///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//
///////////////////////////////////////////////////////////////////////////////
//
// You should define one of the two MATH modes
//		__FIXEDMATH__
//		__FLOATMATH__
//
// You can also use UniCode:
//		__UNICODE__
//
// If you want to use 64bit arithmetics, but your computer/compiler does
// not support such data types, then you can define:
//		__LARGEINT__
//
// It also helps a _lot_, if you defined your target platform plus compiler !
// See GSYSTEM.H for further information.
//

#ifndef __GVETYPES_H
#define __GVETYPES_H

#include "../include/gsystem.h"

//
// First thing we need are some types. There are some machine independant
// types, that should be used, if you need to knoe how many bits they use:
//		(U)TINY		-	8 bit
//		(U)SHORT	-	16 bit
//		(U)LONG		-	32 bit
//		(U)LARGE	-	64 bit (simulated, if not available)
//      BYTE		- 	8 bit
//		WORD		- 	16 bit
//		DWORD		-	32 bit
//		QWORD		-	64 bit
//		WCHAR		-	16 bit (used for UniCode)
//
// Then come the architecture-specific types:
//		CHAR		- 	8/16 bit
//		(U)INT		-	16/32/64 bit
//		PTR			-	32/64 bit pointer
//
// The floating point definitions are definitley machine-dependant, so
// I do not want to tell you anything, because it probably will be wrong
// on a not-so-common-system. But there are two types:
//		FLOAT		-	Single precision
//		DOUBLE		- 	Double precision
//
// Last come some misc. definitions:
//		ERRVAL
//		BOOL
//
#define VOID 					void
typedef          char 			CHAR;
typedef unsigned char 			UCHAR;
typedef   signed char			TINY;
typedef unsigned char			UTINY;
typedef   signed short 			SHORT;
typedef unsigned short 			USHORT;
typedef   signed int 			INT;
typedef unsigned int 			UINT;
typedef   signed long 			LONG;
typedef unsigned long 			ULONG;
typedef 	   double 			DOUBLE;
typedef 	    float 			FLOAT;
typedef unsigned long 			UFIXED;
typedef signed   long 			SFIXED;
typedef          int 			BOOL;
typedef 	     int  			ERRVAL;
typedef 		 void*	   		PTR;

typedef unsigned char 			BYTE;
typedef unsigned short 			WORD;
typedef unsigned long 			DWORD;

typedef BYTE					QBYTE[4];


// Unicode wide character types
#if defined(__FORWIN32__)
 typedef 		wchar_t			WCHAR;
#elif defined(__FORBORLAND__) && (__BORLANDC__>=0x0500)
 typedef		wchar_t			WCHAR;
#elif defined(__FORGNU__)
 typedef		wchar_t			WCHAR;
#else
 typedef		unsigned short	WCHAR;
#endif


// 64bit data type QWORD
#if defined(__64BIT__)
 #define __NATIVEQWORD__
 typedef unsigned __int64		QWORD;
#elif defined(__FORBORLAND__) && (__BORLANDC__ >= 0x0520)
 #define __NATIVEQWORD__
 typedef unsigned __int64		QWORD;
#elif defined(__FORGNU__)
 #define __NATIVEQWORD__
 typedef unsigned long long     QWORD;
#else
 #if defined(__MSBFIRST__)
  typedef struct {
			 DWORD	Hi;
			 DWORD	Lo;
			} QWORD;
 #else
  typedef struct {
			 DWORD	Lo;
			 DWORD	Hi;
			} QWORD;
 #endif
#endif

//
// A nice template for swapping vars...
//
template<class T> inline void SwapVars(T &a,T &b)
{
 T c=b;
 b=a;
 a=c;
}



//
// A nice macro for declaring handles, Windows-like...
//
#ifdef DECLARE_HANDLE
 #undef DECLARE_HANDLE
#endif
#define DECLARE_HANDLE(name)  	struct name##h__ { int unused; }; \
						      	typedef struct name##h__ *name
#define DECLARE_TYPE(name) 		struct name##t__ { int unused; }; \
						   		typedef struct name##t__ *name

//
// The following macros are compiler specific things used to tell
// the compiler to use register variables etc...
//
#if defined(__FORBORLAND__)
 #define FAST	  				register
 #define LOCAL	  				static

 #define FASTCLASS				_fastthis
 #define FASTCALL 				_fastcall
 #define EXPORT	  				_export
 #define IMPORT	  				_export

 #define EXTERN					extern
 #define PUBLIC
#elif defined(__FORGNU__)
 #define FAST	  				register
 #define LOCAL	  				static

 #define FASTCLASS
 #define FASTCALL
 #define EXPORT
 #define IMPORT	  				

 #define EXTERN					extern
 #define PUBLIC
#else
 #define FAST
 #define LOCAL	  				static

 #define FASTCLASS
 #define FASTCALL
 #define EXPORT
 #define IMPORT

 #define EXTERN					extern
 #define PUBLIC
#endif



//
// Supply our applications with 64bit values
//
#if defined(__FORBORLAND__) && defined(__BORLANDC__) && (__BORLANDC__>=0x0520)
  typedef signed __int64 		LARGE;
  typedef unsigned __int64 		ULARGE;
  #ifndef __LARGEINT__
  #define __LARGEINT__
  #endif

  #define LargeMul(a, b) 		((LARGE)((LONG)(a)) * (LARGE)((LONG)(b)))
  #define LargeShl(a, b) 		((LARGE)(a) << (b))
  #define LargeShr(a, b) 		((LARGE)(a) >> (b))
  #define LargeShlLong(a, b)	((LONG)((LARGE)(a) << (b)))
  #define LargeShrLong(a, b)	((LONG)((LARGE)(a) >> (b)))
  #define ULargeMul(a, b) 		((ULARGE)((DWORD)(a)) * (ULARGE)((DWORD)(b)))
  #define ULargeShr(a, b) 		((ULARGE)(a) >> (b))
  #define ULargeShl(a, b) 		((ULARGE)(a) << (b))
  #define ULargeShlULong(a, b)	((ULONG)((ULARGE)(a) << (b)))
  #define ULargeShrULong(a, b)	((ULONG)((ULARGE)(a) >> (b)))
  #define MulDiv(a,b,c)		   	((LONG)(((LARGE)(a)*(LARGE)(b))/(LARGE)(c)))
  #define MulShr(a,b,c)			((LONG)(((LARGE)(a)*(LARGE)(b))>>(INT)(c)))
  #define LARGE2LONG(a)			((LONG)(a))
  #define ULARGE2ULONG(a)		((ULONG)(a))

#elif defined(__FORGNU__)
  typedef signed long long		LARGE;
  typedef unsigned long long	ULARGE;
  #ifndef __LARGEINT__
  #define __LARGEINT__
  #endif

  #define LargeMul(a, b) 		((LARGE)((LONG)(a)) * (LARGE)((LONG)(b)))
  #define LargeShl(a, b) 		((LARGE)(a) << (b))
  #define LargeShr(a, b) 		((LARGE)(a) >> (b))
  #define LargeShlLong(a, b)	((LONG)((LARGE)(a) << (b)))
  #define LargeShrLong(a, b)	((LONG)((LARGE)(a) >> (b)))
  #define ULargeMul(a, b) 		((ULARGE)((DWORD)(a)) * (ULARGE)((DWORD)(b)))
  #define ULargeShr(a, b) 		((ULARGE)(a) >> (b))
  #define ULargeShl(a, b) 		((ULARGE)(a) << (b))
  #define ULargeShlULong(a, b)	((ULONG)((ULARGE)(a) << (b)))
  #define ULargeShrULong(a, b)	((ULONG)((ULARGE)(a) >> (b)))
  #define MulDiv(a,b,c)		   	((LONG)(((LARGE)(a)*(LARGE)(b))/(LARGE)(c)))
  #define MulShr(a,b,c)			((LONG)(((LARGE)(a)*(LARGE)(b))>>(INT)(c)))
  #define LARGE2LONG(a)			((LONG)(a))
  #define ULARGE2ULONG(a)		((ULONG)(a))

#elif defined(__64BIT__)
  typedef signed __int64 		LARGE;
  typedef unsigned __int64 		ULARGE;
  #ifndef __LARGEINT__
  #define __LARGEINT__
  #endif

  #define LargeMul(a, b) 		((LARGE)((LONG)(a)) * (LARGE)((LONG)(b)))
  #define LargeShl(a, b) 		((LARGE)(a) << (b))
  #define LargeShr(a, b) 		((LARGE)(a) >> (b))
  #define LargeShlLong(a, b)	((LONG)((LARGE)(a) << (b)))
  #define LargeShrLong(a, b)	((LONG)((LARGE)(a) >> (b)))
  #define ULargeMul(a, b) 		((ULARGE)((DWORD)(a)) * (ULARGE)((DWORD)(b)))
  #define ULargeShr(a, b) 		((ULARGE)(a) >> (b))
  #define ULargeShl(a, b) 		((ULARGE)(a) << (b))
  #define ULargeShlULong(a, b)	((ULONG)((ULARGE)(a) << (b)))
  #define ULargeShrULong(a, b)	((ULONG)((ULARGE)(a) >> (b)))
  #define MulDiv(a,b,c)		   	((LONG)(((LARGE)(a)*(LARGE)(b))/(LARGE)(c)))
  #define MulShr(a,b,c)			((LONG)(((LARGE)(a)*(LARGE)(b))>>(INT)(c)))
  #define LARGE2LONG(a)			((LONG)(a))
  #define ULARGE2ULONG(a)		((ULONG)(a))

#elif defined(__FORWIN32__) && !(defined(__BORLANDC__) && (__BORLANDC__<=0x0510)) && !defined(__LARGEINT__)
  typedef LONGLONG				LARGE;
  typedef DWORDLONG				ULARGE;

  #define LargeMul(a,b)			Int32x32To64((a),(b))
  #define LargeShl(a,b)			Int64ShllMod32((a),(b))
  #define LargeShr(a,b)			Int64ShraMod32((a),(b))
  #define LargeShlLong(a, b)	((LONG)Int64ShllMod32((a),(b)))
  #define LargeShrLong(a, b)	((LONG)Int64ShraMod32((a),(b))
  #define ULargeMul(a,b) 		UInt32x32To64((a),(b))
  #define ULargeShl(a,b) 		Int64ShllMod32((a),(b))
  #define ULargeShr(a,b) 		Int64ShrlMod32((a),(b))
  #define ULargeShlULong(a, b)	((ULONG)Int64ShllMod32((a),(b)))
  #define ULargeShrULong(a, b)	((ULONG)Int64ShrlMod32((a),(b))
  #define MulDiv(a,b,c)			((LONG)(((LONG)(a)*(LONG)(b))/(LONG)(c)))
  #define MulShr(a,b,c)			((LONG)(((LARGE)(a)*(LARGE)(b))>>(INT)(c)))
  #define LARGE2LONG(a)			((LONG)(a))
  #define ULARGE2ULONG(a)		((ULONG)(a))

#elif defined(__LARGEINT__)
  //
  // O.k., here we define a new 64bit class, that is easy to use, though
  // a bit slow... This class will only be created, if this code couldn't
  // find native 64bit support _and_ you defined __LARGEINT__
  //
  class ULARGE {
		protected:
		#if defined(__LSBFIRST__)
		ULONG Lo;
		ULONG Hi;
		#else
		ULONG Hi;
		ULONG Lo;
		#endif

		public:
		ULARGE();
		ULARGE(ULONG );
		ULARGE(ULONG ,ULONG );
		ULARGE(DOUBLE );

		friend ostream& operator<< (ostream& ,ULARGE& );
		friend istream& operator>> (istream& ,ULARGE& );

		friend ULONG  FASTCALL ULARGE2ULONG(ULARGE &);

		friend ULARGE FASTCALL operator+(ULARGE &,ULARGE &);
		friend ULARGE FASTCALL operator+(ULONG ,ULARGE &);
		friend ULARGE FASTCALL operator+(ULARGE &,ULONG );
		friend ULARGE FASTCALL operator-(ULARGE &,ULARGE &);
		friend ULARGE FASTCALL operator-(ULONG ,ULARGE &);
		friend ULARGE FASTCALL operator-(ULARGE &,ULONG );
		friend ULARGE FASTCALL operator*(ULARGE &,ULARGE &);
		friend ULARGE FASTCALL operator*(ULONG ,ULARGE &);
		friend ULARGE FASTCALL operator*(ULARGE &,ULONG );
		friend ULARGE FASTCALL operator/(ULARGE &,ULARGE &);
		friend ULARGE FASTCALL operator/(ULONG ,ULARGE &);
		friend ULARGE FASTCALL operator/(ULARGE &,ULONG );
		friend ULARGE FASTCALL operator|(ULARGE &,ULARGE &);
		friend ULARGE FASTCALL operator|(ULONG ,ULARGE &);
		friend ULARGE FASTCALL operator|(ULARGE &,ULONG );
		friend ULARGE FASTCALL operator&(ULARGE &,ULARGE &);
		friend ULARGE FASTCALL operator&(ULONG ,ULARGE &);
		friend ULARGE FASTCALL operator&(ULARGE &,ULONG );
		friend ULARGE FASTCALL operator^(ULARGE &,ULARGE &);
		friend ULARGE FASTCALL operator^(ULONG ,ULARGE &);
		friend ULARGE FASTCALL operator^(ULARGE &,ULONG );
		friend ULARGE FASTCALL operator%(ULARGE &,ULARGE &);
		friend ULARGE FASTCALL operator%(ULONG ,ULARGE &);
		friend ULARGE FASTCALL operator%(ULARGE &,ULONG );

		ULARGE FASTCALL operator<<(INT );
		ULARGE FASTCALL operator>>(INT );
		ULARGE& FASTCALL operator+=(ULARGE &);
		ULARGE& FASTCALL operator+=(ULONG );
		ULARGE& FASTCALL operator-=(ULARGE &);
		ULARGE& FASTCALL operator-=(ULONG );
		ULARGE& FASTCALL operator*=(ULARGE &);
		ULARGE& FASTCALL operator*=(ULONG );
		ULARGE& FASTCALL operator/=(ULARGE &);
		ULARGE& FASTCALL operator/=(ULONG );
		ULARGE& FASTCALL operator<<=(INT );
		ULARGE& FASTCALL operator>>=(INT );
		ULARGE& FASTCALL operator|=(ULARGE &);
		ULARGE& FASTCALL operator|=(ULONG );
		ULARGE& FASTCALL operator&=(ULARGE &);
		ULARGE& FASTCALL operator&=(ULONG );
		ULARGE& FASTCALL operator^=(ULARGE &);
		ULARGE& FASTCALL operator^=(ULONG );
		ULARGE& FASTCALL operator%=(ULARGE &);
		ULARGE& FASTCALL operator%=(ULONG );
		ULARGE& FASTCALL operator--();
		ULARGE& FASTCALL operator++();
		ULARGE FASTCALL operator-();
		ULARGE FASTCALL operator~();
		BOOL FASTCALL operator<(ULARGE &);
		BOOL FASTCALL operator<(ULONG );
		BOOL FASTCALL operator>(ULARGE &);
		BOOL FASTCALL operator>(ULONG );
		BOOL FASTCALL operator<=(ULARGE &);
		BOOL FASTCALL operator<=(ULONG );
		BOOL FASTCALL operator>=(ULARGE &);
		BOOL FASTCALL operator>=(ULONG );
		BOOL FASTCALL operator==(ULARGE &);
		BOOL FASTCALL operator==(ULONG );
		BOOL FASTCALL operator!=(ULARGE &);
		BOOL FASTCALL operator!=(ULONG );

		protected:
		DOUBLE FASTCALL ToDouble(void);
	   };


  class LARGE {
		protected:
		#if defined(__LSBFIRST__)
		ULONG Lo;
		ULONG Hi;
		#else
		ULONG Hi;
		ULONG Lo;
		#endif

		public:
		LARGE();
		LARGE(LONG );
		LARGE(DOUBLE );

		friend ostream& operator<< (ostream& ,LARGE& );
		friend istream& operator>> (istream& ,LARGE& );

		friend LONG  FASTCALL LARGE2LONG(LARGE &);

		friend LARGE FASTCALL operator+(LARGE &,LARGE &);
		friend LARGE FASTCALL operator+(LONG ,LARGE &);
		friend LARGE FASTCALL operator+(LARGE &,LONG );
		friend LARGE FASTCALL operator-(LARGE &,LARGE &);
		friend LARGE FASTCALL operator-(LONG ,LARGE &);
		friend LARGE FASTCALL operator-(LARGE &,LONG );
		friend LARGE FASTCALL operator*(LARGE &,LARGE &);
		friend LARGE FASTCALL operator*(LONG ,LARGE &);
		friend LARGE FASTCALL operator*(LARGE &,LONG );
		friend LARGE FASTCALL operator/(LARGE &,LARGE &);
		friend LARGE FASTCALL operator/(LONG ,LARGE &);
		friend LARGE FASTCALL operator/(LARGE &,LONG );
		friend LARGE FASTCALL operator|(LARGE &,LARGE &);
		friend LARGE FASTCALL operator|(LONG ,LARGE &);
		friend LARGE FASTCALL operator|(LARGE &,LONG );
		friend LARGE FASTCALL operator&(LARGE &,LARGE &);
		friend LARGE FASTCALL operator&(LONG ,LARGE &);
		friend LARGE FASTCALL operator&(LARGE &,LONG );
		friend LARGE FASTCALL operator^(LARGE &,LARGE &);
		friend LARGE FASTCALL operator^(LONG ,LARGE &);
		friend LARGE FASTCALL operator^(LARGE &,LONG );
		friend LARGE FASTCALL operator%(LARGE &,LARGE &);
		friend LARGE FASTCALL operator%(LONG ,LARGE &);
		friend LARGE FASTCALL operator%(LARGE &,LONG );

		LARGE FASTCALL operator<<(INT );
		LARGE FASTCALL operator>>(INT );
		LARGE& FASTCALL operator+=(LARGE &);
		LARGE& FASTCALL operator+=(LONG );
		LARGE& FASTCALL operator-=(LARGE &);
		LARGE& FASTCALL operator-=(LONG );
		LARGE& FASTCALL operator*=(LARGE &);
		LARGE& FASTCALL operator*=(LONG );
		LARGE& FASTCALL operator/=(LARGE &);
		LARGE& FASTCALL operator/=(LONG );
		LARGE& FASTCALL operator<<=(INT );
		LARGE& FASTCALL operator>>=(INT );
		LARGE& FASTCALL operator|=(LARGE &);
		LARGE& FASTCALL operator|=(LONG );
		LARGE& FASTCALL operator&=(LARGE &);
		LARGE& FASTCALL operator&=(LONG );
		LARGE& FASTCALL operator^=(LARGE &);
		LARGE& FASTCALL operator^=(LONG );
		LARGE& FASTCALL operator%=(LARGE &);
		LARGE& FASTCALL operator%=(LONG );
		LARGE& FASTCALL operator--();
		LARGE& FASTCALL operator++();
		LARGE FASTCALL operator-();
		LARGE FASTCALL operator~();
		
		BOOL FASTCALL operator<(LARGE &);
		BOOL FASTCALL operator<(LONG );
		BOOL FASTCALL operator>(LARGE &);
		BOOL FASTCALL operator>(LONG );
		BOOL FASTCALL operator<=(LARGE &);
		BOOL FASTCALL operator<=(LONG );
		BOOL FASTCALL operator>=(LARGE &);
		BOOL FASTCALL operator>=(LONG );
		BOOL FASTCALL operator==(LARGE &);
		BOOL FASTCALL operator==(LONG );
		BOOL FASTCALL operator!=(LARGE &);
		BOOL FASTCALL operator!=(LONG );

		protected:
		DOUBLE FASTCALL ToDouble(void);
	   };

  #define LargeMul(a, b) 		((LARGE((a)) * (LARGE((b)))
  #define LargeShl(a, b) 		(LARGE((a)) << (b))
  #define LargeShr(a, b) 		(LARGE((a)) >> (b))
  #define LargeShlLong(a, b)	(LARGE2LONG(LARGE((a)) << (b)))
  #define LargeShrLong(a, b)	(LARGE2LONG(LARGE((a)) >> (b)))
  #define ULargeMul(a, b) 		(ULARGE((a)) * ULARGE((b)))
  #define ULargeShr(a, b) 		(ULARGE((a)) >> (b))
  #define ULargeShl(a, b) 		(ULARGE((a)) << (b))
  #define ULargeShlULong(a, b)	(ULARGE2ULONG(ULARGE((a)) << (b)))
  #define ULargeShrULong(a, b)	(ULARGE2ULONG(ULARGE((a)) >> (b)))
  #define MulDiv(a,b,c)		   	(LARGE2LONG((LARGE((a))*LARGE((b))) /LARGE((c))))
  #define MulShr(a,b,c)			(LARGE2LONG((LARGE((a))*LARGE((b)))>>(INT)(c)))

  ostream& operator<< (ostream& ,LARGE& );
  istream& operator>> (istream& ,LARGE& );
  ostream& operator<< (ostream& ,ULARGE& );
  istream& operator>> (istream& ,ULARGE& );

  inline ostream& operator<< (ostream& s,LARGE& f) {
	s.precision(0);
	s.setf(ios::fixed,ios::floatfield);
	s<<f.ToDouble();
	return s;
  }

  inline istream& operator>> (istream& s,LARGE& f) {
	DOUBLE g;
	s>>g;
	f=LARGE(g);
	return s;
  }

  inline ostream& operator<< (ostream& s,ULARGE& f) {
	s.precision(0);
	s.setf(ios::fixed,ios::floatfield);
	s<<f.ToDouble();
	return s;
  }

  inline istream& operator>> (istream& s,ULARGE& f) {
	DOUBLE g;
	s>>g;
	f=ULARGE(g);
	return s;
  }

  inline ULARGE::ULARGE() {
  	Lo=Hi=0;
  }

  inline ULARGE::ULARGE(ULONG a) {
	Lo=a;
	Hi=0;
  }

  inline ULARGE::ULARGE(ULONG a,ULONG b) {
	Lo=a;
	Hi=b;
  }

  inline ULARGE::ULARGE(DOUBLE a) {
	Lo=(ULONG)a;
	Hi=(ULONG)ldexp(a,-32);
  }

  inline ULONG FASTCALL ULARGE2ULONG(ULARGE &l) {
	return l.Lo;
  }

  inline DOUBLE FASTCALL ULARGE::ToDouble(void) {
	return (DOUBLE)((DOUBLE)Lo+(DOUBLE)ldexp((DOUBLE)Hi,32));
  }

  inline ULARGE FASTCALL operator+(ULARGE &a,ULARGE &b) {
	ULARGE n;
	n.Lo=a.Lo+b.Lo;
	n.Hi=a.Hi+b.Hi;
	if (n.Lo < a.Lo) n.Hi++;
	return n;
  }

  inline ULARGE FASTCALL operator+(ULARGE &a,ULONG b) {
	ULARGE n;
	n.Lo=a.Lo+b;
	n.Hi=a.Hi;
	if (n.Lo < b) n.Hi++;
	return n;
  }

  inline ULARGE FASTCALL operator+(ULONG a,ULARGE &b) {
	ULARGE n;
	n.Lo=b.Lo+a;
	n.Hi=b.Hi;
	if (n.Lo < a) n.Hi++;
	return n;
  }

  inline ULARGE FASTCALL operator-(ULARGE &a,ULARGE &b) {
	ULARGE n;
	n.Lo=a.Lo-b.Lo;
	n.Hi=a.Hi-b.Hi;
	if (n.Lo > a.Lo) n.Hi--;
	return n;
  }

  inline ULARGE FASTCALL operator-(ULARGE &a,ULONG b) {
	ULARGE n;
	n.Lo=a.Lo-b;
	n.Hi=a.Hi;
	if (n.Lo > a.Lo) n.Hi--;
	return n;
  }

  inline ULARGE FASTCALL operator-(ULONG a,ULARGE &b) {
	ULARGE n;
	n.Lo=a-b.Lo;
	n.Hi=-b.Hi;
	if (n.Lo > a) n.Hi--;
	return n;
  }

  inline ULARGE FASTCALL operator*(ULARGE &a,ULARGE &b) {
	return ULARGE((DOUBLE)a.Lo*(DOUBLE)b.Lo+
			 (DOUBLE)ldexp((DOUBLE)a.Lo*(DOUBLE)b.Hi+(DOUBLE)a.Hi*(DOUBLE)b.Lo,32));
  }

  inline ULARGE FASTCALL operator*(ULARGE &a,ULONG b) {
	return ULARGE((DOUBLE)a.Lo*(DOUBLE)b+(DOUBLE)ldexp((DOUBLE)a.Hi*(DOUBLE)b,32));
  }

  inline ULARGE FASTCALL operator*(ULONG a,ULARGE &b) {
	return ULARGE((DOUBLE)b.Lo*(DOUBLE)a+(DOUBLE)ldexp((DOUBLE)b.Hi*(DOUBLE)a,32));
  }

  inline ULARGE FASTCALL operator/(ULARGE &a,ULARGE &b) {
	return ULARGE(((DOUBLE)a.Lo+(DOUBLE)ldexp(a.Hi,32))/((DOUBLE)b.Lo+(DOUBLE)ldexp(b.Hi,32)));
  }

  inline ULARGE FASTCALL operator/(ULARGE &a,ULONG b) {
	return ULARGE(((DOUBLE)a.Lo+(DOUBLE)ldexp(a.Hi,32))/((DOUBLE)b));
  }

  inline ULARGE FASTCALL operator/(ULONG a,ULARGE &b) {
	return ULARGE(((DOUBLE)a)/((DOUBLE)b.Lo+(DOUBLE)ldexp(b.Hi,32)));
  }

  inline ULARGE FASTCALL operator|(ULARGE &a,ULARGE &b) {
	ULARGE n;
	n.Lo=a.Lo|b.Lo;
	n.Hi=a.Hi|b.Hi;
	return n;
  }

  inline ULARGE FASTCALL operator|(ULARGE &a,ULONG b) {
	ULARGE n;
	n.Lo=a.Lo|b;
	n.Hi=a.Hi;
	return n;
  }

  inline ULARGE FASTCALL operator|(ULONG a,ULARGE &b) {
	ULARGE n;
	n.Lo=a|b.Lo;
	n.Hi=b.Hi;
	return n;
  }

  inline ULARGE FASTCALL operator&(ULARGE &a,ULARGE &b) {
	ULARGE n;
	n.Lo=a.Lo&b.Lo;
	n.Hi=a.Hi&b.Hi;
	return n;
  }

  inline ULARGE FASTCALL operator&(ULARGE &a,ULONG b) {
	ULARGE n;
	n.Lo=a.Lo&b;
	n.Hi=0;
	return n;
  }

  inline ULARGE FASTCALL operator&(ULONG a,ULARGE &b) {
	ULARGE n;
	n.Lo=a&b.Lo;
	n.Hi=0;
	return n;
  }

  inline ULARGE FASTCALL operator^(ULARGE &a,ULARGE &b) {
	ULARGE n;
	n.Lo=a.Lo^b.Lo;
	n.Hi=a.Hi^b.Hi;
	return n;
  }

  inline ULARGE FASTCALL operator^(ULARGE &a,ULONG b) {
	ULARGE n;
	n.Lo=a.Lo^b;
	n.Hi=a.Hi;
	return n;
  }

  inline ULARGE FASTCALL operator^(ULONG a,ULARGE &b)  {
	ULARGE n;
	n.Lo=a^b.Lo;
	n.Hi=b.Hi;
	return n;
  }

  inline ULARGE FASTCALL ULARGE::operator<<(INT s) {
	ULARGE n;
	n.Hi=(Hi<<s) | (Lo>>(32-s));
	n.Lo=Lo<<s;
	return n;
  }

  inline ULARGE FASTCALL ULARGE::operator>>(INT s) {
	ULARGE n;
	n.Lo=(Lo>>s) | (Hi<<(32-s));
	n.Hi=Hi>>s;
	return n;
  }

  inline ULARGE FASTCALL ULARGE::operator~() {
	ULARGE n;
	n.Lo=~Lo;
	n.Hi=~Hi;
	return n;
  }

  inline ULARGE& FASTCALL ULARGE::operator++() {
	if (++Lo==0) Hi++;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator--() {
	if (--Lo==(ULONG)(~0)) Hi--;
	return *this;
  }

  inline ULARGE FASTCALL ULARGE::operator-() {
	ULARGE n;
	n.Lo=~Lo;
	n.Hi=~Hi;
	if (++n.Lo==0) n.Hi++;
	return n;
  }

  inline ULARGE& FASTCALL ULARGE::operator+=(ULARGE &l) {
	Lo+=l.Lo;
	Hi+=l.Hi;
	if (Lo < l.Lo) Hi++;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator+=(ULONG l) {
	Lo+=(DWORD)l;
	if (Lo < l) Hi++;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator-=(ULARGE &l) {
	Lo-=l.Lo;
	Hi-=l.Hi;
	if (Lo > l.Lo) Hi--;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator-=(ULONG l) {
	Lo-=l;
	if (Lo > l) Hi--;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator*=(ULARGE &l) {
	return *this=ULARGE((DOUBLE)Lo*(DOUBLE)l.Lo+(DOUBLE)ldexp((DOUBLE)Lo*(DOUBLE)l.Hi+(DOUBLE)Hi*(DOUBLE)l.Lo,32));
  }

  inline ULARGE& FASTCALL ULARGE::operator*=(ULONG l) {
	return *this=ULARGE((DOUBLE)Lo*(DOUBLE)l+(DOUBLE)ldexp((DOUBLE)Hi*(DOUBLE)l,32));
  }

  inline ULARGE& FASTCALL ULARGE::operator/=(ULARGE &l) {
	return *this=ULARGE(((DOUBLE)Lo+(DOUBLE)ldexp(Hi,32))/((DOUBLE)l.Lo+(DOUBLE)ldexp(l.Hi,32)));
  }

  inline ULARGE& FASTCALL ULARGE::operator/=(ULONG l) {
	return *this=ULARGE(((DOUBLE)Lo+(DOUBLE)ldexp(Hi,32))/((DOUBLE)l));
  }

  inline ULARGE& FASTCALL ULARGE::operator|=(ULARGE &l) {
	Lo|=l.Lo;
	Hi|=l.Hi;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator&=(ULARGE &l) {
	Lo&=l.Lo;
	Hi&=l.Hi;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator^=(ULARGE &l) {
	Lo^=l.Lo;
	Hi^=l.Hi;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator|=(ULONG l) {
	Lo|=l;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator&=(ULONG l) {
	Lo&=l;
	Hi=0;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator^=(ULONG l) {
	Lo^=l;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator<<=(INT s) {
	Hi=(Hi<<s) | (Lo>>(32-s));
	Lo<<=s;
	return *this;
  }

  inline ULARGE& FASTCALL ULARGE::operator>>=(INT s) {
	Lo=(Lo>>s) | (Hi<<(32-s));
	Hi>>=s;
	return *this;
  }

  inline BOOL FASTCALL ULARGE::operator==(ULARGE &l) {
	return ((Lo==l.Lo) && (Hi==l.Hi));
  }

  inline BOOL FASTCALL ULARGE::operator==(ULONG l) {
	return ((Lo==l) && (Hi==0));
  }

  inline BOOL FASTCALL ULARGE::operator!=(ULARGE &l) {
	return !((Lo==l.Lo) && (Hi==l.Hi));
  }

  inline BOOL FASTCALL ULARGE::operator!=(ULONG l) {
	return ((Lo!=l) || (Hi!=0));
  }

  inline BOOL FASTCALL ULARGE::operator>(ULARGE &l) {
	return ((Hi>l.Hi) || ((Hi==l.Hi) && (Lo>l.Lo)));
  }

  inline BOOL FASTCALL ULARGE::operator>(ULONG l) {
	return ((Lo>l) || (Hi>0));
  }

  inline BOOL FASTCALL ULARGE::operator<(ULARGE &l) {
	return ((Hi<l.Hi) || ((Hi==l.Hi) && (Lo<l.Lo)));
  }

  inline BOOL FASTCALL ULARGE::operator<(ULONG l) {
	return ((Lo<l) || (Hi==0));
  }

  inline BOOL FASTCALL ULARGE::operator>=(ULARGE &l) {
	return ((Hi>l.Hi) || ((Hi==l.Hi) && (Lo>=l.Lo)));
  }

  inline BOOL FASTCALL ULARGE::operator>=(ULONG l) {
	return ((Lo>=l) || (Hi>0));
  }

  inline BOOL FASTCALL ULARGE::operator<=(ULARGE &l) {
	return ((Hi<l.Hi) || ((Hi==l.Hi) && (Lo<=l.Lo)));
  }

  inline BOOL FASTCALL ULARGE::operator<=(ULONG l) {
	return ((Lo<=l) && (Hi==0));
  }

  inline LARGE::LARGE() {
  	Lo=Hi=0;
  }

  inline LARGE::LARGE(LONG l) {
	Lo=(ULONG)l;
	if (l>=0)
		Hi=0;
	else
		Hi=~0;
  }

  inline LARGE::LARGE(DOUBLE a) {
	if (a<0){
		Hi=~(ULONG)ldexp(-a,-32);
		if ((Lo=(~(ULONG)(-a))+1)==0) Hi++;
	} else {
		Hi=(ULONG)ldexp(a,-32);
		Lo=(ULONG)a;
    }
  }

  inline DOUBLE FASTCALL LARGE::ToDouble(void) {
	ULONG a,b;
	if (((LONG)Hi)>=0)
		return (DOUBLE)((DOUBLE)Lo+(DOUBLE)ldexp((DOUBLE)Hi,32));
	b=~Hi;
	if((a=(~Lo)+1)==0) b++;
	return -((DOUBLE)a+(DOUBLE)ldexp((DOUBLE)b,32));
  }

  inline LONG FASTCALL LARGE2LONG(LARGE &l) {
	return (LONG)l.Lo;
  }

  inline LARGE FASTCALL operator+(LARGE &a,LARGE &b) {
	LARGE n;
	n.Lo=a.Lo+b.Lo;
	n.Hi=a.Hi+b.Hi;
	if ((ULONG)n.Lo < (ULONG)a.Lo) n.Hi++;
	return n;
  }

  inline LARGE FASTCALL operator+(LARGE &a,LONG b) {
	LARGE n;
	n.Lo=a.Lo+b;
	n.Hi=a.Hi;
	if ((ULONG)n.Lo < (ULONG)b) n.Hi++;
	return n;
  }

  inline LARGE FASTCALL operator+(LONG a,LARGE &b) {
	LARGE n;
	n.Lo=b.Lo+a;
	n.Hi=b.Hi;
	if ((ULONG)n.Lo < (ULONG)a) n.Hi++;
	return n;
  }

  inline LARGE FASTCALL operator-(LARGE &a,LARGE &b) {
	LARGE n;
	n.Lo=a.Lo-b.Lo;
	n.Hi=a.Hi-b.Hi;
	if ((ULONG)n.Lo > (ULONG)a.Lo) n.Hi--;
	return n;
  }

  inline LARGE FASTCALL operator-(LARGE &a,LONG b) {
	LARGE n;
	n.Lo=a.Lo-b;
	n.Hi=a.Hi;
	if ((ULONG)n.Lo > (ULONG)a.Lo) n.Hi--;
	return n;
  }

  inline LARGE FASTCALL operator-(LONG a,LARGE &b) {
	LARGE n;
	n.Lo=a-b.Lo;
	n.Hi=-b.Hi;
	if ((ULONG)n.Lo > (ULONG)a) n.Hi--;
	return n;
  }

  inline LARGE FASTCALL operator*(LARGE &a,LARGE &b) {
	return LARGE(a.ToDouble()*b.ToDouble());
  }

  inline LARGE FASTCALL operator*(LARGE &a,LONG b) {
	return LARGE(a.ToDouble()*(DOUBLE)b);
  }

  inline LARGE FASTCALL operator*(LONG a,LARGE &b) {
	return LARGE((DOUBLE)a*b.ToDouble());
  }

  inline LARGE FASTCALL operator/(LARGE &a,LARGE &b) {
	return LARGE(a.ToDouble()/b.ToDouble());
  }

  inline LARGE FASTCALL operator/(LARGE &a,LONG b) {
	return LARGE(a.ToDouble()/(DOUBLE)b);
  }

  inline LARGE FASTCALL operator/(LONG a,LARGE &b) {
	return LARGE((DOUBLE)a/b.ToDouble());
  }

  inline LARGE FASTCALL operator|(LARGE &a,LARGE &b) {
	LARGE n;
	n.Lo=a.Lo|b.Lo;
	n.Hi=a.Hi|b.Hi;
	return n;
  }

  inline LARGE FASTCALL operator|(LARGE &a,LONG b) {
	LARGE n;
	n.Lo=a.Lo|b;
	n.Hi=a.Hi;
	return n;
  }

  inline LARGE FASTCALL operator|(LONG a,LARGE &b) {
	LARGE n;
	n.Lo=a|b.Lo;
	n.Hi=b.Hi;
	return n;
  }

  inline LARGE FASTCALL operator&(LARGE &a,LARGE &b) {
	LARGE n;
	n.Lo=a.Lo&b.Lo;
	n.Hi=a.Hi&b.Hi;
	return n;
  }

  inline LARGE FASTCALL operator&(LARGE &a,LONG b) {
	LARGE n;
	n.Lo=a.Lo&b;
	n.Hi=0;
	return n;
  }

  inline LARGE FASTCALL operator&(LONG a,LARGE &b) {
	LARGE n;
	n.Lo=a&b.Lo;
	n.Hi=0;
	return n;
  }

  inline LARGE FASTCALL operator^(LARGE &a,LARGE &b) {
	LARGE n;
	n.Lo=a.Lo^b.Lo;
	n.Hi=a.Hi^b.Hi;
	return n;
  }

  inline LARGE FASTCALL operator^(LARGE &a,LONG b) {
	LARGE n;
	n.Lo=a.Lo^b;
	n.Hi=a.Hi;
	return n;
  }

  inline LARGE FASTCALL operator^(LONG a,LARGE &b) {
	LARGE n;
	n.Lo=a^b.Lo;
	n.Hi=b.Hi;
	return n;
  }

  inline LARGE FASTCALL LARGE::operator<<(INT s) {
	LARGE n;
	n.Hi=(Hi<<s) | (Lo>>(32-s));
	n.Lo=Lo<<s;
	return n;
  }

  inline LARGE FASTCALL LARGE::operator>>(INT s) {
	LARGE n;
	n.Lo=(Lo>>s) | (Hi<<(32-s));
	(LONG)n.Hi=((LONG)Hi)>>s;
	return n;
  }

  inline LARGE FASTCALL LARGE::operator~() {
	LARGE n;
	n.Lo=~Lo;
	n.Hi=~Hi;
	return n;
  }

  inline LARGE& FASTCALL LARGE::operator++() {
	if (++Lo==0) Hi++;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator--() {
	if ((ULONG)(--Lo)==(ULONG)~0) Hi--;
	return *this;
  }

  inline LARGE FASTCALL LARGE::operator-() {
	LARGE n;
	n.Lo=~Lo;
	n.Hi=~Hi;
	if (++n.Lo==0) n.Hi++;
	return n;
  }

  inline LARGE& FASTCALL LARGE::operator+=(LARGE &l) {
	Lo+=l.Lo;
	Hi+=l.Hi;
	if ((ULONG)Lo < (ULONG)l.Lo) Hi++;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator+=(LONG l) {
	Lo+=l;
	if ((ULONG)Lo < (ULONG)l) Hi++;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator-=(LARGE &l) {
	Lo-=l.Lo;
	Hi-=l.Hi;
	if ((ULONG)Lo > (ULONG)l.Lo) Hi--;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator-=(LONG l) {
	Lo-=(DWORD)l;
	if ((ULONG)Lo > (ULONG)l) Hi--;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator*=(LARGE &l) {
	return *this=LARGE(ToDouble()*l.ToDouble());
  }

  inline LARGE& FASTCALL LARGE::operator*=(LONG l) {
	return *this=LARGE(ToDouble()*(DOUBLE)l);
  }

  inline LARGE& FASTCALL LARGE::operator/=(LARGE &l) {
	return *this=LARGE(ToDouble()/l.ToDouble());
  }

  inline LARGE& FASTCALL LARGE::operator/=(LONG l) {
	return *this=LARGE(ToDouble()/(DOUBLE)l);
  }

  inline LARGE& FASTCALL LARGE::operator|=(LARGE &l) {
	Lo|=l.Lo;
	Hi|=l.Hi;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator&=(LARGE &l) {
	Lo&=l.Lo;
	Hi&=l.Hi;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator^=(LARGE &l) {
	Lo^=l.Lo;
	Hi^=l.Hi;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator|=(LONG l) {
	Lo|=l;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator&=(LONG l) {
	Lo&=l;
	Hi=0;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator^=(LONG l) {
	Lo^=l;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator<<=(INT s) {
	Hi=(Hi<<s) | (Lo>>(32-s));
	Lo<<=s;
	return *this;
  }

  inline LARGE& FASTCALL LARGE::operator>>=(INT s) {
	Lo=(Lo>>s) | (Hi<<(32-s));
	(LONG)Hi>>=s;
	return *this;
  }

  inline BOOL FASTCALL LARGE::operator==(LARGE &l) {
	return ((Lo==l.Lo) && (Hi==l.Hi));
  }

  inline BOOL FASTCALL LARGE::operator==(LONG l) {
	return ((Lo==l) && (((l>=0) && (Hi==0)) ||
						((l<0) && ((ULONG)Hi==(ULONG)~0))));
  }

  inline BOOL FASTCALL LARGE::operator!=(LARGE &l) {
	return !((Lo==l.Lo) && (Hi==l.Hi));
  }

  inline BOOL FASTCALL LARGE::operator!=(LONG l) {
	return ((Lo!=l) || ((l>=0) && (Hi!=0))
					|| ((l<0)  && ((ULONG)Hi!=(ULONG)~0)));
  }

  inline BOOL FASTCALL LARGE::operator>(LARGE &l) {
	return (ToDouble() > l.ToDouble());
  }

  inline BOOL FASTCALL LARGE::operator>(LONG l) {
	return (ToDouble() > (DOUBLE)l);
  }

  inline BOOL FASTCALL LARGE::operator<(LARGE &l) {
	return (ToDouble() < l.ToDouble());
  }

  inline BOOL FASTCALL LARGE::operator<(LONG l) {
	return (ToDouble() < (DOUBLE)l);
  }

  inline BOOL FASTCALL LARGE::operator>=(LARGE &l) {
	return (ToDouble() >= l.ToDouble());
  }

  inline BOOL FASTCALL LARGE::operator>=(LONG l) {
	return (ToDouble() >= (DOUBLE)l);
  }

  inline BOOL FASTCALL LARGE::operator<=(LARGE &l) {
	return (ToDouble() <= l.ToDouble());
  }

  inline BOOL FASTCALL LARGE::operator<=(LONG l) {
	return (ToDouble() <= (DOUBLE)l);
  }

#else
  //
  // Here we simulate a LARGE number by using floating point operations.
  // Not the best, but what the heck. Remember, if you want to do
  // logical operations, just define __LARGEINT__
  //
  #if defined(__FORBORLAND__)
   typedef long double LARGE;
   typedef long double ULARGE;
  #else
   typedef double LARGE;
   typedef double ULARGE;
  #endif

  #define LargeMul(a, b) 		(((LARGE)(LONG)(a)) * ((LARGE)(LONG)(b)))
  #define LargeShl(a, b) 		((LARGE)ldexp((LARGE)(a),(INT)(b)))
  #define LargeShr(a, b) 		((LARGE)ldexp((LARGE)(a),-(INT)(b)))
  #define LargeShlLong(a, b)    ((LONG)ldexp((LARGE)(a),(b))
  #define LargeShrLong(a, b)    ((LONG)ldexp((LARGE)(a),(b))
  #define ULargeMul(a, b) 		(((ULARGE)(ULONG)(a)) * ((ULARGE)(ULONG)(b)))
  #define ULargeShl(a, b) 		((ULARGE)ldexp((ULARGE)(a),(INT)(b)))
  #define ULargeShr(a, b) 		((ULARGE)ldexp((ULARGE)(a),-(INT)(b)))
  #define ULargeShlLong(a, b)   ((ULONG)ldexp((ULARGE)(a),(b))
  #define ULargeShrLong(a, b)   ((ULONG)ldexp((ULARGE)(a),(b))
  #define MulDiv(a,b,c)		   	((LONG)(((LARGE)(a)*(LARGE)(b))/(LARGE)(c)))
  #define MulShr(a,b,c)			((LONG)ldexp((LARGE)(a)*(LARGE)(b),(INT)(c)))
  #define LARGE2LONG(a)			((LONG)(LARGE)(a))
  #define ULARGE2ULONG(a)		((ULONG)(ULARGE)(a))

#endif


#define MINCHAR     			0x80
#define MAXCHAR     			0x7f
#define MINSHORT    			0x8000
#define MAXSHORT    			0x7fff
#define MINLONG     			0x80000000
#define MAXLONG     			0x7fffffff
#define MAXBYTE     			0xff
#define MAXWORD     			0xffff
#define MAXDWORD    			0xffffffff


//
// These macros make it easy to port between fixed-maths systems and
// floating point systems.
//
#if defined __FIXEDMATH__
 typedef signed  long	MATH;
 #define MATH2LONG(value,prec)			((LONG)(((MATH)(value))>>(prec)))
 #define MATH2DOUBLE(value,prec)		((DOUBLE)((DOUBLE)(value)/(DOUBLE)(1 << (prec))))
 #define DOUBLE2MATH(value,prec)		((MATH)((DOUBLE)(value)*(DOUBLE)(1 << (prec))))
 #define LONG2MATH(value,prec)			((MATH)(((LONG)(value))<<(prec)))
 #define FIXED2MATH(value,prec)			((MATH)(value))
 #define MATH2FIXED(value,prec)			((LONG)(value))
 #define MATHSHIFTL(value,prec)			((MATH)((value)<<(prec)))
 #define MATHSHIFTR(value,prec)			((MATH)((value)>>(prec)))
 #define MATHMUL(value,prec)			((value)*(1<<(prec)))
 #define MATHDIV(value,prec)			((value)/(1<<(prec)))
#elif defined __FLOATMATH__
 typedef         double MATH;
 #define MATH2LONG(value,prec)			((LONG)(value))
 #define LONG2MATH(value,prec)			((MATH)(value))
 #define MATH2DOUBLE(value,prec)		((DOUBLE)(value))
 #define DOUBLE2MATH(value,prec)		((MATH)(value))
 #define FIXED2MATH(value,prec)			((DOUBLE)(value)/(DOUBLE)(1<<(prec)))
 #define MATH2FIXED(value,prec)			((LONG)((DOUBLE)(value)*(DOUBLE)(1<<(prec))))
 #define MATHSHIFTL(value,prec)			((MATH)(value))
 #define MATHSHIFTR(value,prec)			((MATH)(value))
 #define MATHMUL(value,prec)			(value)
 #define MATHDIV(value,prec)			(value)
#endif
#define  FIXED2DOUBLE(value,prec)		((DOUBLE)(value)/(DOUBLE)(1<<(prec)))
#define  DOUBLE2FIXED(value,prec)		((LONG)((DOUBLE)(value)*(DOUBLE)(1<<(prec))))




//
// Simple macros for getting low/high-byte/words
//
#ifndef LOBYTE
 #define LOBYTE(w)           	((BYTE)(w))
 #define HIBYTE(w)           	((BYTE)((UINT)(w) >> 8))

 #define LOWORD(l)           	((WORD)(l))
 #define HIWORD(l)           	((WORD)((DWORD)(l) >> 16))

 #define MAKELONG(low, high) 	((LONG)(((WORD)(low)) | (((DWORD)((WORD)(high))) << 16)))
#endif

//
// The following functions are used for fast conversion between
// little/big-endian formats.
//
VOID FASTCALL SwapWord(WORD *);
VOID FASTCALL SwapDWord(DWORD *);
VOID FASTCALL SwapQWord(QWORD *);


inline VOID FASTCALL SwapWord(WORD *w) {
 	*w=(WORD)(((WORD)(*(BYTE *)w))<<8) | (WORD)((BYTE)(*w>>8));
}


inline VOID FASTCALL SwapDWord(DWORD *d) {
 	WORD w1,w2;

	SwapWord((WORD *)d);
 	SwapWord(((WORD *)d)+1);

 	*d=(((DWORD)(*(WORD *)d))<<16) | (DWORD)((WORD)(*d>>16));
}


inline VOID FASTCALL SwapQWord(QWORD *q) {
 	DWORD d1,d2;

 	d1=*((DWORD *)q);
 	d2=*(((DWORD *)q)+1);

 	SwapDWord(&d1);
 	SwapDWord(&d2);

 	*(DWORD *)q=d2;
 	*(((DWORD *)q)+1)=d1;
}


//
// The following routines are used to convert between big endian and
// little endian formats.
//
#if defined(__MSBFIRST__)
 #define MSB2WORD(a)			((WORD)(a))
 #define MSB2DWORD(a)			((DWORD)(a))
 #define MSB2QWORD(a)			((QWORD)(a))

 #define WORD2MSB(a)			((WORD)(a))
 #define DWORD2MSB(a)			((DWORD)(a))
 #define QWORD2MSB(a)			((QWORD)(a))

 inline WORD FASTCALL LSB2WORD(WORD a) {
	return (WORD)((a>>8) | ((WORD)(BYTE)a));
 }

 inline DWORD FASTCALL LSB2DWORD(DWORD a) {
	return ((DWORD)(BYTE)(a>>24)) | (((DWORD)(BYTE)(a>>16))<<8)
		| (((DWORD)(BYTE)(a>>8))<<16) | (((DWORD)(BYTE)(a))<<24);
 }

 inline QWORD FASTCALL LSB2QWORD(QWORD a) {
	#if defined(__64BIT__)
	 return (((QWORD)(BYTE)(a>>56))) | (((QWORD)(BYTE)(a>>48))<<8)
		  | (((QWORD)(BYTE)(a>>40))<<16) | (((QWORD)(BYTE)(a>>32))<<24)
		  | (((QWORD)(BYTE)(a>>24))<<32) | (((QWORD)(BYTE)(a>>16))<<40)
		  | (((QWORD)(BYTE)(a>>8))<<48) | (((QWORD)(BYTE)(a))<<56);
	#else
	 QWORD q;
	 q.Hi=LSB2DWORD(a.Lo);
	 q.Lo=LSB2DWORD(a.Hi);
	 return q;
	#endif
 }

 inline WORD FASTCALL WORD2LSB(WORD a) {
	return (WORD)((a>>8) | ((WORD)(BYTE)a));
 }

 inline DWORD FASTCALL DWORD2LSB(DWORD a) {
	return ((DWORD)(BYTE)(a>>24)) | (((DWORD)(BYTE)(a>>16))<<8)
		| (((DWORD)(BYTE)(a>>8))<<16) | (((DWORD)(BYTE)(a))<<24);
 }

 inline QWORD FASTCALL QWORD2LSB(QWORD a) {
	#if defined(__64BIT__)
	 return (((QWORD)(BYTE)(a>>56))) | (((QWORD)(BYTE)(a>>48))<<8)
		  | (((QWORD)(BYTE)(a>>40))<<16) | (((QWORD)(BYTE)(a>>32))<<24)
		  | (((QWORD)(BYTE)(a>>24))<<32) | (((QWORD)(BYTE)(a>>16))<<40)
		  | (((QWORD)(BYTE)(a>>8))<<48) | (((QWORD)(BYTE)(a))<<56);
	#else
	 QWORD q;
	 q.Hi=DWORD2LSB(a.Lo);
	 q.Lo=DWORD2LSB(a.Hi);
	 return q;
	#endif
 }

#elif defined(__LSBFIRST__)
 #define LSB2WORD(a)			((WORD)(a))
 #define LSB2DWORD(a)			((DWORD)(a))
 #define LSB2QWORD(a)			((QWORD)(a))

 #define WORD2LSB(a)			((WORD)(a))
 #define DWORD2LSB(a)			((DWORD)(a))
 #define QWORD2LSB(a)			((QWORD)(a))

 inline WORD FASTCALL MSB2WORD(WORD a) {
	return (WORD)((a>>8) | ((WORD)(BYTE)a));
 }

 inline DWORD FASTCALL MSB2DWORD(DWORD a) {
	return ((DWORD)(BYTE)(a>>24)) | (((DWORD)(BYTE)(a>>16))<<8)
		| (((DWORD)(BYTE)(a>>8))<<16) | (((DWORD)(BYTE)(a))<<24);
 }

 inline QWORD FASTCALL MSB2QWORD(QWORD a) {
	#if defined(__NATIVEQWORD__)
	 return (((QWORD)(BYTE)(a>>56))) | (((QWORD)(BYTE)(a>>48))<<8)
		  | (((QWORD)(BYTE)(a>>40))<<16) | (((QWORD)(BYTE)(a>>32))<<24)
		  | (((QWORD)(BYTE)(a>>24))<<32) | (((QWORD)(BYTE)(a>>16))<<40)
		  | (((QWORD)(BYTE)(a>>8))<<48) | (((QWORD)(BYTE)(a))<<56);
	#else
	 QWORD q;
	 q.Hi=MSB2DWORD(a.Lo);
	 q.Lo=MSB2DWORD(a.Hi);
	 return q;
	#endif
 }

 inline WORD FASTCALL WORD2MSB(WORD a) {
	return (WORD)((a>>8) | ((WORD)(BYTE)a));
 }

 inline DWORD FASTCALL DWORD2MSB(DWORD a) {
	return ((DWORD)(BYTE)(a>>24)) | (((DWORD)(BYTE)(a>>16))<<8)
		| (((DWORD)(BYTE)(a>>8))<<16) | (((DWORD)(BYTE)(a))<<24);
 }

 inline QWORD FASTCALL QWORD2MSB(QWORD a) {
	#if defined(__NATIVEQWORD__)
	 return (((QWORD)(BYTE)(a>>56))) | (((QWORD)(BYTE)(a>>48))<<8)
		  | (((QWORD)(BYTE)(a>>40))<<16) | (((QWORD)(BYTE)(a>>32))<<24)
		  | (((QWORD)(BYTE)(a>>24))<<32) | (((QWORD)(BYTE)(a>>16))<<40)
		  | (((QWORD)(BYTE)(a>>8))<<48) | (((QWORD)(BYTE)(a))<<56);
	#else
	 QWORD q;
	 q.Hi=DWORD2MSB(a.Lo);
	 q.Lo=DWORD2MSB(a.Hi);
	 return q;
	#endif
 }


#else
 #define LSB2WORD(a)			((WORD)(a))
 #define LSB2DWORD(a)			((DWORD)(a))
 #define LSB2QWORD(a)			((QWORD)(a))
 #define MSB2WORD(a)			((WORD)(a))
 #define MSB2DWORD(a)           ((DWORD)(a))
 #define MSB2QWORD(a)           ((QWORD)(a))

 #define WORD2LSB(a)			((WORD)(a))
 #define DWORD2LSB(a)			((DWORD)(a))
 #define QWORD2LSB(a)			((QWORD)(a))
 #define WORD2MSB(a)			((WORD)(a))
 #define DWORD2MSB(a)           ((DWORD)(a))
 #define QWORD2MSB(a)           ((QWORD)(a))
#endif


//
// These macros are used to convert between the various floating point
// formats. Please note, that byte ordering is NOT accomplished by these
// macros, so you may have to call SwapDWord first.
//
#if defined (__FPIEEE__)
 #define FLOAT2IEEE(a)			((FLOAT)(a))
 #define FLOAT2IBM(a)			((FLOAT)(a))
 #define FLOAT2VAX(a)			((FLOAT)(a))
 #define FLOAT2CRAY(a)			((FLOAT)(a))

 #define IEEE2FLOAT(a)			((FLOAT)(a))
 #define VAX2FLOAT(a)			((FLOAT)(a))
 #define IBM2FLOAT(a)			((FLOAT)(a))
 #define CRAY2FLOAT(a)			((FLOAT)(a))
#elif defined (__FPIBM__)
 #define FLOAT2IEEE(a)			((FLOAT)(a))
 #define FLOAT2IBM(a)			((FLOAT)(a))
 #define FLOAT2VAX(a)			((FLOAT)(a))
 #define FLOAT2CRAY(a)			((FLOAT)(a))

 #define IEEE2FLOAT(a)			((FLOAT)(a))
 #define VAX2FLOAT(a)			((FLOAT)(a))
 #define IBM2FLOAT(a)			((FLOAT)(a))
 #define CRAY2FLOAT(a)			((FLOAT)(a))
#elif defined (__FPVAX__)
 #define FLOAT2IEEE(a)			((FLOAT)(a))
 #define FLOAT2IBM(a)			((FLOAT)(a))
 #define FLOAT2VAX(a)			((FLOAT)(a))
 #define FLOAT2CRAY(a)			((FLOAT)(a))

 #define IEEE2FLOAT(a)			((FLOAT)(a))
 #define VAX2FLOAT(a)			((FLOAT)(a))
 #define IBM2FLOAT(a)			((FLOAT)(a))
 #define CRAY2FLOAT(a)			((FLOAT)(a))
#elif defined (__FPCRAY__)
 #define FLOAT2IEEE(a)			((FLOAT)(a))
 #define FLOAT2IBM(a)			((FLOAT)(a))
 #define FLOAT2VAX(a)			((FLOAT)(a))
 #define FLOAT2CRAY(a)			((FLOAT)(a))

 #define IEEE2FLOAT(a)			((FLOAT)(a))
 #define VAX2FLOAT(a)			((FLOAT)(a))
 #define IBM2FLOAT(a)			((FLOAT)(a))
 #define CRAY2FLOAT(a)			((FLOAT)(a))
#endif


#ifdef MAKETYPE
 #undef MAKETYPE
#endif
#define MAKETYPE(ch0,ch1,ch2,ch3)	(((DWORD)(BYTE)(ch0)) | \
									 ((DWORD)(BYTE)(ch1)<<8) | \
									 ((DWORD)(BYTE)(ch2)<<16) | \
									 ((DWORD)(BYTE)(ch3)<<24))
#ifndef max
 #define max(a,b)            	(((a) > (b)) ? (a) : (b))
#endif
#ifndef min
 #define min(a,b)            	(((a) < (b)) ? (a) : (b))
#endif


#ifndef TRUE
 #define TRUE					1
 #define FALSE					0
#endif

#ifndef NULL
 #define NULL					0L
#endif

#endif
