///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#ifndef __GVE3DS_H
#define __GVE3DS_H

#include "../include/gvepro.h"


class G3dsFile {
		private:
		enum   {
			CHUNK_RGBF      					= 0x0010,
			CHUNK_RGBB      					= 0x0011,
			CHUNK_RBGB2     					= 0x0012,

			CHUNK_PRJ       					= 0xC23D,
			CHUNK_MLI       					= 0x3DAA,

			CHUNK_MAIN      					= 0x4D4D,
				CHUNK_OBJMESH    				= 0x3D3D,
				CHUNK_BKGCOLOR  				= 0x1200,
				CHUNK_AMBCOLOR  				= 0x2100,
				CHUNK_OBJBLOCK  				= 0x4000,
				CHUNK_TRIMESH   				= 0x4100,
					CHUNK_VERTLIST  			= 0x4110,
					CHUNK_FACELIST  			= 0x4120,
					CHUNK_FACEMAT   			= 0x4130,
					CHUNK_MAPLIST   			= 0x4140,
								CHUNK_SMOOLIST  = 0x4150,
								CHUNK_TRMATRIX  = 0x4160,
						CHUNK_LIGHT     		= 0x4600,
								CHUNK_SPOTLIGHT = 0x4610,
						CHUNK_CAMERA    		= 0x4700,
						CHUNK_HIERARCHY 		= 0x4F00,
				CHUNK_VIEWPORT  				= 0x7001,
					CHUNK_MATERIAL  			= 0xAFFF,
							CHUNK_MATNAME   	= 0xA000,
							CHUNK_AMBIENT   	= 0xA010,
							CHUNK_DIFFUSE   	= 0xA020,
						CHUNK_SPECULAR  		= 0xA030,
				CHUNK_TEXTURE   				= 0xA200,
						CHUNK_BUMPMAP   		= 0xA230,
						CHUNK_MAPFILE   		= 0xA300,
				   CHUNK_KEYFRAMER 				= 0xB000,
						CHUNK_AMBIENTKEY    	= 0xB001,
				CHUNK_TRACKINFO 				= 0xB002,
							CHUNK_TRACKOBJNAME  = 0xB010,
							CHUNK_TRACKPIVOT    = 0xB013,
							CHUNK_TRACKPOS      = 0xB020,
							CHUNK_TRACKROTATE   = 0xB021,
							CHUNK_TRACKSCALE    = 0xB022,
					CHUNK_OBJNUMBER     		= 0xB030,
							CHUNK_TRACKCAMERA 	= 0xB003,
							CHUNK_TRACKFOV  	= 0xB023,
							CHUNK_TRACKROLL 	= 0xB024,
						CHUNK_TRACKCAMTGT 		= 0xB004,
						CHUNK_TRACKLIGHT  		= 0xB005,
						CHUNK_TRACKLIGTGT 		= 0xB006,
						CHUNK_TRACKSPOTL  		= 0xB007,
						CHUNK_FRAMES    		= 0xB008
			   };

		#pragma option -a2
		typedef struct {
			WORD    id;
			DWORD   len;
		} ChunkHeader;
		#pragma option -a.

		private:
		int				Handle;
		BOOL			Verbose;
		CHAR			ObjectName[64];
		GVectorObject	*Head;
		GLightObject	*CurrentLight;
		GCameraObject	*CurrentCamera;
		GMeshObject		*CurrentMesh;
		GVertexList 	*CurrentVertexList;
		GFaceList   	*CurrentFaceList;

		public:
		GVectorObject*	FASTCALL Import(char *,BOOL v=FALSE);

		private:
		INT 			FASTCALL ASCIIZReader(CHAR *);
		BOOL 			FASTCALL ChunkReader(LONG p);
		BOOL 			FASTCALL RGBFReader(LONG p);
		BOOL 			FASTCALL RGBBReader(LONG p);
		BOOL 			FASTCALL ObjBlockReader(LONG p);
		BOOL 			FASTCALL TriMeshReader(LONG p);
		BOOL 			FASTCALL VertListReader(LONG p);
		BOOL 			FASTCALL FaceListReader(LONG p);
		BOOL 			FASTCALL FaceMatReader(LONG p);
		BOOL 			FASTCALL MapListReader(LONG p);
		BOOL 			FASTCALL SmooListReader(LONG p);
		BOOL 			FASTCALL TrMatrixReader(LONG p);
		BOOL 			FASTCALL LightReader(LONG p);
		BOOL 			FASTCALL SpotLightReader(LONG p);
		BOOL 			FASTCALL CameraReader(LONG p);
	   };

#endif
