///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//

#ifndef __GVEMESH_H
#define __GVEMESH_H

#include "../include/gvepro.h"


#define GVE_OBJ_BACKCULLING			((DWORD)0x00000001L)
#define GVE_OBJ_NOSORT				((DWORD)0x00000002L)
#define GVE_OBJ_MESH				((DWORD)0x00000004L)
#define GVE_OBJ_VERTEX_INDEX		((DWORD)0x00000008L)
#define GVE_OBJ_VERTEX_ADDRESS		((DWORD)0x00000010L)
#define GVE_OBJ_VERTEXBOX			((DWORD)0x00000020L)
#define GVE_OBJ_CHUNKBOX			((DWORD)0x00000040L)
#define GVE_OBJ_NORMALS				((DWORD)0x00000080L)
#define GVE_OBJ_VERTEX_SHADING		((DWORD)0x00000100L)
#define GVE_OBJ_HIDE				((DWORD)0x00000200L)
#define GVE_OBJ_NEWPOS				((DWORD)0x00000400L)
#define GVE_OBJ_NEWROT				((DWORD)0x00000800L)

#define GVE_TMAP_U_WRAP				0x04000
#define GVE_TMAP_V_WRAP             0x08000


class GMapping {
		public:
		LONG	u;
		LONG	v;

		public:
		GMapping() {};
		GMapping(LONG a,LONG b) {u=a; v=b;};
		~GMapping() {};
		GMapping	FASTCALL operator+(GMapping &m) {return GMapping(u+m.u,v+m.v);};
		GMapping	FASTCALL operator-(GMapping &m) {return GMapping(u-m.u,v-m.v);};
		GMapping& 	FASTCALL operator+=(GMapping &m) {u+=m.u; v+=m.v; return *this;};
		GMapping& 	FASTCALL operator-=(GMapping &m) {u-=m.u; v-=m.v; return *this;};
		BOOL 		FASTCALL operator==(GMapping &m) {return ((u==m.u) && (v==m.v));};
		BOOL 		FASTCALL operator!=(GMapping &m) {return ((u!=m.u) || (v!=m.v));};
	   };

typedef struct {
		GVector		Position;
		GNormal		Normal;
		GVector		TPosition;
		GPointL   	SPosition;
		GNormal		TNormal;
		LONG		Shading;
		LONG		InverseZ;
		GMapping	Chrome;
		GMapping	Texture;
	   } GVertex;

#pragma option -a1
typedef struct {
		WORD		Size;
		BYTE		Type;
		BYTE		VertexCount;
		BYTE		BumpMap;
		BYTE		LightMap;
		WORD   		TextureMap;
		LONG  		Normal[3];
		LONG  		Plane;
		GVertex*   	Vertex[];
	   } GFace;

typedef struct {
		WORD		Size;
		BYTE		Type;
		BYTE		VertexCount;
		BYTE		BumpMap;
		BYTE		LightMap;
		WORD   		TextureMap;
		LONG  		Normal[3];
		LONG		Plane;
		GVertex*	Vertex[3];
	   } GTriangle;
#pragma option -a.

typedef struct {
		INT			Count;
		GVertex  	Vertex[];
	   } GVertexList;

typedef struct {
		INT			Count;
		GFace   	Face[];
	   } GFaceList;


class GVectorObject:public GObjectKnot {
		friend class 	GVectorSystem;
		friend class 	G3dsFile;
		friend class 	G3doFile;

		public:
		#pragma option -a1
		typedef struct {
				DOUBLE 	   	x,y,z;
			   } FileVector;
		typedef struct {
				CHAR		Name[GMZ_NAME_LENGTH];
				DWORD		Flags;
				DWORD		Type;
				DWORD		Magic;
				FileVector	Position;
				FileVector	Rotation;
			   } FileContainerHead;
		#pragma option -a.

		protected:
		GVector			Rotation;
		GVector			Position;

		GMatrix			LocalMatrix;
		GMatrix			FinalMatrix;

		public:
		GVectorObject(CHAR *name,DWORD typ=GObject::EMPTY_TYPE);
		GVectorObject(GVectorObject *f=NULL,DWORD t=GObject::EMPTY_TYPE);
		~GVectorObject();

        CHAR*			FASTCALL SetDefaultName();

		GVector			FASTCALL SetPosition(GVector &);
		GVector			FASTCALL SetPosition(DOUBLE ,DOUBLE ,DOUBLE );
		GVector 	   	FASTCALL GetPosition(void);
		GVector			FASTCALL GetPosition(DOUBLE *,DOUBLE *,DOUBLE *);
		GVector			FASTCALL SetRotation(GVector &);
		GVector			FASTCALL SetRotation(DOUBLE ,DOUBLE ,DOUBLE );
		GVector 	   	FASTCALL GetRotation(void);
		GVector			FASTCALL GetRotation(DOUBLE *,DOUBLE *,DOUBLE *);
		BOOL 			FASTCALL RealizePosition(BOOL f=FALSE);

		INT     		FASTCALL GetFaceCount(BOOL subobjects=TRUE);
		INT     		FASTCALL GetVertexCount(BOOL subobjects=TRUE);
		INT     		FASTCALL GetObjectCount(BOOL subobjects=TRUE);
		CHAR*			FASTCALL GetFlagsName(CHAR *);

		BOOL			FASTCALL Write(GFile &,DWORD );
		BOOL          	FASTCALL Read(GFile &);
	   };

class GCameraObject:public GVectorObject {
		friend class GVectorSystem;
		friend class GVectorObject;
		friend class G3dsFile;
		friend class G3doFile;

        public:
        enum {
        	CAMERA_TYPE = MAKETYPE('C','A','M',' ')
        };

		public:
		#pragma option -a1
		typedef struct {
				CHAR		Name[GMZ_NAME_LENGTH];
				DWORD		Flags;
				DWORD		Type;
				DWORD		Magic;
				FileVector	Position;
				FileVector	Rotation;
				DOUBLE		Zoom;
				DOUBLE		DistanceX;
				DOUBLE		DistanceY;
				DOUBLE		NearPlane;
				DOUBLE		FarPlane;
				LONG		OffsetX;
				LONG		OffsetY;
			   } FileCameraHead;
		#pragma option -a.

		protected:
		MATH			Zoom;
		MATH			NearPlane;
		MATH			FarPlane;
		GPointL    		Offset;
		GPointM			Distance;

		public:
		GCameraObject(CHAR *name,DWORD typ=GCameraObject::CAMERA_TYPE);
		GCameraObject(GVectorObject *f=NULL,DWORD typ=GCameraObject::CAMERA_TYPE);
		~GCameraObject();

        CHAR*			FASTCALL SetDefaultName();

		GPointD			FASTCALL SetDistance(DOUBLE ,DOUBLE );
		GPointD			FASTCALL SetDistance(GPointD& );
		GPointL			FASTCALL SetOffset(LONG ,LONG );
		GPointL			FASTCALL SetOffset(GPointL& );
		DOUBLE  		FASTCALL SetNearPlane(DOUBLE );
		DOUBLE  		FASTCALL SetFarPlane(DOUBLE );
		DOUBLE			FASTCALL SetZoom(DOUBLE );
		GPointD			FASTCALL GetDistance(void);
		GPointL			FASTCALL GetOffset(void);
		DOUBLE  		FASTCALL GetNearPlane(void);
		DOUBLE  		FASTCALL GetFarPlane(void);
		DOUBLE			FASTCALL GetZoom(void);

		BOOL			FASTCALL Write(GFile &,DWORD );
		BOOL          	FASTCALL Read(GFile &);
	   };


class GMeshObject:public GVectorObject {
		friend class GVectorSystem;
		friend class GVectorObject;
		friend class G3dsFile;
		friend class G3doFile;

        public:
        enum {
        	MESH_TYPE = MAKETYPE('M','E','S','H'),
            VERTICES_TYPE = MAKETYPE('V','E','R','T'),
            FACES_TYPE = MAKETYPE('F','A','C','E')
        };

		public:
		typedef struct {
				MATH	minx,miny,minz;
				MATH	maxx,maxy,maxz;
			   } GBoundingBox;

		#pragma option -a1
		typedef struct {
				LONG	   	u;
				LONG	   	v;
			   } FileMapping;
		typedef struct {
				DOUBLE  	minx,miny,minz;
				DOUBLE  	maxx,maxy,maxz;
			   } FileBoundingBox;
		typedef struct {
				FileVector 	Position;
				FileVector	Normal;
				FileMapping	Texture;
			   } FileVertex;
		typedef struct {
				WORD		Size;
				BYTE		Type;
				BYTE		VertexCount;
				BYTE		BumpMap;
				BYTE		LightMap;
				WORD		TextureMap;
				LONG  		Normal[3];
				LONG  		Plane;
				WORD		Vertex[];
			   } FileFace;
		typedef struct {
				WORD		Size;
				BYTE		Type;
				BYTE		VertexCount;
				BYTE		BumpMap;
				BYTE		LightMap;
				WORD		TextureMap;
				LONG  		Normal[3];
				LONG		Plane;
				WORD		Vertex[3];
			   } FileTriangle;
		typedef struct {
				DWORD		Count;
				DWORD		Flags;
			   } FileVertexHead;
		typedef struct {
				DWORD		Count;
				DWORD		Flags;
			   } FileFaceHead;
		typedef struct {
				CHAR		Name[GMZ_NAME_LENGTH];
				DWORD		Flags;
				DWORD		Type;
				DWORD		Magic;
				FileVector	Position;
				FileVector	Rotation;
				FileBoundingBox BoundingBox;
			   } FileMeshHead;
		#pragma option -a.


		protected:
		GBoundingBox 	BoundingThis;
		GBoundingBox   	BoundingAll;

		GVertexList 	*VertexList;
		GFaceList 		*FaceList;

		INT				MeshWidth;
		INT				MeshHeight;

		public:
		GMeshObject(CHAR *name,DWORD typ=GMeshObject::MESH_TYPE);
		GMeshObject(GVectorObject *f=NULL,DWORD typ=GMeshObject::MESH_TYPE);
		~GMeshObject();

        CHAR*			FASTCALL SetDefaultName();

		BOOL 			FASTCALL ReferenceFaces(void);
		BOOL 			FASTCALL DeReferenceFaces(void);
		BOOL			FASTCALL Optimize(void);

		INT  			FASTCALL SetShading(INT );
		INT 			FASTCALL GetShading(void);
		CHAR*			FASTCALL GetShadingName(CHAR *);
		DWORD			FASTCALL GetShadingFlags(void);

		INT  		    FASTCALL SetColour(INT );
		INT  			FASTCALL SetTextureMap(INT );
		INT  			FASTCALL SetLightMap(INT );
		INT  			FASTCALL SetBumpMap(INT );
		INT 			FASTCALL SetPhong(INT );

		INT 			FASTCALL GetColour(void);
		INT 			FASTCALL GetTextureMap(void);
		INT 			FASTCALL GetLightMap(void);
		INT 			FASTCALL GetBumpMap(void);
		INT				FASTCALL GetPhong(void);

		GVertexList* 	FASTCALL GetVertexList(void);
		GFaceList* 		FASTCALL GetFaceList(void);

		BOOL 			FASTCALL MakeTriangles(void);
		BOOL 			FASTCALL InitNormals(void);
		BOOL 			FASTCALL InitBoundingBox(void);

		GVertexList* 	FASTCALL AllocVertexList(INT );
		GFaceList*  	FASTCALL AllocTriangleList(INT );
		BOOL 			FASTCALL FreeVertexList(void);
		BOOL 			FASTCALL FreeFaceList(void);
		BOOL 			FASTCALL FreeTriangleList(void) {return FreeFaceList();};
		BOOL 			FASTCALL FreeObject(void) {return FreeMesh();};

		BOOL 			FASTCALL AllocMesh(INT ,INT ,DOUBLE );
		BOOL 			FASTCALL FreeMesh(void);
		BOOL 			FASTCALL UpdateMesh(void);

		BOOL			FASTCALL Write(GFile &,DWORD );
		BOOL        	FASTCALL Read(GFile &);

		private:
		BOOL			FASTCALL ReadVertexList(GFile &);
		BOOL			FASTCALL ReadFaceList(GFile &);
	   };


class GLightObject:public GVectorObject {
		friend class GVectorSystem;
		friend class GVectorObject;
		friend class G3dsFile;
		friend class G3doFile;

        public:
        enum {
        	LIGHT_TYPE = MAKETYPE('L','I','G','H'),
        	SPOTLIGHT_TYPE = MAKETYPE('S','P','O','T'),
        	POINTLIGHT_TYPE = MAKETYPE('P','O','N','T'),
        	AMBIENT_TYPE = MAKETYPE('A','M','B','T')
        };
        
		public:
		#pragma option -a1
		typedef struct {
				BYTE		R,G,B,A;
			   } FileColour;
		typedef struct {
				CHAR		Name[GMZ_NAME_LENGTH];
				DWORD		Flags;
				DWORD		Type;
				DWORD		Magic;
				FileVector	Position;
				FileVector	Rotation;
				FileColour  Colour;
				LONG		Specularity;
			   } FileLightHead;
		#pragma option -a.

		protected:
		GColour			Colour;
		INT				Specularity;

		public:
		GLightObject(CHAR *name,DWORD typ=GLightObject::LIGHT_TYPE);
		GLightObject(GVectorObject *f=NULL,DWORD typ=GLightObject::LIGHT_TYPE);
		~GLightObject();

        CHAR*			FASTCALL SetDefaultName();

		GColour 		FASTCALL GetColour(void);
		GColour			FASTCALL SetColour(GColour& );
		INT    			FASTCALL GetSpecularity(void);
		INT 			FASTCALL SetSpecularity(INT );

		BOOL			FASTCALL Write(GFile &,DWORD );
		BOOL        	FASTCALL Read(GFile &);
	   };



inline GVectorObject::GVectorObject(CHAR *name,DWORD typ) : GObjectKnot(name,typ) {
 	Position=GVector(0.0,0.0,0.0);
 	Rotation=GVector(0.0,0.0,0.0);
 	Flags|=GVE_OBJ_NEWPOS | GVE_OBJ_NEWROT;
 	RealizePosition(TRUE);
}

inline GVectorObject::GVectorObject(GVectorObject *f,DWORD t): GObjectKnot(f,t) {
 	Position=GVector(0.0,0.0,0.0);
 	Rotation=GVector(0.0,0.0,0.0);
 	Flags|=GVE_OBJ_NEWPOS | GVE_OBJ_NEWROT;
 	RealizePosition(TRUE);
}

inline GVectorObject::~GVectorObject() {
}

inline CHAR* FASTCALL GVectorObject::SetDefaultName() {
 	strcpy(Name,"vector object");
    return Name;
}

inline GVector FASTCALL GVectorObject::GetPosition(void) {
 	return Position;
}

inline GVector FASTCALL GVectorObject::GetPosition(DOUBLE *x,DOUBLE *y,DOUBLE *z) {
 	*x=Position.GetX();
 	*y=Position.GetY();
 	*z=Position.GetZ();
 	return Position;
}

inline GVector FASTCALL GVectorObject::GetRotation(void) {
 	return Rotation;
}

inline GVector FASTCALL GVectorObject::GetRotation(DOUBLE *x,DOUBLE *y,DOUBLE *z) {
 	*x=Rotation.GetX();
 	*y=Rotation.GetY();
 	*z=Rotation.GetZ();
 	return Rotation;
}

inline GVector FASTCALL GVectorObject::SetPosition(GVector &v) {
 	Flags|=GVE_OBJ_NEWPOS;
 	return Position=v;
}

inline GVector FASTCALL GVectorObject::SetPosition(DOUBLE x,DOUBLE y,DOUBLE z) {
 	Flags|=GVE_OBJ_NEWPOS;
 	return Position=GVector(x,y,z);
}

inline GVector FASTCALL GVectorObject::SetRotation(GVector &v) {
 	Flags|=GVE_OBJ_NEWROT;
 	return Rotation=v;
}

inline GVector FASTCALL GVectorObject::SetRotation(DOUBLE x,DOUBLE y,DOUBLE z) {
 	Flags|=GVE_OBJ_NEWROT;
 	return Rotation=GVector(x,y,z);
}




inline GMeshObject::GMeshObject(GVectorObject *f,DWORD typ) : GVectorObject(f,typ) {
 	VertexList=NULL;
 	FaceList=NULL;
 	Flags&=~(GVE_OBJ_CHUNKBOX | GVE_OBJ_VERTEXBOX);
}

inline GMeshObject::GMeshObject(CHAR *name,DWORD typ) : GVectorObject(name,typ) {
 	VertexList=NULL;
 	FaceList=NULL;
 	Flags&=~(GVE_OBJ_CHUNKBOX | GVE_OBJ_VERTEXBOX);
}

inline GMeshObject::~GMeshObject(void) {
 	if(VertexList) {
	 	free(VertexList);
	 	VertexList=NULL;
	}
 	if(FaceList) {
	 	free(FaceList);
	 	FaceList=NULL;
	}
 	Flags&=~(GVE_OBJ_VERTEX_ADDRESS | GVE_OBJ_VERTEX_INDEX);
}

inline CHAR* FASTCALL GMeshObject::SetDefaultName() {
 	strcpy(Name,"mesh object");
    return Name;
}

inline GVertexList* FASTCALL GMeshObject::GetVertexList(void) {
 	return VertexList;
}

inline GFaceList* FASTCALL GMeshObject::GetFaceList(void) {
 	return FaceList;
}




inline GLightObject::GLightObject(GVectorObject *f,DWORD typ) : GVectorObject(f,typ) {
 	Colour=GColour(255,255,255);
 	Specularity=255;
}

inline GLightObject::GLightObject(CHAR *name,DWORD typ) : GVectorObject(name,typ) {
 	Colour=GColour(255,255,255);
 	Specularity=255;
}

inline GLightObject::~GLightObject(void) {
}

inline CHAR* FASTCALL GLightObject::SetDefaultName() {
 	strcpy(Name,"light object");
    return Name;
}

inline GColour FASTCALL GLightObject::SetColour(GColour &c) {
 	return Colour=c;
}

inline GColour FASTCALL GLightObject::GetColour(void) {
 	return Colour;
}

inline INT FASTCALL GLightObject::SetSpecularity(INT s)	{
 	return Specularity=s;
}

inline INT FASTCALL GLightObject::GetSpecularity(void) {
 	return Specularity;
}




inline GCameraObject::GCameraObject(GVectorObject *f,DWORD typ) : GVectorObject(f,typ) {
 	Distance.x=Distance.y=LONG2MATH(100,GVE_VECTOR_SHIFT);
 	Zoom=LONG2MATH(1,GVE_ZOOM_SHIFT);
}

inline GCameraObject::GCameraObject(CHAR *name,DWORD typ) : GVectorObject(name,typ) {
 	Distance.x=Distance.y=LONG2MATH(100,GVE_VECTOR_SHIFT);
 	Zoom=LONG2MATH(1,GVE_ZOOM_SHIFT);
}

inline GCameraObject::~GCameraObject() {
}

inline CHAR* FASTCALL GCameraObject::SetDefaultName() {
 	strcpy(Name,"camera object");
    return Name;
}

inline GPointD FASTCALL GCameraObject::SetDistance(GPointD &d) {
 	Distance.x=DOUBLE2MATH(d.x,GVE_VECTOR_SHIFT);
 	Distance.y=DOUBLE2MATH(d.y,GVE_VECTOR_SHIFT);
 	return d;
}

inline GPointD FASTCALL GCameraObject::SetDistance(DOUBLE d,DOUBLE e) {
 	return GPointD(Distance.x=DOUBLE2MATH(d,GVE_VECTOR_SHIFT),Distance.y=DOUBLE2MATH(e,GVE_VECTOR_SHIFT));
}

inline GPointD FASTCALL GCameraObject::GetDistance(void) {
 	return GPointD(MATH2DOUBLE(Distance.x,GVE_VECTOR_SHIFT),MATH2DOUBLE(Distance.x,GVE_VECTOR_SHIFT));
}

inline DOUBLE FASTCALL GCameraObject::SetZoom(DOUBLE d) {
 	Zoom=DOUBLE2MATH(d,GVE_ZOOM_SHIFT);
 	return d;
}

inline DOUBLE FASTCALL GCameraObject::GetZoom(void) {
 	return MATH2DOUBLE(Zoom,GVE_ZOOM_SHIFT);
}

inline DOUBLE FASTCALL GCameraObject::SetNearPlane(DOUBLE d) {
 	NearPlane=DOUBLE2MATH(d,GVE_VECTOR_SHIFT);
 	return d;
}

inline DOUBLE FASTCALL GCameraObject::GetNearPlane(void) {
 	return MATH2DOUBLE(NearPlane,GVE_VECTOR_SHIFT);
}

inline DOUBLE FASTCALL GCameraObject::GetFarPlane(void) {
 	return MATH2DOUBLE(FarPlane,GVE_VECTOR_SHIFT);
}

inline DOUBLE FASTCALL GCameraObject::SetFarPlane(DOUBLE d) {
 	FarPlane=DOUBLE2MATH(d,GVE_VECTOR_SHIFT);
 	return d;
}

inline GPointL FASTCALL GCameraObject::SetOffset(GPointL& s) {
 	return Offset=s;
}

inline GPointL FASTCALL GCameraObject::SetOffset(LONG x,LONG y) {
 	Offset.x=x<<GVE_MATH_SHIFT;
 	Offset.y=y<<GVE_MATH_SHIFT;
 	return Offset;
}

inline GPointL FASTCALL GCameraObject::GetOffset(void) {
 	return Offset;
}


#endif
