///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
///////////////////////////////////////////////////////////////////////////////
//
//
// The following #defines may be done in this file:
//  	GVE_PRO					- The pro version. Always set
//		GVE_SPHERICAL_CHROME    - Use spherical chrome mapping
// 		GVE_CULL_AFTER
// 		GVE_TRIVIAL_REJECT
// 		GVE_BOUNDING_BOX
// 		GVE_ZBUFFER
// 		GVE_PROFILER
// 		GVE_EXACT_FRAME_SPEED
// 		GVE_DEBUG
//
// 		GVE_RASTER8
// 		GVE_RASTER16
//		GVE_RASTER24
// 		GVE_RASTER32
//
// 		GVE_FIXED_MATH
// 		GVE_FLOAT_MATH
//
// 		GVE_CULL_FIRST
// 		GVE_CULL_AFTER
//
// 		GVE_SORT_LIST
// 		GVE_SORT_TREE
//		GVE_SORT_HEAP
//
//
// Some important constants, you would like to change:
//  	GVE_MAX_VISIBLE_POLYS		(9000)
//  	GVE_MAX_TEXTUREMAPS			(2048)
//  	GVE_MAX_BUMPMAPS			(64)
//  	GVE_MAX_LIGHTMAPS			(64)
//  	GVE_MAX_PHONG				(16)
//
// You can also change the values in the maths section, refering to
// the shift values. BUT: You really should know what you are doing there,
// or the system may crash !
//
//					Have fun !
//


#ifndef __GVEPRO_H
#define __GVEPRO_H

/*
 * The following #defines enable/disable some options of the vector engine.
 * You may (un)define any constants here.
 */
#define GVE_PRO
#define GVE_SPHERICAL_CHROME
#define GVE_CULL_AFTER
#define GVE_TRIVIAL_REJECT
#define GVE_BOUNDING_BOX
#define GVE_SORT_LIST
#define GVE_WALK_PREORDER
#define GVE_RASTER8
#define GVE_RASTER16
#define GVE_FLOAT_MATH
#define GVE_DEBUG

//#define GVE_FIXED_MATH
//#define GVE_WALK_POSTORDER
//#define GVE_CULL_AFTER
//#define GVE_SORT_TREE
//#define GVE_SORT_HEAP
//#define GVE_ZBUFFER
//#define GVE_PROFILER
//#define GVE_EXACT_FRAME_SPEED
//#define GVE_RASTER32



/*
 * These constants control some limits of the engine. You may change all
 * these #defines according to your needs.
 */
#define GVE_MAX_VISIBLE_POLYS		9000
#define GVE_MAX_TEXTUREMAPS			2048
#define GVE_MAX_BUMPMAPS			64
#define GVE_MAX_LIGHTMAPS			64
#define GVE_MAX_PHONG				16
#define GVE_MAX_METHODS				512
#define GVE_TIMER_SPEED				4000
#define GVE_BLURR_FACTOR			(32+16)
#define GVE_3DS_FACTOR	     		((DOUBLE)1.0)
#define GVE_FILE_VERSION			0x10000L
#define GVE_MINIMAL_Z(v)     		(1 << GVE_VECTOR_SHIFT)
#define GVE_LOG_NAME				"gvepro.log"


/*
 * The following #defines are used to control the fixedpoint precisions
 * used in the engine. You should not change them, unless you are sure what
 * you are doing, because they are related to each other.
 */
#define GVE_MATH_SHIFT				10
#define GVE_MATRIX_SHIFT			20
#define GVE_NORMAL_SHIFT			14
#define GVE_VECTOR_SHIFT			GVE_MATH_SHIFT
#define GVE_SHAD_SHIFT				(2*GVE_NORMAL_SHIFT-6)
#define GVE_SCRIPT_SHIFT			12
#define GVE_MAPPING_SHIFT			16
#define GVE_ASPECT_SHIFT			16
#define GVE_ZOOM_SHIFT				GVE_MATH_SHIFT
#define GVE_RASTERX_SHIFT			GVE_ASPECT_SHIFT
#define GVE_RASTERY_SHIFT			0
#define GVE_DITHER_SHIFT			(2*GVE_NORMAL_SHIFT-6)

#define GVE_COSINE_SHIFT			GVE_MATRIX_SHIFT
#define GVE_COSINE_TABLE_SHIFT		10
#define GVE_COSINE_TABLE_SIZE		(1 << GVE_COSINE_TABLE_SHIFT)
#define GVE_ASINE_SHIFT				(GVE_MAPPING_SHIFT+7)
#define GVE_ASINE_TABLE_SHIFT		10
#define GVE_ASINE_TABLE_SIZE		(1 << GVE_ASINE_TABLE_SHIFT)

#define GVE_SHADINGLUT_ENTRIES		(64*256)
#define GVE_GLASSLUT_ENTRIES		(256*256)
#define GVE_BLURRLUT_ENTRIES		256
#define GVE_BUMPLUT_ENTRIES			GVE_SHADINGLUT_ENTRIES

/*
 * Now follow some constants used for textures etc...
 */
#define GVE_BITMAP_WIDTH			256
#define GVE_BITMAP_HEIGHT			256
#define GVE_BITMAP_DEPTH			8

#define GVE_TEXTURE_WIDTH			GVE_BITMAP_WIDTH
#define GVE_TEXTURE_HEIGHT			GVE_BITMAP_HEIGHT
#define GVE_TEXTURE_DEPTH		    GVE_BITMAP_DEPTH

#define GVE_LIGHTMAP_WIDTH			GVE_BITMAP_WIDTH
#define GVE_LIGHTMAP_HEIGHT			GVE_BITMAP_HEIGHT
#define GVE_LIGHTMAP_DEPTH			GVE_BITMAP_DEPTH

#define GVE_BUMPMAP_WIDTH			GVE_BITMAP_WIDTH
#define GVE_BUMPMAP_HEIGHT			GVE_BITMAP_HEIGHT
#define GVE_BUMPMAP_DEPTH			16

#define GVE_VIDEO_WIDTH				GVE_BITMAP_WIDTH
#define GVE_VIDEO_HEIGHT			GVE_BITMAP_HEIGHT
#define GVE_VIDEO_BUFFER_SIZE		100


#ifndef GVE_PRO
 #define GVE_MAX_EDGES				24
#else
 #define GVE_MAX_EDGES				3
#endif

#ifdef GVE_RASTER15
 #define GVE_RASTER16
#endif

#if defined(__FIXEDMATH__)
 #define GVE_FIXED_MATH
#elif defined(__FLOATMATH__)
 #define GVE_FLOAT_MATH
#endif

#if ((defined GVE_SORT_AVLTREE) && (defined GVE_SORT_TREE))
 #undef	GVE_SORT_TREE
#endif


/*
 * Now a quick option-error-checking will be performed
 */
#if ((!defined GVE_CULL_FIRST) && (!defined GVE_CULL_AFTER))
#error You must define GVE_CULL_FIRST or GVE_CULL_AFTER !
#elif ((defined GVE_CULL_FIRST) && (defined GVE_CULL_AFTER))
#error You cannot use both GVE_CULL_FIRST and GVE_CULL_AFTER !
#endif

#if ((!defined(GVE_SORT_TREE)) && (!defined(GVE_SORT_LIST)) && (!defined(GVE_SORT_HEAP)))
#error You must define GVE_SORT_TREE or GVE_SORT_LIST or GVE_SORT_HEAP
#elif (defined GVE_SORT_TREE) && (defined GVE_SORT_LIST)
#error You cannot use both GVE_SORT_TREE and GVE_SORT_LIST
#elif (defined GVE_SORT_TREE) && (defined GVE_SORT_HEAP)
#error You cannot use both GVE_SORT_TREE and GVE_SORT_HEAP
#elif (defined GVE_SORT_LIST) && (defined GVE_SORT_HEAP)
#error You cannot use both GVE_SORT_LIST and GVE_SORT_HEAP
#endif

#if ((!defined GVE_WALK_PREORDER) && (!defined GVE_WALK_POSTORDER))
#error You must define GVE_WALK_PREORDER or GVE_WALK_POSTORDER
#elif ((defined GVE_WALK_PREORDER) && (defined GVE_WALK_POSTORDER))
#error You cannot use both GVE_WALK_PREORDER and GVE_WALK_POSTORDER
#endif

#if ((!defined GVE_FIXED_MATH) && (!defined GVE_FLOAT_MATH))
#error You must define GVE_FIXED_MATH or GVE_FLOAT_MATH
#elif ((defined GVE_FIXED_MATH) && (defined GVE_FLOAT_MATH))
#error You cannot use both GVE_FIXED_MATH and GVE_FLOAT_MATH
#endif

#if (GVE_RASTERY_SHIFT!=0)
#error GVE_RASTERY_SHIFT must be set to zero !
#endif



/*
 *  The following constants are flags used for the different classes and
 *  objects of the GVE.
 */
#define GVE_SYS_READY				((DWORD)0x00001L)
#define GVE_SYS_8BIT				((DWORD)0x00002L)
#define GVE_SYS_15BIT				((DWORD)0x00004L)
#define GVE_SYS_16BIT				((DWORD)0x00006L)
#define GVE_SYS_24BIT				((DWORD)0x00008L)
#define GVE_SYS_32BIT				((DWORD)0x00009L)
#define GVE_SYS_INSTANCE			((DWORD)0x00010L)
#define GVE_SYS_BUSY				((DWORD)0x00020L)


#define GVE_LOG_QUIET				0
#define GVE_LOG_ERROR				1
#define GVE_LOG_WARNING				2
#define GVE_LOG_MESSAGE				3

#define GVE_VID_LOOP				((DWORD)1L)
#define GVE_VID_EXTERNAL			((DWORD)2L)
#define GVE_VID_INTERNAL			((DWORD)4L)
#define GVE_VID_REMAP				((DWORD)8L)
#define GVE_VID_RESCALE				((DWORD)16L)


#include <values.h>
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <dir.h>
#include <math.h>
#include <mem.h>
#include <alloc.h>
#include <time.h>
#include <sys/stat.h>


#include "../include/gomez.h"
#include "../include/gveerr.h"
#include "../include/gvemath.h"
#include "../include/gvemap.h"
#include "../include/gveobj.h"
#include "../include/gveworld.h"
#include "../include/gve3ds.h"
#include "../include/gve3do.h"
#include "../include/gverst.h"


#define GDeg2Rad(a)		((DOUBLE)(a)/180.0)
#define GRad2Deg(a)		((DOUBLE)(a)*180.0)


class GFaceOrderList {
		friend class GVectorSystem;

		private:
		typedef struct GEntry {
				MATH		Key;
				GFace    	*Face;
			   } GEntry;

		private:
		INT			Size;
		GEntry 		*FirstEntry;
		GEntry 		*LastEntry;
		GEntry 		*CurrentEntry;

		public:
		GFaceOrderList();
		~GFaceOrderList();
		void 		FASTCALL AddEntry(MATH ,GFace *);
		void 		FASTCALL DeleteEntry(MATH ,GFace *);
		void 		FASTCALL DeleteAll(void);
		void 		FASTCALL Sort(void);
		BOOL 		FASTCALL Alloc(INT );
		BOOL 		FASTCALL Free(void);

		GFace* 		FASTCALL Start(void);
		GFace* 		FASTCALL Next(void);

		private:
		static int	_USERENTRY cmpfnc(const void *,const void *);
	   };


class GFaceOrderTree {
		friend class GVectorSystem;

		private:
		typedef struct GEntry {
				MATH		Key;
				GFace		*Face;
				GEntry 		*LeftSon;
				GEntry 		*RightSon;
			   } GEntry;

		public:
		GEntry 		*Head;

		private:
		GEntry 		*FirstEntry;
		GEntry 		*LastEntry;
		GEntry 		*CurrentEntry;
		INT			Size;

		public:
		GFaceOrderTree();
		~GFaceOrderTree();
		void 		FASTCALL AddEntry(MATH ,GFace *);
		void 		FASTCALL DeleteEntry(MATH ,GFace *);
		void 		FASTCALL DeleteAll(void);
		void 		FASTCALL Sort(void) {};
		BOOL 		FASTCALL Alloc(INT );
		BOOL 		FASTCALL Free(void);

		GFace* 		FASTCALL Start(void);
		GFace* 		FASTCALL Next(void);
	   };

class GFaceOrderHeap {
		friend class GVectorSystem;

		private:
		typedef struct GEntry {
				MATH		Key;
				GFace    	*Face;
			   } GEntry;

		private:
		INT			Size;
		GEntry 		*FirstEntry;
		GEntry 		*LastEntry;
		GEntry 		*CurrentEntry;

		public:
		GFaceOrderHeap();
		~GFaceOrderHeap();
		void 		FASTCALL AddEntry(MATH ,GFace *);
		void 		FASTCALL DeleteEntry(MATH ,GFace *);
		void 		FASTCALL DeleteAll(void);
		void 		FASTCALL Sort(void);
		BOOL 		FASTCALL Alloc(INT );
		BOOL 		FASTCALL Free(void);

		GFace* 		FASTCALL Start(void);
		GFace* 		FASTCALL Next(void);
	   };

class GVectorSystem : public GObject {
		friend class 	GVectorWorld;
		private:
        GRasterEngine*	RasterEngine;
        GVectorWorld*	World;
        GAlphaBuffer*	CurrentFrame;
        GAlphaBuffer*	LastFrame;
        INT				FrameWidth;
        INT				FrameHeight;
        DOUBLE			FrameAspectRatio;

		#if defined GVE_SORT_LIST
		GFaceOrderList 	OrderList;
		#elif defined GVE_SORT_TREE
		GFaceOrderTree 	OrderList;
		#elif defined GVE_SORT_HEAP
		GFaceOrderHeap 	OrderList;
		#endif

		typedef struct GRasterIndex {
				INT		Type;
				DWORD	Flags;
			   } GRasterIndex;

		static CHAR     BuildInfo[];
		static GRasterIndex RasterIndex[];
		static MATH 	SinusTable[GVE_COSINE_TABLE_SIZE+GVE_COSINE_TABLE_SIZE/4];
		#ifdef GVE_SPHERICAL_CHROME
		static LONG 	ASinusTable[GVE_ASINE_TABLE_SIZE+1];
		#endif

		public:
		GVectorSystem();
		~GVectorSystem();

        CHAR*			FASTCALL SetDefaultName();

		BOOL 			FASTCALL Init(GRasterEngine* );
		BOOL 			FASTCALL Exit(void);
		BOOL			FASTCALL SetBuffers(GAlphaBuffer *n,GAlphaBuffer *o=NULL);
		BOOL			FASTCALL SetWorld(GVectorWorld *);
		void 			FASTCALL DrawWorld(void);

		BOOL   			FASTCALL SetBumpMaps(GBitmapList* );
		BOOL   			FASTCALL SetTextureMaps(GBitmapList* );
		BOOL   			FASTCALL SetLightMaps(GBitmapList* );
		BOOL			FASTCALL SetPalette(GPalette *);
		BOOL			FASTCALL RealizePalette(void);

		static CHAR* 	FASTCALL GetShadingName(INT ,CHAR *);
		static DWORD 	FASTCALL GetShadingFlags(INT );
		static INT	 	FASTCALL FindShading(DWORD );

		CHAR*			FASTCALL GetBuildInfo(void);
		CHAR*			FASTCALL GetBuildInfo(CHAR *);
	   };



///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GVectorSystem
//
inline GVectorSystem::GVectorSystem() : GObject() {
 	(PTR)RasterEngine=
 	(PTR)World=
 	(PTR)CurrentFrame=
 	(PTR)LastFrame=NULL;
}

inline GVectorSystem::~GVectorSystem() {
}

inline CHAR* FASTCALL GVectorSystem::SetDefaultName() {
 	strcpy(Name,"vector system");
    return Name;
}

inline BOOL FASTCALL GVectorSystem::SetTextureMaps(GBitmapList *l) {
 	return RasterEngine->SetTextureMaps(l);
}

inline BOOL FASTCALL GVectorSystem::SetLightMaps(GBitmapList *l) {
 	return RasterEngine->SetLightMaps(l);
}

inline BOOL FASTCALL GVectorSystem::SetBumpMaps(GBitmapList *l) {
 	return RasterEngine->SetBumpMaps(l);
}

inline BOOL FASTCALL GVectorSystem::SetPalette(GPalette *pal) {
 	return RasterEngine->SetPalette(pal);
}

inline BOOL FASTCALL GVectorSystem::RealizePalette() {
 	return RasterEngine->RealizePalette();
}

inline CHAR* FASTCALL GVectorSystem::GetBuildInfo(void) {
 	return BuildInfo;
}

inline CHAR* FASTCALL GVectorSystem::GetBuildInfo(CHAR *s) {
 	return strcpy(s,BuildInfo);
}

inline BOOL FASTCALL GVectorSystem::SetBuffers(GAlphaBuffer *n,GAlphaBuffer *o) {
 	CurrentFrame=n;
 	LastFrame=o ? o : n;
 	RasterEngine->SetBuffers(CurrentFrame,LastFrame);
 	FrameWidth=CurrentFrame->GetWidth();
 	FrameHeight=CurrentFrame->GetHeight();
 	FrameAspectRatio=CurrentFrame->GetAspectRatio();
 	return TRUE;
}

inline BOOL FASTCALL GVectorSystem::SetWorld(GVectorWorld *w) {
 	World=w;
 	RasterEngine->SetLight(w->GetLight());
 	RasterEngine->SetAmbientLight(w->GetAmbientLight());
 	RasterEngine->SetTextureMaps(&w->TextureMapList);
 	RasterEngine->SetLightMaps(&w->LightMapList);
 	RasterEngine->SetBumpMaps(&w->BumpMapList);
 	RasterEngine->SetPalette(&w->Palette);
 	RasterEngine->RealizePalette();
 	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GFaceOrderTree
//
inline GFaceOrderTree::GFaceOrderTree(void) {
 	memset(this,0,sizeof(GFaceOrderTree));
}

inline GFaceOrderTree::~GFaceOrderTree(void) {
}

inline void FASTCALL GFaceOrderTree::DeleteAll(void) {
 	Head=LastEntry=FirstEntry;
}

inline GFace* FASTCALL GFaceOrderTree::Next(void) {
 	return NULL;
}



///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GFaceOrderList
//
inline GFaceOrderList::GFaceOrderList(void) {
 	memset(this,0,sizeof(GFaceOrderList));
}

inline GFaceOrderList::~GFaceOrderList(void) {
}

inline void FASTCALL GFaceOrderList::DeleteAll(void) {
 	LastEntry=FirstEntry;
}

inline void FASTCALL GFaceOrderList::AddEntry(MATH key,GFace *face) {
 	LastEntry->Key=key;
 	LastEntry->Face=face;
 	LastEntry++;
}

inline int _USERENTRY GFaceOrderList::cmpfnc(const void *a,const void *b) {
 	MATH d;
 	d=((GFaceOrderList::GEntry *)a)->Key - ((GFaceOrderList::GEntry *)b)->Key;
 	if (d>0) return -1;
 	if (d<0) return +1;
 	return 0;
}

inline void FASTCALL GFaceOrderList::Sort(void) {
 	if (FirstEntry!=LastEntry)
		qsort(FirstEntry,LastEntry-FirstEntry,sizeof(GEntry),cmpfnc);
}

inline GFace* FASTCALL GFaceOrderList::Start(void) {
 	if (LastEntry==FirstEntry)
		return  (GFace *)(CurrentEntry=NULL);
 	else
		return (CurrentEntry=FirstEntry)->Face;
}

inline GFace* FASTCALL GFaceOrderList::Next(void) {
	if (CurrentEntry>=LastEntry)
		return (GFace *)(CurrentEntry=NULL);
 	else
		return (++CurrentEntry)->Face;
}



///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GFaceOrderHeap
//
inline GFaceOrderHeap::GFaceOrderHeap(void) {
 	memset(this,0,sizeof(GFaceOrderHeap));
}

inline GFaceOrderHeap::~GFaceOrderHeap(void) {
}

inline void FASTCALL GFaceOrderHeap::DeleteAll(void) {
 	LastEntry=FirstEntry;
}

inline void FASTCALL GFaceOrderHeap::AddEntry(MATH key,GFace *face) {
 	LastEntry->Key=key;
 	LastEntry->Face=face;
 	LastEntry++;
}

inline void FASTCALL GFaceOrderHeap::Sort(void) {
}

inline GFace* FASTCALL GFaceOrderHeap::Start(void) {
 	if (LastEntry==FirstEntry)
		return  (GFace *)(CurrentEntry=NULL);
 	else
		return (CurrentEntry=FirstEntry)->Face;
}

inline GFace* FASTCALL GFaceOrderHeap::Next(void) {
 	if (CurrentEntry>=LastEntry)
		return (GFace *)(CurrentEntry=NULL);
 	else
		return (++CurrentEntry)->Face;
}



///////////////////////////////////////////////////////////////////////////////
//
// Parts of the implementation of GMeshObject
//
inline CHAR* FASTCALL GMeshObject::GetShadingName(CHAR *b) {
 	return GVectorSystem::GetShadingName(GetShading(),b);
}

inline DWORD FASTCALL GMeshObject::GetShadingFlags(void) {
 	return GVectorSystem::GetShadingFlags(GetShading());
}

#endif
