///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//

#ifndef __GVERST_H
#define __GVERST_H

#include "../include/gvepro.h"

#define GVE_SHAD_NOTHING						0
#define GVE_SHAD_POINT                          1
#define GVE_SHAD_WIREFRAME                      2
#define GVE_SHAD_SOFT_WIRE                      3
#define GVE_SHAD_GLASS_WIRE                     4
#define GVE_SHAD_FILLED                         5
#define GVE_SHAD_LAMBERT                        6
#define GVE_SHAD_GOURAUD                        7
#define GVE_SHAD_PHONG                          8
#define GVE_SHAD_GLASS                          9
#define GVE_SHAD_LAMBERT_GLASS                  10
#define GVE_SHAD_GOURAUD_GLASS                  11
#define GVE_SHAD_PHONG_GLASS                    12
#define GVE_SHAD_TEXTURE                        13
#define GVE_SHAD_TEXTURE_LAMBERT                14
#define GVE_SHAD_TEXTURE_GOURAUD                15
#define GVE_SHAD_TEXTURE_PHONG                  16
#define GVE_SHAD_TEXTURE_GLASS                  17
#define GVE_SHAD_TEXTURE_LIGHTMAP               18
#define GVE_SHAD_TEXTURE_CHROME                 19
#define GVE_SHAD_TEXTURE_BUMP_LIGHTMAP          20
#define GVE_SHAD_TEXTURE_BUMP_CHROME            21
#define GVE_SHAD_PTEXTURE                       22
#define GVE_SHAD_PTEXTURE_LAMBERT               23
#define GVE_SHAD_PTEXTURE_GOURAUD               24
#define GVE_SHAD_PTEXTURE_PHONG                 25
#define GVE_SHAD_PTEXTURE_GLASS                 26
#define GVE_SHAD_PTEXTURE_LIGHTMAP              27
#define GVE_SHAD_PTEXTURE_CHROME                28
#define GVE_SHAD_PTEXTURE_BUMP_LIGHTMAP         29
#define GVE_SHAD_PTEXTURE_BUMP_CHROME           30
#define GVE_SHAD_CHROME                         31
#define GVE_SHAD_CHROME_LAMBERT                 32
#define GVE_SHAD_CHROME_GOURAUD                 33
#define GVE_SHAD_CHROME_PHONG                   34
#define GVE_SHAD_CHROME_GLASS                   35
#define GVE_SHAD_BUMP                           36
#define GVE_SHAD_BUMP_LAMBERT                   37
#define GVE_SHAD_BUMP_GOURAUD                   38
#define GVE_SHAD_BUMP_PHONG			            39
#define GVE_SHAD_BUMP_GLASS			            40
#define GVE_SHAD_BUMP_CHROME		            41
#define GVE_SHAD_BUMP_LIGHTMAP		            42
#define GVE_SHAD_CARVA                          43
#define GVE_SHAD_CARVA_LAMBERT		            44
#define GVE_SHAD_CARVA_GOURAUD		            45
#define GVE_SHAD_CARVA_PHONG		            46
#define GVE_SHAD_CARVA_GLASS		            47
#define GVE_SHAD_CARVA_LAMBERT_GLASS		    48
#define GVE_SHAD_CARVA_GOURAUD_GLASS		    49
#define GVE_SHAD_CARVA_PHONG_GLASS              50
#define GVE_SHAD_MAX_METHOD						50

#define GVE_SF_WIRE					((DWORD)0x00000001LU)
#define GVE_SF_FILLED				((DWORD)0x00000002LU)
#define GVE_SF_LAMBERT				((DWORD)0x00000004LU)
#define GVE_SF_GOURAUD				((DWORD)0x00000008LU)
#define GVE_SF_PHONG				((DWORD)0x00000010LU)
#define GVE_SF_TEXTURE				((DWORD)0x00000020LU)
#define GVE_SF_BUMP 				((DWORD)0x00000040LU)
#define GVE_SF_LIGHTMAP				((DWORD)0x00000080LU)
#define GVE_SF_CHROME				((DWORD)0x00000100LU)
#define GVE_SF_PERSPECTIVE			((DWORD)0x00000200LU)
#define GVE_SF_CARVA      			((DWORD)0x00000400LU)
#define GVE_SF_GLASS      			((DWORD)0x00000800LU)
#define GVE_SF_SOFT       			((DWORD)0x00001000LU)
#define GVE_SF_VERTEX     			((DWORD)0x00002000LU)

#define GVE_SF_LOWDETAIL			((DWORD)0x01000000LU)
#define GVE_SF_MEDIUMDETAIL			((DWORD)0x02000000LU)
#define GVE_SF_HIGHDETAIL			((DWORD)0x04000000LU)
#define GVE_SF_ALLDETAIL			(GVE_SF_LOWDETAIL | GVE_SF_MEDIUMDETAIL | GVE_SF_HIGHDETAIL)
#define GVE_SF_8BIT					((DWORD)0x08000000LU)
#define GVE_SF_15BIT		 		((DWORD)0x10000000LU)
#define GVE_SF_16BIT	   			((DWORD)0x20000000LU)
#define GVE_SF_24BIT	   			((DWORD)0x40000000LU)
#define GVE_SF_32BIT	 			((DWORD)0x80000000LU)



class GRasterEngine : public GObject {
		protected:
        GAlphaBuffer* 	CurrentFrame;
        GAlphaBuffer* 	LastFrame;
        GBitmapList*	TextureMapList;
        GBitmapList*	BumpMapList;
		GBitmapList*	LightMapList;
        GPalette		Palette;
		GLightObject*	Light[8];
		GLightObject*	Ambient;
		INT				FrameWidth;
		INT				FrameHeight;
		INT				FrameDepth;
        GRectL			FrameClipWindow;
        void			(FASTCALL GRasterEngine::*DrawFaceFunc[64])(GFace *);

        public:
		GRasterEngine();
		~GRasterEngine();

        CHAR*			FASTCALL SetDefaultName();

		virtual BOOL	FASTCALL Init(void);
		virtual BOOL	FASTCALL Exit(void);

		BOOL 			FASTCALL SetBuffers(GAlphaBuffer* ,GAlphaBuffer *);
		virtual BOOL   	FASTCALL SetBumpMaps(GBitmapList* );
		virtual BOOL   	FASTCALL SetTextureMaps(GBitmapList* );
		virtual BOOL   	FASTCALL SetLightMaps(GBitmapList* );

		virtual BOOL  	FASTCALL SetLight(GLightObject* ,INT n=0);
		virtual BOOL	FASTCALL SetAmbientLight(GLightObject* );
		BOOL			FASTCALL SetPalette(GPalette* );
		virtual BOOL	FASTCALL RealizePalette(void);

		virtual void	FASTCALL Begin(void);
		virtual void	FASTCALL End(void);
		virtual void	FASTCALL Postprocess(DWORD );
		virtual void	FASTCALL Preprocess(DWORD );

		void			FASTCALL DrawFace(GFace *);

		protected:
		virtual void   	FASTCALL DrawNothing(GFace *);
		virtual void   	FASTCALL DrawVertex(GFace *);
		virtual void   	FASTCALL DrawWire(GFace *);
		virtual void   	FASTCALL DrawWireGlass(GFace *);
		virtual void   	FASTCALL DrawWireSoft(GFace *);
		virtual void   	FASTCALL DrawFilled(GFace *);
		virtual void   	FASTCALL DrawFilledLambert(GFace *);
		virtual void   	FASTCALL DrawFilledGouraud(GFace *);
		virtual void   	FASTCALL DrawFilledPhong(GFace *);
		virtual void   	FASTCALL DrawFilledGlass(GFace *);
		virtual void   	FASTCALL DrawFilledGlassLambert(GFace *);
		virtual void   	FASTCALL DrawFilledGlassGouraud(GFace *);
		virtual void   	FASTCALL DrawFilledGlassPhong(GFace *);
		virtual void   	FASTCALL DrawCarva(GFace *);
        virtual void   	FASTCALL DrawCarvaLambert(GFace *);
		virtual void   	FASTCALL DrawCarvaGouraud(GFace *);
		virtual void   	FASTCALL DrawCarvaPhong(GFace *);
		virtual void   	FASTCALL DrawCarvaGlass(GFace *);
		virtual void   	FASTCALL DrawCarvaLambertGlass(GFace *);
		virtual void   	FASTCALL DrawCarvaGouraudGlass(GFace *);
		virtual void   	FASTCALL DrawCarvaPhongGlass(GFace *);
        virtual void   	FASTCALL DrawTextured(GFace *);
        virtual void   	FASTCALL DrawTexturedLambert(GFace *);
        virtual void   	FASTCALL DrawTexturedGouraud(GFace *);
		virtual void   	FASTCALL DrawTexturedPhong(GFace *);
        virtual void   	FASTCALL DrawTexturedGlass(GFace *);
        virtual void   	FASTCALL DrawTexturedLightmap(GFace *);
        virtual void   	FASTCALL DrawTexturedChrome(GFace *);
        virtual void   	FASTCALL DrawTexturedBumpLightmap(GFace *);
		virtual void   	FASTCALL DrawTexturedBumpChrome(GFace *);
        virtual void   	FASTCALL DrawPersTextured(GFace *);
        virtual void   	FASTCALL DrawPersTexturedLambert(GFace *);
		virtual void   	FASTCALL DrawPersTexturedGouraud(GFace *);
        virtual void   	FASTCALL DrawPersTexturedPhong(GFace *);
        virtual void   	FASTCALL DrawPersTexturedGlass(GFace *);
        virtual void   	FASTCALL DrawPersTexturedLightmap(GFace *);
		virtual void   	FASTCALL DrawPersTexturedChrome(GFace *);
		virtual void   	FASTCALL DrawPersTexturedBumpLightmap(GFace *);
        virtual void   	FASTCALL DrawPersTexturedBumpChrome(GFace *);
		virtual void   	FASTCALL DrawChrome(GFace *);
		virtual void   	FASTCALL DrawChromeLambert(GFace *);
		virtual void   	FASTCALL DrawChromeGouraud(GFace *);
		virtual void   	FASTCALL DrawChromePhong(GFace *);
		virtual void   	FASTCALL DrawChromeGlass(GFace *);
		virtual void   	FASTCALL DrawBump(GFace *);
		virtual void   	FASTCALL DrawBumpLambert(GFace *);
		virtual void   	FASTCALL DrawBumpGouraud(GFace *);
		virtual void   	FASTCALL DrawBumpPhong(GFace *);
		virtual void   	FASTCALL DrawBumpGlass(GFace *);
		virtual void   	FASTCALL DrawBumpChrome(GFace *);
		virtual void   	FASTCALL DrawBumpLightmap(GFace *);
	   };


class GRasterEngine8 : public GRasterEngine {
		protected:
		BYTE			CarvaTable[256];
		BYTE			GlassLUT[256*256];
		BYTE			ShadingLUT[GVE_SHADINGLUT_ENTRIES];
		BYTE			BumpLUT[GVE_BUMPLUT_ENTRIES];
		BYTE			BlurrLUT[GVE_BLURRLUT_ENTRIES];

		public:
		GRasterEngine8();
		~GRasterEngine8();

        CHAR*			FASTCALL SetDefaultName();

		virtual BOOL	FASTCALL Init(void);
		virtual BOOL	FASTCALL Exit(void);
		virtual BOOL  	FASTCALL RealizePalette(void);
		virtual void	FASTCALL Postprocess(DWORD );
		virtual void	FASTCALL Preprocess(DWORD );

		protected:
		virtual void    FASTCALL DrawVertex(GFace *);
		virtual void    FASTCALL DrawFilled(GFace *);
		virtual void   	FASTCALL DrawFilledLambert(GFace *);
		virtual void    FASTCALL DrawFilledGouraud(GFace *);
		virtual void    FASTCALL DrawTextured(GFace *);

		private:
		void			FASTCALL MakeShadingLUT(void);
		void			FASTCALL MakeBumpLUT(void);
		void			FASTCALL MakeBlurrLUT(void);
		void			FASTCALL MakeGlassLUT(void);
	   };

class GRasterEngine16 : public GRasterEngine {
		protected:
		DWORD			SoftenMask;
		BYTE			CarvaTable[256];
		WORD			GlassLUT[256*256];
		WORD			ShadingLUT[GVE_SHADINGLUT_ENTRIES];
		WORD			BumpLUT[GVE_BUMPLUT_ENTRIES];

		public:
		GRasterEngine16();
		~GRasterEngine16();

        CHAR*			FASTCALL SetDefaultName();

		virtual BOOL	FASTCALL Init(void);
		virtual BOOL	FASTCALL Exit(void);
		virtual BOOL  	FASTCALL RealizePalette(void);
		virtual void	FASTCALL Postprocess(DWORD );
		virtual void	FASTCALL Preprocess(DWORD );

		protected:
		virtual void    FASTCALL DrawVertex(GFace *);
		virtual void    FASTCALL DrawFilled(GFace *);
		virtual void   	FASTCALL DrawFilledLambert(GFace *);
		virtual void    FASTCALL DrawFilledGouraud(GFace *);
		virtual void    FASTCALL DrawFilledGlass(GFace *);
		virtual void   	FASTCALL DrawFilledGlassLambert(GFace *);
		virtual void    FASTCALL DrawFilledGlassGouraud(GFace *);
		virtual void    FASTCALL DrawTextured(GFace *);

		private:
		void			FASTCALL MakeShadingLUT(void);
		void			FASTCALL MakeBumpLUT(void);
		void			FASTCALL MakeGlassLUT(void);
	   };

class GRasterEngine15 : public GRasterEngine16 {
		public:
		GRasterEngine15();
		~GRasterEngine15();

        CHAR*		FASTCALL SetDefaultName();
	   };



///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GRasterEngine
//
inline GRasterEngine::~GRasterEngine() {
}

inline CHAR* FASTCALL GRasterEngine::SetDefaultName() {
	strcpy(Name,"raster engine");
    return Name;
}

inline BOOL FASTCALL GRasterEngine::Init(void) {
 	return TRUE;
}

inline BOOL FASTCALL GRasterEngine::Exit(void) {
 	return TRUE;
}

inline BOOL FASTCALL GRasterEngine::SetBuffers(GAlphaBuffer *n,GAlphaBuffer *o) {
 	CurrentFrame=n;
 	LastFrame=o;
	FrameWidth=CurrentFrame->GetWidth();
	FrameHeight=CurrentFrame->GetHeight();
 	FrameDepth=CurrentFrame->GetDepth();
 	FrameClipWindow=CurrentFrame->GetClipWindow();
 	return TRUE;
}

inline BOOL FASTCALL GRasterEngine::SetLight(GLightObject *l,INT n) {
 	if (n>8) return FALSE;
 	Light[n]=l;
 	return TRUE;
}

inline BOOL FASTCALL GRasterEngine::SetAmbientLight(GLightObject *a) {
 	Ambient=a;
 	return TRUE;
}

inline BOOL FASTCALL GRasterEngine::SetTextureMaps(GBitmapList *l) {
 	TextureMapList=l;
 	return TRUE;
}

inline BOOL FASTCALL GRasterEngine::SetLightMaps(GBitmapList *l) {
 	LightMapList=l;
 	return TRUE;
}

inline BOOL FASTCALL GRasterEngine::SetBumpMaps(GBitmapList *l) {
 	BumpMapList=l;
 	return TRUE;
}

inline BOOL FASTCALL GRasterEngine::SetPalette(GPalette *p) {
 	Palette=*p;
 	return TRUE;
}

inline BOOL FASTCALL GRasterEngine::RealizePalette() {
 	return TRUE;
}

inline void FASTCALL GRasterEngine::DrawFace(GFace *f) {
 	(this->*DrawFaceFunc[f->Type])(f);
}

inline void FASTCALL GRasterEngine::Begin(void) {
 	FrameWidth=CurrentFrame->GetWidth();
 	FrameHeight=CurrentFrame->GetHeight();
 	FrameDepth=CurrentFrame->GetDepth();
 	FrameClipWindow=CurrentFrame->GetClipWindow();
}

inline void FASTCALL GRasterEngine::End(void) {
}



///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GRasterEngine8
//
inline GRasterEngine8::GRasterEngine8(void) : GRasterEngine() {
 	FrameDepth=8;
}

inline GRasterEngine8::~GRasterEngine8() {
}

inline CHAR* FASTCALL GRasterEngine8::SetDefaultName() {
	strcpy(Name,"raster engine 8 bit");
    return Name;
}

inline BOOL FASTCALL GRasterEngine8::RealizePalette(void) {
 	MakeShadingLUT();
 	MakeBumpLUT();
 	MakeGlassLUT();
 	MakeBlurrLUT();
 	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GRasterEngine16
//
inline GRasterEngine16::GRasterEngine16(void) : GRasterEngine() {
 	FrameDepth=16;
 	SoftenMask=0xf7def7d;
}

inline GRasterEngine16::~GRasterEngine16() {
}

inline CHAR* FASTCALL GRasterEngine16::SetDefaultName() {
	strcpy(Name,"raster engine 16 bit");
    return Name;
}

inline BOOL FASTCALL GRasterEngine16::RealizePalette(void) {
 	MakeShadingLUT();
 	MakeBumpLUT();
 	MakeGlassLUT();
 	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
// Implementation of GRasterEngine15
//
inline GRasterEngine15::GRasterEngine15(void) : GRasterEngine16() {
 	FrameDepth=15;
 	SoftenMask=0x7bde7bde;
}

inline GRasterEngine15::~GRasterEngine15() {
}

inline CHAR* FASTCALL GRasterEngine15::SetDefaultName() {
	strcpy(Name,"raster engine 15 bit");
    return Name;
}


///////////////////////////////////////////////////////////////////////////////
//
// Dummy raster procs
//
#pragma argsused
inline void FASTCALL GRasterEngine::Preprocess(DWORD f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::Postprocess(DWORD f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawNothing(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawVertex(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawWire(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawWireGlass(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawWireSoft(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawFilled(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawFilledLambert(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawFilledGouraud(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawFilledPhong(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawFilledGlass(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawFilledGlassLambert(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawFilledGlassGouraud(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawFilledGlassPhong(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawCarva(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawCarvaLambert(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawCarvaGouraud(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawCarvaPhong(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawCarvaGlass(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawCarvaLambertGlass(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawCarvaGouraudGlass(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawCarvaPhongGlass(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawTextured(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawTexturedLambert(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawTexturedGouraud(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawTexturedPhong(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawTexturedGlass(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawTexturedLightmap(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawTexturedChrome(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawTexturedBumpLightmap(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawTexturedBumpChrome(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawPersTextured(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawPersTexturedLambert(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawPersTexturedGouraud(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawPersTexturedPhong(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawPersTexturedGlass(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawPersTexturedLightmap(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawPersTexturedChrome(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawPersTexturedBumpLightmap(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawPersTexturedBumpChrome(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawChrome(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawChromeLambert(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawChromeGouraud(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawChromePhong(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawChromeGlass(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawBump(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawBumpLambert(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawBumpGouraud(GFace *f)
{
}

#pragma argsused
inline void FASTCALL GRasterEngine::DrawBumpPhong(GFace *f)
{
}

#pragma argsused
inline void	FASTCALL GRasterEngine::DrawBumpGlass(GFace *f)
{
}

#pragma argsused
inline void	FASTCALL GRasterEngine::DrawBumpChrome(GFace *f)
{
}

#pragma argsused
inline void	FASTCALL GRasterEngine::DrawBumpLightmap(GFace *f)
{
}

#endif
