///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#ifndef __GVEWORLD_H
#define __GVEWORLD_H

#include "../include/gvepro.h"

#define GVE_WRLD_MOTION_BLURR			((DWORD)0x000001)
#define GVE_WRLD_NOSORT					((DWORD)0x000002)
#define GVE_WRLD_SOFTEN_HORIZONTAL		((DWORD)0x000004)
#define GVE_WRLD_SOFTEN_VERTICAL		((DWORD)0x000008)
#define GVE_WRLD_MOTION_TRACE			((DWORD)0x000010)
#define GVE_WRLD_CLEAR_FRAME			((DWORD)0x000020)
#define GVE_WRLD_FEED_BACK				((DWORD)0x000040)
#define GVE_WRLD_RENDERING				((DWORD)0x000080)
#define GVE_WRLD_BUSY     				((DWORD)0x000100)


class GVectorWorld:public GObject {
		friend class 	GVectorSystem;

		public:
        enum {
        	WORLD_TYPE = MAKETYPE('W','R','L','D')
        };

		#pragma option -a1
		typedef struct {
				CHAR		Name[GMZ_NAME_LENGTH];
				DWORD		Flags;
				DWORD		Type;
				DWORD		Magic;
				DWORD		CameraMagic;
				DWORD		LightMagic;
				DWORD		AmbientMagic;
	    } FileWorldHead;
		#pragma option -a.

		public:
		GObjectTree 	ObjectList;
		GBitmapList 	TextureMapList;
		GBitmapList 	BumpMapList;
		GBitmapList 	LightMapList;
		GPalette 		Palette;
		GInfoObject		Info;

		private:
		GCameraObject 	*Camera;
		GLightObject 	*Light;
		GLightObject 	*Ambient;

		public:
		GVectorWorld(CHAR *name=NULL,DWORD typ=GVectorWorld::WORLD_TYPE);
		~GVectorWorld();

        CHAR*			FASTCALL SetDefaultName();

		BOOL			FASTCALL Alloc(void);
		BOOL			FASTCALL Free(void);

		GInfoObject*	FASTCALL SetInfo(GInfoObject *);
		GCameraObject* 	FASTCALL SetCamera(GCameraObject *);
		GLightObject*  	FASTCALL SetLight(GLightObject *);
		GLightObject*  	FASTCALL SetAmbientLight(GLightObject *);
		GVectorObject* 	FASTCALL FindObject(GVectorObject *);
		GVectorObject* 	FASTCALL FindObject(DWORD );
		GVectorObject* 	FASTCALL FindType(GVectorObject *,DWORD );
		void 	    	FASTCALL RealizePosition(BOOL f=FALSE);
		BOOL			FASTCALL Optimize(void);

		GInfoObject*	FASTCALL GetInfo(void);
		GCameraObject* 	FASTCALL GetCamera(void);
		GLightObject* 	FASTCALL GetLight(void);
		GLightObject* 	FASTCALL GetAmbientLight(void);
		GVectorObject*	FASTCALL GetRootObject(void);
		CHAR*			FASTCALL GetFlagsName(CHAR *);

		GPalette*		FASTCALL GetPalette(void);
		GPalette*		FASTCALL GetPalette(GPalette *);
		GPalette*	    FASTCALL SetPalette(GPalette *);

		INT				FASTCALL GetObjectCount(void);
		INT				FASTCALL GetFaceCount(void);
		INT				FASTCALL GetVertexCount(void);

		BOOL			FASTCALL Write(GFile &,DWORD );
		BOOL			FASTCALL Read(GFile &);

		BOOL			FASTCALL Load(char *);
		BOOL			FASTCALL Save(char *);

		private:
		BOOL			FASTCALL WriteObjects(GFile &,GVectorObject *);
		BOOL			FASTCALL ReadObjects(GFile &,GVectorObject *);
	   };


inline GVectorWorld::GVectorWorld(CHAR* name,DWORD typ) : GObject(name,typ) {
 	(PTR)Camera=
 	(PTR)Light=
 	(PTR)Ambient=NULL;
}

inline GVectorWorld::~GVectorWorld() {
 	Free();
}

inline CHAR* FASTCALL GVectorWorld::SetDefaultName() {
	strcpy(Name,"vector world");
    return Name;
}

inline GInfoObject* FASTCALL GVectorWorld::GetInfo(void) {
 	return &Info;
}

inline GInfoObject* FASTCALL GVectorWorld::SetInfo(GInfoObject *i) {
 	Info=*i;
 	return &Info;
}

inline GPalette* FASTCALL GVectorWorld::GetPalette(void) {
 	return &Palette;
}

inline GPalette* FASTCALL GVectorWorld::GetPalette(GPalette *pal) {
 	*pal=Palette;
 	return pal;
}

inline GPalette* FASTCALL GVectorWorld::SetPalette(GPalette *pal) {
 	Palette=*pal;
 	return &Palette;
}

inline GCameraObject* FASTCALL GVectorWorld::GetCamera(void) {
 	return Camera;
}

inline GLightObject* FASTCALL GVectorWorld::GetLight(void) {
 	return Light;
}

inline GLightObject* FASTCALL GVectorWorld::GetAmbientLight(void) {
 	return Ambient;
}

inline GVectorObject* FASTCALL GVectorWorld::GetRootObject(void) {
 	return (GVectorObject *)ObjectList.GetRoot();
}

inline GVectorObject* FASTCALL GVectorWorld::FindObject(GVectorObject *o) {
 	return (GVectorObject *)ObjectList.FindEntry(o);
}

inline GVectorObject* FASTCALL GVectorWorld::FindObject(DWORD magic) {
 	return (GVectorObject *)ObjectList.FindEntry(magic);
}

inline GVectorObject* FASTCALL GVectorWorld::FindType(GVectorObject *start,DWORD type) {
 	return (GVectorObject *)ObjectList.FindType(start,type);
}

#endif
