///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#include "../include/gomez.h"
const CHAR GDisplay::CLASSNAME[]="GOSDISPLAY";

BOOL FASTCALL GDisplay::Open(CHAR *dtitle,INT width,INT height,INT depth,DWORD flags) {
 	Width=Height=Depth=LineSize=0;
 	AspectRatio=1.0;

 	width=(width+3) & (~3);
 	strcpy(Name,dtitle);

 	Width=width;
 	Height=height;
 	Depth=depth;
 	Flags|=flags;
 	LineSize=Width*((depth+7)>>3);

 	ClipWindow.Left=0;
 	ClipWindow.Top=0;
 	ClipWindow.Right=width-1;
 	ClipWindow.Bottom=height-1;

   #if defined(__FORWINDOWS__)
 	RECT rect;
  	DWORD style;

  	if (Flags & RESIZEABLE)
		style=WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MAXIMIZEBOX | WS_SIZEBOX;
  	else
		style=WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU;

  	rect.left=1;
  	rect.top=1;
  	rect.bottom=height;
  	rect.right=width;

  	AdjustWindowRectEx(&rect,style,FALSE,WS_EX_CLIENTEDGE);
  	if ((WindowHandle = CreateWindowEx(WS_EX_CLIENTEDGE,CLASSNAME,
							   dtitle, style,
							   CW_USEDEFAULT, CW_USEDEFAULT,
							   rect.right-rect.left+1,rect.bottom-rect.top+1,
							   NULL, NULL, GWindowInstance, (LPVOID)this))==NULL) return FALSE;
  	if ((DeviceHandle=GetDC(WindowHandle))==NULL)
    	return FALSE;
  	SetMapMode(DeviceHandle,MM_TEXT);
  	AspectRatio=(DOUBLE)GetDeviceCaps(DeviceHandle,ASPECTX)/(DOUBLE)GetDeviceCaps(DeviceHandle,ASPECTX);
  	if ((depth<=8) || (GetDeviceCaps(DeviceHandle,RASTERCAPS) & RC_PALETTE)) {
		Flags|=PALETTE;
		if((Palette=(LOGPALETTE *)malloc(sizeof(LOGPALETTE)+256*sizeof(PALETTEENTRY)))==NULL)
        	return FALSE;
		Palette->palVersion=0x300;
		Palette->palNumEntries=256;
		PaletteHandle=CreatePalette(Palette);
		OldPaletteHandle=SelectPalette(DeviceHandle,PaletteHandle,FALSE);
   	}

  	ShowWindow(WindowHandle,SW_SHOW);
  	UpdateWindow(WindowHandle);
   #elif defined(__FORX11__)
  	INT i;
  	GraphicHandle=DefaultGC(GDisplayHandle,GScreenHandle);
  	WindowHandle=XCreateSimpleWindow(GDisplayHandle,GRootWindowHandle,0,0,
			       width,height,CopyFromParent,CopyFromParent,CopyFromParent);

  	PaletteHandle=XCreateColormap(GDisplayHandle,GRootWindowHandle,GDisplayVisual,AllocNone);
  	Palette=malloc(sizeof(XColor)*256);
  	for(i=0;i<256;i++) {
  		if(!XAllocColorCells(GDisplayHandle,PaletteHandle,0,0,0,Colours+i,1)) {
   			return(0);
  		} else {
	    	Palette[i].flags=DoRed|DoGreen|DoBlue;
	    	Palette[i].red=i;
	    	Palette[i].green=i;
	    	Palette[i].blue=i;
	    	Palette[i].pixel=Colours[i];
	   	}
   	}
   #else
  	return FALSE;
   #endif

 	return TRUE;
}


BOOL FASTCALL GDisplay::Close(void)
{
 #if defined(__FORWINDOWS__)
  if (WindowHandle==NULL) return TRUE;
  if (PaletteHandle)
   {
	if(SelectPalette(DeviceHandle,OldPaletteHandle,TRUE)==NULL) return FALSE;
    ::UnrealizeObject(PaletteHandle);
	if(DeleteObject(PaletteHandle)==FALSE) return FALSE;
	PaletteHandle=NULL;
   }
  if (Palette)
   {
	free(Palette);
	Palette=NULL;
   }
  if (TimerSpeed)
   {
	KillTimer(WindowHandle,0);
	TimerSpeed=0;
   }
  ReleaseDC(WindowHandle,DeviceHandle);
  DeviceHandle=NULL;
  if(DestroyWindow(WindowHandle)==FALSE) return FALSE;
  WindowHandle=NULL;
 #elif defined(__FORX11__)
  if (Palette)
   {
	free(Palette);
	Palette=NULL;
   }
  WindowHandle=NULL;
  PaletteHandle=NULL;
  GraphicHandle=NULL;
 #endif

 Flags=Width=Height=Depth=LineSize=0;
 return TRUE;
}


BOOL FASTCALL GDisplay::Suspend(void)
{
 if (!SuspendCount++)
	return OnSuspend();
 else
	return TRUE;
}


BOOL FASTCALL GDisplay::Resume(void)
{
 if (--SuspendCount)
	return OnResume();
 else
	return TRUE;
}


BOOL FASTCALL GDisplay::Start(LONG p1)
{
 if (Flags & STARTED) return TRUE;
 SuspendCount=0;
 Flags|=STARTED;
 MouseButtons=RBUTTONUP | MBUTTONUP | LBUTTONUP;
 if(!OnStart(p1)) return FALSE;

 #if defined(__FORWINDOWS__)
  MSG Message;
  while (::GetMessage(&Message, NULL, 0, 0))
   {
	TranslateMessage(&Message);
	DispatchMessage(&Message);
   }
 #elif defined(__FORX11__)
  XEvent report;
  while((!SuspendCount) && (Flags & STARTED))
   {
  	if(XCheckWindowEvent(GDisplayHandle,WindowHandle,KeyPressMask,&report)==1)
   		OnKey(((XKeyEvent*)&report)->keycode);
   }
 #endif

 return TRUE;
}


BOOL FASTCALL GDisplay::Terminate(void)
{
 if (!(Flags & STARTED)) return FALSE;
 SuspendCount=0;
 if (!OnTerminate()) return FALSE;
 Flags&=~STARTED;

 #if defined(__FORWINDOWS__)
  PostQuitMessage(0);
 #elif defined(__FORX11__)

 #else

 #endif
 return TRUE;
}


LONG FASTCALL GDisplay::SetTimerSpeed(LONG s)
{
 #if defined(__FORWINDOWS__)
  if (s)
	SetTimer(WindowHandle,0,(UINT)s,NULL);
  else
	KillTimer(WindowHandle,0);
  TimerSpeed=s;
  return s;
 #elif defined(__FORX11__)

 #else
  return 0;
 #endif
}


GPalette* FASTCALL GDisplay::SetPalette(GPalette *pal,INT first,INT count)
{
 if(!Palette) return NULL;

 #if defined(__FORWINDOWS__)
  INT i;
  for(i=first;i<first+count;i++)
   {
	Palette->palPalEntry[i].peRed=pal->GetColour(i).R;
	Palette->palPalEntry[i].peGreen=pal->GetColour(i).G;
	Palette->palPalEntry[i].peBlue=pal->GetColour(i).B;
	Palette->palPalEntry[i].peFlags=PC_NOCOLLAPSE;
   }
  if(SetPaletteEntries(PaletteHandle,first,count,&Palette->palPalEntry[first])==0) return NULL;
 #elif defined(__FORX11__)
  INT i;
  for(i=first;i<first+count;i++)
   {
    Palette[i].flags=DoRed|DoGreen|DoBlue;
    Palette[i].red=i;
    Palette[i].green=i;
    Palette[i].blue=i;
    Palette[i].pixel=Colours[i];
   }
 #else
  Palette->SetPalette(pal,first,count);
 #endif

 return pal;
}


GPalette* FASTCALL GDisplay::GetPalette(GPalette *pal,INT first,INT count)
{
 if(!Palette) return NULL;

 #if defined(__FORWINDOWS__)
  INT i;
  for(i=first;i<first+count;i++)
	pal->SetColour(i,Palette->palPalEntry[i].peRed,
    			     Palette->palPalEntry[i].peGreen,
	                 Palette->palPalEntry[i].peBlue);
 #elif defined(__FORX11__)
  INT i;
  for(i=first;i<first+count;i++)
	pal->SetColour(i,Palette[i].red,
    			     Palette[i].green,
	                 Palette[i].blue);
 #else
  pal->SetPalette(Palette,first,count);
 #endif

 return pal;
}


BOOL FASTCALL GDisplay::RealizePalette(void)
{
 if(!Palette) return FALSE;

 #if defined(__FORWINDOWS__)
  SelectPalette(DeviceHandle,PaletteHandle,FALSE);
  if(::RealizePalette(DeviceHandle)==GDI_ERROR) return FALSE;
 #elif defined(__FORX11__)
  XStoreColors(GDisplayHandle,PaletteHandle,Palette,256);
 #else

 #endif
 return TRUE;
}


BOOL FASTCALL GDisplay::UnrealizePalette(void)
{
 if (!Palette) return FALSE;

 #if defined(__FORWINDOWS__)
  UnrealizeObject(PaletteHandle);
  SelectPalette(DeviceHandle,OldPaletteHandle,FALSE);
 #elif defined(__FORX11__)

 #else

 #endif
 return TRUE;
}


GColour FASTCALL GDisplay::SetColour(INT i,GColour &c)
{
 if(!Palette) return GColour(0,0,0);

 #if defined(__FORWINDOWS__)
  Palette->palPalEntry[i].peRed=c.R;
  Palette->palPalEntry[i].peGreen=c.G;
  Palette->palPalEntry[i].peBlue=c.B;
  Palette->palPalEntry[i].peFlags=PC_NOCOLLAPSE;
  SetPaletteEntries(PaletteHandle,i,1,&Palette->palPalEntry[i]);
 #elif defined(__FORX11__)
  Palette[i].flags=DoRed|DoGreen|DoBlue;
  Palette[i].red=c.R;
  Palette[i].green=c.G;
  Palette[i].blue=c.B;
  Palette[i].pixel=Colours[i];
 #else
  Palette->SetColour(i,c);
 #endif

 return c;
}


CHAR* FASTCALL GDisplay::GetFlagsName(CHAR *buffer)
{
 buffer[0]=0;

 if (Flags & PALETTE) 		strcat(buffer,"PALETTE ");

 return buffer;
}


#if defined(__FORWINDOWS__)
LRESULT CALLBACK GDisplay::WindowProc(HWND hwnd, UINT Message,WPARAM wParam, LPARAM lParam)
{
 GDisplay *disp=(GDisplay *)GetWindowLong(hwnd,GWL_USERDATA);
 if (disp->SuspendCount)
	return DefWindowProc(hwnd, Message, wParam, lParam);

 switch(Message)
   {
	case WM_CLOSE:
		if (disp->Flags & STARTED)
			disp->OnTerminate();
		disp->Flags&=~(STARTED);
		disp->SuspendCount=0;
		return 0;
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	case WM_TIMER:
		disp->OnTimer((LONG)wParam,0);
		return 0;
	case WM_KEYDOWN:
		disp->OnKeyDown((LONG)wParam);
		return 0;
	case WM_KEYUP:
		disp->OnKeyUp((LONG)wParam);
		return 0;
	case WM_CHAR:
		disp->OnChar((LONG)wParam);
		return 0;
	case WM_PAINT:
		PAINTSTRUCT PaintStruct;
		BeginPaint(hwnd, &PaintStruct);
		disp->OnPaint();
		EndPaint(hwnd, &PaintStruct);
		return 0;
	case WM_MOUSEMOVE:
		disp->MousePosition.x=LOWORD(lParam);
		disp->MousePosition.y=HIWORD(lParam);
		disp->OnMouseMove(disp->MouseButtons,LOWORD(lParam),HIWORD(lParam));
		return 0;
	case WM_MBUTTONDOWN:
		disp->MousePosition.x=LOWORD(lParam);
		disp->MousePosition.y=HIWORD(lParam);
		disp->MouseButtons|=MBUTTONDOWN;
		disp->MouseButtons&=~MBUTTONUP;
		disp->OnMouseButton(MBUTTONDOWN,LOWORD(lParam),HIWORD(lParam));
		return 0;
	case WM_MBUTTONUP:
		disp->MousePosition.x=LOWORD(lParam);
		disp->MousePosition.y=HIWORD(lParam);
		disp->MouseButtons|=MBUTTONUP;
		disp->MouseButtons&=~MBUTTONDOWN;
		disp->OnMouseButton(MBUTTONUP,LOWORD(lParam),HIWORD(lParam));
		return 0;
	case WM_LBUTTONDOWN:
		disp->MousePosition.x=LOWORD(lParam);
		disp->MousePosition.y=HIWORD(lParam);
		disp->MouseButtons|=LBUTTONDOWN;
		disp->MouseButtons&=~LBUTTONUP;
		disp->OnMouseButton(LBUTTONDOWN,LOWORD(lParam),HIWORD(lParam));
		return 0;
	case WM_LBUTTONUP:
		disp->MousePosition.x=LOWORD(lParam);
		disp->MousePosition.y=HIWORD(lParam);
		disp->MouseButtons|=LBUTTONUP;
		disp->MouseButtons&=~LBUTTONDOWN;
		disp->OnMouseButton(LBUTTONUP,LOWORD(lParam),HIWORD(lParam));
		return 0;
	case WM_RBUTTONDOWN:
		disp->MousePosition.x=LOWORD(lParam);
		disp->MousePosition.y=HIWORD(lParam);
		disp->MouseButtons|=RBUTTONDOWN;
		disp->MouseButtons&=~RBUTTONUP;
		disp->OnMouseButton(RBUTTONDOWN,LOWORD(lParam),HIWORD(lParam));
		return 0;
	case WM_RBUTTONUP:
		disp->MousePosition.x=LOWORD(lParam);
		disp->MousePosition.y=HIWORD(lParam);
		disp->MouseButtons|=RBUTTONUP;
		disp->MouseButtons&=~RBUTTONDOWN;
		disp->OnMouseButton(RBUTTONUP,LOWORD(lParam),HIWORD(lParam));
		return 0;
	case WM_SIZE:
		disp->Width=LOWORD(lParam);
		disp->Height=HIWORD(lParam);
		disp->LineSize=disp->Width*((disp->Depth+7)>>3);
		disp->SetClipWindow(0,0,disp->Width-1,disp->Height-1);
		disp->OnResize(LOWORD(lParam),HIWORD(lParam));
		return 0;
    case WM_MESSAGE_NOTIFY:
    	disp->OnMessage();
        return 0;
    case WM_ACTIVATE:
    	if ((LOWORD(wParam)==WA_ACTIVE) || (LOWORD(wParam)==WA_CLICKACTIVE))
	    	disp->OnActivate();
        else
        	disp->OnDeactivate();
    	return 0;
  	case WM_ERASEBKGND:
      	return 1;
    case WM_QUERYNEWPALETTE:
    	if (!disp->PaletteHandle) return 0;
  		SelectPalette(disp->DeviceHandle,disp->PaletteHandle,FALSE);
  		::RealizePalette(disp->DeviceHandle);
		return 1;
    case WM_PALETTECHANGED:
    	if ((hwnd!=(HWND)wParam) && (disp->PaletteHandle))
           {
 			SelectPalette(disp->DeviceHandle,disp->PaletteHandle,TRUE);
	  		::RealizePalette(disp->DeviceHandle);
            return 1;
           }
		return 0;
	default:
		return DefWindowProc(hwnd, Message, wParam, lParam);
   }
}


LRESULT CALLBACK GDisplay::WindowStartProc(HWND hwnd, UINT Message,WPARAM wParam, LPARAM lParam) {
 	GDisplay *tc;

 	if (Message==WM_CREATE) {
		tc=(GDisplay *)(((CREATESTRUCT *)lParam)->lpCreateParams);
		SetWindowLong(hwnd,GWL_USERDATA,(LONG)tc);
		SetWindowLong(hwnd,GWL_WNDPROC,(LONG)tc->GetWindowProc());
		return 0;
   	}
 	return DefWindowProc(hwnd, Message, wParam, lParam);
}

void FASTCALL GDisplay::UpdateWindowStyle(void) {
	LONG style;
	if (Flags & RESIZEABLE)
		style=WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MAXIMIZEBOX | WS_SIZEBOX;
	else
		style=WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU;
	if (WindowHandle)
	    SetWindowLong(WindowHandle,GWL_STYLE,style);
}
#endif
