///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#include "../include/gomez.h"


BOOL FASTCALL GFile::Open(char *fname)
{
 DWORD ver;

 if((Handle=fopen(fname,"rb"))==NULL) return FALSE;
 if(fread(FileType,sizeof(FileType),1,Handle)!=1) return FALSE;
 if(fread(&ver,sizeof(ver),1,Handle)!=1) return FALSE;

 Mode=GMZ_MODE_READ;
 FileVersion=LSB2DWORD(ver);
 ChunkLevel=0;
 ChunkStart[0]=0;
 ChunkSize[0]=12;
 ChunkPosition[0]=12;
 ChunkCompression[0]=COMPRESSION_SUBS;
 ChunkType[0]=MAIN_TYPE;

 return TRUE;
}


BOOL FASTCALL GFile::Close(void)
{
 if (!Handle) return TRUE;
 if (Mode==GMZ_MODE_READ)
	{
	 if (fclose(Handle)!=0) return FALSE;
	}
 if (Mode==GMZ_MODE_WRITE)
	{
	 while(ChunkLevel) CloseChunk();
	 CreateChunk(END_TYPE,COMPRESSION_NONE);
	 CloseChunk();
	 if (fclose(Handle)!=0) return FALSE;
	}

 Handle=NULL;
 Mode=0;

 return TRUE;
}


BOOL FASTCALL GFile::Create(char *fname,DWORD comp)
{
 DWORD ver=DWORD2LSB(FileVersion);

 if((Handle=fopen(fname,"w+b"))==NULL) return FALSE;
 if(fwrite(FileType,sizeof(FileType),1,Handle)!=1) return FALSE;
 if(fwrite(&ver,sizeof(ver),1,Handle)!=1) return FALSE;

 Mode=GMZ_MODE_WRITE;
 StandardCompression=comp;
 ChunkLevel=0;
 ChunkStart[0]=0;
 ChunkSize[0]=12;
 ChunkPosition[0]=12;
 ChunkCompression[0]=COMPRESSION_SUBS;
 ChunkType[0]=MAIN_TYPE;

 return TRUE;
}


DWORD FASTCALL GFile::OpenChunk(void)
{
 ChunkHeader H;

 if(Mode!=GMZ_MODE_READ) return NULL;
 if(ChunkCompression[ChunkLevel]!=COMPRESSION_SUBS) return NULL;

 ChunkLevel++;
 ChunkStart[ChunkLevel]=ChunkStart[ChunkLevel-1]+ChunkPosition[ChunkLevel-1];

 fseek(Handle,ChunkStart[ChunkLevel],SEEK_SET);
 if(fread(&H,sizeof(ChunkHeader),1,Handle)!=1) return NULL;
 ChunkSize[ChunkLevel]=LSB2DWORD(H.Size);
 ChunkCompression[ChunkLevel]=LSB2DWORD(H.Compression);
 ChunkType[ChunkLevel]=LSB2DWORD(H.Type);
 ChunkPosition[ChunkLevel]=12;
 ChunkPosition[ChunkLevel-1]+=ChunkSize[ChunkLevel];

 if(ChunkType[ChunkLevel]==END_TYPE)
	CloseChunk();

 return LSB2DWORD(H.Type);
}


BOOL FASTCALL GFile::CloseChunk(void)
{
 if(Mode==GMZ_MODE_READ)
	{
	 if(ChunkLevel)  ChunkLevel--;
	 return TRUE;
	}

 if(Mode==GMZ_MODE_WRITE)
	{
	 if(ChunkLevel)
	   {
		if (ChunkCompression[ChunkLevel]==COMPRESSION_SUBS)
		   {
			OpenChunk(END_TYPE,COMPRESSION_NONE);
			CloseChunk();
		   }

		DWORD size=DWORD2LSB(ChunkSize[ChunkLevel]);

		fseek(Handle,ChunkStart[ChunkLevel]+4,SEEK_SET);
		fwrite(&size,sizeof(size),1,Handle);
		fseek(Handle,ChunkStart[ChunkLevel]+ChunkSize[ChunkLevel],SEEK_SET);
		ChunkSize[ChunkLevel-1]+=ChunkSize[ChunkLevel];
		ChunkPosition[ChunkLevel-1]+=ChunkSize[ChunkLevel];
		ChunkLevel--;
	   }
	 return TRUE;
	}

 return FALSE;
}


BOOL FASTCALL GFile::CreateChunk(DWORD typ,DWORD comp)
{
 ChunkHeader H;

 if(Mode!=GMZ_MODE_WRITE) return FALSE;
 if(ChunkCompression[ChunkLevel]!=COMPRESSION_SUBS) return FALSE;

 ChunkLevel++;
 ChunkStart[ChunkLevel]=ChunkStart[ChunkLevel-1]+ChunkPosition[ChunkLevel-1];

 if (comp==COMPRESSION_STANDARD) comp=StandardCompression;
 H.Type=DWORD2LSB(typ);
 H.Compression=DWORD2LSB(comp);
 H.Size=DWORD2LSB(12);
 ChunkSize[ChunkLevel]=12;
 ChunkPosition[ChunkLevel]=12;
 ChunkCompression[ChunkLevel]=comp;
 ChunkType[ChunkLevel]=typ;

 fseek(Handle,ChunkStart[ChunkLevel],SEEK_SET);
 if(fwrite(&H,sizeof(ChunkHeader),1,Handle)!=1) return FALSE;

 return TRUE;
}


INT FASTCALL GFile::Write(void *src,INT size)
{
 if(Mode!=GMZ_MODE_WRITE) return NULL;

 switch(ChunkCompression[ChunkLevel])
	{
	 case COMPRESSION_NONE:
		if (fwrite(src,size,1,Handle)!=1) return 0;
		ChunkSize[ChunkLevel]+=size;
		ChunkPosition[ChunkLevel]+=size;
		return size;
	 case COMPRESSION_SUBS:
		return NULL;
	 default:
		return NULL;
	}
}


INT FASTCALL GFile::Read(void *dst,INT size)
{
 if(Mode!=GMZ_MODE_READ) return NULL;

 switch(ChunkCompression[ChunkLevel])
	{
	 case COMPRESSION_NONE:
		if(fread(dst,size,1,Handle)!=1) return 0;
		ChunkPosition[ChunkLevel]+=size;
		return size;
	 case COMPRESSION_SUBS:
		return NULL;
	 default:
		return NULL;
	}
}


INT FASTCALL GFile::WriteChunk(void *src,INT size)
{
 LZWCodec *LZW;
 void *buf;
 INT ret;
 if(Mode!=GMZ_MODE_WRITE) return NULL;

 switch(ChunkCompression[ChunkLevel])
	{
	 case COMPRESSION_NONE:
		fseek(Handle,ChunkStart[ChunkLevel]+12,SEEK_SET);
		if (fwrite(src,size,1,Handle)!=1) return 0;
		ChunkSize[ChunkLevel]+=size;
		ChunkPosition[ChunkLevel]+=size;
		return size;
	 case COMPRESSION_SUBS:
		return NULL;
	 case COMPRESSION_LZW:
		LZW=new LZWCodec;
		ret=size;
		size=LZW->Compress(src,buf,ret);
		fseek(Handle,ChunkStart[ChunkLevel]+12,SEEK_SET);
		if(fwrite(buf,size,1,Handle)!=1) return 0;
		ChunkSize[ChunkLevel]+=size;
		ChunkPosition[ChunkLevel]+=size;
		delete LZW;
		if(buf) free(buf);
		return ret;
	 default:
		return NULL;
	}
}


INT FASTCALL GFile::ReadChunk(void **dst)
{
 INT s;
 void *buf;
 LZWCodec *LZW;
 if(Mode!=GMZ_MODE_READ) return NULL;

 switch(ChunkCompression[ChunkLevel])
	{
	 case COMPRESSION_NONE:
		if((*dst=malloc(ChunkSize[ChunkLevel]-12))==NULL) return 0;
		fseek(Handle,ChunkStart[ChunkLevel]+12,SEEK_SET);
		if(fread(*dst,ChunkSize[ChunkLevel]-12,1,Handle)!=1) return 0;
		ChunkPosition[ChunkLevel]+=ChunkSize[ChunkLevel]-12;
		return ChunkSize[ChunkLevel]-12;
	 case COMPRESSION_SUBS:
		return NULL;
	 case COMPRESSION_LZW:
		LZW=new LZWCodec;
		if((buf=malloc(ChunkSize[ChunkLevel]-12))==NULL) return 0;
		fseek(Handle,ChunkStart[ChunkLevel]+12,SEEK_SET);
		if(fread(buf,ChunkSize[ChunkLevel]-12,1,Handle)!=1) return 0;
		s=LZW->Decompress(buf,*dst,ChunkSize[ChunkLevel]-12);
		ChunkPosition[ChunkLevel]+=ChunkSize[ChunkLevel]-12;
		delete LZW;
		free(buf);
		return s;
	 default:
		return NULL;
	}
}


/******************************************************************************
 **
 ** These are the LZW compression routines...
 ** They are adapted from Steve Rimmers Book "Supercharged Bitmapped
 ** Graphics".
 **
 **/
BYTE LZWCodec::MaskTable[8]={0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};
BYTE LZWCodec::BitTable[8]= {0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80};
UINT LZWCodec::WordMaskTable[16] = {0x0000,0x0001,0x0003,0x0007,
				  0x000f,0x001f,0x003f,0x007f,
				  0x00ff,0x01ff,0x03ff,0x07ff,
				  0x0fff,0x1fff,0x3fff,0x7fff };


void FASTCALL LZWCodec::InitWriteTable(void)
{
 INT i;

 CodeSize=LZW_MIN_CODE_SIZE+1;
 ClearCode=(1<<LZW_MIN_CODE_SIZE);
 EofCode=ClearCode+1;
 FreeCode=ClearCode+2;
 MaxCode=(1<<CodeSize);

 for(i=0;i<LZW_TABLE_SIZE;i++) CurrentTable[i]=0;
}



void FASTCALL LZWCodec::WriteCode(int code)
{
 UINT temp;
 UINT ByteOffset;

 ByteOffset = BitOffset >> 3;
 BitsLeft = BitOffset & 7;

 if(BitsLeft > 0)
   {
	temp = ((int) code << BitsLeft) | WriteBuffer[ByteOffset];
	WriteBuffer[ByteOffset]=(BYTE)temp;
	WriteBuffer[ByteOffset+1]=(BYTE)(temp >> 8);
	WriteBuffer[ByteOffset+2]=(BYTE)(temp >> 16);
   }
 else
   {
	WriteBuffer[ByteOffset]=(BYTE)code;
	WriteBuffer[ByteOffset+1]=(BYTE)(code >> 8);
   }
 BitOffset += CodeSize;
}




INT FASTCALL LZWCodec::Compress(void *src,void *&dst,INT size)
{
 INT hx,d;
 INT csize,usize;

 csize=16;
 usize=size;
 Source=src;
 if((Destination=malloc(size*2))==NULL)	return NULL;

 *((DWORD *)Destination)=DWORD2LSB(csize);
 *((DWORD *)Destination+1)=DWORD2LSB(usize);
 WriteBuffer=(BYTE *)Destination+16;
 ReadBuffer=(BYTE *)Source;

 BitOffset=0;
 SuffixChar=0x0ffff;

 InitWriteTable();
 WriteCode(ClearCode);

 SuffixChar=(int)*ReadBuffer++;
 size--;
 PrefixCode = SuffixChar;

 while (size)
   {
	SuffixChar=(int)*ReadBuffer++;
	size--;

	/* derive an index into the code table */
	hx=(PrefixCode ^ (SuffixChar << 5)) % LZW_TABLE_SIZE;
	d=1;

	for(;;)
	   {
		if(CurrentTable[hx] == 0)
		   {
			WriteCode(PrefixCode);
			d = FreeCode;

			/* find the next free code */
			if(FreeCode <= LZW_LARGEST_CODE)
			   {
				PrefixTable[hx] = PrefixCode;
				SuffixTable[hx] = (BYTE)SuffixChar;
				CurrentTable[hx] = FreeCode;
				FreeCode++;
			   }

			/* expand the code size or scrap the table */
			if(d == MaxCode)
			   {
				if(CodeSize < LZW_MAX_CODE_SIZE)
				   {
					CodeSize++;
					MaxCode <<= 1;
				   }
				else
				   {
					WriteCode(ClearCode);
					InitWriteTable();
				   }
			   }
			PrefixCode = SuffixChar;
			break;
		   }
		if(PrefixTable[hx] == PrefixCode && SuffixTable[hx] == SuffixChar)
		   {
			PrefixCode = CurrentTable[hx];
			break;
		   }
		hx += d;
		d += 2;
		if(hx >= LZW_TABLE_SIZE) hx -= LZW_TABLE_SIZE;
	   }
   }

 WriteCode(PrefixCode);
 WriteCode(EofCode);

 csize=((BitOffset+7)/8)+16;
 *((INT *)Destination)=csize;
 *((INT *)Destination+1)=usize;
 *((INT *)Destination+2)=0;
 *((INT *)Destination+3)=0;

 if((dst=realloc(dst,csize))==NULL) return NULL;

 return csize;
}




void FASTCALL LZWCodec::InitReadTable()
{
 CodeSize = LZW_MIN_CODE_SIZE + 1;
 TopSlot = 1 << CodeSize;
 ClearCode = 1 << LZW_MIN_CODE_SIZE;
 EofCode = ClearCode + 1;
 Slot = FreeCode = EofCode + 1;
 BitsLeft = 0;
}




int FASTCALL LZWCodec::ReadCode(void)
{
 FAST INT ret;
 BYTE b;

 if (BitsLeft == 0)
   {
	b = *ReadBuffer++;
	BitsLeft = 8;
   }

 ret = b >> (8 - BitsLeft);
 while (CodeSize > BitsLeft)
   {
	b = *ReadBuffer++;
	ret |= b << BitsLeft;
	BitsLeft += 8;
   }
 BitsLeft -= CodeSize;
 ret &= WordMaskTable[CodeSize];
 return(ret);
}



#pragma argsused
INT FASTCALL LZWCodec::Decompress(void *src,void *&dst,INT size)
{
 FAST BYTE *sp;
 FAST INT code, fc, oc;
 INT c;
 INT usize;

 Source=src;
 if((dst=Destination=malloc(LSB2DWORD(*(DWORD *)Source+1)))==NULL) return NULL;

 InitReadTable();
 ReadBuffer=(BYTE *)Source+16;
 WriteBuffer=(BYTE *)Destination;
 usize=LSB2DWORD(*(DWORD *)Source+1);

 oc = fc = 0;
 sp = CodeStack;

 while ((c = ReadCode()) != EofCode)
   {
	if (c == ClearCode)
	   {
		CodeSize = LZW_MIN_CODE_SIZE + 1;
		Slot = FreeCode;
		TopSlot = 1 << CodeSize;

		while ((c = ReadCode()) == ClearCode);
		if (c == EofCode) break;
		if (c >= Slot) c = 0;
		oc = fc = c;
		*sp++ = (BYTE)c;
	   }
	else
	   {
		code = c;

		if (code >= Slot)
		   {
			code = oc;
			*sp++ = (BYTE)fc;
		   }

		while (code >= FreeCode)
		   {
			*sp++ = SuffixTable[code];
			code = PrefixTable[code];
		   }

		*sp++ = (BYTE)code;
		if (Slot < TopSlot)
		   {
			SuffixTable[Slot] = (BYTE)(fc = code);
			PrefixTable[Slot++] = oc;
			oc = c;
		   }
		if ((Slot >= TopSlot) && (CodeSize < LZW_MAX_CODE_SIZE))
		   {
			TopSlot <<= 1;
			++CodeSize;
		   }
	   }

	while (sp > CodeStack)
	   {
		--sp;
		*WriteBuffer++ = *sp;
	   }
   }

 return usize;
}



