///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#include "../include/gomez.h"


GFrame::GFrame(GDisplay *disp) : GAlphaBuffer(NULL,FRAME_TYPE)
{
 Display=disp;

 #if defined(__FORWINDOWS__)
  (PTR)DeviceHandle=
  (PTR)BitmapHandle=
  (PTR)OldBitmapHandle=
  (PTR)Palette=
  (PTR)BitmapInfo=NULL;
 #elif defined(__FORX11__)
  (PTR)Palette=
  (PTR)WindowHandle=
  (PTR)GraphicHandle=
  (PTR)PaletteHandle=NULL;
 #else
  Palette=NULL;
 #endif

 #if defined(__FORWINDOWS__)
  if (!disp)
	DeviceHandle=CreateCompatibleDC(NULL);
  else
	DeviceHandle=CreateCompatibleDC(disp->DeviceHandle);
  SetMapMode(DeviceHandle,MM_TEXT);
 #endif
}


BOOL FASTCALL GFrame::Alloc(INT width,INT height,INT depth)
{
 Width=Height=Depth=LineSize=0;
 AspectRatio=1.0;
 ImageBuffer=NULL;
 AlphaBuffer=NULL;

 #if defined(__FORWINDOWS__)
  if (depth==0) depth=GetDeviceCaps(DeviceHandle,BITSPIXEL);
  if (depth==24) depth=32;
  if (depth<=8)
	Flags|=PALETTE;
  else
	Flags&=~PALETTE;
 #else
  if (depth==0) depth=8;
  if (depth<=8) Flags|=PALETTE;
 #endif
 width=(width+3) & ~3;

 #if defined(__FORWINDOWS__)
  if((BitmapInfo=(BITMAPINFO *)malloc(sizeof(BITMAPINFO)+256*sizeof(RGBQUAD)))==NULL) return FALSE;
  Palette=&BitmapInfo->bmiColors[0];

  BitmapInfo->bmiHeader.biSize=sizeof(BITMAPINFO)-sizeof(RGBQUAD);
  BitmapInfo->bmiHeader.biWidth=width;
  BitmapInfo->bmiHeader.biHeight=-height;
  BitmapInfo->bmiHeader.biPlanes=1;
  BitmapInfo->bmiHeader.biBitCount=(WORD)((depth+7) & 0xf8);
  BitmapInfo->bmiHeader.biCompression=(depth==15 || depth==16) ? BI_BITFIELDS : BI_RGB;
  BitmapInfo->bmiHeader.biSizeImage=width*height*((depth+7)>>3);
  BitmapInfo->bmiHeader.biXPelsPerMeter=0;
  BitmapInfo->bmiHeader.biYPelsPerMeter=0;
  BitmapInfo->bmiHeader.biClrUsed=0;
  BitmapInfo->bmiHeader.biClrImportant=0;
  if(depth==15)
	{
	 ((DWORD *)Palette)[0]=0x7c00;
	 ((DWORD *)Palette)[1]=0x03e0;
	 ((DWORD *)Palette)[2]=0x001f;
	}
  else if (depth==16)
	{
	 ((DWORD *)Palette)[0]=0xf800;
	 ((DWORD *)Palette)[1]=0x07e0;
	 ((DWORD *)Palette)[2]=0x001f;
	}
  else if (depth<=8)
	{
	 for (INT i=0;i<256;i++)
	   {
		Palette[i].rgbRed=
		Palette[i].rgbGreen=
		Palette[i].rgbBlue=(BYTE)i;
		Palette[i].rgbReserved=0;
	   }
	}

  if((BitmapHandle=CreateDIBSection(DeviceHandle,BitmapInfo,DIB_RGB_COLORS,(void **)&ImageBuffer,NULL,NULL))==NULL) return FALSE;
  if((OldBitmapHandle=(HBITMAP)SelectObject(DeviceHandle,BitmapHandle))==NULL) return FALSE;

  Width=BitmapInfo->bmiHeader.biWidth;
  Height=-BitmapInfo->bmiHeader.biHeight;
  Depth=BitmapInfo->bmiHeader.biBitCount;
  AspectRatio=(DOUBLE)GetDeviceCaps(DeviceHandle,ASPECTX)/(DOUBLE)GetDeviceCaps(DeviceHandle,ASPECTX);
  LineSize=Width*((Depth+7)>>3);
 #elif defined(__FORX11__)

 #else
  if((ImageBuffer=(BYTE *)malloc(width*height*((depth+7)>>3)))==NULL) return FALSE;
  if(Flags & PALETTE)
	 if((Palette=new GPalette)==NULL) return FALSE;
  Width=width;
  Height=height;
  Depth=depth;
  LineSize=Width*((Depth+7)>>3);
 #endif

 if((AlphaBuffer=(GAlpha *)malloc(Width*Height*sizeof(GAlpha)))==NULL) return FALSE;
 Flags|=DEPTHBUFFER | STENCILBUFFER;
 SetPixelFunctions();
 SetAlphaFunctions();

 ClipWindow.Left=0;
 ClipWindow.Top=0;
 ClipWindow.Right=Width-1;
 ClipWindow.Bottom=Height-1;

 return TRUE;
}


BOOL FASTCALL GFrame::Resize(INT width,INT height)
{
 width=(width+3) & ~3;

 #if defined(__FORWINDOWS__)
  if(SelectObject(DeviceHandle,OldBitmapHandle)==NULL) return FALSE;
  if(DeleteObject(BitmapHandle)==FALSE) return FALSE;
  BitmapInfo->bmiHeader.biWidth=width;
  BitmapInfo->bmiHeader.biHeight=-height;
  BitmapInfo->bmiHeader.biSizeImage=width*height*((Depth+7)>>3);
  if(Depth==15)
	{
	 ((DWORD *)Palette)[0]=0x7c00;
	 ((DWORD *)Palette)[1]=0x03e0;
	 ((DWORD *)Palette)[2]=0x001f;
	}
  else if (Depth==16)
	{
	 ((DWORD *)Palette)[0]=0xf800;
	 ((DWORD *)Palette)[1]=0x07e0;
	 ((DWORD *)Palette)[2]=0x001f;
	}

  if((BitmapHandle=CreateDIBSection(DeviceHandle,BitmapInfo,DIB_RGB_COLORS,(void **)&ImageBuffer,NULL,NULL))==NULL) return FALSE;
  if((OldBitmapHandle=(HBITMAP)SelectObject(DeviceHandle,BitmapHandle))==NULL) return FALSE;
  Width=BitmapInfo->bmiHeader.biWidth;
  Height=-BitmapInfo->bmiHeader.biHeight;
  Depth=BitmapInfo->bmiHeader.biBitCount;
  LineSize=Width*((Depth+7)>>3);
 #elif defined(__FORX11__)

 #else
  Width=width;
  Height=height;
  LineSize=Width*((Depth+7)>>3);
  if((ImageBuffer=realloc(ImageBuffer,Width*Height*((Depth+7)>>3)))==NULL) return FALSE;
 #endif

 if((AlphaBuffer=(GAlpha *)realloc(AlphaBuffer,Width*Height*sizeof(GAlpha)))==NULL) return FALSE;

 ClipWindow.Left=0;
 ClipWindow.Top=0;
 ClipWindow.Right=Width-1;
 ClipWindow.Bottom=Height-1;

 return TRUE;
}


BOOL FASTCALL GFrame::Free(void)
{
 #if defined(__FORWINDOWS__)
  if (BitmapHandle)
   {
	if(SelectObject(DeviceHandle,OldBitmapHandle)==NULL) return FALSE;
	if(DeleteObject(BitmapHandle)==FALSE) return FALSE;
	BitmapHandle=NULL;
    ImageBuffer=NULL;
   }
  if (BitmapInfo)
   {
	free(BitmapInfo);
	BitmapInfo=NULL;
	Palette=NULL;
   }
 #elif defined(__FORX11__)

 #else
  if (ImageBuffer)
   {
	free(ImageBuffer);
	ImageBuffer=NULL;
   }
  if (Palette)
   {
	delete Palette;
	Palette=NULL;
   }
 #endif
 if (AlphaBuffer)
   {
	free(AlphaBuffer);
	AlphaBuffer=NULL;
   }

 Flags=Width=Height=Depth=LineSize=0;
 return TRUE;
}


BOOL FASTCALL GFrame::Put(GDisplay *disp,INT x,INT y)
{
 #if defined(__FORWINDOWS__)
  if (!disp->DeviceHandle) return FALSE;
  BitBlt(disp->DeviceHandle,x,y,Width,Height,DeviceHandle,0,0,SRCCOPY);
 #elif defined(__FORX11__)

 #else

 #endif
 return TRUE;
}


BOOL FASTCALL GFrame::Put(PTR hdc,INT x,INT y)
{
 #if defined(__FORWINDOWS__)
  BitBlt((HDC)hdc,x,y,Width,Height,DeviceHandle,0,0,SRCCOPY);
 #elif defined(__FORX11__)

 #else

 #endif
 return TRUE;
}


GPalette* FASTCALL GFrame::SetPalette(GPalette *pal,INT first,INT count)
{
 INT i;
 if(!Palette) return NULL;

 #if defined(__FORWINDOWS__)
  for(i=first;i<first+count;i++)
   {
	Palette[i].rgbRed=pal->GetColour(i).R;
	Palette[i].rgbGreen=pal->GetColour(i).G;
	Palette[i].rgbBlue=pal->GetColour(i).B;
	Palette[i].rgbReserved=0;
   }
  if(SetDIBColorTable(DeviceHandle,first,count,&Palette[first])==0) return NULL;
 #elif defined(__FORX11__)

 #else
  Palette->SetPalette(pal,first,count);
 #endif

 return pal;
}


GColour FASTCALL GFrame::SetColour(INT i,GColour &c)
{
 if(!Palette) return GColour(0,0,0);

 #if defined(__FORWINDOWS__)
  Palette[i].rgbRed=c.R;
  Palette[i].rgbGreen=c.G;
  Palette[i].rgbBlue=c.B;
  Palette[i].rgbReserved=0;
 #elif defined(__FORX11__)

 #else
  Palette->SetColour(i,c);
 #endif

 return c;
}


GPalette* FASTCALL GFrame::GetPalette(GPalette *pal,INT first,INT count)
{
 if(!Palette) return NULL;

 #if defined(__FORWINDOWS__)
  INT i;
  for(i=first;i<first+count;i++)
	pal->SetColour(i,Palette[i].rgbRed,Palette[i].rgbGreen,Palette[i].rgbBlue);
 #elif defined(__FORX11__)

 #else
  pal->SetPalette(Palette,first,count);
 #endif

 return pal;
}


BOOL FASTCALL GFrame::RealizePalette(void)
{
 return TRUE;
}


BOOL FASTCALL GFrame::UnrealizePalette(void)
{
 return TRUE;
}


