///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#include "../include/gvepro.h"



GVectorObject* FASTCALL G3dsFile::Import(char *fname,BOOL v)
{
 LONG p;
 CHAR oname[MAXPATH];

 Verbose=v;
 fnsplit(fname,NULL,NULL,oname,NULL);
 if((Head=new GVectorObject(oname,GObject::CONTAINER_TYPE))==NULL) return NULL;

 if((Handle=open(fname,O_RDONLY|O_BINARY))<0) return NULL;

 lseek(Handle, 0, SEEK_END);
 p = tell(Handle);
 lseek(Handle, 0, SEEK_SET);

 ChunkReader(p);
 close(Handle);
 Head->RealizePosition(TRUE);

 return Head;
}



BOOL FASTCALL G3dsFile::ChunkReader(LONG p)
{
 ChunkHeader h;
 LONG CurrentPosition,NextPosition;
 CHAR const *ChnkMsg="Chunk type \"%s\", offset 0x%lX, size %d bytes\n";

 #define HANDLE_CHUNK(a,txt,fn)	case a: 								\
			if (Verbose) printf(ChnkMsg,txt,CurrentPosition,h.len);		\
				if (fn(NextPosition)!=TRUE) return FALSE;             	\
		break;
 #define FORWARD_CHUNK(a,txt,fn) case a: 								\
		if (Verbose) printf(ChnkMsg,txt,CurrentPosition,h.len);			\
				if (ChunkReader(NextPosition)!=TRUE) return FALSE;		\
				break;


 while ((CurrentPosition=tell(Handle)) < p)
   {
	if (read(Handle,&h, sizeof(h)) != sizeof(h)) return FALSE;
	#if defined(__MSBFIRST__)
	 SwapDWord(&h.len);
	 SwapWord(&h.id);
	#endif
	if (h.len == 0) return TRUE;
	NextPosition=CurrentPosition+h.len;
	switch(h.id)
	   {
		HANDLE_CHUNK(CHUNK_RGBF,	"RGB float",					RGBFReader)
		HANDLE_CHUNK(CHUNK_RGBB,	"RGB byte",						RGBBReader)

		FORWARD_CHUNK(CHUNK_PRJ,        "Project",          		NULL)
		FORWARD_CHUNK(CHUNK_MLI,        "Material Library", 		NULL)

		FORWARD_CHUNK(CHUNK_MAIN,       "Main",             		NULL)
		FORWARD_CHUNK(CHUNK_OBJMESH,    "Object Mesh",      		NULL)
		FORWARD_CHUNK(CHUNK_BKGCOLOR,   "Background color", 		NULL)
		FORWARD_CHUNK(CHUNK_AMBCOLOR,   "Ambient color",    		NULL)
		HANDLE_CHUNK(CHUNK_OBJBLOCK,    "Object Block",     		ObjBlockReader)
		HANDLE_CHUNK(CHUNK_TRIMESH,     "Tri-Mesh",         		TriMeshReader)
		HANDLE_CHUNK(CHUNK_VERTLIST,    "Vertex list",      		VertListReader)
		HANDLE_CHUNK(CHUNK_FACELIST,    "Face list",        		FaceListReader)
		HANDLE_CHUNK(CHUNK_FACEMAT,     "Face material",    		FaceMatReader)
		HANDLE_CHUNK(CHUNK_MAPLIST,     "Mappings list",    		MapListReader)
		HANDLE_CHUNK(CHUNK_SMOOLIST,    "Smoothings",       		SmooListReader)
		HANDLE_CHUNK(CHUNK_TRMATRIX,    "Matrix",           		TrMatrixReader)
		HANDLE_CHUNK(CHUNK_LIGHT,       "Light",            		LightReader)
		HANDLE_CHUNK(CHUNK_SPOTLIGHT,   "Spotlight",        		SpotLightReader)
		HANDLE_CHUNK(CHUNK_CAMERA,      "Camera",           		CameraReader)
		FORWARD_CHUNK(CHUNK_HIERARCHY,  "Hierarchy",        		NULL)

		FORWARD_CHUNK(CHUNK_VIEWPORT,   "Viewport info",    		NULL)
		FORWARD_CHUNK(CHUNK_MATERIAL,   "Material",         		NULL)
		FORWARD_CHUNK(CHUNK_MATNAME,    "Material name",    		NULL)
		FORWARD_CHUNK(CHUNK_AMBIENT,    "Ambient color",    		NULL)
		FORWARD_CHUNK(CHUNK_DIFFUSE,    "Diffuse color",    		NULL)
		FORWARD_CHUNK(CHUNK_SPECULAR,   "Specular color",   		NULL)
		FORWARD_CHUNK(CHUNK_TEXTURE,    "Texture map",      		NULL)
		FORWARD_CHUNK(CHUNK_BUMPMAP,    "Bump map",         		NULL)
		FORWARD_CHUNK(CHUNK_MAPFILE,    "Map filename",     		NULL)

		FORWARD_CHUNK(CHUNK_KEYFRAMER,  "Keyframer data",   		NULL)
		FORWARD_CHUNK(CHUNK_AMBIENTKEY, "Ambient key",      		NULL)
		FORWARD_CHUNK(CHUNK_TRACKINFO,  "Track info",       		NULL)
		FORWARD_CHUNK(CHUNK_FRAMES,     "Frames",           		NULL)
		FORWARD_CHUNK(CHUNK_TRACKOBJNAME,"Track Obj. Name",  		NULL)
		FORWARD_CHUNK(CHUNK_TRACKPIVOT, "Pivot point",      		NULL)
		FORWARD_CHUNK(CHUNK_TRACKPOS,   "Position keys",    		NULL)
		FORWARD_CHUNK(CHUNK_TRACKROTATE,"Rotation keys",    		NULL)
		FORWARD_CHUNK(CHUNK_TRACKSCALE, "Scale keys",       		NULL)
		FORWARD_CHUNK(CHUNK_OBJNUMBER,  "Object number",    		NULL)

		FORWARD_CHUNK(CHUNK_TRACKCAMERA,"Camera track",        		NULL)
		FORWARD_CHUNK(CHUNK_TRACKCAMTGT,"Camera target track",     	NULL)
		FORWARD_CHUNK(CHUNK_TRACKLIGHT, "Pointlight track",        	NULL)
		FORWARD_CHUNK(CHUNK_TRACKLIGTGT,"Pointlight target track", 	NULL)
		FORWARD_CHUNK(CHUNK_TRACKSPOTL, "Spotlight track",         	NULL)
		FORWARD_CHUNK(CHUNK_TRACKFOV,   "FOV track",               	NULL)
		FORWARD_CHUNK(CHUNK_TRACKROLL,  "Roll track",              	NULL)

		default:
			if (Verbose) printf("Unknown chunk: 0x%04X, offset 0x%lX, size: %d bytes.\n",
					h.id,CurrentPosition,h.len);
				break;
	   }
	lseek(Handle,NextPosition,SEEK_SET);
   }

 return TRUE;
 #undef HANDLE_CHUNK
 #undef FORWARD_CHUNK
}




INT FASTCALL G3dsFile::ASCIIZReader(CHAR *buffer)
{
 INT s=0;

 do
   {
	if(read(Handle,buffer,1)!=1) break;
	s++;
   }
 while (*buffer++!=0);


 return s;
}


#pragma argsused
BOOL FASTCALL G3dsFile::RGBFReader(LONG p)
{
 FLOAT c[3];
 if (read(Handle,&c, sizeof(c)) != sizeof(c)) return FALSE;
 #if defined (__MSBFIRST__)
  SwapDWord ( (DWORD*)(&c[0]) );
  SwapDWord ( (DWORD*)(&c[1]) );
  SwapDWord ( (DWORD*)(&c[2]) );
 #endif

 if (Verbose) printf("    Red: %f, Green: %f, Blue: %f\n", IEEE2FLOAT(c[0]),IEEE2FLOAT(c[1]),IEEE2FLOAT(c[2]));
 return TRUE;
}


#pragma argsused
BOOL FASTCALL G3dsFile::RGBBReader(LONG p)
{
 BYTE c[3];
 if (read(Handle,&c,sizeof(c)) != sizeof(c)) return FALSE;
 if (Verbose) printf("    Red: %d, Green: %d, Blue: %d\n", c[0], c[1], c[2]);
 return TRUE;
}



BOOL FASTCALL G3dsFile::ObjBlockReader(LONG p)
{
 ASCIIZReader(ObjectName);
 return ChunkReader(p);
}



BOOL FASTCALL G3dsFile::TriMeshReader(LONG p)
{
 if ((CurrentMesh=new GMeshObject)==NULL) return FALSE;
 Head->AddSon(CurrentMesh);
 CurrentMesh->SetName(ObjectName);
 CurrentMesh->Flags|=GVE_OBJ_VERTEX_INDEX | GVE_OBJ_NEWPOS | GVE_OBJ_NEWROT;

 if (ChunkReader(p)==TRUE)
	return CurrentMesh->ReferenceFaces();
 else
	return FALSE;
}


#pragma argsused
BOOL FASTCALL G3dsFile::VertListReader(LONG p)
{
 INT i;
 WORD nv;
 FLOAT c[3];

 if (read(Handle,&nv, sizeof(nv)) != sizeof(nv)) return FALSE;
 if ((CurrentVertexList=CurrentMesh->AllocVertexList((LONG)LSB2WORD(nv)))==NULL) return FALSE;

 for(i=0;i<LSB2WORD(nv);i++)
   {
	if (read(Handle,&c, sizeof(c)) != sizeof(c)) return FALSE;
	#if defined (__MSBFIRST__)
	 SwapDWord ( (DWORD*)(&c[0]) );
	 SwapDWord ( (DWORD*)(&c[1]) );
	 SwapDWord ( (DWORD*)(&c[2]) );
	#endif

	CurrentVertexList->Vertex[i].Position.x=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(c[0])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);
	CurrentVertexList->Vertex[i].Position.y=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(c[1])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);
	CurrentVertexList->Vertex[i].Position.z=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(c[2])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);
   }

 return TRUE;
}



#pragma argsused
BOOL FASTCALL G3dsFile::FaceListReader(LONG p)
{
 WORD nv;
 WORD c[3];
 WORD flags;
 GTriangle *tri;

 if (read(Handle,&nv,sizeof(nv)) != sizeof(nv)) return FALSE;
 if ((CurrentFaceList=CurrentMesh->AllocTriangleList((LONG)LSB2WORD(nv)))==NULL) return FALSE;
 tri=(GTriangle *)&CurrentFaceList->Face[0];

 #if defined(__MSBFIRST__)
  SwapWord(&nv);
 #endif

 while (nv-- > 0)
   {
	if (read(Handle,&c,sizeof(c)) != sizeof(c)) return FALSE;
	if (read(Handle,&flags,sizeof(flags)) != sizeof(flags)) return FALSE;

	tri->Type=GVE_SHAD_FILLED;
	tri->VertexCount=3;
	tri->Size=sizeof(GTriangle);
	tri->TextureMap=
	tri->LightMap=
	tri->BumpMap=0;
	tri->Vertex[0]=(GVertex *)LSB2WORD(c[0]);
	tri->Vertex[1]=(GVertex *)LSB2WORD(c[1]);
	tri->Vertex[2]=(GVertex *)LSB2WORD(c[2]);

	#if defined(__MSBFIRST__)
	 SwapWord(&flags);
	#endif
	if ((flags & 0x08) != 0) tri->TextureMap|=GVE_TMAP_U_WRAP;
	if ((flags & 0x10) != 0) tri->TextureMap|=GVE_TMAP_V_WRAP;
//        printf(" AB %d BC %d CA %d UWrap %d VWrap %d\n",
//               (flags & 0x04) != 0, (flags & 0x02) != 0, (flags & 0x01) != 0,
//               (flags & 0x08) != 0, (flags & 0x10) != 0);
	((BYTE *)tri)+=sizeof(GTriangle);
   }

 return TRUE;
}


#pragma argsused
BOOL FASTCALL G3dsFile::FaceMatReader(LONG p)
{
 WORD n,nf;
 CHAR name[64];

 ASCIIZReader(name);										// Material name
 if (read(Handle,&n,sizeof(n)) != sizeof(n)) return FALSE; 	// Faces with this material

 #if defined (__MSBFIRST__)
  SwapWord(&n);
 #endif
 while (n-- > 0)
   {
	if (read(Handle,&nf,sizeof(nf)) != sizeof(nf)) return FALSE;
	// printf("%*s    Face %d\n",ind, "", nf);
   }

 return TRUE;
}


#pragma argsused
BOOL FASTCALL G3dsFile::MapListReader(LONG p)
{
 INT i;
 WORD nv;
 FLOAT c[2];

 if (read(Handle,&nv,sizeof(nv)) != sizeof(nv)) return FALSE;
 for (i=0;i<LSB2WORD(nv);i++)
   {
	if (read(Handle,&c,sizeof(c)) != sizeof(c)) return FALSE;
	#if defined (__MSBFIRST__)
	 SwapDWord ( (DWORD*)(&c[0]) );
	 SwapDWord ( (DWORD*)(&c[1]) );
	#endif

	CurrentVertexList->Vertex[i].Texture.u=(LONG)((DOUBLE)IEEE2FLOAT(c[0])*(DOUBLE)(1<<GVE_MAPPING_SHIFT));
	CurrentVertexList->Vertex[i].Texture.v=(LONG)((DOUBLE)IEEE2FLOAT(c[1])*(DOUBLE)(1<<GVE_MAPPING_SHIFT));
   }

 return TRUE;
}


#pragma argsused
BOOL FASTCALL G3dsFile::SmooListReader(LONG p)
{
 DWORD s;

 while (tell(Handle) < p)
   {
	if (read(Handle,&s, sizeof(s)) != sizeof(s)) return FALSE;
//        printf("Smoothing groups: ");
//        for (i = 0; i < 32; i++)
//            if (s & (1 << i))
//                printf("%d, ", i + 1);
//        printf("\n");
   }

 return TRUE;
}


#pragma argsused
BOOL FASTCALL G3dsFile::TrMatrixReader(LONG p)
{
 FLOAT rot[9];
 FLOAT trans[3];

 if (read(Handle,&rot,sizeof(rot)) != sizeof(rot)) return FALSE;
 if (read(Handle,&trans,sizeof(trans)) != sizeof(trans)) return FALSE;

 #if defined (__MSBFIRST__)
  SwapDWord ( (DWORD*)(&rot[0]) );
  SwapDWord ( (DWORD*)(&rot[1]) );
  SwapDWord ( (DWORD*)(&rot[2]) );
  SwapDWord ( (DWORD*)(&rot[3]) );
  SwapDWord ( (DWORD*)(&rot[4]) );
  SwapDWord ( (DWORD*)(&rot[5]) );
  SwapDWord ( (DWORD*)(&rot[6]) );
  SwapDWord ( (DWORD*)(&rot[7]) );
  SwapDWord ( (DWORD*)(&rot[8]) );
  SwapDWord ( (DWORD*)(&trans[0]) );
  SwapDWord ( (DWORD*)(&trans[1]) );
  SwapDWord ( (DWORD*)(&trans[2]) );
 #endif


 CurrentMesh->LocalMatrix.m[0][0]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(rot[0]),GVE_MATRIX_SHIFT);
 CurrentMesh->LocalMatrix.m[0][1]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(rot[1]),GVE_MATRIX_SHIFT);
 CurrentMesh->LocalMatrix.m[0][2]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(rot[2]),GVE_MATRIX_SHIFT);
 CurrentMesh->LocalMatrix.m[0][3]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(trans[0])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);

 CurrentMesh->LocalMatrix.m[1][0]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(rot[3]),GVE_MATRIX_SHIFT);
 CurrentMesh->LocalMatrix.m[1][1]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(rot[4]),GVE_MATRIX_SHIFT);
 CurrentMesh->LocalMatrix.m[1][2]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(rot[5]),GVE_MATRIX_SHIFT);
 CurrentMesh->LocalMatrix.m[1][3]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(trans[1])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);

 CurrentMesh->LocalMatrix.m[2][0]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(rot[6]),GVE_MATRIX_SHIFT);
 CurrentMesh->LocalMatrix.m[2][1]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(rot[7]),GVE_MATRIX_SHIFT);
 CurrentMesh->LocalMatrix.m[2][2]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(rot[8]),GVE_MATRIX_SHIFT);
 CurrentMesh->LocalMatrix.m[2][3]=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(trans[2])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);

 return TRUE;
}



BOOL FASTCALL G3dsFile::LightReader(LONG p)
{
 FLOAT c[3];

 if (read(Handle,&c,sizeof(c)) != sizeof(c)) return FALSE;
 if ((CurrentLight=new GLightObject(Head,GLightObject::POINTLIGHT_TYPE))==NULL) return FALSE;
 Head->AddSon(CurrentLight);
 #if defined (__MSBFIRST__)
  SwapDWord ( (DWORD*)(&c[0]) );
  SwapDWord ( (DWORD*)(&c[1]) );
  SwapDWord ( (DWORD*)(&c[2]) );
 #endif

 CurrentLight->SetName(ObjectName);
 CurrentLight->Position.x=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(c[0])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);
 CurrentLight->Position.y=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(c[1])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);
 CurrentLight->Position.z=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(c[2])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);

 return ChunkReader(p);
}


#pragma argsused
BOOL FASTCALL G3dsFile::SpotLightReader(LONG p)
{
 FLOAT c[5];
 if (read(Handle,&c,sizeof(c)) != sizeof(c)) return FALSE;
// printf("    Target X: %f, Y: %f, Z: %f; Hotspot %f, Falloff %f\n",
//           , c[0], c[1], c[2], c[3], c[4]);

 return TRUE;
}


#pragma argsused
BOOL FASTCALL G3dsFile::CameraReader(LONG p)
{
 FLOAT c[8];
 if (read(Handle,&c, sizeof(c)) != sizeof(c)) return FALSE;
 if ((CurrentCamera=new GCameraObject)==NULL) return FALSE;
 Head->AddSon(CurrentCamera);
 #if defined (__MSBFIRST__)
  SwapDWord ( (DWORD*)(&c[0]) );
  SwapDWord ( (DWORD*)(&c[1]) );
  SwapDWord ( (DWORD*)(&c[2]) );
  SwapDWord ( (DWORD*)(&c[3]) );
  SwapDWord ( (DWORD*)(&c[4]) );
  SwapDWord ( (DWORD*)(&c[5]) );
  SwapDWord ( (DWORD*)(&c[6]) );
  SwapDWord ( (DWORD*)(&c[7]) );
 #endif

 CurrentCamera->SetName(ObjectName);
 CurrentCamera->Position.x=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(c[0])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);
 CurrentCamera->Position.y=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(c[1])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);
 CurrentCamera->Position.z=DOUBLE2MATH((DOUBLE)IEEE2FLOAT(c[2])*(DOUBLE)GVE_3DS_FACTOR,GVE_VECTOR_SHIFT);

// printf("%*s    Target: X: %f, Y: %f, Z: %f\n", ind, "", c[3], c[4], c[5]);
// printf("%*s    Bank: %f, Lens: %f\n", ind, "", c[6], c[7]);
 return TRUE;
}

