///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#include "../include/gvepro.h"

// +-                                                                                               -+
// | cos(z)*cos(y)-sin(z)*sin(x)*sin(y)   -sin(z)*cos(x)    cos(z)*sin(y)+sin(z)*sin(x)*cos(y)   Tx  |
// | sin(z)*cos(y)+cos(z)*sin(x)*sin(y)    cos(z)*cos(x)    sin(z)*sin(y)-cos(z)*sin(x)*cos(y)   Ty  |
// |       -cos(x)*sin(y)                     sin(x)               cos(x)*cos(y)                 Tz  |
// |             0                               0                       0                       1   |
// +-                                                                                               -+
GMatrix::GMatrix(GVector p,GVector r)
{
 #define fcos(a)	((DOUBLE)cos(MATH2DOUBLE(a,GVE_VECTOR_SHIFT)*M_PI))
 #define fsin(a)	((DOUBLE)sin(MATH2DOUBLE(a,GVE_VECTOR_SHIFT)*M_PI))

 m[0][0]=DOUBLE2MATH(fcos(p.z)*fcos(p.y)-fsin(p.z)*fsin(p.x)*fsin(p.y),GVE_MATRIX_SHIFT);
 m[0][1]=DOUBLE2MATH(-fsin(p.z)*fcos(p.x),GVE_MATRIX_SHIFT);
 m[0][2]=DOUBLE2MATH(fcos(p.z)*fsin(p.y)+fsin(p.z)*fsin(p.x)*fcos(p.y),GVE_MATRIX_SHIFT);
 m[0][3]=r.x;
 m[1][0]=DOUBLE2MATH(fsin(p.z)*fcos(p.y)+fcos(p.z)*fsin(p.x)*fsin(p.y),GVE_MATRIX_SHIFT);
 m[1][1]=DOUBLE2MATH(fcos(p.z)*fcos(p.x),GVE_MATRIX_SHIFT);
 m[1][2]=DOUBLE2MATH(fsin(p.z)*fsin(p.y)-fcos(p.z)*fsin(p.x)*fcos(p.y),GVE_MATRIX_SHIFT);
 m[1][3]=r.y;
 m[2][0]=DOUBLE2MATH(-fcos(p.x)*fsin(p.y),GVE_MATRIX_SHIFT);
 m[2][1]=DOUBLE2MATH(fsin(p.x),GVE_MATRIX_SHIFT);
 m[2][2]=DOUBLE2MATH(fcos(p.x)*fcos(p.y),GVE_MATRIX_SHIFT);
 m[2][3]=r.z;
 m[3][0]=
 m[3][1]=
 m[3][2]=0;
 m[3][3]=LONG2MATH(1,GVE_MATRIX_SHIFT);

 #undef fcos
 #undef fsin
}



GMatrix FASTCALL GMatrix::operator *(GMatrix &a)
{
 GMatrix n;
 INT i,j;

 #if defined(FIXED_MATH)
  for (i=0;i<3;i++)
   {
	for (j=0;j<4;j++)
		n.m[i][j]=(MATH)(MulShr(m[i][0],a.m[0][j],GVE_MATRIX_SHIFT)+
						 MulShr(m[i][1],a.m[1][j],GVE_MATRIX_SHIFT)+
						 MulShr(m[i][2],a.m[2][j],GVE_MATRIX_SHIFT));
	n.m[i][3]+=m[i][3];
   }
  #else
  for (i=0;i<3;i++)
   {
	for (j=0;j<4;j++)
		n.m[i][j]=(MATH)((DOUBLE)m[i][0]*(DOUBLE)a.m[0][j]+
						 (DOUBLE)m[i][1]*(DOUBLE)a.m[1][j]+
						 (DOUBLE)m[i][2]*(DOUBLE)a.m[2][j]);
	n.m[i][3]+=m[i][3];
   }
 #endif

 n.m[3][0]=n.m[3][1]=n.m[3][2]=0; n.m[3][3]=LONG2MATH(1,GVE_MATRIX_SHIFT);
 return n;
}



GMatrix FASTCALL GMatrix::operator /(GMatrix &a)
{
 GMatrix n;
 INT i,j;

 #if defined(FIXED_MATH)
  for (i=0;i<3;i++)
   {
	for (j=0;j<3;j++)
		n.m[i][j]=(MATH)(MulShr(m[i][0],a.m[j][0],GVE_MATRIX_SHIFT)+
						 MulShr(m[i][1],a.m[j][1],GVE_MATRIX_SHIFT)+
						 MulShr(m[i][2],a.m[j][2],GVE_MATRIX_SHIFT));
	n.m[i][3]= m[i][3] -
	  (MATH)(MulShr(n.m[i][0],a.m[0][3],GVE_MATRIX_SHIFT)+
			 MulShr(n.m[i][1],a.m[1][3],GVE_MATRIX_SHIFT)+
			 MulShr(n.m[i][2],a.m[2][3],GVE_MATRIX_SHIFT));
   }
 #else
  for (i=0;i<3;i++)
   {
	for (j=0;j<3;j++)
		n.m[i][j]=(MATH)((DOUBLE)m[i][0]*(DOUBLE)a.m[j][0]+
						 (DOUBLE)m[i][1]*(DOUBLE)a.m[j][1]+
						 (DOUBLE)m[i][2]*(DOUBLE)a.m[j][2]);
	n.m[i][3]=m[i][3]-
		   ((DOUBLE)n.m[i][0]*(DOUBLE)a.m[0][3]+
			(DOUBLE)n.m[i][1]*(DOUBLE)a.m[1][3]+
			(DOUBLE)n.m[i][2]*(DOUBLE)a.m[2][3]);
   }
 #endif

 n.m[3][0]=n.m[3][1]=n.m[3][2]=0; n.m[3][3]=LONG2MATH(1,GVE_MATRIX_SHIFT);
 return n;
}




GMatrix FASTCALL GMatrix::operator ^(GMatrix &a)
{
 GMatrix n;
 INT i,j;

 #if defined(FIXED_MATH)
  for (i=0;i<3;i++)
   {
	for (j=0;j<4;j++)
		n.m[i][j]=(MATH)(MulShr(m[0][i],a.m[0][j],GVE_MATRIX_SHIFT)+
						 MulShr(m[1][i],a.m[1][j],GVE_MATRIX_SHIFT)+
						 MulShr(m[2][i],a.m[2][j],GVE_MATRIX_SHIFT));
	n.m[i][3]-=(MATH)(MulShr(m[0][i],m[0][3],GVE_MATRIX_SHIFT)+
					  MulShr(m[1][i],m[1][3],GVE_MATRIX_SHIFT)+
					  MulShr(m[2][i],m[2][3],GVE_MATRIX_SHIFT));
   }
 #else
  for (i=0;i<3;i++)
   {
	for (j=0;j<4;j++)
		n.m[i][j]=(MATH)((DOUBLE)m[0][i]*(DOUBLE)a.m[0][j]+
						 (DOUBLE)m[1][i]*(DOUBLE)a.m[1][j]+
						 (DOUBLE)m[2][i]*(DOUBLE)a.m[2][j]);
	n.m[i][3]-=(MATH)((DOUBLE)m[0][i]*(DOUBLE)m[0][3]+
					  (DOUBLE)m[1][i]*(DOUBLE)m[1][3]+
					  (DOUBLE)m[2][i]*(DOUBLE)m[2][3]);
   }
 #endif

 n.m[3][0]=n.m[3][1]=n.m[3][2]=0; n.m[3][3]=LONG2MATH(1,GVE_MATRIX_SHIFT);
 return n;
}



GMatrix FASTCALL GMatrix::operator ~(void)
{
 GMatrix n;
 INT i,j;

 #if defined(FIXED_MATH)
  for(i=0;i<3;i++)
	for (j=0;j<3;j++)
		n.m[i][j]=m[j][i];
  for(i=0;i<3;i++)
	n.m[i][3]=-(MATH)(MulShr(m[0][i],m[0][3],GVE_MATRIX_SHIFT)+
					  MulShr(m[1][i],m[1][3],GVE_MATRIX_SHIFT)+
					  MulShr(m[2][i],m[2][3],GVE_MATRIX_SHIFT));
 #else
  for(i=0;i<3;i++)
	for (j=0;j<3;j++)
		n.m[i][j]=m[j][i];
  for(i=0;i<3;i++)
	n.m[i][3]=-(MATH)((DOUBLE)m[0][i]*(DOUBLE)m[0][3]+
					  (DOUBLE)m[1][i]*(DOUBLE)m[1][3]+
					  (DOUBLE)m[2][i]*(DOUBLE)m[2][3]);
 #endif

 n.m[3][0]=n.m[3][1]=n.m[3][2]=0; n.m[3][3]=LONG2MATH(1,GVE_MATRIX_SHIFT);
 return n;
}



void FASTCALL GMatrix::Rotation(GVector p)
{
 #define fcos(a)	((DOUBLE)cos(MATH2DOUBLE(a,GVE_VECTOR_SHIFT)))
 #define fsin(a)	((DOUBLE)sin(MATH2DOUBLE(a,GVE_VECTOR_SHIFT)))

 m[0][0]=DOUBLE2MATH(fcos(p.z)*fcos(p.y)-fsin(p.z)*fsin(p.x)*fsin(p.y),GVE_MATRIX_SHIFT);
 m[0][1]=DOUBLE2MATH(-fsin(p.z)*fcos(p.x),GVE_MATRIX_SHIFT);
 m[0][2]=DOUBLE2MATH(fcos(p.z)*fsin(p.y)+fsin(p.z)*fsin(p.x)*fcos(p.y),GVE_MATRIX_SHIFT);
 m[0][3]=0;
 m[1][0]=DOUBLE2MATH(fsin(p.z)*fcos(p.y)+fcos(p.z)*fsin(p.x)*fsin(p.y),GVE_MATRIX_SHIFT);
 m[1][1]=DOUBLE2MATH(fcos(p.z)*fcos(p.x),GVE_MATRIX_SHIFT);
 m[1][2]=DOUBLE2MATH(fsin(p.z)*fsin(p.y)-fcos(p.z)*fsin(p.x)*fcos(p.y),GVE_MATRIX_SHIFT);
 m[1][3]=0;
 m[2][0]=DOUBLE2MATH(fcos(p.x)*fsin(p.y),GVE_MATRIX_SHIFT);
 m[2][1]=DOUBLE2MATH(fsin(p.x),GVE_MATRIX_SHIFT);
 m[2][2]=DOUBLE2MATH(fcos(p.x)*fcos(p.y),GVE_MATRIX_SHIFT);
 m[2][3]=
 m[3][0]=
 m[3][1]=
 m[3][2]=0;
 m[3][3]=LONG2MATH(1,GVE_MATRIX_SHIFT);

 #undef fsin
 #undef fcos
}



void FASTCALL GMatrix::Translation(GVector t)
{
 m[0][0]=LONG2MATH(1,GVE_MATRIX_SHIFT);
 m[0][1]=m[0][2]=0;
 m[0][3]=t.x;

 m[1][1]=LONG2MATH(1,GVE_MATRIX_SHIFT);
 m[1][0]=m[1][2]=0;
 m[1][3]=t.y;

 m[2][2]=LONG2MATH(1,GVE_MATRIX_SHIFT);
 m[2][0]=m[2][1]=0;
 m[2][3]=t.y;

 m[3][0]=m[3][1]=m[3][2]=0;
 m[3][3]=LONG2MATH(1,GVE_MATRIX_SHIFT);
}
