///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#include "../include/gvepro.h"


CHAR* FASTCALL GVectorObject::GetFlagsName(CHAR *buffer)
{
 buffer[0]=0;

 if (Flags & GVE_OBJ_BACKCULLING)  	strcat(buffer,"BACK_CULLING ");
 if (Flags & GVE_OBJ_NOSORT)  		strcat(buffer,"NOSORT ");
 if (Flags & GVE_OBJ_MESH)  		strcat(buffer,"MESH ");
 if (Flags & GVE_OBJ_VERTEX_INDEX)  strcat(buffer,"VERTEX_INDEX ");
 if (Flags & GVE_OBJ_VERTEX_ADDRESS)strcat(buffer,"VERTEX_ADDRESS ");
 if (Flags & GVE_OBJ_NORMALS)  		strcat(buffer,"TRANSFORM_NORMALS ");
 if (Flags & GVE_OBJ_VERTEX_SHADING)strcat(buffer,"VERTEX_SHADING ");
 if (Flags & GVE_OBJ_VERTEXBOX)  	strcat(buffer,"VERTEX_BOX ");
 if (Flags & GVE_OBJ_CHUNKBOX)  	strcat(buffer,"CHUNK_BOX ");
 if (Flags & GVE_OBJ_HIDE)  		strcat(buffer,"HIDE ");

 return buffer;
}


BOOL FASTCALL GVectorObject::RealizePosition(BOOL forceupdate)
{
 GVectorObject *o=(GVectorObject *)Son;

 if (Flags & (GVE_OBJ_NEWROT | GVE_OBJ_NEWPOS))
   {
	Flags&=~(GVE_OBJ_NEWROT | GVE_OBJ_NEWPOS);
	LocalMatrix=GMatrix(Rotation,Position);
	forceupdate=TRUE;
   }

 if (forceupdate)
   {
	if (Father==NULL)
		FinalMatrix=LocalMatrix;
	else
		FinalMatrix=((GVectorObject *)Father)->FinalMatrix*LocalMatrix;
   }

 while(o)
	{
	 o->RealizePosition(forceupdate);
	 o=(GVectorObject *)o->Brother;
	}

 return TRUE;
}


INT FASTCALL GVectorObject::GetVertexCount(BOOL sub)
{
 if (sub==FALSE)
   {
	if (typeid(*this)==typeid(GMeshObject))
		return (((GMeshObject *)this)->VertexList ? ((GMeshObject *)this)->VertexList->Count : NULL);
	else
		return 0;
   }
 else
   {
	INT ret;
	GVectorObject *son=(GVectorObject *)Son;
	if (typeid(*this)==typeid(GMeshObject))
		ret=(((GMeshObject *)this)->VertexList ? ((GMeshObject *)this)->VertexList->Count : NULL);
	else
		ret=0;

	while(son)
	   {
		ret+=son->GetVertexCount(sub);
		son=(GVectorObject *)son->Brother;
	   }

	return ret;
   }
}


INT FASTCALL GVectorObject::GetFaceCount(BOOL sub)
{
 if (sub==FALSE)
   {
	if (typeid(*this)==typeid(GMeshObject))
		return (((GMeshObject *)this)->FaceList ? ((GMeshObject *)this)->FaceList->Count : NULL);
	else
		return 0;
   }
 else
   {
	INT ret;
	GVectorObject *son=(GVectorObject *)Son;
	if (typeid(*this)==typeid(GMeshObject))
		ret=(((GMeshObject *)this)->FaceList ? ((GMeshObject *)this)->FaceList->Count : NULL);
	else
		ret=0;


	while(son)
	   {
		ret+=son->GetFaceCount(sub);
		son=(GVectorObject *)son->Brother;
	   }

	return ret;
   }
}


INT FASTCALL GVectorObject::GetObjectCount(BOOL sub)
{
 if (sub==FALSE)
	return 1;
 else
   {
	INT ret=1;
	GVectorObject *son=(GVectorObject *)Son;

	while(son)
	   {
		ret+=son->GetObjectCount(sub);
		son=(GVectorObject *)son->Brother;
	   }

	return ret;
   }
}




GVertexList* FASTCALL GMeshObject::AllocVertexList(INT n)
{
 if(VertexList) free(VertexList);
 VertexList=(GVertexList *)malloc(n*sizeof(GVertex)+sizeof(GVertexList));
 if (VertexList==NULL) return NULL;
 memset(&VertexList->Vertex[0],0,n*sizeof(GVertex));
 VertexList->Count=n;
 return VertexList;
}


GFaceList* FASTCALL GMeshObject::AllocTriangleList(INT n)
{
 if(FaceList) free(FaceList);
 FaceList=(GFaceList *)malloc(n*sizeof(GTriangle)+sizeof(GFaceList));
 if (FaceList==NULL) return NULL;
 memset(&FaceList->Face[0],0,n*sizeof(GTriangle));
 FaceList->Count=n;
 return FaceList;
}


BOOL FASTCALL GMeshObject::FreeVertexList(void)
{
 if(VertexList)
	{
	 free(VertexList);
	 VertexList=NULL;
	}

 return TRUE;
}


BOOL FASTCALL GMeshObject::FreeFaceList(void)
{
 if(FaceList)
	{
	 free(FaceList);
	 FaceList=NULL;
	}

 return TRUE;
}


BOOL FASTCALL GMeshObject::FreeMesh(void)
{
 if(VertexList)
	{
	 free(VertexList);
	 VertexList=NULL;
	}
 if(FaceList)
	{
	 free(FaceList);
	 FaceList=NULL;
	}

 Flags&=~(GVE_OBJ_VERTEX_ADDRESS | GVE_OBJ_VERTEX_INDEX);
 return TRUE;
}


BOOL FASTCALL GMeshObject::AllocMesh(INT width,INT height,DOUBLE step)
{
 INT i,j;
 GVertex *v;
 GTriangle *t;
 MATH x,y;
 SFIXED tx,ty,txinc,tyinc;

 FreeMesh();

 if(AllocVertexList(width*height)==NULL) return FALSE;
 if(AllocTriangleList(2*(width-1)*(height-1))==NULL) return FALSE;

 MeshWidth=width;
 MeshHeight=height;
 Type=GMeshObject::MESH_TYPE;
 Flags&=~GVE_OBJ_VERTEX_INDEX;
 Flags|=GVE_OBJ_VERTEX_ADDRESS | GVE_OBJ_MESH;

 VertexList->Count=width*height;
 FaceList->Count=2*(width-1)*(height-1);

 v=&VertexList->Vertex[0];
 t=(GTriangle *)&FaceList->Face[0];

 txinc=(256<<GVE_MAPPING_SHIFT)/width;
 tyinc=(256<<GVE_MAPPING_SHIFT)/height;

 y=DOUBLE2MATH(-(height/2)*step,GVE_VECTOR_SHIFT);
 ty=0;

 for(i=0;i<height;i++)
   {
	tx=0;
	x=DOUBLE2MATH(-(width/2)*step,GVE_VECTOR_SHIFT);
	for (j=0;j<width;j++)
	   {
		v->Position.x=x;
		v->Position.y=y;
		v->Position.z=0;
		v->Texture.u=tx;
		v->Texture.v=ty;
		(BYTE *)v+=sizeof(GVertex);
		x+=step;
		tx+=txinc;
	   }
	y+=step;
	ty+=tyinc;
   }

 for (i=0;i<(height-1);i++)
	for (j=0;j<(width-1);j++)
	   {
		t->Size=sizeof(GTriangle);
		t->Type=GVE_SHAD_FILLED;
		t->VertexCount=3;
		t->BumpMap=
		t->TextureMap=
		t->LightMap=0;
		t->Vertex[0]=&VertexList->Vertex[i*width+j];
		t->Vertex[1]=&VertexList->Vertex[(i+1)*width+j];
		t->Vertex[2]=&VertexList->Vertex[(i+1)*width+j+1];
		(BYTE *)t+=sizeof(GTriangle);

		t->Size=sizeof(GTriangle);
		t->Type=GVE_SHAD_FILLED;
		t->VertexCount=3;
		t->BumpMap=
		t->TextureMap=
		t->LightMap=0;
		t->Vertex[0]=&VertexList->Vertex[i*width+j];
		t->Vertex[1]=&VertexList->Vertex[(i+1)*width+j+1];
		t->Vertex[2]=&VertexList->Vertex[i*width+j+1];
		(BYTE *)t+=sizeof(GTriangle);
	   }

 return TRUE;
}


BOOL FASTCALL GMeshObject::ReferenceFaces(void)
{
 INT i,j;
 GFace *face;

 if(Flags & GVE_OBJ_VERTEX_ADDRESS) return TRUE;
 if((VertexList==NULL) || (FaceList==NULL)) return TRUE;

 face=(GFace *)&FaceList->Face[0];

 for (i=0;i<FaceList->Count;i++)
   {
	for(j=0;j<face->VertexCount;j++)
		face->Vertex[j]=&VertexList->Vertex[(INT)face->Vertex[j]];
	(BYTE *)face+=face->Size;
   }

 Flags&=~GVE_OBJ_VERTEX_INDEX;
 Flags|=GVE_OBJ_VERTEX_ADDRESS;

 return TRUE;
}


BOOL FASTCALL GMeshObject::DeReferenceFaces(void)
{
 INT i,j;
 GFace *face;

 if(Flags & GVE_OBJ_VERTEX_INDEX) return TRUE;
 if((VertexList==NULL) || (FaceList==NULL)) return TRUE;

 face=(GFace *)&FaceList->Face[0];

 for (i=0;i<FaceList->Count;i++)
   {
	for(j=0;j<face->VertexCount;j++)
		face->Vertex[j]=(GVertex *)(((LONG)((BYTE *)face->Vertex[j]-(BYTE *)VertexList)-4L)/sizeof(GVertex));
	(BYTE *)face+=face->Size;
   }

 Flags&=~GVE_OBJ_VERTEX_ADDRESS;
 Flags|=GVE_OBJ_VERTEX_INDEX;

 return TRUE;
}


INT FASTCALL GMeshObject::SetShading(INT s)
{
 INT i;
 GFace *face;

 if (FaceList==NULL)	return 0;
 face=(GFace *)&FaceList->Face[0];

 for (i=0;i<FaceList->Count;i++)
   {
	face->Type=(BYTE)s;
	(BYTE *)face+=face->Size;
   }

 return s;
}


INT FASTCALL GMeshObject::SetColour(INT c)
{
 INT i;
 GFace *face;

 if (FaceList==NULL)	return 0;
 face=(GFace *)&FaceList->Face[0];
 c&=0x03fff;

 for (i=0;i<FaceList->Count;i++)
   {
	face->TextureMap&=0x0c000;
	face->TextureMap|=(WORD)c;
	(BYTE *)face+=face->Size;
   }

 return c;
}


INT FASTCALL GMeshObject::SetTextureMap(INT m)
{
 INT i;
 GFace *face;

 if (FaceList==NULL)	return 0;
 face=(GFace *)&FaceList->Face[0];
 m&=0x03fff;

 for (i=0;i<FaceList->Count;i++)
   {
	face->TextureMap&=0x0c000;
	face->TextureMap|=(WORD)m;
	(BYTE *)face+=face->Size;
   }

 return m;
}


INT FASTCALL GMeshObject::SetLightMap(INT m)
{
 INT i;
 GFace *face;

 if (FaceList==NULL)	return 0;
 face=(GFace *)&FaceList->Face[0];

 for (i=0;i<FaceList->Count;i++)
   {
	face->LightMap=(BYTE)m;
	(BYTE *)face+=face->Size;
   }

 return m;
}


INT FASTCALL GMeshObject::SetPhong(INT m)
{
 INT i;
 GFace *face;
 if (m>GVE_MAX_PHONG) m=GVE_MAX_PHONG;

 if (FaceList==NULL)	return 0;
 face=(GFace *)&FaceList->Face[0];

 for (i=0;i<FaceList->Count;i++)
   {
	face->LightMap=(BYTE)m;
	(BYTE *)face+=face->Size;
   }

 return m;
}


INT FASTCALL GMeshObject::SetBumpMap(INT m)
{
 INT i;
 GFace *face;

 if (FaceList==NULL)	return 0;
 face=(GFace *)&FaceList->Face[0];

 for (i=0;i<FaceList->Count;i++)
   {
	face->BumpMap=(BYTE)m;
	(BYTE *)face+=face->Size;
   }

 return m;
}


INT FASTCALL GMeshObject::GetShading(void)
{
 if ((FaceList==NULL) || (FaceList->Count==0L)) return 0L;
 return (INT)(FaceList->Face[0].Type);
}


INT FASTCALL GMeshObject::GetColour(void)
{
 if ((FaceList==NULL) || (FaceList->Count==0L)) return 0L;
 return (INT)(FaceList->Face[0].TextureMap & 0x0ff);
}


INT FASTCALL GMeshObject::GetTextureMap(void)
{
 if ((FaceList==NULL) || (FaceList->Count==0L)) return 0L;
 return (INT)(FaceList->Face[0].TextureMap & 0x03fff);
}


INT FASTCALL GMeshObject::GetLightMap(void)
{
 if ((FaceList==NULL) || (FaceList->Count==0L)) return 0L;
 return (INT)(FaceList->Face[0].LightMap);
}


INT FASTCALL GMeshObject::GetPhong(void)
{
 if ((FaceList==NULL) || (FaceList->Count==0L)) return 0L;
 return (INT)(FaceList->Face[0].LightMap);
}


INT FASTCALL GMeshObject::GetBumpMap(void)
{
 if ((FaceList==NULL) || (FaceList->Count==0L)) return 0L;
 return (INT)(FaceList->Face[0].BumpMap);
}


BOOL FASTCALL GMeshObject::InitNormals(void)
{
 INT i,j;
 GFace *face;
 GVector n,*v0,*v1,*v2;

 if ((FaceList==NULL) || (VertexList==NULL)) return TRUE;
 if (!(Flags & GVE_OBJ_VERTEX_ADDRESS)) return FALSE;

 for (i=0;i<VertexList->Count;i++)
	{
	 VertexList->Vertex[i].Normal.x=
	 VertexList->Vertex[i].Normal.y=
	 VertexList->Vertex[i].Normal.z=0;
	}

 face=(GFace *)&FaceList->Face[0];
 for(i=0;i<FaceList->Count;i++)
	{
	 if(face->VertexCount>2)
		{
		 v0=&face->Vertex[0]->Position;
		 v1=&face->Vertex[1]->Position;
		 v2=&face->Vertex[2]->Position;
		 n=~((*v0-*v1)%(*v2-*v1));

		 face->Normal[0]=MATH2FIXED(n.x,GVE_NORMAL_SHIFT);
		 face->Normal[1]=MATH2FIXED(n.y,GVE_NORMAL_SHIFT);
		 face->Normal[2]=MATH2FIXED(n.z,GVE_NORMAL_SHIFT);
		 face->Plane=MATH2FIXED(-(*v0*n),2*GVE_NORMAL_SHIFT);

		 for(j=0;j<face->VertexCount;j++)
		   {
			face->Vertex[j]->Normal.x+=n.x;
			face->Vertex[j]->Normal.y+=n.y;
			face->Vertex[j]->Normal.z+=n.z;
		   }
		}
	 (BYTE *)face+=face->Size;
	}

 for(i=0;i<VertexList->Count;i++)
	 VertexList->Vertex[i].Normal=~VertexList->Vertex[i].Normal;

 return TRUE;
}


BOOL FASTCALL GMeshObject::MakeTriangles(void)
{
 INT i,j,current_face=0;
 INT needed_faces=0;
 GFace *face;
 GFaceList *list;

 if (FaceList==NULL) return TRUE;
 face=(GFace *)&FaceList->Face[0];

 for (i=0;i<FaceList->Count;i++)
	{
	 needed_faces+=(face->VertexCount-2);
	 (BYTE *)face+=face->Size;
	}

 if((list=(GFaceList *)malloc(needed_faces*sizeof(GTriangle)+4))==NULL) return FALSE;
 list->Count=needed_faces;

 for(i=0;i<FaceList->Count;i++)
	{
	 for(j=0;j<(face->VertexCount-2);j++)
		{
		 list->Face[current_face].Size=sizeof(GTriangle);
		 list->Face[current_face].Type=face->Type;
		 list->Face[current_face].VertexCount=3;
		 list->Face[current_face].TextureMap=face->TextureMap;
		 list->Face[current_face].LightMap=face->LightMap;
		 list->Face[current_face].BumpMap=face->BumpMap;
		 list->Face[current_face].Normal[0]=face->Normal[0];
		 list->Face[current_face].Normal[1]=face->Normal[1];
		 list->Face[current_face].Normal[2]=face->Normal[2];
		 list->Face[current_face].Plane=face->Plane;
		 list->Face[current_face].Vertex[0]=face->Vertex[0];
		 list->Face[current_face].Vertex[1]=face->Vertex[j+1];
		 list->Face[current_face].Vertex[2]=face->Vertex[j+2];
		 current_face++;
		}
	 (BYTE *)face+=face->Size;
	}

 free(FaceList);
 FaceList=list;

 return TRUE;
}


BOOL FASTCALL GMeshObject::Optimize(void)
{
 LONG			i,j,k;
 BOOL			isequal;
 GFace			*face,*iface;
 GFaceList 		*newfaces;
 GVertexList 	*newvertices;
 LONG			newfacecount=0,newfacesize=sizeof(GFaceList);
 LONG			newvertexcount=0,newvertexsize=sizeof(GVertexList);

 if ((FaceList==NULL) || (VertexList==NULL)) return TRUE;
 if (!(Flags & GVE_OBJ_VERTEX_ADDRESS)) return FALSE;

 for (i=0;i<VertexList->Count;i++)
	VertexList->Vertex[i].InverseZ=0;

 // Find double vertices. Save in InverseZ the address of same vertex.
 // If there hasn't been an equal vertex before, InverseZ is set to 0.
 for (i=0;i<VertexList->Count;i++)
   {
	if (VertexList->Vertex[i].InverseZ==0)
	   {
		for (j=i+1;j<VertexList->Count;j++)
		   {
			if ((VertexList->Vertex[j].Position==VertexList->Vertex[i].Position)
			  &&(VertexList->Vertex[j].Texture==VertexList->Vertex[i].Texture))
				VertexList->Vertex[j].InverseZ=(LONG)&VertexList->Vertex[i];
		   }
	   }
   }

 // Substitue all addresses of double vertices in our face list
 face=&FaceList->Face[0];
 for (i=0;i<FaceList->Count;i++)
   {
	for (j=0;j<face->VertexCount;j++)
		if(face->Vertex[j]->InverseZ!=NULL)
			face->Vertex[j]=(GVertex *)face->Vertex[j]->InverseZ;
	(BYTE *)face+=(LONG)face->Size;
   }


 // Find empty faces. These faces will be marked by setting the
 // VertexCount to zero.
 face=&FaceList->Face[0];
 for (i=0;i<FaceList->Count;i++)
   {
	INT numvert= face->VertexCount;
	if (face->VertexCount<3)
		face->VertexCount=0;
	else
	   {
		for (j=0;j<face->VertexCount;j++)
			for (k=j+1;k<face->VertexCount;k++)
				if (face->Vertex[i]==face->Vertex[j])
				   {
					numvert--;
					break;
				   }
		if (numvert<3)
			face->VertexCount=0;
	   }
   }


 // Mark all vertices as "unused".
 for (i=0;i<VertexList->Count;i++)
	VertexList->Vertex[i].InverseZ=FALSE;

 // Find double faces. Double faces will be marked by VertexCount=0.
 face=&FaceList->Face[0];
 for (i=0;i<FaceList->Count;i++)
   {
	if (face->VertexCount!=0)
	   {
		// This face surely is needed, so save some space for it
		// and increment the face-counter.
		newfacecount++;
		newfacesize+=face->Size;

		// Mark used vertices as "used".
		for (j=0;j<face->VertexCount;j++)
		   {
			if (face->Vertex[j]->InverseZ==FALSE)
			   {
				newvertexcount++;
				newvertexsize+=sizeof(GVertex);
				face->Vertex[j]->InverseZ=TRUE;
			   }
		   }


		// Search for other faces with same vertex-list
		(BYTE *)iface=(BYTE *)face+(LONG)face->Size;
		for (j=i+1;j<FaceList->Count;j++)
		   {
			if (face->VertexCount==iface->VertexCount)
			   {
				isequal=TRUE;
				for (k=0;k<face->VertexCount;k++)
				   {
					if (face->Vertex[k]!=iface->Vertex[k])
						isequal=FALSE;
				   }
				if (isequal)
					face->VertexCount=0;
			   }
			// Skip to next face
			(BYTE *)iface+=iface->Size;
		   }
	   }

	// Skip to next face
	(BYTE *)face+=face->Size;
   }


 // Now we build a new vertex-list. Note that we store the new
 // vertex-addresses in the famous InverseZ-field.
 if((newvertices=(GVertexList *)malloc(newvertexsize))==NULL) return FALSE;
 memset(&newvertices->Vertex[0],0,newvertexsize-sizeof(GVertexList));
 newvertices->Count=newvertexcount;
 j=0;
 for (i=0;i<VertexList->Count;i++)
	if (VertexList->Vertex[i].InverseZ==TRUE)
	   {
		newvertices->Vertex[j]=VertexList->Vertex[i];
		VertexList->Vertex[i].InverseZ=(LONG)&newvertices->Vertex[j++];
	   }

 // Second last we build a new face-list.
 if((newfaces=(GFaceList *)malloc(newfacesize))==NULL) return FALSE;
 memset(&newfaces->Face[0],0,newfacesize-sizeof(GFaceList));
 newfaces->Count=newfacecount;
 face=&FaceList->Face[0];
 iface=&newfaces->Face[0];
 for (i=0;i<FaceList->Count;i++)
	if (face->VertexCount!=0)
	   {
		memcpy(iface,face,face->Size);
		for (j=0;j<iface->VertexCount;j++)
			iface->Vertex[j]=(GVertex *)iface->Vertex[j]->InverseZ;
	   }


 // Really last, we free the old face- and vertex-list.
 free(VertexList);
 free(FaceList);
 VertexList=newvertices;
 FaceList=newfaces;

 return TRUE;
}


BOOL FASTCALL GMeshObject::InitBoundingBox(void)
{
 INT i;
 MATH minx=0,miny=0,minz=0;
 MATH maxx=0,maxy=0,maxz=0;
 GVectorObject *o;
 GMeshObject *mesh;

 if (VertexList)
	{
	 maxx=minx=VertexList->Vertex[0].Position.x;
	 maxy=miny=VertexList->Vertex[0].Position.y;
	 maxz=minz=VertexList->Vertex[0].Position.z;

	 for (i=0;i<VertexList->Count;i++)
		{
		 if (maxx<VertexList->Vertex[i].Position.x) maxx=VertexList->Vertex[i].Position.x;
		 if (maxy<VertexList->Vertex[i].Position.y) maxy=VertexList->Vertex[i].Position.y;
		 if (maxz<VertexList->Vertex[i].Position.z) maxx=VertexList->Vertex[i].Position.z;

		 if (minx>VertexList->Vertex[i].Position.x) minx=VertexList->Vertex[i].Position.x;
		 if (miny>VertexList->Vertex[i].Position.y) miny=VertexList->Vertex[i].Position.y;
		 if (minz>VertexList->Vertex[i].Position.z) minx=VertexList->Vertex[i].Position.z;
		}
	 Flags|=GVE_OBJ_VERTEXBOX | GVE_OBJ_CHUNKBOX;
	}
 else
	{
	 Flags&=~(GVE_OBJ_VERTEXBOX | GVE_OBJ_CHUNKBOX);
	}

 BoundingAll.minx=BoundingThis.minx=minx;
 BoundingAll.miny=BoundingThis.miny=miny;
 BoundingAll.minz=BoundingThis.minz=minz;
 BoundingAll.maxx=BoundingThis.maxx=maxx;
 BoundingAll.maxy=BoundingThis.maxy=maxy;
 BoundingAll.maxz=BoundingThis.maxz=maxz;


 o=(GVectorObject *)Son;
 while (o)
   {
	if (typeid(*o)!=typeid(GMeshObject))
	   {
		o=(GVectorObject *)o->GetBrother();
		continue;
	   }

	mesh=(GMeshObject *)o;
	mesh->InitBoundingBox();
	if ((mesh->Flags & GVE_OBJ_CHUNKBOX) && (Flags & (GVE_OBJ_VERTEXBOX | GVE_OBJ_CHUNKBOX)))
	   {
		if(BoundingAll.minx > mesh->BoundingAll.minx) BoundingAll.minx = mesh->BoundingAll.minx;
		if(BoundingAll.miny > mesh->BoundingAll.miny) BoundingAll.miny = mesh->BoundingAll.miny;
		if(BoundingAll.minz > mesh->BoundingAll.minz) BoundingAll.minz = mesh->BoundingAll.minz;
		if(BoundingAll.maxx < mesh->BoundingAll.maxx) BoundingAll.maxx = mesh->BoundingAll.maxx;
		if(BoundingAll.maxy < mesh->BoundingAll.maxy) BoundingAll.maxy = mesh->BoundingAll.maxy;
		if(BoundingAll.maxz < mesh->BoundingAll.maxz) BoundingAll.maxz = mesh->BoundingAll.maxz;
		Flags|=GVE_OBJ_CHUNKBOX;
	   }
	else if (mesh->Flags & GVE_OBJ_CHUNKBOX)
	   {
		BoundingAll.minx = mesh->BoundingAll.minx;
		BoundingAll.miny = mesh->BoundingAll.miny;
		BoundingAll.minz = mesh->BoundingAll.minz;
		BoundingAll.maxx = mesh->BoundingAll.maxx;
		BoundingAll.maxy = mesh->BoundingAll.maxy;
		BoundingAll.maxz = mesh->BoundingAll.maxz;
		Flags|=GVE_OBJ_CHUNKBOX;
	   }

	o=(GVectorObject *)o->GetBrother();
   }

 return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
// File I/O routines
//
BOOL FASTCALL GVectorObject::Write(GFile &file,DWORD ctype)
{
 FileContainerHead fcont;

 GetName(fcont.Name);
 fcont.Flags=DWORD2LSB(Flags);
 fcont.Type=DWORD2LSB(Type);
 fcont.Magic=DWORD2LSB(Magic);
 fcont.Rotation.x=Rotation.GetX();
 fcont.Rotation.y=Rotation.GetY();
 fcont.Rotation.z=Rotation.GetZ();
 fcont.Position.x=Position.GetX();
 fcont.Position.y=Position.GetY();
 fcont.Position.z=Position.GetZ();

 if(file.CreateChunk(ctype)!=TRUE) return FALSE;
 if(file.WriteChunk(&fcont,sizeof(fcont))!=sizeof(fcont)) return FALSE;
 return file.CloseChunk();
}


BOOL FASTCALL GVectorObject::Read(GFile &file)
{
 void *data;

 if(file.ReadChunk(&data)!=sizeof(FileContainerHead)) return FALSE;
 if(file.CloseChunk()!=TRUE) return FALSE;
 SetName(((FileContainerHead *)data)->Name);
 Magic=LSB2DWORD(((FileContainerHead *)data)->Magic);
 Flags=LSB2DWORD(((FileContainerHead *)data)->Flags);
 Type=LSB2DWORD(((FileContainerHead *)data)->Type);
 Rotation.SetX(((FileContainerHead *)data)->Rotation.x);
 Rotation.SetY(((FileContainerHead *)data)->Rotation.y);
 Rotation.SetZ(((FileContainerHead *)data)->Rotation.z);
 Position.SetX(((FileContainerHead *)data)->Position.x);
 Position.SetY(((FileContainerHead *)data)->Position.y);
 Position.SetZ(((FileContainerHead *)data)->Position.z);
 free(data);
 Flags|=GVE_OBJ_NEWROT | GVE_OBJ_NEWPOS;
 RealizePosition();

 return TRUE;
}


BOOL FASTCALL GLightObject::Write(GFile &file,DWORD ctype)
{
 FileLightHead flight;

 GetName(flight.Name);
 flight.Flags=DWORD2LSB(Flags);
 flight.Type=DWORD2LSB(Type);
 flight.Magic=DWORD2LSB(Magic);
 flight.Colour.R=Colour.R;
 flight.Colour.G=Colour.G;
 flight.Colour.B=Colour.B;
 flight.Specularity=Specularity;
 flight.Rotation.x=Rotation.GetX();
 flight.Rotation.y=Rotation.GetY();
 flight.Rotation.z=Rotation.GetZ();
 flight.Position.x=Position.GetX();
 flight.Position.y=Position.GetY();
 flight.Position.z=Position.GetZ();

 if(file.CreateChunk(ctype)!=TRUE) return FALSE;
 if(file.WriteChunk(&flight,sizeof(flight))!=sizeof(flight)) return FALSE;
 return file.CloseChunk();
}


BOOL FASTCALL GLightObject::Read(GFile &file)
{
 void *data;

 if(file.ReadChunk(&data)!=sizeof(FileLightHead)) return FALSE;
 if(file.CloseChunk()!=TRUE) return FALSE;
 SetName(((FileLightHead *)data)->Name);
 Magic=LSB2DWORD(((FileLightHead *)data)->Magic);
 Flags=LSB2DWORD(((FileLightHead *)data)->Flags);
 Type=LSB2DWORD(((FileLightHead *)data)->Type);
 Colour.R=((FileLightHead *)data)->Colour.R;
 Colour.G=((FileLightHead *)data)->Colour.G;
 Colour.B=((FileLightHead *)data)->Colour.B;
 Specularity=((FileLightHead *)data)->Specularity;
 Rotation.SetX(((FileLightHead *)data)->Rotation.x);
 Rotation.SetY(((FileLightHead *)data)->Rotation.y);
 Rotation.SetZ(((FileLightHead *)data)->Rotation.z);
 Position.SetX(((FileLightHead *)data)->Position.x);
 Position.SetY(((FileLightHead *)data)->Position.y);
 Position.SetZ(((FileLightHead *)data)->Position.z);
 free(data);
 Flags|=GVE_OBJ_NEWROT | GVE_OBJ_NEWPOS;
 RealizePosition();

 return TRUE;
}


BOOL FASTCALL GCameraObject::Write(GFile &file,DWORD ctype)
{
 FileCameraHead fcam;

 GetName(fcam.Name);
 fcam.Flags=DWORD2LSB(Flags);
 fcam.Type=DWORD2LSB(Type);
 fcam.Magic=DWORD2LSB(Magic);
 fcam.Rotation.x=Rotation.GetX();
 fcam.Rotation.y=Rotation.GetY();
 fcam.Rotation.z=Rotation.GetZ();
 fcam.Position.x=Position.GetX();
 fcam.Position.y=Position.GetY();
 fcam.Position.z=Position.GetZ();
 fcam.OffsetX=DWORD2LSB(Offset.x);
 fcam.OffsetY=DWORD2LSB(Offset.y);
 fcam.DistanceX=MATH2DOUBLE(Distance.x,GVE_VECTOR_SHIFT);
 fcam.DistanceY=MATH2DOUBLE(Distance.y,GVE_VECTOR_SHIFT);
 fcam.NearPlane=GetNearPlane();
 fcam.FarPlane=GetFarPlane();
 fcam.Zoom=GetZoom();

 if(file.CreateChunk(ctype)!=TRUE) return FALSE;
 if(file.WriteChunk(&fcam,sizeof(fcam))!=sizeof(fcam)) return FALSE;
 return file.CloseChunk();
}


BOOL FASTCALL GCameraObject::Read(GFile &file)
{
 void *data;

 if(file.ReadChunk(&data)!=sizeof(FileCameraHead)) return FALSE;
 if(file.CloseChunk()!=TRUE) return FALSE;
 SetName(((FileCameraHead *)data)->Name);
 Magic=LSB2DWORD(((FileCameraHead *)data)->Magic);
 Flags=LSB2DWORD(((FileCameraHead *)data)->Flags);
 Type=LSB2DWORD(((FileCameraHead *)data)->Type);
 Offset.x=LSB2DWORD(((FileCameraHead *)data)->OffsetX);
 Offset.y=LSB2DWORD(((FileCameraHead *)data)->OffsetY);
 Distance.x=DOUBLE2MATH(((FileCameraHead *)data)->DistanceX,GVE_VECTOR_SHIFT);
 Distance.y=DOUBLE2MATH(((FileCameraHead *)data)->DistanceY,GVE_VECTOR_SHIFT);
 SetZoom(((FileCameraHead *)data)->Zoom);
 SetNearPlane(((FileCameraHead *)data)->NearPlane);
 SetFarPlane(((FileCameraHead *)data)->FarPlane);
 Rotation.SetX(((FileCameraHead *)data)->Rotation.x);
 Rotation.SetY(((FileCameraHead *)data)->Rotation.y);
 Rotation.SetZ(((FileCameraHead *)data)->Rotation.z);
 Position.SetX(((FileCameraHead *)data)->Position.x);
 Position.SetY(((FileCameraHead *)data)->Position.y);
 Position.SetZ(((FileCameraHead *)data)->Position.z);
 free(data);
 Flags|=GVE_OBJ_NEWROT | GVE_OBJ_NEWPOS;
 RealizePosition();

 return TRUE;
}


BOOL FASTCALL GMeshObject::Write(GFile &file,DWORD ctype)
{
 FileMeshHead fmesh;

 GetName(fmesh.Name);
 fmesh.Flags=DWORD2LSB(Flags);
 fmesh.Type=DWORD2LSB(Type);
 fmesh.Magic=DWORD2LSB(Magic);
 fmesh.Rotation.x=Rotation.GetX();
 fmesh.Rotation.y=Rotation.GetY();
 fmesh.Rotation.z=Rotation.GetZ();
 fmesh.Position.x=Position.GetX();
 fmesh.Position.y=Position.GetY();
 fmesh.Position.z=Position.GetZ();
 fmesh.BoundingBox.minx=MATH2DOUBLE(BoundingThis.minx,GVE_VECTOR_SHIFT);
 fmesh.BoundingBox.miny=MATH2DOUBLE(BoundingThis.miny,GVE_VECTOR_SHIFT);
 fmesh.BoundingBox.minz=MATH2DOUBLE(BoundingThis.minz,GVE_VECTOR_SHIFT);
 fmesh.BoundingBox.maxx=MATH2DOUBLE(BoundingThis.maxx,GVE_VECTOR_SHIFT);
 fmesh.BoundingBox.maxy=MATH2DOUBLE(BoundingThis.maxy,GVE_VECTOR_SHIFT);
 fmesh.BoundingBox.maxz=MATH2DOUBLE(BoundingThis.maxz,GVE_VECTOR_SHIFT);

 if(file.CreateMetaChunk(ctype)!=TRUE) return FALSE;
 if(file.CreateChunk(GFile::HEAD_TYPE)!=TRUE) return FALSE;
 if(file.WriteChunk(&fmesh,sizeof(fmesh))!=sizeof(fmesh)) return FALSE;
 if(file.CloseChunk()!=TRUE) return FALSE;

 if (FaceList)
   {
	INT i,j;
	LONG size=0;
	PTR buffer;
	GFace *gface=&FaceList->Face[0];
	GVertex *gvert=&VertexList->Vertex[0];
	FileFaceHead h;
	FileFace *fface;

	h.Count=DWORD2LSB(FaceList->Count);
	h.Flags=0;

	for (i=0;i<FaceList->Count;i++)
	   {
		size+=sizeof(FileFace)+gface->VertexCount*sizeof(fface->Vertex[0]);
		(BYTE *)gface+=gface->Size;
	   }

	(PTR)fface=buffer=(PTR)malloc(size);
	gface=&FaceList->Face[0];
	for (i=0;i<FaceList->Count;i++)
	   {
		fface->Size=WORD2LSB(sizeof(FileFace)+gface->VertexCount*sizeof(fface->Vertex[0]));
		fface->Type=gface->Type;
		fface->VertexCount=gface->VertexCount;
		fface->BumpMap=gface->BumpMap;
		fface->LightMap=gface->LightMap;
		fface->TextureMap=WORD2LSB(gface->TextureMap);
		fface->Normal[0]=DWORD2LSB(gface->Normal[0]);
		fface->Normal[1]=DWORD2LSB(gface->Normal[1]);
		fface->Normal[2]=DWORD2LSB(gface->Normal[2]);
		fface->Plane=DWORD2LSB(gface->Plane);

		for (j=0;j<gface->VertexCount;j++)
			fface->Vertex[j]=WORD2LSB((WORD)(((BYTE *)gface->Vertex[j]-(BYTE *)gvert)/sizeof(GVertex)));

		(BYTE *)gface+=gface->Size;
		(BYTE *)fface+=LSB2WORD(fface->Size);
	   }

	if(file.CreateMetaChunk(GMeshObject::FACES_TYPE)!=TRUE) return FALSE;

	if(file.CreateChunk(GFile::HEAD_TYPE,GFile::COMPRESSION_NONE)!=TRUE) return FALSE;
	if(file.WriteChunk(&h,sizeof(h))!=sizeof(h)) return FALSE;
	if(file.CloseChunk()!=TRUE) return FALSE;

	if(file.CreateChunk(GFile::DATA_TYPE)!=TRUE) return FALSE;
	if(file.WriteChunk(buffer,size)!=size) return FALSE;
	if(file.CloseChunk()!=TRUE) return FALSE;

	if(file.CloseMetaChunk()!=TRUE) return FALSE;
	free(buffer);
   }

 if (VertexList)
   {
	INT i;
	LONG size;
	PTR buffer;
	GVertex *gvert=&VertexList->Vertex[0];
	FileVertex *fvert;
	FileVertexHead h;

	h.Count=DWORD2LSB(VertexList->Count);
	h.Flags=0;

	size=VertexList->Count*sizeof(FileVertex);
	(PTR)fvert=buffer=(PTR)malloc(size);

	for (i=0;i<VertexList->Count;i++)
	   {
		fvert[i].Position.x=gvert[i].Position.GetX();
		fvert[i].Position.y=gvert[i].Position.GetY();
		fvert[i].Position.z=gvert[i].Position.GetZ();
		fvert[i].Normal.x=gvert[i].Normal.GetX();
		fvert[i].Normal.y=gvert[i].Normal.GetY();
		fvert[i].Normal.z=gvert[i].Normal.GetZ();
		fvert[i].Texture.u=DWORD2LSB(gvert[i].Texture.u);
		fvert[i].Texture.v=DWORD2LSB(gvert[i].Texture.v);
	   }

	if(file.CreateMetaChunk(GMeshObject::VERTICES_TYPE)!=TRUE) return FALSE;

	if(file.CreateChunk(GFile::HEAD_TYPE,GFile::COMPRESSION_NONE)!=TRUE) return FALSE;
	if(file.WriteChunk(&h,sizeof(h))!=sizeof(h)) return FALSE;
	if(file.CloseChunk()!=TRUE) return FALSE;

	if(file.CreateChunk(GFile::DATA_TYPE)!=TRUE) return FALSE;
	if(file.WriteChunk(buffer,size)!=size) return FALSE;
	if(file.CloseChunk()!=TRUE) return FALSE;

	if(file.CloseMetaChunk()!=TRUE) return FALSE;
	free(buffer);
   }

 return file.CloseMetaChunk();
}


BOOL FASTCALL GMeshObject::Read(GFile &file)
{
 DWORD typ;
 PTR data;

 while ((typ=file.OpenChunk())!=GFile::END_TYPE)
   {
	switch (typ)
	   {
		case GFile::HEAD_TYPE:
			if(file.ReadChunk(&data)!=sizeof(FileMeshHead)) return FALSE;
			if(file.CloseChunk()!=TRUE) return FALSE;
			SetName(((FileMeshHead *)data)->Name);
			Flags=DWORD2LSB(((FileMeshHead *)data)->Flags);
			Type=DWORD2LSB(((FileMeshHead *)data)->Type);
			Magic=DWORD2LSB(((FileMeshHead *)data)->Magic);
			Rotation.SetX(((FileMeshHead *)data)->Rotation.x);
			Rotation.SetY(((FileMeshHead *)data)->Rotation.y);
			Rotation.SetZ(((FileMeshHead *)data)->Rotation.z);
			Position.SetX(((FileMeshHead *)data)->Position.x);
			Position.SetY(((FileMeshHead *)data)->Position.y);
			Position.SetZ(((FileMeshHead *)data)->Position.z);
			BoundingThis.minx=DOUBLE2MATH(((FileMeshHead *)data)->BoundingBox.minx,GVE_VECTOR_SHIFT);
			BoundingThis.miny=DOUBLE2MATH(((FileMeshHead *)data)->BoundingBox.miny,GVE_VECTOR_SHIFT);
			BoundingThis.minz=DOUBLE2MATH(((FileMeshHead *)data)->BoundingBox.minz,GVE_VECTOR_SHIFT);
			BoundingThis.maxx=DOUBLE2MATH(((FileMeshHead *)data)->BoundingBox.maxx,GVE_VECTOR_SHIFT);
			BoundingThis.maxy=DOUBLE2MATH(((FileMeshHead *)data)->BoundingBox.maxy,GVE_VECTOR_SHIFT);
			BoundingThis.maxz=DOUBLE2MATH(((FileMeshHead *)data)->BoundingBox.maxz,GVE_VECTOR_SHIFT);
			free(data);
			break;
		case GMeshObject::FACES_TYPE:
			if(ReadFaceList(file)!=TRUE) return FALSE;
			break;
		case GMeshObject::VERTICES_TYPE:
			if(ReadVertexList(file)!=TRUE) return FALSE;
			break;
		default:
			if(file.CloseChunk()!=TRUE) return FALSE;
			break;
	   }
   }

 Flags&=~GVE_OBJ_VERTEX_ADDRESS;
 Flags|=GVE_OBJ_VERTEX_INDEX |GVE_OBJ_NEWROT | GVE_OBJ_NEWPOS;
 ReferenceFaces();
 RealizePosition();

 return TRUE;
}



BOOL FASTCALL GMeshObject::ReadVertexList(GFile &file)
{
 DWORD typ;
 INT i;
 LONG count;
 PTR data;
 GVertex *gvert;
 FileVertex *fvert;

 FreeVertexList();

 while ((typ=file.OpenChunk())!=GFile::END_TYPE)
   {
	switch (typ)
	   {
		case GFile::HEAD_TYPE:
			if(file.ReadChunk(&data)!=sizeof(FileVertexHead)) return FALSE;
			if(file.CloseChunk()!=TRUE) return FALSE;
			count=LSB2DWORD(((FileVertexHead *)data)->Count);
			free(data);
			break;
		case GFile::DATA_TYPE:
			if(file.ReadChunk(&data)==0) return FALSE;
			if(file.CloseChunk()!=TRUE) return FALSE;

			AllocVertexList(count);
			gvert=&VertexList->Vertex[0];
			fvert=(FileVertex *)data;

			for (i=0;i<VertexList->Count;i++)
			   {
				gvert[i].Position.SetX(fvert[i].Position.x);
				gvert[i].Position.SetY(fvert[i].Position.y);
				gvert[i].Position.SetZ(fvert[i].Position.z);
				gvert[i].Normal.SetX(fvert[i].Normal.x);
				gvert[i].Normal.SetY(fvert[i].Normal.y);
				gvert[i].Normal.SetZ(fvert[i].Normal.z);
				gvert[i].Texture.u=LSB2DWORD(fvert[i].Texture.u);
				gvert[i].Texture.v=LSB2DWORD(fvert[i].Texture.v);
			   }

			free(data);
			break;
		default:
			if(file.CloseChunk()!=TRUE) return FALSE;
			break;
	   }
   }

 return TRUE;
}



BOOL FASTCALL GMeshObject::ReadFaceList(GFile &file)
{
 DWORD typ;
 INT i,j;
 LONG count,size;
 PTR data;
 GFace *gface;
 FileFace *fface;

 FreeFaceList();

 while ((typ=file.OpenChunk())!=GFile::END_TYPE)
   {
	switch (typ)
	   {
		case GFile::HEAD_TYPE:
			if(file.ReadChunk(&data)!=sizeof(FileFaceHead)) return FALSE;
			if(file.CloseChunk()!=TRUE) return FALSE;
			count=LSB2DWORD(((FileFaceHead *)data)->Count);
			free(data);
			break;
		case GFile::DATA_TYPE:
			if(file.ReadChunk(&data)==0) return FALSE;
			if(file.CloseChunk()!=TRUE) return FALSE;

			size=sizeof(GFaceList);
			fface=(FileFace *)data;
			for (i=0;i<count;i++)
			   {
				size+=sizeof(GFace)+fface->VertexCount*sizeof(GVertex *);
				(BYTE *)fface+=LSB2WORD(fface->Size);
			   }

			FaceList=(GFaceList *)malloc(size);
			FaceList->Count=count;
			gface=&FaceList->Face[0];
			fface=(FileFace *)data;

			for (i=0;i<count;i++)
			   {
				gface->VertexCount=fface->VertexCount;
				gface->Size=(WORD)(sizeof(GFace)+gface->VertexCount*sizeof(GVertex *));
				gface->Type=fface->Type;
				gface->BumpMap=fface->BumpMap;
				gface->LightMap=fface->LightMap;
				gface->TextureMap=LSB2WORD(fface->TextureMap);
				gface->Normal[0]=LSB2DWORD(fface->Normal[0]);
				gface->Normal[1]=LSB2DWORD(fface->Normal[1]);
				gface->Normal[2]=LSB2DWORD(fface->Normal[2]);
				gface->Plane=LSB2DWORD(fface->Plane);

				for (j=0;j<gface->VertexCount;j++)
					gface->Vertex[j]=(GVertex *)(DWORD)LSB2WORD(fface->Vertex[j]);

				(BYTE *)gface+=gface->Size;
				(BYTE *)fface+=LSB2WORD(fface->Size);
			   }

			free(data);
			break;
		default:
			if(file.CloseChunk()!=TRUE) return FALSE;
			break;
	   }
   }

 return TRUE;
}

