///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#include "../include/gvepro.h"


///////////////////////////////////////////////////////////////////////////////
//
// Vertex drawing
//
void FASTCALL GRasterEngine16::DrawVertex(GFace *face)
{
 INT i;
 FAST LONG x,y,z;
 FAST WORD col=ShadingLUT[GVE_SHADINGLUT_ENTRIES/2+LOBYTE(face->TextureMap)];

 for (i=0;i<face->VertexCount;i++)
   {
	x=face->Vertex[i]->SPosition.x>>GVE_RASTERX_SHIFT;
	y=face->Vertex[i]->SPosition.y>>GVE_RASTERY_SHIFT;
	z=face->Vertex[i]->InverseZ;
	CurrentFrame->PutPixel(x,y,col,z,0);
   }
}


///////////////////////////////////////////////////////////////////////////////
//
// Simple single-colour unlit filled polygons
//
void FASTCALL GRasterEngine16::DrawFilled(GFace *face)
{
 #define PIXEL_TYPE				WORD

 #define PIXEL_COLOUR()			\
	colour

 #define SETUP_CODE()			\
	WORD colour=ShadingLUT[GVE_SHADINGLUT_ENTRIES/2+LOBYTE(face->TextureMap)];

 #include "../source/gvetritm.cpp"
}

///////////////////////////////////////////////////////////////////////////////
//
// Simple single-colour lit filled polygons
//
void FASTCALL GRasterEngine16::DrawFilledLambert(GFace *face)
{
 #define PIXEL_TYPE				WORD

 #define PIXEL_COLOUR()			\
	colour

 #define SETUP_CODE()			\
	WORD colour=ShadingLUT[LOBYTE(face->TextureMap)+((face->Vertex[0]->Shading>>(2*GVE_NORMAL_SHIFT-5-8))&~0xff)];

 #include "../source/gvetritm.cpp"
}


///////////////////////////////////////////////////////////////////////////////
//
// Gouraud shaded filled triangle
//
void FASTCALL GRasterEngine16::DrawFilledGouraud(GFace *face)
{
 #define PIXEL_TYPE				WORD
 #define INTERPOLATE_SHADING
 #define DITHER_PIXEL

 #define PIXEL_COLOUR()			\
	ShadingLUT[(((hs+hd)>>(2*GVE_NORMAL_SHIFT-5-8))&~0xff)+colour]

 #define SETUP_CODE()													   	\
	INT colour=(INT)LOBYTE(face->TextureMap);

 #include "../source/gvetritm.cpp"
}


///////////////////////////////////////////////////////////////////////////////
//
// Simple single-colour unlit filled transparent polygons
//
void FASTCALL GRasterEngine16::DrawFilledGlass(GFace *face)
{
 #define PIXEL_TYPE				WORD

 #define PIXEL_COLOUR()			\
	((frame[hx1]&SoftenMask)+colour)>>1

 #define SETUP_CODE()			\
	WORD colour=ShadingLUT[GVE_SHADINGLUT_ENTRIES/2+LOBYTE(face->TextureMap)] & SoftenMask;

 #include "../source/gvetritm.cpp"
}

///////////////////////////////////////////////////////////////////////////////
//
// Simple single-colour lit filled transparent polygons
//
void FASTCALL GRasterEngine16::DrawFilledGlassLambert(GFace *face)
{
 #define PIXEL_TYPE				WORD

 #define PIXEL_COLOUR()			\
	frame[hx1]=((frame[hx1]&SoftenMask)+colour)>>1

 #define SETUP_CODE()			\
	WORD colour=ShadingLUT[LOBYTE(face->TextureMap)+((face->Vertex[0]->Shading>>(2*GVE_NORMAL_SHIFT-5-8))&~0xff)] & SoftenMask;

 #include "../source/gvetritm.cpp"
}


///////////////////////////////////////////////////////////////////////////////
//
// Gouraud shaded filled transparent triangle
//
void FASTCALL GRasterEngine16::DrawFilledGlassGouraud(GFace *face)
{
 #define PIXEL_TYPE				WORD
 #define INTERPOLATE_SHADING
 #define DITHER_PIXEL

 #define PIXEL_COLOUR()			\
	((frame[hx1]&SoftenMask)+(ShadingLUT[(((hs+hd)>>(2*GVE_NORMAL_SHIFT-5-8))&~0xff)+colour]&SoftenMask))>>1

 #define SETUP_CODE()													   	\
	INT colour=(INT)LOBYTE(face->TextureMap);

 #include "../source/gvetritm.cpp"
}


///////////////////////////////////////////////////////////////////////////////
//
// Simple texturemapped filled triangle
//
void FASTCALL GRasterEngine16::DrawTextured(GFace *face)
{
 #define PIXEL_TYPE				WORD
 #define INTERPOLATE_TEXTURE_UV

 #define PIXEL_COLOUR()			\
	ShadingLUT[mapdata[(((htv*GVE_TEXTURE_WIDTH)>>GVE_MAPPING_SHIFT)&0xff)+     \
						((htu>>GVE_MAPPING_SHIFT)&0xff)]]

 #define SETUP_CODE()													   	\
	GBitmap* map=TextureMapList->GetEntry(face->TextureMap & 0x3fff);       \
	BYTE* mapdata=(BYTE *)map->GetImage();

 #include "../source/gvetritm.cpp"
}




///////////////////////////////////////////////////////////////////////////////
//
// Postprocessor
//
void FASTCALL GRasterEngine16::Postprocess(DWORD flags)
{
 WORD *nbuf=(WORD *)CurrentFrame->GetImage();
 WORD *obuf=(WORD *)LastFrame->GetImage();

 if (flags & GVE_WRLD_MOTION_BLURR)
   {
	WORD *nptr=(WORD *)nbuf,*optr=(WORD *)obuf;
	WORD *last=(WORD *)nbuf+FrameWidth*FrameHeight;

	while (nptr<last)
	   {
		*nptr=(WORD)(((DWORD)(*nptr & SoftenMask) + (DWORD)(*optr & SoftenMask))>>1);
		nptr++;
		optr++;
	   }
   }
 if (flags & GVE_WRLD_SOFTEN_HORIZONTAL)
   {
	WORD *nptr=(WORD *)nbuf;
	WORD *last=(WORD *)nbuf+FrameWidth*FrameHeight-1;

	while (nptr<last)
	   {
		*nptr=(WORD)(((DWORD)(*nptr & SoftenMask) + (DWORD)(*(nptr+1) & SoftenMask))>>1);
		nptr++;
	   }
	}
 if (flags & GVE_WRLD_SOFTEN_VERTICAL)
   {
	WORD *nptr=(WORD *)nbuf;
	WORD *last=(WORD *)nbuf+FrameWidth*(FrameHeight-1);

	while (nptr<last)
	   {
		*nptr=(WORD)(((DWORD)(*nptr & SoftenMask) + (DWORD)(*(nptr+FrameWidth) & SoftenMask))>>1);
		nptr++;
	   }
   }
}


///////////////////////////////////////////////////////////////////////////////
//
// Simple texturemapped filled triangle
//
void FASTCALL GRasterEngine16::Preprocess(DWORD flags)
{
 WORD *nbuf=(WORD *)CurrentFrame->GetImage();
 WORD *obuf=(WORD *)LastFrame->GetImage();

 if (flags & GVE_WRLD_CLEAR_FRAME)
	memset(nbuf,0,FrameWidth*FrameHeight*2);
}


