///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//
///////////////////////////////////////////////////////////////////////////////
//
// The following macros can be defined
//		INTERPOLATE_TEXTURE_UV
//		INTERPOLATE_CHROME_UV
//		INTERPOLATE_DEPTH
//		INTERPOLATE_SHADING
//		DITHER_PIXEL
//		DEPTH_TEST
//		DEPTH_WRITE
//
// For drawing the horizontal spans, you can use one or more of the
// following methods:
//		PIXEL_TYPE				- BYTE,WORD or DWORD
//      DRAW_PIXEL(x1,x2)		- Draw span at frame[x1] to fram[x2]
// or:
//		PIXEL_TYPE				- BYTE,WORD or DWORD
//		PIXEL_COLOUR()			- Colour value at current position
// or:
//		DRAW_PIXEL(x,y,len)		- Draw span at (x,y) with length len
//
// You can also specify a setup code:
//		SETUP_CODE()
//
///////////////////////////////////////////////////////////////////////////////
//
// Triangle drawing template
//
{
 LONG lx,rx,y,hx1,hx2,x2,d;
 LONG inc_x1x2,inc_x2x3,inc_x1x3;
 LONG delta_y1y2,delta_y2y3,delta_y1y3;
 GVertex *v1=face->Vertex[0],*v2=face->Vertex[1],*v3=face->Vertex[2];

 #if defined(DEPTH_TEST) && !defined(INTERPOLATE_DEPTH)
 #define INTERPOLATE_DEPTH
 #endif
 #if defined(DEPTH_WRITE) && !defined(INTERPOLATE_DEPTH)
 #define INTERPOLATE_DEPTH
 #endif

 #if defined(INTERPOLATE_TEXTURE_UV) || defined(INTERPOLATE_CHROME_UV) \
  || defined(INTERPOLATE_DEPTH) || defined(INTERPOLATE_SHADING)
 #define DENOM
 LONG denom;
 #endif
 #if defined(INTERPOLATE_SHADING)
 LONG inc_s1s2,inc_s2s3,inc_hs,ls,s2,hs;
 #endif
 #if defined(INTERPOLATE_DEPTH)
 LONG inc_z1z2,inc_z2z3,inc_hz,lz,z2,hz;
 #endif
 #if defined(DITHER_PIXEL)
 LONG ld,hd;
 #endif
 #if defined(INTERPOLATE_TEXTURE_UV)
 LONG inc_tu1tu2,inc_tu2tu3,inc_htu,ltu,tu2,htu;
 LONG inc_tv1tv2,inc_tv2tv3,inc_htv,ltv,tv2,htv;
 #endif
 #if defined(INTERPOLATE_CHROME_UV)
 LONG inc_cu1cu2,inc_cu2cu3,inc_hcu,lcu,cu2,hcu;
 LONG inc_cv1cv2,inc_cv2cv3,inc_hcv,lcv,cv2,hcv;
 #endif

 //////////////////////////////////////////////////////////////////////////////
 //
 // First sort all vertices (same for all shaders)
 //
 if (v1->SPosition.y < v2->SPosition.y)
   {
	if (v2->SPosition.y > v3->SPosition.y)
	   {
		SwapVars(v2,v3);
		if (v2->SPosition.y < v1->SPosition.y)
				SwapVars(v2,v1);
	   }
   }
 else
   {
	if (v2->SPosition.y > v3->SPosition.y)
	   {
		SwapVars(v1,v3);
	   }
	else
	   {
		SwapVars(v1,v2);
		if (v2->SPosition.y > v3->SPosition.y)
		SwapVars(v2,v3);
	   }
   }

 // Perform super-simple rejecting
 if (v3->SPosition.y < FrameClipWindow.Top) return;
 if (v1->SPosition.y > FrameClipWindow.Bottom) return;

 #if defined(SETUP_CODE)
 SETUP_CODE();
 #endif

 // Now calc the slopes and deltas
 if (!(delta_y1y3=v3->SPosition.y - v1->SPosition.y)) return;
	inc_x1x3=(v3->SPosition.x - v1->SPosition.x)/delta_y1y3;
 if (delta_y2y3=v3->SPosition.y - v2->SPosition.y)
	inc_x2x3=(v3->SPosition.x - v2->SPosition.x)/delta_y2y3;
 if (delta_y1y2=v2->SPosition.y - v1->SPosition.y)
	inc_x1x2=(v2->SPosition.x - v1->SPosition.x)/delta_y1y2;

 lx=rx=v1->SPosition.x;
 x2=v2->SPosition.x;
 y=v1->SPosition.y;

 #if defined DENOM
 if(!(denom=LARGE2LONG((LARGE(v1->SPosition.x-v3->SPosition.x)*delta_y2y3-
	LARGE(v2->SPosition.x-v3->SPosition.x)*delta_y1y3)>>GVE_RASTERX_SHIFT))) return;
 #endif

 //////////////////////////////////////////////////////////////////////////////
 //
 // Get the horizontal increments for the shaders
 //
 #if defined(INTERPOLATE_SHADING)
 inc_hs=LARGE2LONG((LARGE(v1->Shading-v3->Shading)*delta_y2y3-
			LARGE(v2->Shading-v3->Shading)*delta_y1y3)/denom);
 ls=v1->Shading;
 #endif
 #if defined(INTERPOLATE_DEPTH)
 inc_hz=LARGE2LONG((LARGE(v1->InverseZ-v3->InverseZ)*delta_y2y3-
			LARGE(v2->InverseZ-v3->InverseZ)*delta_y1y3)/denom);
 lz=v1->InverseZ;
 #endif
 #if defined(INTERPOLATE_TEXTURE_UV)
 inc_htu=LARGE2LONG((LARGE(v1->Texture.u-v3->Texture.u)*delta_y2y3-
			LARGE(v2->Texture.u-v3->Texture.u)*delta_y1y3)/denom);
 inc_htv=LARGE2LONG((LARGE(v1->Texture.v-v3->Texture.v)*delta_y2y3-
			LARGE(v2->Texture.v-v3->Texture.v)*delta_y1y3)/denom);
 ltu=v1->Texture.u;
 ltv=v1->Texture.v;
 #endif
 #if defined(INTERPOLATE_CHROME_UV)
 inc_hcu=LARGE2LONG((LARGE(v1->Chrome.u-v3->Chrome.u)*delta_y2y3-
			LARGE(v2->Chrome.u-v3->Chrome.u)*delta_y1y3)/denom);
 inc_hcv=LARGE2LONG((LARGE(v1->Chrome.v-v3->Chrome.v)*delta_y2y3-
			LARGE(v2->Chrome.v-v3->Chrome.v)*delta_y1y3)/denom);
 lcu=v1->Chrome.u;
 lcv=v1->Chrome.v;
 #endif

 // Get vertical slopes for interpolation shaders
 #if defined(DENOM)
 if (denom>0)
   {
	if (delta_y1y2)
	   {
		#if defined(INTERPOLATE_SHADING)
		inc_s1s2=(v2->Shading - v1->Shading)/delta_y1y2;
		s2=v2->Shading;
		#endif
		#if defined(INTERPOLATE_DEPTH)
		inc_z1z2=(v2->InverseZ - v1->InverseZ)/delta_y1y2;
		z2=v2->InverseZ;
		#endif
		#if defined(INTERPOLATE_TEXTURE_UV)
		inc_tu1tu2=(v2->Texture.u - v1->Texture.u)/delta_y1y2;
		inc_tv1tv2=(v2->Texture.v - v1->Texture.v)/delta_y1y2;
		tu2=v2->Texture.u;
		tv2=v2->Texture.v;
		#endif
		#if defined(INTERPOLATE_CHROME_UV)
		inc_cu1cu2=(v2->Chrome.u - v1->Chrome.u)/delta_y1y2;
		inc_cv1cv2=(v2->Chrome.v - v1->Chrome.v)/delta_y1y2;
		cu2=v2->Chrome.u;
		cv2=v2->Chrome.v;
		#endif
	   }
	if (delta_y2y3)
	   {
		#if defined(INTERPOLATE_SHADING)
		inc_s2s3=(v3->Shading - v2->Shading)/delta_y2y3;
		s2=v2->Shading;
		#endif
		#if defined(INTERPOLATE_DEPTH)
		inc_z2z3=(v3->InverseZ - v2->InverseZ)/delta_y2y3;
		s2=v2->InverseZ;
		#endif
		#if defined(INTERPOLATE_TEXTURE_UV)
		inc_tu1tu2=(v3->Texture.u - v2->Texture.u)/delta_y2y3;
		inc_tv1tv2=(v3->Texture.v - v2->Texture.v)/delta_y2y3;
		tu2=v2->Texture.u;
		tv2=v2->Texture.v;
		#endif
		#if defined(INTERPOLATE_CHROME_UV)
		inc_cu1cu2=(v3->Chrome.u - v2->Chrome.u)/delta_y2y3;
		inc_cv1cv2=(v3->Chrome.v - v2->Chrome.v)/delta_y2y3;
		cu2=v2->Chrome.u;
		cv2=v2->Chrome.v;
		#endif
	   }
   }
 else
   {
	#if defined(INTERPOLATE_SHADING)
	inc_s1s2=inc_s2s3=(v3->Shading - v1->Shading)/delta_y1y3;
	s2=ls+delta_y1y2*inc_s1s2;
	#endif
	#if defined(INTERPOLATE_DEPTH)
	inc_z1z2=inc_z2z3=(v3->InverseZ - v1->InverseZ)/delta_y1y3;
	z2=ls+delta_y1y2*inc_z1z2;
	#endif
	#if defined(INTERPOLATE_TEXTURE_UV)
	inc_tu1tu2=inc_tu2tu3=(v3->Texture.u - v1->Texture.u)/delta_y1y3;
	inc_tv1tv2=inc_tv2tv3=(v3->Texture.v - v1->Texture.v)/delta_y1y3;
	tu2=ltu+delta_y1y2*inc_tu1tu2;
	tv2=ltv+delta_y1y2*inc_tv1tv2;
	#endif
	#if defined(INTERPOLATE_CHROME_UV)
	inc_cu1cu2=inc_cu2cu3=(v3->Chrome.u - v1->Chrome.u)/delta_y1y3;
	inc_cv1cv2=inc_cv2cv3=(v3->Chrome.v - v1->Chrome.v)/delta_y1y3;
	cu2=lcu+delta_y1y2*inc_cu1cu2;
	cv2=lcv+delta_y1y2*inc_cv1cv2;
	#endif
   }
 #endif


 //////////////////////////////////////////////////////////////////////////////
 //
 // Perform top/down clipping
 //
 if ((d=y-FrameClipWindow.Top)<0)
   {
	rx-=d*inc_x1x3;
	if((delta_y1y2+=d)<0)
	   {
		#if defined(INTERPOLATE_SHADING)
		ls=s2 - delta_y1y2*inc_s2s3;
		inc_s1s2=inc_s2s3;
		#endif
		#if defined(INTERPOLATE_DEPTH)
		lz=z2 - delta_y1y2*inc_z2z3;
		inc_z1z2=inc_z2z3;
		#endif
		#if defined(INTERPOLATE_TEXTURE_UV)
		ltu=tu2 - delta_y1y2*inc_tu2tu3;
		ltv=tv2 - delta_y1y2*inc_tv2tv3;
		inc_tu1tu2=inc_tu2tu3;
		inc_tv1tv2=inc_tv2tv3;
		#endif
		#if defined(INTERPOLATE_CHROME_UV)
		lcu=cu2 - delta_y1y2*inc_cu2cu3;
		lcv=cv2 - delta_y1y2*inc_cv2cv3;
		inc_cu1cu2=inc_cu2cu3;
		inc_cv1cv2=inc_cv2cv3;
		#endif
		lx=x2 - delta_y1y2*inc_x2x3;
		inc_x1x2=inc_x2x3;
		delta_y1y2+=delta_y2y3;
		delta_y2y3=0;
	   }
	else
	   {
		#if defined(INTERPOLATE_SHADING)
		ls-=d*inc_s1s2;
		#endif
		#if defined(INTERPOLATE_DEPTH)
		lz-=d*inc_z1z2;
		#endif
		#if defined(INTERPOLATE_TEXTURE_UV)
		ltu-=d*inc_tu1tu2;
		ltv-=d*inc_tv1tv2;
		#endif
		#if defined(INTERPOLATE_CHROME_UV)
		lcu-=d*inc_cu1cu2;
		lcv-=d*inc_cv1cv2;
		#endif
		lx-=d*inc_x1x2;
	   }
	y = FrameClipWindow.Top;
   }

 // Peform vertical clipping at bottom
 if (y+delta_y1y2 > FrameClipWindow.Bottom)
   {
	delta_y1y2 = FrameClipWindow.Bottom - y + 1;
	delta_y2y3 = 0;
   }
 else if (y+delta_y1y2+delta_y2y3 > FrameClipWindow.Bottom)
   {
	delta_y2y3 = FrameClipWindow.Bottom - y - delta_y1y2 + 1;
   }

 //////////////////////////////////////////////////////////////////////////////
 //
 // Start drawing
 //
 #if defined(PIXEL_TYPE)
 #if !defined(SPAN_BUFFER)
  PIXEL_TYPE* frame=&((PIXEL_TYPE *)CurrentFrame->GetImage())[y*FrameWidth];
  #if defined(INTERPOLATE_DEPTH)
  #define ALPHA
  GAlpha* alpha=&CurrentFrame->GetAlpha[y*FrameWidth];
  #endif
 #else
  GExtPixel span[GMZ_MAX_FRAMEWIDTH];
  GExtPixel *spanptr=span;
 #endif
 #endif

 #if defined(DITHER_PIXEL)
 ld=(y & 1)<<GVE_DITHER_SHIFT;
 #endif

 // Draw triangle !
 for(;;)
   {
	// Check if we need to start second part of triangle
	if (!delta_y1y2)
	   {
		if(!(delta_y1y2=delta_y2y3)) return;
		inc_x1x2=inc_x2x3;
		delta_y2y3=0;
		lx=x2;
		#if defined(INTERPOLATE_SHADING)
		inc_s1s2=inc_s2s3;
		ls=s2;
		#endif
		#if defined(INTERPOLATE_DEPTH)
		inc_z1z2=inc_z2z3;
		lz=z2;
		#endif
		#if defined(INTERPOLATE_TEXTURE_UV)
		inc_tu1tu2=inc_tu2tu3;
		inc_tv1tv2=inc_tv2tv3;
		ltu=tu2;
		ltv=tv2;
		#endif
		#if defined(INTERPOLATE_CHROME_UV)
		inc_cu1cu2=inc_cu2cu3;
		inc_cv1cv2=inc_cv2cv3;
		lcu=cu2;
		lcv=cv2;
		#endif
	   }

	// Make sure that we draw from left to right
	if ((hx1=lx>>GVE_RASTERX_SHIFT)>(hx2=rx>>GVE_RASTERX_SHIFT))
		SwapVars(hx1,hx2);

	if ((hx1<=FrameClipWindow.Right) &&
		(hx2>=FrameClipWindow.Left))
	   {
		// Horizontal clipping
		if (hx1<FrameClipWindow.Left)
		   {
			#if defined(INTERPOLATE_SHADING)
			hs=ls+(FrameClipWindow.Left-hx1)*inc_hs;
			#endif
			#if defined(INTERPOLATE_DEPTH)
			hz=lz+(FrameClipWindow.Left-hx1)*inc_hz;
			#endif
			#if defined(INTERPOLATE_TEXTURE_UV)
			htu=ltu+(FrameClipWindow.Left-hx1)*inc_htu;
			htv=ltv+(FrameClipWindow.Left-hx1)*inc_htv;
			#endif
			#if defined(INTERPOLATE_CHROME_UV)
			hcu=lcu+(FrameClipWindow.Left-hx1)*inc_hcu;
			hcv=lcv+(FrameClipWindow.Left-hx1)*inc_hcv;
			#endif
			hx1=FrameClipWindow.Left;
		   }
		else
		   {
			#if defined(INTERPOLATE_SHADING)
			hs=ls;
			#endif
			#if defined(INTERPOLATE_DEPTH)
			hz=lz;
			#endif
			#if defined(INTERPOLATE_TEXTURE_UV)
			htu=ltu;
			htv=ltv;
			#endif
			#if defined(INTERPOLATE_CHROME_UV)
			hcu=lcu;
			hcv=lcv;
			#endif
		   }
		if (hx2>FrameClipWindow.Right)
			hx2=FrameClipWindow.Right;

		// Setup dither values;
		#if defined(DITHER_PIXEL)
		hd=ld^((hx1 & 1)<<GVE_DITHER_SHIFT);
		#endif


		// The variables are now set up like this:
		//		frame			- Pointer to first pixel in row
		//		hx1				- Leftmost pixel
		//		hx2 			- Rightmost pixel (not to be drawn !)
		//		hs  			- Shading value
		//		hz  			- Depth value
		//		hd		    	- Dither value
		//		inc_hs			- Hor. shading increment
		//		htu,htv     	- Texture coordinates
		//		inc_htu,inc_htv - Hor. texture increment
		//		hcu,hcv     	- Chrome coordinates
		//		inc_hcu,inc_hcv - Hor. chrome increment
		#if defined(PIXEL_TYPE) && defined(PIXEL_COLOUR)
		#if defined(SPAN_BUFFER)
		for(d=hx2-hx1;d;d--)
		#else
		for (;hx1<hx2;hx1++)
		#endif
		   {
			#if defined(SPAN_BUFFER)
			spanptr->Depth=hd;
			spanptr->Colour=(DWORD)(PIXEL_COLOUR());
			spanptr->Stencil=stencil;
			spanptr++;
			#elif defined(DEPTH_TEST)
			if (alpha[hx1].Depth<hd)
			   {
				alpha[hx1].Depth=hd;
				frame[hx1]=(PIXEL_COLOUR());
			   }
			#elif defined(DEPTH_WRITE)
			alpha[hx1].Depth=hd;
			frame[hx1]=(PIXEL_COLOUR());
			#else
			frame[hx1]=(PIXEL_COLOUR());
			#endif
			#if defined(INTERPOLATE_SHADING)
			hs+=inc_hs;
			#endif
			#if defined(DITHER_PIXEL)
			hd^=1<<GVE_DITHER_SHIFT;
			#endif
			#if defined(INTERPOLATE_DEPTH)
			hz+=inc_hz;
			#endif
			#if defined(INTERPOLATE_TEXTURE_UV)
			htu+=inc_htu;
			htv+=inc_htv;
			#endif
			#if defined(INTERPOLATE_CHROME_UV)
			hcu+=inc_hcu;
			hcv+=inc_hcv;
			#endif
		   }
		#if defined(SPAN_BUFFER)
		CurrentFrame->PutSpan(hx1,y,hx2-hx1,span);
		y++;
		#endif
		#elif defined(PIXEL_TYPE)
		DRAW_PIXEL(hx1,hx2);
		#else
		d=hx2-hx1;
		DRAW_PIXEL(y,hx1,d);
		#endif
	   }

	// Skip to next row
	lx+=inc_x1x2;
	rx+=inc_x1x3;
	#if defined(ALPHA)
	alpha+=FrameWidth;
	#endif
	#if defined(PIXEL_TYPE)
	(PIXEL_TYPE *)frame+=FrameWidth;
	#else
	y++;
	#endif
	#if defined(DITHER_PIXEL)
	ld^=1<<GVE_DITHER_SHIFT;
	#endif
	#if defined(INTERPOLATE_SHADING)
	ls+=inc_s1s2;
	#endif
	#if defined(INTERPOLATE_DEPTH)
	lz+=inc_z1z2;
	#endif
	#if defined(INTERPOLATE_TEXTURE_UV)
	ltu+=inc_tu1tu2;
	ltv+=inc_tv1tv2;
	#endif
	#if defined(INTERPOLATE_CHROME_UV)
	lcu+=inc_cu1cu2;
	lcv+=inc_cv1cv2;
	#endif
	delta_y1y2--;
   }
}

#undef INTERPOLATE_TEXTURE_UV
#undef INTERPOLATE_CHROME_UV
#undef INTERPOLATE_DEPTH
#undef INTERPOLATE_SHADING
#undef DITHER_PIXEL
#undef PIXEL_TYPE
#undef PIXEL_COLOUR
#undef DRAW_PIXEL
#undef SETUP_CODE
#undef DEPTH_TEST
#undef DEPTH_WRITE
#undef DENOM
#undef ALPHA
