///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#include "../include/gvepro.h"


BOOL FASTCALL GVectorWorld::Alloc(void)
{
 GVectorObject *root;

 Flags=GVE_WRLD_CLEAR_FRAME;
 root=new GVectorObject((GVectorObject *)NULL,GObject::CONTAINER_TYPE);
 delete ObjectList.GetRoot();
 ObjectList.SetRoot(root);

 Camera=(GCameraObject *)root->AddSon(new GCameraObject);
 Light=(GLightObject *)root->AddSon(new GLightObject(root,GLightObject::POINTLIGHT_TYPE));
 Ambient=(GLightObject *)root->AddSon(new GLightObject(root,GLightObject::AMBIENT_TYPE));
 Palette.Reset();

 Camera->SetPosition(0.0,0.0,-100.0);
 Camera->SetRotation(0.0,0.0,0.0);

 Light->SetPosition(50.0,50.0,0.0);
 Light->SetRotation(10.0,20.0,30.0);

 Ambient->SetColour(GColour(20,20,20));

 RealizePosition(TRUE);
 return TRUE;
}


BOOL FASTCALL GVectorWorld::Free(void)
{
 ObjectList.Clear();
 TextureMapList.Clear();
 LightMapList.Clear();
 BumpMapList.Clear();
 Info.Clear();

 (PTR)Flags=
 (PTR)Camera=
 (PTR)Light=
 (PTR)Ambient=NULL;
 return TRUE;
}


GCameraObject* FASTCALL GVectorWorld::SetCamera(GCameraObject *cam)
{
 if (!FindObject(cam)) return NULL;
 if ((typeid(*cam)!=typeid(GCameraObject)) || (cam->GetType()!=GCameraObject::CAMERA_TYPE)) return NULL;
 return Camera=cam;
}


GLightObject* FASTCALL GVectorWorld::SetLight(GLightObject *ligh)
{
 if (!FindObject(ligh)) return NULL;
 if ((typeid(*ligh)!=typeid(GLightObject)) ||
	 ((ligh->GetType()!=GLightObject::LIGHT_TYPE) &&
	  (ligh->GetType()!=GLightObject::POINTLIGHT_TYPE) &&
	  (ligh->GetType()!=GLightObject::SPOTLIGHT_TYPE))) return NULL;
 return Light=ligh;
}


GLightObject* FASTCALL GVectorWorld::SetAmbientLight(GLightObject *ligh)
{
 if (!FindObject(ligh)) return NULL;
 if ((typeid(*ligh)!=typeid(GLightObject)) || (ligh->GetType()!=GLightObject::AMBIENT_TYPE)) return NULL;
 return Ambient=ligh;
}


CHAR* FASTCALL GVectorWorld::GetFlagsName(CHAR *buffer)
{
 buffer[0]=0;

 if (Flags & GVE_WRLD_NOSORT) strcat(buffer,"NOSORT ");
 if (Flags & GVE_WRLD_MOTION_BLURR) strcat(buffer,"MOTION_BLURR ");
 if (Flags & GVE_WRLD_MOTION_TRACE) strcat(buffer,"MOTION_TRACE ");
 if (Flags & GVE_WRLD_SOFTEN_HORIZONTAL) strcat(buffer,"SOFTEN_HORIZONTAL ");
 if (Flags & GVE_WRLD_SOFTEN_VERTICAL) strcat(buffer,"SOFTEN_VERTICAL ");
 if (Flags & GVE_WRLD_CLEAR_FRAME) strcat(buffer,"CLEAR_FRAME ");
 if (Flags & GVE_WRLD_FEED_BACK) strcat(buffer,"FEED_BACK ");

 return buffer;
}


void FASTCALL GVectorWorld::RealizePosition(BOOL f)
{
 GVectorObject *b=GetRootObject();

 while(b)
   {
	b->RealizePosition(f);
	b=(GVectorObject *)b->GetBrother();
   }
}


BOOL FASTCALL GVectorWorld::Optimize(void)
{
 GVectorObject *o=GetRootObject();
 GObjectStack *stack=new GObjectStack;

 do
   {
	if (o->GetSon()) stack->Push(o->GetSon());
	if (typeid(*o)==typeid(GMeshObject))
	   {
		if (((GMeshObject *)o)->Optimize()==FALSE)
		   {
			delete stack;
			return FALSE;
		   }
	   }
	if ((o=(GVectorObject *)o->GetBrother())==NULL) o=(GVectorObject *)stack->Pop();
   }
 while (o);

 delete stack;
 return TRUE;
}


INT FASTCALL GVectorWorld::GetVertexCount(void)
{
 GVectorObject *b=GetRootObject();
 INT ret=0;

 while(b)
   {
	ret+=b->GetVertexCount(TRUE);
	b=(GVectorObject *)b->GetBrother();
   }
 return ret;
}


INT FASTCALL GVectorWorld::GetFaceCount(void)
{
 GVectorObject *b=GetRootObject();
 INT ret=0;

 while(b)
   {
	ret+=b->GetFaceCount(TRUE);
	b=(GVectorObject *)b->GetBrother();
   }
 return ret;

}


INT FASTCALL GVectorWorld::GetObjectCount(void)
{
 GVectorObject *b=GetRootObject();
 INT ret=0;

 while(b)
   {
	ret+=b->GetObjectCount(TRUE);
	b=(GVectorObject *)b->GetBrother();
   }
 return ret;
}



///////////////////////////////////////////////////////////////////////////////
//
// File I/O routines
//
BOOL FASTCALL GVectorWorld::Save(char *fname)
{
 GFile file("VECTFILE",GVE_FILE_VERSION);

 if (file.Create(fname,GFile::COMPRESSION_NONE)!=TRUE) return FALSE;
 if (Write(file,GFile::MAIN_TYPE)!=TRUE) return FALSE;
 return file.Close();
}


BOOL FASTCALL GVectorWorld::Load(char *fname)
{
 GFile file("VECTFILE",GVE_FILE_VERSION);

 if (file.Open(fname)!=TRUE) return FALSE;
 if (file.OpenMetaChunk()!=GFile::MAIN_TYPE) return FALSE;
 if (Read(file)!=TRUE) return FALSE;
 if (file.CloseMetaChunk()!=TRUE) return FALSE;
 return file.Close();
}


BOOL FASTCALL GVectorWorld::Write(GFile &file,DWORD ctype)
{
 FileWorldHead fworld;

 GetName(fworld.Name);
 fworld.Flags=LSB2DWORD(Flags & ~(GVE_WRLD_BUSY | GVE_WRLD_RENDERING));
 fworld.Magic=LSB2DWORD(Magic);
 fworld.Type=LSB2DWORD(Type);
 fworld.CameraMagic=DWORD2LSB(Camera->GetMagic());
 fworld.LightMagic=DWORD2LSB(Light->GetMagic());
 fworld.AmbientMagic=DWORD2LSB(Ambient->GetMagic());


 if(file.CreateMetaChunk(ctype)!=TRUE) return FALSE;

 if(file.CreateMetaChunk(GObject::OBJECT_TYPE)!=TRUE) return FALSE;
 if(WriteObjects(file,GetRootObject())!=TRUE) return FALSE;
 if(file.CloseMetaChunk()!=TRUE) return FALSE;

 if(Info.Write(file,GInfoObject::INFO_TYPE)!=TRUE) return FALSE;
 if(Palette.Write(file,GPalette::PALETTE_TYPE)!=TRUE) return FALSE;
 if(TextureMapList.Write(file,GBitmap::TEXTUREMAP_TYPE)!=TRUE) return FALSE;
 if(LightMapList.Write(file,GBitmap::LIGHTMAP_TYPE)!=TRUE) return FALSE;
 if(BumpMapList.Write(file,GBitmap::BUMPMAP_TYPE)!=TRUE) return FALSE;

 if(file.CreateChunk(GObject::STATE_TYPE)!=TRUE) return FALSE;
 if(file.WriteChunk(&fworld,sizeof(fworld))!=sizeof(fworld)) return FALSE;
 if(file.CloseChunk()!=TRUE) return FALSE;

 return file.CloseMetaChunk();
}


BOOL FASTCALL GVectorWorld::Read(GFile &file)
{
 DWORD typ;
 PTR data;

 if(Free()!=TRUE) return FALSE;
 if(Alloc()!=TRUE) return FALSE;

 while ((typ=file.OpenChunk())!=GFile::END_TYPE)
   {
	switch(typ)
	   {
		case GPalette::PALETTE_TYPE:
			if(Palette.Read(file)!=TRUE) return FALSE;
			break;
		case GInfoObject::INFO_TYPE:
			if(Info.Read(file)!=TRUE) return FALSE;
			break;
		case GBitmap::TEXTUREMAP_TYPE:
			if(TextureMapList.Read(file)!=TRUE) return FALSE;
			break;
		case GBitmap::BUMPMAP_TYPE:
			if(BumpMapList.Read(file)!=TRUE) return FALSE;
			break;
		case GBitmap::LIGHTMAP_TYPE:
			if(LightMapList.Read(file)!=TRUE) return FALSE;
			break;
		case GObject::OBJECT_TYPE:
			if(ReadObjects(file,GetRootObject())!=TRUE) return FALSE;
			break;
		case GObject::STATE_TYPE:
			if(file.ReadChunk(&data)!=sizeof(FileWorldHead)) return FALSE;
			if(file.CloseChunk()!=TRUE) return FALSE;
			SetName(((FileWorldHead *)data)->Name);
			Flags=LSB2DWORD(((FileWorldHead *)data)->Flags);
			Magic=LSB2DWORD(((FileWorldHead *)data)->Magic);
			Type=LSB2DWORD(((FileWorldHead *)data)->Type);
			Camera=(GCameraObject *)LSB2DWORD(((FileWorldHead *)data)->CameraMagic);
			Light=(GLightObject *)LSB2DWORD(((FileWorldHead *)data)->LightMagic);
			Ambient=(GLightObject *)LSB2DWORD(((FileWorldHead *)data)->AmbientMagic);
			free(data);
			break;
		default:
			if(file.CloseChunk()!=TRUE) return FALSE;
			break;
	   }
   }

 Camera=(GCameraObject *)FindObject((DWORD)Camera);
 Light=(GLightObject *)FindObject((DWORD)Light);
 Ambient=(GLightObject *)FindObject((DWORD)Ambient);

 return TRUE;
}


BOOL FASTCALL GVectorWorld::WriteObjects(GFile &file,GVectorObject *o)
{
 while (o)
   {
	if (typeid(*o)==typeid(GLightObject))
		o->Write(file,GLightObject::LIGHT_TYPE);
	else if (typeid(*o)==typeid(GMeshObject))
		o->Write(file,GMeshObject::MESH_TYPE);
	else if (typeid(*o)==typeid(GCameraObject))
		o->Write(file,GCameraObject::CAMERA_TYPE);
	else
		o->Write(file,GObject::CONTAINER_TYPE);

	if (o->GetSon())
	   {
		if(file.OpenMetaChunk(GFile::SONS_TYPE)!=TRUE) return FALSE;
		if(WriteObjects(file,(GVectorObject *)o->GetSon())!=TRUE) return FALSE;
		if(file.CloseMetaChunk()!=TRUE) return FALSE;
	   }
	o=(GVectorObject *)o->GetBrother();
   }
 return TRUE;
}


BOOL FASTCALL GVectorWorld::ReadObjects(GFile &file,GVectorObject *o)
{
 DWORD typ;
 GVectorObject* p=o;

 while ((typ=file.OpenChunk())!=GFile::END_TYPE)
   {
	switch (typ)
	   {
		case GFile::SONS_TYPE:
			if(ReadObjects(file,(GVectorObject *)o->AddSon())!=TRUE) return FALSE;
			break;
		case GLightObject::LIGHT_TYPE:
			o=(GVectorObject *)o->AddBrother(new GLightObject);
			if(o->Read(file)!=TRUE) return FALSE;
			break;
		case GMeshObject::MESH_TYPE:
			o=(GVectorObject *)o->AddBrother(new GMeshObject);
			if(o->Read(file)!=TRUE) return FALSE;
			break;
		case GCameraObject::CAMERA_TYPE:
			o=(GVectorObject *)o->AddBrother(new GCameraObject);
			if(o->Read(file)!=TRUE) return FALSE;
			break;
		case GObject::CONTAINER_TYPE:
			o=(GVectorObject *)o->AddBrother(new GVectorObject);
			if(o->Read(file)!=TRUE) return FALSE;
			break;
		default:
			if(file.CloseChunk()!=TRUE) return FALSE;
			break;
	   }
   }

 return ObjectList.DeleteEntry(p);
}

