///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#define STRICT
#include <windows.h>
#include <windowsx.h>
#include "..\include\gvepro.h"
#include "gve4win.h"


EXTERN HINSTANCE    	hMainInstance;
EXTERN HWND 			hMainWindow,hObjectList,hObjCtrl,hCamCtrl,hLightCtrl;
EXTERN char				AppName[];
EXTERN char 			AppTitle[];
EXTERN char 			ErrTitle[];

EXTERN RECT 		 	ClientRect,MainRect;
EXTERN RECT				ObjControlRect,CamControlRect,LightControlRect;
EXTERN GVectorSystem 	VSystem;
EXTERN CHAR				VWorldName[MAXPATH];
EXTERN GVectorWorld  	*VWorld;
EXTERN GVectorObject 	*VObject;
EXTERN GVectorObject 	*VMesh;
EXTERN GLightObject  	*VLight;
EXTERN GCameraObject 	*VCamera;
EXTERN GFrame		 	*VFrame;


LOCAL BOOL File_New(HWND hwnd);
LOCAL BOOL File_Import(HWND hwnd);
LOCAL BOOL File_Load(HWND hwnd);
LOCAL BOOL File_Save(HWND hwnd);
LOCAL BOOL File_SaveAs(HWND hwnd);
LOCAL BOOL Obj_Info(HWND hwnd);
LOCAL BOOL Obj_Import(HWND hwnd);
LOCAL BOOL Colour_Load(HWND hwnd);
LOCAL BOOL Colour_Save(HWND hwnd);
LOCAL BOOL Opt_Options(HWND hwnd);
LOCAL BOOL Opt_Info(HWND hwnd);
LOCAL BOOL Opt_Load(HWND hwnd);
LOCAL BOOL Opt_Save(HWND hwnd);




void MainMenu(HWND hwnd, int id, HWND hwndCtl,UINT codeNotify)
{
 switch(id)
   {
	case CM_HLP_ABOUT:
		MessageBox(hwnd,"Gomez Vector Engine / Pro   1.0\n"
			"WIN32 Version \n\n"
			"(C) 1998  by Marc van Shaney \n"
			"               aka Kaya Memisoglu",
			"About GVEpro",MB_OK|MB_ICONINFORMATION);
		break;
	case CM_HLP_BUILD:
		MessageBox(hwnd,VSystem.GetBuildInfo(),"About GVEpro",MB_OK|MB_ICONINFORMATION);
		break;
	case CM_FILE_NEW:
		File_New(hwnd);
		break;
	case CM_FILE_OPEN:
		File_Load(hwnd);
		break;
	case CM_FILE_SAVE:
		File_Save(hwnd);
		break;
	case CM_FILE_SAVEAS:
		File_SaveAs(hwnd);
		break;
	case CM_FILE_IMPORT:
		File_Import(hwnd);
		break;
	case CM_FILE_EXIT:
		DestroyWindow(hwnd);
		break;
	case CM_OBJ_NEWCAMERA:
		VObject->AddSon(new GCameraObject);
		AppRealizeObjectList();
		break;
	case CM_OBJ_NEWLIGHT:
		VObject->AddSon(new GLightObject);
		AppRealizeObjectList();
		break;
	case CM_OBJ_DELETE:
		if (VObject!=VWorld->GetRootObject())
			VWorld->ObjectList.DeleteEntry(VObject);
		AppRealizeObjectList();
		break;
	case CM_OBJ_ATTRIBUTES:
		DisplayObjectProps(hwnd);
		break;
	case CM_OBJ_INFO:
		Obj_Info(hwnd);
		break;
	case CM_OBJ_IMPORT:
		Obj_Import(hwnd);
		break;
	case CM_COLOUR_LOAD:
		Colour_Load(hwnd);
		break;
	case CM_COLOUR_SAVE:
		Colour_Save(hwnd);
		break;
	case CM_COLOUR_TABLES:
		VSystem.RealizePalette();
		break;
	case CM_SHOW_OBJECT:
		DisplayObjControl(hwnd);
		break;
	case CM_SHOW_CAMERA:
		DisplayCamControl(hwnd);
		break;
	case CM_SHOW_LIGHT:
//		DisplayLightCtrl(hwnd);
		break;
	case CM_SHOW_OBJLIST:
		DisplayObjectList(hwnd);
		break;
	case CM_OPT_DISPLAY:
		Opt_Options(hwnd);
		break;
	case CM_OPT_INFO:
		Opt_Info(hwnd);
		break;
	case CM_OPT_LOAD:
		Opt_Load(hwnd);
		break;
	case CM_OPT_SAVE:
		Opt_Save(hwnd);
		break;
  }
}



LOCAL BOOL File_New(HWND hwnd)
{
 if (MessageBox(hwnd,"\nAre you sure you want to discard the current world ?\n","New World",
		 MB_OKCANCEL | MB_DEFBUTTON2 |
		 MB_ICONQUESTION | MB_APPLMODAL)==IDCANCEL) return TRUE;
 VWorld->Free();
 VWorld->Alloc();
 AppRealizeWorld();
 return TRUE;
}



LOCAL BOOL File_Import(HWND hwnd)
{
 GVectorObject *head;
 G3dsFile Import;
 char szFile[MAXPATH]="";
 char szFilter[]= "3D-Studio files (*.3DS)\0*.3DS\0"
		  "All files (*.*)\0*.*\0";


 if(GetLoadFileName(hwnd,"Import File",szFilter,szFile,MAXPATH)==NULL) return FALSE;
 if((head=Import.Import(szFile))==NULL)
   {
	MessageBox(hwnd,"3DS-Import failed.",ErrTitle,MB_ICONSTOP|MB_OK);
	return FALSE;
   }

 VWorld->Free();
 VWorld->Alloc();
 VObject=VWorld->GetRootObject();
 VObject->AddSon(head);
 AppRealizeWorld();
 return TRUE;
}



LOCAL BOOL File_Load(HWND hwnd)
{
 char szFilter[]="Gomez Vector Engine files (*.GVE)\0*.GVE\0"
		 "All files (*.*)\0*.*\0";

 if(GetLoadFileName(hwnd,"Load Vector File",szFilter,VWorldName,MAXPATH)==NULL) return FALSE;
 if (VWorld->Load(VWorldName)==FALSE)
   {
	MessageBox(hwnd,"Error while reading GVE file",ErrTitle,MB_ICONSTOP|MB_OK);
	VWorld->Free();
	VWorld->Alloc();
	AppRealizeWorld();
	return FALSE;
   }

 AppRealizeWorld();
 return TRUE;
}



LOCAL BOOL File_Save(HWND hwnd)
{
 if (VWorld->Save(VWorldName)==FALSE)
   {
	MessageBox(hwnd,"Error while writing GVE file",ErrTitle,MB_ICONSTOP|MB_OK);
	return FALSE;
   }

 return TRUE;
}



LOCAL BOOL File_SaveAs(HWND hwnd)
{
 char szFilter[]="Gomez Vector Engine files (*.GVE)\0*.GVE\0"
		 "All files (*.*)\0*.*\0";

 if (GetSaveFileName(hwnd,"Save Vector File",szFilter,VWorldName,MAXPATH)==NULL) return FALSE;
 if (VWorld->Save(VWorldName)==FALSE)
   {
	MessageBox(hwnd,"Error while writing GVE file",ErrTitle,MB_ICONSTOP|MB_OK);
	return FALSE;
   }

 return TRUE;
}



LOCAL BOOL Obj_Info(HWND hwnd)
{
 char text[1000];
 CHAR buffer1[64],buffer2[128];

 sprintf(text,"Object Name\t:%s\n"
		  "Object Class\t:%s\n"
		  "Object Type\t:%s\n"
		  "Object Flags\t:%s\n\n"
		  "Number of Faces\t:%d\n"
		  "Number of Vertices\t:%d\n\n"
		  "Number of Objects\t:%d\n"
		  "Number of Faces\t:%d\n"
		  "Number of Vertices\t:%d\n",
		  VObject->GetName(),
		  typeid(*VObject).name(),
		  VObject->GetTypeName(buffer1),
		  VObject->GetFlagsName(buffer2),
		  VObject->GetFaceCount(FALSE),
		  VObject->GetVertexCount(FALSE),
		  VObject->GetObjectCount(TRUE),
		  VObject->GetFaceCount(TRUE),
		  VObject->GetVertexCount(TRUE));

 MessageBox(hwnd,text,"Object Info",MB_OK|MB_ICONINFORMATION);

 return TRUE;

}




LOCAL BOOL Obj_Import(HWND hwnd)
{
 GVectorObject *head;
 G3dsFile Import;
 char szFile[MAXPATH]="";
 char szFilter[]="3D-Studio files (*.3DS)\0*.3DS\0"
		 "All files (*.*)\0*.*\0";

 if(GetLoadFileName(hwnd,"Import Objects",szFilter,szFile,MAXPATH)==NULL) return FALSE;
 if((head=Import.Import(szFile))==NULL)
   {
	MessageBox(hwnd,"3DS-Import failed.",ErrTitle,MB_ICONSTOP|MB_OK);
	return FALSE;
   }

 VObject->AddSon(head);
 AppRealizeObjectList();
 return TRUE;
}



LOCAL BOOL Colour_Load(HWND hwnd)
{
 char szFile[MAXPATH]="";
 char szFilter[]="18 bit PAL files (*.PAL)\0*.PAL\0"
		 "All files (*.*)\0*.*\0";

 if(GetLoadFileName(hwnd,"Import Palette",szFilter,szFile,MAXPATH)==NULL) return FALSE;

 VWorld->Palette.ImportPAL(szFile);
 VFrame->SetPalette(&VWorld->Palette);
 VFrame->RealizePalette();

 return TRUE;
}



LOCAL BOOL Colour_Save(HWND hwnd)
{
 char szFile[MAXPATH]="";
 char szFilter[]="18 bit PAL files (*.PAL)\0*.PAL\0"
		 "All files (*.*)\0*.*\0";

 if(GetSaveFileName(hwnd,"Export Palette",szFilter,szFile,MAXPATH)==NULL) return TRUE;

 VWorld->Palette.ExportPAL(szFile);

 return TRUE;
}




LOCAL BOOL Opt_Options(HWND hwnd)
{
 return TRUE;
}



LOCAL BOOL Opt_Info(HWND hwnd)
{
 char text[1000];
 sprintf(text,"Frame Size\t:%dx%d\n"
		  "Frame Depth\t:%d\n",
		  VFrame->GetWidth(),VFrame->GetHeight(),
		  VFrame->GetDepth());

 MessageBox(hwnd,text,"Frame Info",MB_OK|MB_ICONINFORMATION);

 return TRUE;
}



LOCAL BOOL Opt_Load(HWND hwnd)
{
 AppLoadOptions();

 SetWindowPos(hwnd,NULL,ClientRect.left,ClientRect.top,
			 ClientRect.right-ClientRect.left+1,
			 ClientRect.bottom-ClientRect.top+1,SWP_NOZORDER|SWP_NOCOPYBITS);
 AppSetFrameRect();
 return TRUE;
}



LOCAL BOOL Opt_Save(HWND hwnd)
{
 AppSaveOptions();
 return TRUE;
}


