///////////////////////////////////////////////////////////////////////////////
//
// (c) 1997 by Marc van Shaney , aka Kaya Memisoglu
//
// This Source-Code was written by and is copyrighted by Kaya Memisoglu
// You may use the whole code or only parts of it in your own programs
// whether they are commercial or not. BUT you must give credits for these
// parts to me (Kaya Memisoglu  or  Marc van Shaney).
// I am not responsible for any damage or loss caused by this program, so
// if you success in erasing your harddisk with this code, then it is your
// own fault and you should not claim me to be responsible...
// I would be very pleased if you also sent me a postcard or even a letter
// if you can use this code. My address:
//
//        Kaya Memisoglu
//        Reichenberger Ring 50
//        63512 Hainburg
//        Germany
//
//


#define STRICT
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include "..\include\gvepro.h"
#include "gve4win.h"
#include "treeview.h"


EXTERN HINSTANCE    	hMainInstance;
EXTERN HWND 			hMainWindow,hObjectList,hObjCtrl,hCamCtrl,hLightCtrl;
EXTERN char				AppName[];
EXTERN char 			AppTitle[];
EXTERN char 			ErrTitle[];

EXTERN RECT 		 	ClientRect,MainRect,ObjectListRect;
EXTERN RECT				ObjControlRect,CamControlRect,LightControlRect;
EXTERN GVectorSystem 	VSystem;
EXTERN GVectorWorld  	*VWorld;
EXTERN GVectorObject 	*VObject;
EXTERN GVectorObject 	*VMesh;
EXTERN GLightObject  	*VLight;
EXTERN GCameraObject 	*VCamera;


LOCAL HWND        	hwndTreeView;          	// hwnd of TreeView control
LOCAL HIMAGELIST  	hImageList;            	// handle to ImageList

LOCAL int			iRootIcon;
LOCAL int			iObjectIcon;
LOCAL int			iCameraIcon;
LOCAL int			iContainerIcon;
LOCAL int			iLightIcon;
LOCAL int			iMeshIcon;



LOCAL BOOL 			ObjectListAddObject(HWND hwnd,HTREEITEM hParent,GVectorObject *o);;
LOCAL HIMAGELIST 	ObjectListCreateImageList(void);
LOCAL BOOL 			ObjectListOnCreate(HWND hwnd, CREATESTRUCT* lpCreateStruct);
LOCAL void 			ObjectListOnDestroy(HWND hwnd);
LOCAL void 			ObjectListOnSize(HWND hwnd,UINT state,int cx,int cy);
LOCAL LRESULT		ObjectListOnNotify(HWND hwnd, INT idForm, NMHDR* pnmhdr);





BOOL DisplayObjectList(HWND hwnd)
{
 if (hObjectList)
   {
	SetFocus(hObjectList);
	return TRUE;
   }	

 hObjectList = CreateWindowEx(WS_EX_TOOLWINDOW,
				"ObjectList",
				"Objects",
				WS_OVERLAPPEDWINDOW | WS_POPUP,
				ObjectListRect.left, ObjectListRect.top,
				ObjectListRect.right-ObjectListRect.left+1,
				ObjectListRect.bottom-ObjectListRect.top+1,
				hwnd,
				NULL,
				hMainInstance,
				NULL   );

 if (!hObjectList) return FALSE;

 ShowWindow(hObjectList, SW_SHOWDEFAULT);
 UpdateWindow(hObjectList);

 return TRUE;
}




BOOL DestroyObjectList(HWND hwnd)
{
 if (hObjectList) DestroyWindow(hObjectList);
 hObjectList=NULL;
 return TRUE;
}



LRESULT CALLBACK ObjectListProc(HWND hwnd, UINT Message,WPARAM wParam, LPARAM lParam)
{
 switch(Message)
   {
	HANDLE_MSG(hwnd, WM_CREATE, ObjectListOnCreate);
	HANDLE_MSG(hwnd, WM_DESTROY, ObjectListOnDestroy);
	HANDLE_MSG(hwnd, WM_SIZE, ObjectListOnSize);
	HANDLE_MSG(hwnd, WM_NOTIFY, ObjectListOnNotify);
	case PM_CHANGE_OBJECTLIST:
		TreeView_DeleteAllItems(hwndTreeView);
		ObjectListAddObject(hwndTreeView,TVI_ROOT,VWorld->GetRootObject());
		break;
	case PM_CHANGE_OBJECT:
		break;
	default:
		return DefWindowProc(hwnd, Message, wParam, lParam);
   }
 return 0;
}




BOOL ObjectListOnCreate(HWND hwnd, CREATESTRUCT* lpCreateStruct)
{
 hImageList = ObjectListCreateImageList();
 hwndTreeView = CreateTreeView(hMainInstance,
				  hwnd,
				  hImageList,
				  0,
				  TVS_HASLINES | TVS_HASBUTTONS |
				  TVS_LINESATROOT | TVS_HASLINES | TVS_SHOWSELALWAYS |
				  WS_VISIBLE | WS_CHILD | WS_BORDER);

 TreeView_DeleteAllItems(hwndTreeView);
 ObjectListAddObject(hwndTreeView,TVI_ROOT,VWorld->GetRootObject());

 return TRUE;
}



void ObjectListOnDestroy(HWND hwnd)
{
 GetWindowRect(hwnd,&ObjectListRect);

 if (hImageList)  ImageList_Destroy(hImageList);
 hObjectList=NULL;
}







void ObjectListOnSize(HWND hwnd,UINT state,int cx,int cy)
{
 RECT rc;

 GetClientRect(hwnd, &rc);

 if (hwndTreeView)
	SetWindowPos(hwndTreeView,
			 NULL,
			 0, 0,
			 rc.right, rc.bottom,
			 SWP_NOZORDER);
}


LRESULT ObjectListOnNotify(HWND hwnd, INT idForm, NMHDR* lparam)
{
 if (lparam->code==TVN_SELCHANGED)
	{
	TV_ITEM tviNew = ((LPNM_TREEVIEW)lparam)->itemNew;

	if (tviNew.hItem)
	   {
		if(!VWorld->FindObject((GVectorObject *)tviNew.lParam))
			return 0;
		VObject=(GVectorObject *)tviNew.lParam;
		if (typeid(*VObject)==typeid(GCameraObject))
		   {
			VCamera=(GCameraObject *)VObject;
			VWorld->SetCamera(VCamera);
		   }
		else if (typeid(*VObject)==typeid(GLightObject))
		   {
			VLight=(GLightObject *)VObject;
			VWorld->SetLight(VLight);
		   }
		else VMesh=(GMeshObject *)VObject;
		AppRealizeObject();
		return 0;
	   }
   }
 return 0;
}





BOOL ObjectListAddObject(HWND hwnd,HTREEITEM hParent,GVectorObject *o)
{
 TV_ITEM			tvi;
 TV_INSERTSTRUCT  	tvins;
 HTREEITEM			hPrev=TVI_FIRST;
 INT i;

 while (o)
   {
	switch (o->GetType())
	   {
		case GObject::CONTAINER_TYPE:
			i=iContainerIcon;
			break;
		case GMeshObject::MESH_TYPE:
			i=iMeshIcon;
			break;
		case GVectorObject::ROOT_TYPE:
			i=iRootIcon;
			break;
		case GObject::OBJECT_TYPE:
			i=iObjectIcon;
			break;
		case GCameraObject::CAMERA_TYPE:
			i=iCameraIcon;
			break;
		case GLightObject::LIGHT_TYPE:
			i=iLightIcon;
			break;
		case GLightObject::POINTLIGHT_TYPE:
			i=iLightIcon;
			break;
		case GLightObject::SPOTLIGHT_TYPE:
			i=iLightIcon;
			break;
		case GLightObject::AMBIENT_TYPE:
			i=iLightIcon;
			break;
	   }

	tvi.iImage          =
	tvi.iSelectedImage  = i;

	tvi.mask            = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
	tvi.pszText         = o->GetName();
	tvi.cchTextMax      = 32;
	tvi.lParam	   		= (LPARAM)o;

	tvins.item         	= tvi;
	tvins.hInsertAfter 	= hPrev;
	tvins.hParent      	= hParent;

	hPrev = TreeView_InsertItem(hwndTreeView, &tvins);

	if (o->GetSon()) ObjectListAddObject(hwnd,hPrev,(GVectorObject *)o->GetSon());
	o=(GVectorObject *)o->GetBrother();
   }

 return TRUE;
}



HIMAGELIST ObjectListCreateImageList(void)
{
 HIMAGELIST hIL;
 HICON      hIcon;

 hIL = ImageList_Create(16,     // X size of one image
			16,     // Y size of one image
			TRUE,   // Masked images
			6,      // Four images in this list
			0);     // No "growing" bitmaps in this list.
 if (!hIL) return NULL;

 hIcon = LoadIcon(hMainInstance, MAKEINTRESOURCE(IDI_CAMERAICON));
 iCameraIcon = ImageList_AddIcon(hIL, hIcon);
 DeleteObject (hIcon);

 hIcon = LoadIcon(hMainInstance, MAKEINTRESOURCE(IDI_OBJECTICON));
 iObjectIcon = ImageList_AddIcon(hIL, hIcon);
 DeleteObject (hIcon);

 hIcon = LoadIcon(hMainInstance, MAKEINTRESOURCE(IDI_LIGHTICON));
 iLightIcon = ImageList_AddIcon(hIL, hIcon);
 DeleteObject (hIcon);

 hIcon = LoadIcon(hMainInstance, MAKEINTRESOURCE(IDI_CONTAINERICON));
 iContainerIcon = ImageList_AddIcon(hIL, hIcon);
 DeleteObject (hIcon);

 hIcon = LoadIcon(hMainInstance, MAKEINTRESOURCE(IDI_ROOTICON));
 iRootIcon = ImageList_AddIcon(hIL, hIcon);
 DeleteObject (hIcon);

 hIcon = LoadIcon(hMainInstance, MAKEINTRESOURCE(IDI_MESHICON));
 iMeshIcon = ImageList_AddIcon(hIL, hIcon);
 DeleteObject (hIcon);

 return hIL;
}


